*********************************************
* Pooling data 
********************************************

set more off
log using "Pooling.log", replace text


use childdata_population98.dta, clear
gen wave=1998 

append using childdata_population04.dta
replace wave=2004 if wave==.

* create a 1 for each hh in each wave
bysort wave HHIDSUB: gen inx= _n
tab inx
* for multi-child family, only count once
replace inx=. if inx!=1
* count the number of times a household appears in appended data
egen inx2=sum(inx), by(HHIDSUB)
tab inx2

* 1998 + non-overlapping 2004 

keep if (wave==2004 & inx2==1 ) | ( wave==1998 )
drop inx inx2

append using childdata_population10.dta
replace wave=2010 if wave==.

* create a 1 for each hh in each wave
bysort wave HHIDSUB: gen inx= _n
* for multi-child family, only count once
replace inx=. if inx!=1
* count the number of times a household appears in appended data
egen inx2=sum(inx), by(HHIDSUB)

* 1998 + non-overlapping 2004 + non-overlapping 2010

keep if (wave==2010 & inx2==1 ) | (wave==2004) | ( wave==1998 )
drop inx inx2

drop C_step* *NURSEHOME indage index final

foreach var of varlist metro5 P_cohort5 P_cohort6 {
replace `var'=0 if `var'==.
}

gen wave1998=(wave==1998)
gen wave2004=(wave==2004)
gen wave2010=(wave==2010)

order wave* HHID- FPN_NFIN  JSUBHH- JPN_NFIN metro- metro4  metro5NewHH2010 P_cohort1- P_cohort4  P_cohort5  P_cohort6

bysort HHIDSUB: gen index= _n
tab TOT if index==1
tab wave TOT
tab wave TOT if index==1

****************************
* Construct variables *
****************************
gen metro_missing= metro4+ metro5NewHH2010
gen C_CollegeP= C_College+ C_PostCollege

* parent health information after pooling
rename P_health P_health0
rename R_health R_health0
rename S_health S_health0

factor R_sah R_ADLall R_IADLsum R_diabetes R_highbp R_stroke
predict R_health

factor S_sah S_ADLall S_IADLsum S_diabetes S_highbp S_stroke if P_father==0 & P_mother==0
predict S_health

gen P_health = -R_health
replace P_health = -(R_health+S_health)/2 if P_father==0 & P_mother==0

gen R_widowed_single=R_widowed==1 & SINGLE==1
gen R_widowed_cohab=R_widowed==1 & SINGLE==0
gen R_nonwidowed_single=(R_married==1 | R_sepdiv==1| R_nevermarried==1 |R_othermarstat==1)& SINGLE==1
gen R_nonwidowed_cohab=(R_married==1 | R_sepdiv==1| R_nevermarried==1 |R_othermarstat==1)& SINGLE==0
gen R_married_single=R_married==1 & SINGLE==1
gen R_married_cohab=R_married==1 & SINGLE==0

gen P_father_ws=P_father*R_widowed_single
gen P_father_nws=P_father*R_nonwidowed_single
gen P_mother_ws=P_mother*R_widowed_single
gen P_mother_nws=P_mother*R_nonwidowed_single

sum

save childdata_pooled.dta, replace

log close
