MODULE MATRIX
! VERSION 1.1, DEC 1 2003
IMPLICIT NONE
PRIVATE :: SWAP
CONTAINS

FUNCTION MATINV(AIN)
IMPLICIT NONE
REAL(8), DIMENSION(:,:), INTENT(IN) :: AIN
INTEGER, DIMENSION(SIZE(AIN,1)) :: ipiv,indxr,indxc
LOGICAL, DIMENSION(SIZE(AIN,1)) :: lpiv
REAL(8) :: pivinv,A(SIZE(AIN,1),SIZE(AIN,2)),MATINV(SIZE(AIN,1),SIZE(AIN,2))
REAL(8), DIMENSION(SIZE(AIN,1)) :: dumc
INTEGER, TARGET :: irc(2)
INTEGER :: i,l,n
INTEGER, POINTER :: irow,icol
IF (SIZE(AIN,1).NE.SIZE(AIN,2)) THEN
    PRINT *, 'FATAL ERROR: matrix is not square (MATINV)***'
    STOP
END IF    
A=AIN
irow => irc(1)
icol => irc(2)
ipiv=0
n=SIZE(A,1)
DO i=1,n
    lpiv = (ipiv == 0)
    irc=MAXLOC(ABS(A),outerand(lpiv,lpiv))
    ipiv(icol)=ipiv(icol)+1
    IF (ipiv(icol) > 1) THEN
        PRINT *, '***FATAL ERROR: MATINV: singular matrix (1)'
        STOP
    END IF     
    IF (irow /= icol) THEN
        CALL swap(A(irow,:),A(icol,:))
    END IF
    indxr(i)=irow
    indxc(i)=icol
    IF (a(icol,icol) == 0.0) THEN
        PRINT *, '***FATAL ERROR: MATINV: singular matrix (2)'
        STOP
    END IF     
    pivinv=1.0d0/A(icol,icol)
    A(icol,icol)=1.0d0
    A(icol,:)=A(icol,:)*pivinv
    dumc=A(:,icol)
    A(:,icol)=0.0
    A(icol,icol)=pivinv
    A(1:icol-1,:)=A(1:icol-1,:)-outerprod(dumc(1:icol-1),A(icol,:))
    A(icol+1:,:)=A(icol+1:,:)-outerprod(dumc(icol+1:),A(icol,:))
END DO
DO l=n,1,-1
    CALL swap(A(:,indxr(l)),A(:,indxc(l)))
END DO
MATINV=A
END FUNCTION MATINV




FUNCTION CHOL(AIN)
IMPLICIT NONE
REAL(8), DIMENSION(:,:), INTENT(IN) :: AIN
REAL(8) :: A(SIZE(AIN,1),SIZE(AIN,2)),p(SIZE(AIN,1)),CHOL(SIZE(AIN,1),SIZE(AIN,2))
INTEGER :: i,j,n
REAL(8) :: summ
IF (SIZE(AIN,1).NE.SIZE(AIN,2)) THEN
    PRINT *, 'FATAL ERROR: matrix is not square (CHOL)***'
    STOP
END IF    
A=AIN
CHOL=0.0d0
n=SIZE(A,1)
DO i=1,n
    summ=A(i,i)-DOT_PRODUCT(A(i,1:i-1),A(i,1:i-1))
    IF (summ <= 0.0) THEN
!        PAUSE 'CHOL failed'
        PRINT *, 'CHOL failed'
    END IF      
    p(i)=DSQRT(summ)
    A(i+1:n,i)=(A(i,i+1:n)-matmul(A(i+1:n,1:i-1),A(i,1:i-1)))/p(i)
    A(i,i)=p(i)
END DO
FORALL (i=1:n,j=1:n,j<=i) CHOL(i,j)=A(i,j)
END FUNCTION CHOL


PURE FUNCTION outerand(a,b)
LOGICAL, DIMENSION(:), INTENT(IN) :: a,b
LOGICAL, DIMENSION(size(a),size(b)) :: outerand
outerand = spread(a,dim=2,ncopies=size(b)) .and. &
spread(b,dim=1,ncopies=size(a))
END FUNCTION outerand

PURE FUNCTION outerprod(a,b) ! Gets the outerproduct of two vectors
IMPLICIT NONE	
REAL(8), DIMENSION(:), INTENT(IN) :: a(:),b(:)
REAL(8) :: outerprod(SIZE(a),SIZE(b))
outerprod = SPREAD(a,dim=2,ncopies=SIZE(b)) * &
SPREAD(b,dim=1,ncopies=size(a))
END FUNCTION outerprod

PURE SUBROUTINE swap(a,b)
REAL(8), INTENT(INOUT) :: a(:),b(:)
REAL(8) :: dum(SIZE(A))
dum=a
a=b
b=dum
END SUBROUTINE swap

END MODULE MATRIX
