MODULE MINIMIZATION
USE MATRIX
USE Object
IMPLICIT NONE
PUBLIC dfpmin 

CONTAINS
SUBROUTINE dfpmin(theta,n,gtol,func)
IMPLICIT NONE
REAL(8), INTENT(IN)    :: gtol
INTEGER, INTENT(IN)    :: n	        ! number of elements in the parameter vector
REAL(8), INTENT(INOUT) :: theta(n)	! parameter vector over which you which to maximize
INTERFACE	                        ! This lets the subroutine know that func is not a variable, it is a function
	REAL(8) FUNCTION func(theta,n,gradflag)   ! This is not the likelihood itself, it just let the program know
    IMPLICIT NONE		  
    INTEGER, INTENT(IN) :: n
    INTEGER, INTENT(IN) :: gradflag
    REAL(8), INTENT(IN) :: theta(n)
    END FUNCTION func
END INTERFACE
INTEGER, PARAMETER :: ITMAX=300	! Maximum number of iterations (in case it does not converge)
REAL(8), PARAMETER :: STPMX=100.0,EPS=EPSILON(theta),TOLD=EPS ! Tolerance criterions
INTEGER :: j,its ! indices
LOGICAL :: check 
REAL(8) :: den,fac,fad,fae,ftheta,sumdg,sumd ! working matrices
REAL(8) :: dg(n),g(n),hdg(n),thetanew(n),d(n)
REAL(8) :: B(n,n),fret,stpmax

print *,'check2-1! - Dfpmin In!'
ftheta=func(theta,n,0)	        ! initial value of the function
g=gradient(theta,n,ftheta,func)	! initial value of the gradient, if analytical gradient unknown here

!print *,theta2
!print *,gmmobj(theta2,ntheta2,0)
!print *,g
!STOP

print *,'check2-2!'
B=0.0d0				! initialize the matrix to a negative definite matrix
DO j = 1, n
	B(j,j)=1.0d0
END DO
d=-MATMUL(B,g)	! initial line direction
!	print *,'B: ',B
!	print *,'g: ',g
!	print *,'d: ',d
stpmax=STPMX*MAX(vabs(theta),REAL(SIZE(theta),8))
DO its=1,ITMAX	! Main loop over iterations
print *,'check2-3!', its
	CALL lnsrch(theta,ftheta,g,d,thetanew,fret,stpmax,check,func)
print *, 'Lnsrch Out!'
	ftheta=fret	! update the value of the fuction
    d=thetanew-theta	! update the direction of movement
    theta=thetanew	! update the point
    IF (MAXVAL(ABS(d)/MAX(ABS(theta),1.0D0)) < TOLD) RETURN     ! Check convergence in Delta-d
print *,'check2-4-Loop!'
    dg=g				! save the old gradient
	g=gradient(theta,n,ftheta,func)	! and get the new gradient
    den=MAX(ftheta,1.0d0)	
    IF (MAXVAL(ABS(g)*MAX(ABS(theta),1.0d0)/den) < gtol) RETURN ! Check for convergence on zero gradient
    dg=g-dg			! compute difference in gradients
    hdg=MATMUL(B,dg)	        ! and difference times current matrix
    fac=DOT_PRODUCT(dg,d)	! Dot products for denominators
    fae=DOT_PRODUCT(dg,hdg)
    sumdg=DOT_PRODUCT(dg,dg)
    sumd=DOT_PRODUCT(d,d)
    IF (fac > DSQRT(EPS*sumdg*sumd)) THEN	! Skip update if fac not positive enough	
		fac=1.0d0/fac
        fad=1.0d0/fae
        dg=fac*d-fad*hdg			! Vector that makes BFGS different from DFP
        ! BFGS updating formula
        B=B+fac*outerprod(d,d)-fad*outerprod(hdg,hdg)+fae*outerprod(dg,dg)
    END IF
    d=-MATMUL(B,g)	! Next direction to go
END DO		! Go back for another iteration
PAUSE 'dfpmin: too many iterations'
STOP
END SUBROUTINE dfpmin





FUNCTION GRADIENT(x0,k,f0,func)
! Routine to get the numerical gradient of a function using forward differences
IMPLICIT NONE
INTEGER, INTENT(IN) :: k
REAL(8), INTENT(IN) :: x0(k),f0
INTERFACE	                       ! This lets the subroutine know that func is not a variable, it is a function
	REAL(8) FUNCTION func(theta,k,gradflag)  ! This is not the likelihood itself, it just let the program know
    IMPLICIT NONE		       ! the arguments and type of arguments it accepts
    INTEGER, INTENT(IN) :: k
    INTEGER, INTENT(IN) :: gradflag
    REAL(8), INTENT(IN) :: theta(k)
    END FUNCTION func
END INTERFACE
REAL(8) :: GRADIENT(k),grdd(k),dh(k),ax0(k),xdh(k)
REAL(8) :: arg(k,k),dax0(k),one
INTEGER :: i
one  = 1.0D0	! This parameter affects the length of dh
grdd = 0.0D0
! Computation of stepsize (dh) for gradient
ax0  = ABS(x0)
dax0 = 1.0D0
WHERE (x0.ne.0.0D0) dax0 = x0/ax0
dh = (1.0D-6)*(1.0D-2)*one*dax0
WHERE (ax0.gt.(1.0D-2)*one) dh = (1.0D-8)*ax0*dax0
xdh = x0+dh
arg = SPREAD(x0,DIM=2,NCOPIES=k)
DO i = 1, k
	print *,'Calculating Gradient', i
	arg(i,i)=xdh(i)
    grdd(i)=func(arg(:,i),k,1)
END DO
GRADIENT = (grdd-f0)/dh
END FUNCTION GRADIENT
  
SUBROUTINE lnsrch(xold,fold,g,p,x,f,stpmax,check,func)
! Given an N-dimensional point xold, the value of the function and gradient there, fold
! and g, and a direction p, finds a new point x along the direction p from xold where the
! function func has decreased sufficiently. xold, g, p, and x are all arrays of length N.
! The new function value is returned in f. stpmax is an input quantity that limits the length
! of the steps so that you do not try to evaluate the function in regions where it is undefined
! or subject to overflow. p is usually the Newton direction. The output quantity check is
! false on a normal exit. It is true when x is too close to xold. In a minimization algorithm,
! this usually signals convergence and can be ignored. However, in a zero-finding algorithm
! the calling program should check whether the convergence is spurious.
! Parameters: ALF ensures su.cient decrease in function value; TOLX is the convergence
! criterion on .x.
IMPLICIT NONE
REAL(8), DIMENSION(:), INTENT(IN) :: xold,g
REAL(8), DIMENSION(:), INTENT(INOUT) :: p(SIZE(xold))
REAL(8), INTENT(IN) :: fold,stpmax
REAL(8), INTENT(OUT) :: x(SIZE(xold))
REAL(8), INTENT(OUT) :: f
LOGICAL, INTENT(OUT) :: check
INTERFACE	! This lets the subroutine know that func is not a variable, it is a function
	REAL(8) FUNCTION func(theta,k,gradflag)  ! This is not the likelihood itself, it just let the program know
    IMPLICIT NONE		
    INTEGER, INTENT(IN) :: k
    INTEGER, INTENT(IN) :: gradflag
    REAL(8), INTENT(IN) :: theta(k)
    END FUNCTION func
END INTERFACE
REAL(8), PARAMETER :: ALF=1.0e-4,TOLX=epsilon(x)
INTEGER :: ndum
REAL(8) :: a,alam,alam2,alamin,b,disc,f2,pabs,rhs1,rhs2,slope,tmplam,zero
	print *, 'Lnsrch In!'
zero=0.0D0
ndum=SIZE(xold)
IF (SIZE(g).NE.size(xold)) THEN
	PAUSE '***FATAL ERROR: Dimensions do not agree (LNSRCH)***'
    STOP
END IF
check=.false.
pabs=vabs(p(:))
!	print *, 'g: ',g
!	print *, 'p: ',p
!	print *, 'pabs: ',pabs
DO WHILE (pabs > HUGE(pabs))
	p(:)=p(:)/10
	pabs=vabs(p(:))
END DO
!	print *, 'p: ',p
IF (pabs > stpmax) p(:)=p(:)*stpmax/pabs ! Scale if attempted step is too big.
!	print *, 'p: ',p
slope=DOT_PRODUCT(g,p)
IF (slope >= zero) THEN
	print *, 'g: ',g
	print *, 'p: ',p
	print *, 'slope: ',slope
	PAUSE '***FATAL ERROR: roundoff problem in lnsrch***'
    STOP
END IF
alamin=TOLX/MAXVAL(ABS(p(:))/MAX(ABS(xold(:)),1.0D0)) !Compute .min.
alam=1.0D0 ! Always try full Newton step first.
print *,'check3-3! - Start lnsearch iteration loop'
DO ! Start of iteration loop.
	x(:)=xold(:)+alam*p(:)
    F=FUNC(x,size(x),0)
print *,'check3-4!'
    IF (alam < alamin) THEN                 ! Convergence on .x. For zero finding,
print *,'check3-4-a!'
		!the calling program should verify the convergence.
		x(:)=xold(:)
		check=.true.
		RETURN
    ELSE IF (f <= fold+ALF*alam*slope) THEN ! Sufficient function decrease.
print *,'check3-4-b!'
		RETURN
	ELSE !Backtrack.
		IF (alam == 1.0D0) THEN !First time.
print *,'check3-4-c!'
			tmplam=-slope/(2.0D0*(f-fold-slope))
        ELSE !Subsequent backtracks.
print *,'check3-4-d!'
			rhs1=f-fold-alam*slope
            rhs2=f2-fold-alam2*slope
            a=(rhs1/alam**2-rhs2/alam2**2)/(alam-alam2)
            b=(-alam2*rhs1/alam**2+alam*rhs2/alam2**2)/&
              (alam-alam2)
            IF (a == 0.0D0) THEN
				tmplam=-slope/(2.0D0*b)
            ELSE
				disc=b*b-3.0*a*slope
                IF (disc < 0.0D0) THEN
					tmplam=0.5D0*alam
                ELSE IF (b <= 0.0D0) THEN
					tmplam=(-b+SQRT(disc))/(3.0D0*a)
                ELSE
					tmplam=-slope/(b+SQRT(disc))
                END IF
             END IF
             IF (tmplam > 0.5D0*alam) tmplam=0.5D0*alam
		END IF
	END IF
    alam2=alam
    f2=f
    alam=MAX(tmplam,0.1D0*alam) 
END DO !Try again.
END SUBROUTINE lnsrch

FUNCTION vabs(v)
! Return the length (ordinary L2 norm) of a vector.
REAL(8), DIMENSION(:), INTENT(IN) :: v
REAL(8) :: vabs
vabs=sqrt(dot_product(v,v))
END FUNCTION vabs

END MODULE MINIMIZATION
