MODULE Globvar
IMPLICIT NONE
! This module declares the variables that are going to be used throughout the program.
! To avoid stack overflow error --> Stack Reserve: 50,000,000

INTEGER, PARAMETER :: nchildmax=4  ! max number of children
INTEGER, PARAMETER :: sm2=2**(1+1)  ! necessary amount for strategy patterns
INTEGER, PARAMETER :: sm3=2**(1+1)*2**(2+1)  ! necessary amount for strategy patterns
INTEGER, PARAMETER :: sm4=2**(1+1)*2**(2+1)*2**(3+1)  ! necessary amount for strategy patterns
	! The 4th child in a 4-child family has 2**(3+1) strategies, because 
	! the number of possible elder siblings in the upstream from 0 to 3, i.e. 4.
INTEGER :: nobs
INTEGER :: nfam
INTEGER :: ns              ! number of simulation draws
REAL(8) :: objsaved
INTEGER :: numeval
INTEGER :: ntheta
INTEGER :: ntheta1
INTEGER, ALLOCATABLE :: famid(:)   ! famid(nobs)
INTEGER, ALLOCATABLE :: nchild(:), famindex(:)   ! nchild(nfam), famindex(nfam)
REAL(8), ALLOCATABLE :: theta(:) ! theta(ntheta)
REAL(8), ALLOCATABLE :: x(:,:)   ! x(ntheta1,nobs)
INTEGER, ALLOCATABLE :: y(:)     ! y(nobs)
INTEGER, ALLOCATABLE :: near(:)   ! near(nfam)
REAL(8), ALLOCATABLE :: utilsub(:) ! utilsub(nobs)
INTEGER, ALLOCATABLE :: predobs(:)  !nearpred(nobs)

REAL(8), ALLOCATABLE :: draws2(:,:)    ! nchild=2. draws(ns,necessary amount)
REAL(8), ALLOCATABLE :: draws3(:,:,:)    ! nchild=3. draws(ns,necessary amount,num player-1)
REAL(8), ALLOCATABLE :: draws4(:,:,:)    ! nchild=3. draws(ns,necessary amount,num player-1)

CHARACTER(10) :: path                          ! Specify the data directory. To be read from setting.prn

END MODULE Globvar






MODULE INIT
	USE Globvar
	IMPLICIT NONE

CONTAINS

	SUBROUTINE READDATA(nobs)
		USE Globvar
		IMPLICIT NONE
		INTEGER, INTENT(IN) :: nobs
		INTEGER :: i,j,k,m,io

        CHARACTER(60) :: readfile_x      ! File names to be read

readfile_x='data_short00.raw'

x=0.0d0; y=0.0d0
x(1,:)=1.000000d0

print *,'Reading Data ... '

	OPEN(1,file=(readfile_x),status='old')
    DO i = 1, nobs
	READ(1,fmt=*,iostat=io) y(i), famid(i), x(2:,i)
    IF (io<0) EXIT
    END DO
    CLOSE(1)

nchild=0
j=1
nchild(1)=1
DO i=2,nobs
	IF (famid(i)/=famid(i-1)) THEN
		j=j+1
	ENDIF
		nchild(j)=nchild(j)+1
ENDDO

famindex(1)=nchild(1)
DO i=2,nfam
	famindex(i)=famindex(i-1)+nchild(i)
ENDDO


! Convert "fam outcome" to binary number
DO m=1,nfam
near(m)=0
DO i=1,nchild(m)
near(m)=near(m)*2
near(m)=near(m)+y(famindex(m)-nchild(m)+i)
END DO
END DO


print *,'Reading Data ... Finished '

	END SUBROUTINE READDATA


END MODULE INIT

MODULE Object
	USE Globvar
	USE MATRIX
    USE PROBABILITY
	IMPLICIT NONE

CONTAINS


FUNCTION lnmlobj(param,n,gradflag)	! This function computes the ML objective function
USE Globvar
IMPLICIT NONE
INTEGER, INTENT(IN) :: n 
INTEGER, INTENT(IN) :: gradflag
REAL(8), INTENT(IN) :: param(n)
REAL(8) :: mlobj
REAL(8) :: lnmlobj
REAL(8) :: famprob(nfam)			! The family-level likelihood
INTEGER :: i, j, k, m
REAL(4) :: tms

tms = SECNDS(0.0)

numeval=numeval+1

print ('(a16,I8)'),'mle evaluation: ',numeval


utilsub(1:nobs)=MATMUL(param(1:ntheta1),x(:,1:nobs))

!$OMP PARALLEL DO SCHEDULE(STATIC)
DO m = 1, nfam
!	IF (numeval==95) THEN
!    WRITE(4,fmt=*) m,param,n
!    ENDIF
	famprob(m)=DLOG(GHK(m,param,n))
!	IF (numeval==95) THEN
!    WRITE(4,fmt=*) m,nchild(m),famprob(m)
!    ENDIF
END DO
!$OMP END PARALLEL DO
!STOP

lnmlobj=(0-1)*SUM(famprob)
!WRITE(4,fmt=*) numeval,lnmlobj
!print *,'Chcek2: lnmlobj:',lnmlobj



print *,'ln ML objective: ',-lnmlobj
!IF (gradflag==0) objsaved=mlobj

print *,'Time spent for Likelihood evaluation: ',SECNDS(tms)


END FUNCTION lnmlobj


FUNCTION GHK(m,param,n) RESULT(ghkval)
	! This function calculates the GHK simulator for each market*year
USE Globvar
USE PROBABILITY
IMPLICIT NONE
INTEGER, INTENT(IN) :: m			! market id
INTEGER, INTENT(IN) :: n 
REAL(8), INTENT(IN) :: param(n)
INTEGER :: i, j, k, l, ienta, ientb
INTEGER :: s1, s2, s3, s4, s5
! Specify the array sizes based on the number of players (i.e. possible number of strategy combinations)
REAL(8) :: wg(ns,0:sm4-1)
REAL(8) :: ta(ns), tb(ns)
REAL(8) :: v(ns,0:sm4-1,nchild(m))   ! GHK simulation draws constructed from draws()
REAL(8) :: utiltemp
REAL(8) :: ghkval		! The calculated GHK value
REAL(8) :: lchol(nchild(m),nchild(m))


wg=0.0d0
v=0.0d0
ghkval=0.0d0

k=famindex(m)

! PRINT *, 'GHK-in'

SELECT CASE(nchild(m))

CASE(1)

	IF (IBITS(near(m),0,1)==1) THEN
    ghkval=1.0d0-NORMCDF(-utilsub(k),0.0d0,1.0d0)
	ELSE
    ghkval=NORMCDF(-utilsub(k),0.0d0,1.0d0)
    ENDIF



CASE(2)

lchol=lcholbase(m,param,n)

DO s1=0,2**2-1       ! Strategy loop
IF (ghkcheck(s1,2,m)==1) THEN

    IF (s1==3) THEN
        
        ta=NORMCDF((-utilsub(k)),0.0d0,1.0d0)

    wg(:,s1)=1.0-ta

        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,s1,1)=HUGE(v(j,s1,1))
            ELSE
            	v(j,s1,1)=INVNORMCDF((ta(j)*(1-draws2(j,s1))+draws2(j,s1)),0.0d0,1.0d0)
	        ENDIF
        END DO

    ELSE IF (s1==0) THEN

        tb=NORMCDF(-utilsub(k),0.0d0,1.0d0)
        wg(:,s1)=tb

        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,s1,1)=-HUGE(v(j,s1,1))
   	        ELSE
   	        v(j,s1,1)=INVNORMCDF((tb(j)*draws2(j,s1)),0.0d0,1.0d0)
            ENDIF
        END DO


    ELSE IF (s1==1) THEN       ! s1==1
        ta=NORMCDF((-utilsub(k)),0.0d0,1.0d0)
        tb=NORMCDF((-utilsub(k)),0.0d0,1.0d0)
        IF (tb(1)-ta(1) > 0.0d0) THEN
        wg(:,s1)=tb-ta

        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,s1,1)=-HUGE(v(j,s1,1))
   	        ELSE
          	v(j,s1,1)=INVNORMCDF((ta(j)*(1-draws2(j,s1))+tb(j)*draws2(j,s1)),0.0d0,1.0d0)
            ENDIF
        END DO

        ENDIF

    ELSE              ! s1==2
        ta=NORMCDF((-utilsub(k)),0.0d0,1.0d0)
        tb=NORMCDF((-utilsub(k)),0.0d0,1.0d0)
        IF (tb(1)-ta(1) > 0.0d0) THEN
        wg(:,s1)=tb-ta

        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,s1,1)=-HUGE(v(j,s1,1))
   	        ELSE
          	v(j,s1,1)=INVNORMCDF((ta(j)*(1-draws2(j,s1))+tb(j)*draws2(j,s1)),0.0d0,1.0d0)
            ENDIF
        END DO
        
        ENDIF

    ENDIF

ENDIF
END DO         ! Strategy loop



! player 1

DO s1=0,2**2-1       ! Strategy loop
IF (ghkcheck(s1,2,m)==1 .and. wg(1,s1)>0.0d0) THEN

	IF (IBITS(near(m),1,1)==1) THEN
    utiltemp=utilsub(k-1)
	DO j=1, ns
    ta(j)=NORMCDF((-utiltemp-DOT_PRODUCT(lchol(nchild(m),1:nchild(m)-1),v(j,s1,1:nchild(m)-1)))/lchol(nchild(m),nchild(m)),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:,s1)=wg(:,s1)*(1.0-ta)
	ELSE
    utiltemp=utilsub(k-1)
	DO j=1, ns
    tb(j)=NORMCDF((-utiltemp-DOT_PRODUCT(lchol(nchild(m),1:nchild(m)-1),v(j,s1,1:nchild(m)-1)))/lchol(nchild(m),nchild(m)),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:,s1)=wg(:,s1)*tb
    ENDIF
ENDIF
END DO         ! Strategy loop

DO s1=0,2**2-1       ! Strategy loop
ghkval=ghkval+SUM(wg(:,s1))
END DO         ! Strategy loop
ghkval=ghkval/ns



CASE(3)

lchol=lcholbase(m,param,n)

DO s2=0,2**3-1       ! Strategy loop
IF (ghkcheck(s2,3,m)==1) THEN

    IF (s2==7) THEN
        
        ta=NORMCDF((-utilsub(k)),0.0d0,1.0d0)

        DO s1=0,2**2-1
        IF (ghkcheck(s1,2,m)==1) THEN
        wg(:,s1*8+s2)=1.0-ta

        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,s1*8+s2,1)=HUGE(v(j,s1*8+s2,1))
            ELSE
            	v(j,s1*8+s2,1)=INVNORMCDF((ta(j)*(1-draws3(j,s1*8+s2,1))+draws3(j,s1*8+s2,1)),0.0d0,1.0d0)
	        ENDIF
        END DO
        ENDIF
        END DO

    ELSE IF (s2==0) THEN

        tb=NORMCDF(-utilsub(k),0.0d0,1.0d0)
        DO s1=0,2**2-1
        IF (ghkcheck(s1,2,m)==1) THEN
        wg(:,s1*8+s2)=tb

        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,s1*8+s2,1)=-HUGE(v(j,s1*8+s2,1))
   	        ELSE
   	            v(j,s1*8+s2,1)=INVNORMCDF((tb(j)*draws3(j,s1*8+s2,1)),0.0d0,1.0d0)
            ENDIF
        END DO
        ENDIF
        END DO

    ELSE IF (s2==1) THEN       ! s2==1
        ta=NORMCDF((-utilsub(k)),0.0d0,1.0d0)
        tb=NORMCDF((-utilsub(k)),0.0d0,1.0d0)
        IF (tb(1)-ta(1) > 0.0d0) THEN
        DO s1=0,2**2-1
        IF (ghkcheck(s1,2,m)==1) THEN
        wg(:,s1*8+s2)=tb-ta

        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,s1*8+s2,1)=-HUGE(v(j,s1*8+s2,1))
   	        ELSE
          	v(j,s1*8+s2,1)=INVNORMCDF((ta(j)*(1-draws3(j,s1*8+s2,1))+tb(j)*draws3(j,s1*8+s2,1)),0.0d0,1.0d0)
            ENDIF
        END DO
        ENDIF
        ENDDO
        ENDIF

    ELSE IF (s2==6) THEN       ! s2==1
        ta=NORMCDF((-utilsub(k)),0.0d0,1.0d0)
        tb=NORMCDF((-utilsub(k)),0.0d0,1.0d0)
        IF (tb(1)-ta(1) > 0.0d0) THEN
        DO s1=0,2**2-1
        IF (ghkcheck(s1,2,m)==1) THEN
        wg(:,s1*8+s2)=tb-ta

        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,s1*8+s2,1)=-HUGE(v(j,s1*8+s2,1))
   	        ELSE
          	v(j,s1*8+s2,1)=INVNORMCDF((ta(j)*(1-draws3(j,s1*8+s2,1))+tb(j)*draws3(j,s1*8+s2,1)),0.0d0,1.0d0)
            ENDIF
        END DO
        ENDIF
        ENDDO
        ENDIF

    ENDIF

ENDIF
END DO         ! Strategy loop


! The second to the last prayer
DO s1=0,2**2-1       ! Strategy loop
IF (ghkcheck(s1,2,m)==1) THEN
DO s2=0,2**3-1
IF (ghkcheck(s2,3,m)==1) THEN

    IF (s1==0) THEN
        utiltemp=-utilsub(k-1)
    	DO j=1, ns
        tb(j)=NORMCDF((utiltemp-DOT_PRODUCT(lchol(2,1:1),v(j,s1*8+s2,1:1)))/lchol(2,2),0.0d0,1.0d0)
    	END DO         ! Simulation loop
        wg(:,s1*8+s2)=wg(:,s1*8+s2)*tb

        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,s1*8+s2,2)=-HUGE(v(j,s1*8+s2,2))
   	        ELSE
!            PRINT *,'k,s1,s2',k,s1,s2
   	        v(j,s1*8+s2,2)=INVNORMCDF((tb(j)*draws3(j,s1*8+s2,2)),0.0d0,1.0d0)
            ENDIF
        END DO
    
    ELSE IF (s1==3) THEN
        utiltemp=-utilsub(k-1)
    	DO j=1, ns
        ta(j)=NORMCDF((utiltemp-DOT_PRODUCT(lchol(2,1:1),v(j,s1*8+s2,1:1)))/lchol(2,2),0.0d0,1.0d0)
    	END DO         ! Simulation loop
        wg(:,s1*8+s2)=wg(:,s1*8+s2)*(1.0d0-ta)

        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,s1*8+s2,2)=HUGE(v(j,s1*8+s2,2))
            ELSE
            	v(j,s1*8+s2,2)=INVNORMCDF((ta(j)*(1-draws3(j,s1*8+s2,2))+draws3(j,s1*8+s2,2)),0.0d0,1.0d0)
	        ENDIF
        END DO


    ELSE IF (s1==1) THEN       ! s1==1
        utiltemp=-utilsub(k-1)
    	DO j=1, ns
        ta(j)=NORMCDF((utiltemp-DOT_PRODUCT(lchol(2,1:1),v(j,s1*8+s2,1:1)))/lchol(2,2),0.0d0,1.0d0)
    	END DO         ! Simulation loop

        utiltemp=-utilsub(k-1)
    	DO j=1, ns
        tb(j)=NORMCDF((utiltemp-DOT_PRODUCT(lchol(2,1:1),v(j,s1*8+s2,1:1)))/lchol(2,2),0.0d0,1.0d0)
    	END DO         ! Simulation loop


        DO j=1, ns
            IF (tb(j)-ta(j) > 0.0d0) THEN
            wg(j,s1*8+s2)=wg(j,s1*8+s2)*(tb(j)-ta(j))
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,s1*8+s2,2)=HUGE(v(j,s1*8+s2,2))
            ELSE IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,s1*8+s2,2)=-HUGE(v(j,s1*8+s2,2))
   	        ELSE
          	    v(j,s1*8+s2,2)=INVNORMCDF((ta(j)*(1-draws3(j,s1*8+s2,2))+tb(j)*draws3(j,s1*8+s2,2)),0.0d0,1.0d0)
            ENDIF
            ELSE
                wg(j,s1*8+s2)=0
            ENDIF
        END DO

    ELSE IF (s1==2) THEN       ! s1==2
        utiltemp=-utilsub(k-1)

    	DO j=1, ns
        ta(j)=NORMCDF((utiltemp-DOT_PRODUCT(lchol(2,1:1),v(j,s1*8+s2,1:1)))/lchol(2,2),0.0d0,1.0d0)
    	END DO         ! Simulation loop

        utiltemp=-utilsub(k-1)
    	DO j=1, ns
        tb(j)=NORMCDF((utiltemp-DOT_PRODUCT(lchol(2,1:1),v(j,s1*8+s2,1:1)))/lchol(2,2),0.0d0,1.0d0)
    	END DO         ! Simulation loop


        DO j=1, ns
            IF (tb(j)-ta(j) > 0.0d0) THEN
            wg(j,s1*8+s2)=wg(j,s1*8+s2)*(tb(j)-ta(j))
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,s1*8+s2,2)=HUGE(v(j,s1*8+s2,2))
            ELSE IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,s1*8+s2,2)=-HUGE(v(j,s1*8+s2,2))
   	        ELSE
          	    v(j,s1*8+s2,2)=INVNORMCDF((ta(j)*(1-draws3(j,s1*8+s2,2))+tb(j)*draws3(j,s1*8+s2,2)),0.0d0,1.0d0)
            ENDIF
            ELSE
                wg(j,s1*8+s2)=0
            ENDIF
        END DO

    ELSE

    	wg(:,s1*8+s2)=0.0d0
    	v(:,s1*8+s2,2)=0.0d0
        
    ENDIF


ELSE
    wg(:,s1*8+s2)=0.0d0
ENDIF
END DO         ! Strategy loop
ELSE
    DO s2=0,2**3-1
    wg(:,s1*8+s2)=0.0d0
    END DO
ENDIF
END DO         ! Strategy loop




! player 1

DO s1=0,2**2-1       ! Strategy loop
IF (ghkcheck(s1,2,m)==1) THEN
DO s2=0,2**3-1       ! Strategy loop
IF (ghkcheck(s2,3,m)==1 .and. wg(1,s1*8+s2)>0.0d0) THEN

	IF (IBITS(near(m),2,1)==1) THEN
    utiltemp=utilsub(k-2)
	DO j=1, ns
    ta(j)=NORMCDF((-utiltemp-DOT_PRODUCT(lchol(nchild(m),1:nchild(m)-1),v(j,s1*8+s2,1:nchild(m)-1)))/lchol(nchild(m),nchild(m)),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:,s1*8+s2)=wg(:,s1*8+s2)*(1.0-ta)
	ELSE
    utiltemp=utilsub(k-2)
	DO j=1, ns
    tb(j)=NORMCDF((-utiltemp-DOT_PRODUCT(lchol(nchild(m),1:nchild(m)-1),v(j,s1*8+s2,1:nchild(m)-1)))/lchol(nchild(m),nchild(m)),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:,s1*8+s2)=wg(:,s1*8+s2)*tb
    ENDIF

ENDIF
END DO         ! Strategy loop
ENDIF
END DO         ! Strategy loop


DO s1=0,2**2-1       ! Strategy loop
DO s2=0,2**3-1
ghkval=ghkval+SUM(wg(:,s1*8+s2))
END DO
END DO         ! Strategy loop
ghkval=ghkval/ns






CASE(4)

lchol=lcholbase(m,param,n)

DO s3=0,2**4-1       ! Strategy loop
IF (ghkcheck(s3,4,m)==1) THEN

    IF (s3==15) THEN
        
        ta=NORMCDF((-utilsub(k)),0.0d0,1.0d0)

        DO s1=0,2**2-1
        IF (ghkcheck(s1,2,m)==1) THEN
        DO s2=0,2**3-1
        IF (ghkcheck(s2,3,m)==1) THEN
        wg(:,s1*128+s2*16+s3)=1.0-ta

        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,s1*128+s2*16+s3,1)=HUGE(v(j,s1*128+s2*16+s3,1))
            ELSE
            	v(j,s1*128+s2*16+s3,1)=INVNORMCDF((ta(j)*(1-draws4(j,s1*128+s2*16+s3,1))+draws4(j,s1*128+s2*16+s3,1)),0.0d0,1.0d0)
	        ENDIF
        END DO
        ENDIF
        END DO
        ENDIF
        END DO

    ELSE IF (s3==0) THEN

        tb=NORMCDF(-utilsub(k),0.0d0,1.0d0)
        DO s1=0,2**2-1
        IF (ghkcheck(s1,2,m)==1) THEN
        DO s2=0,2**3-1
        IF (ghkcheck(s2,3,m)==1) THEN
        wg(:,s1*128+s2*16+s3)=tb

        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,s1*128+s2*16+s3,1)=-HUGE(v(j,s1*128+s2*16+s3,1))
   	        ELSE
   	            v(j,s1*128+s2*16+s3,1)=INVNORMCDF((tb(j)*draws4(j,s1*128+s2*16+s3,1)),0.0d0,1.0d0)
            ENDIF
        END DO
        ENDIF
        END DO
        ENDIF
        END DO

    ELSE IF (s3==1) THEN       ! s3==1
        ta=NORMCDF((-utilsub(k)),0.0d0,1.0d0)
        tb=NORMCDF((-utilsub(k)),0.0d0,1.0d0)
        IF (tb(1)-ta(1) > 0.0d0) THEN
        DO s1=0,2**2-1
        IF (ghkcheck(s1,2,m)==1) THEN
        DO s2=0,2**3-1
        IF (ghkcheck(s2,3,m)==1) THEN
        wg(:,s1*128+s2*16+s3)=tb-ta

        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,s1*128+s2*16+s3,1)=-HUGE(v(j,s1*128+s2*16+s3,1))
   	        ELSE
          	v(j,s1*128+s2*16+s3,1)=INVNORMCDF((ta(j)*(1-draws4(j,s1*128+s2*16+s3,1))+tb(j)*draws4(j,s1*128+s2*16+s3,1)),0.0d0,1.0d0)
            ENDIF
        END DO
        ENDIF
        ENDDO
        ENDIF
        ENDDO
        ENDIF

    ELSE IF (s3==14) THEN       ! s3==14
        ta=NORMCDF((-utilsub(k)),0.0d0,1.0d0)
        tb=NORMCDF((-utilsub(k)),0.0d0,1.0d0)
        IF (tb(1)-ta(1) > 0.0d0) THEN
        DO s1=0,2**2-1
        IF (ghkcheck(s1,2,m)==1) THEN
        DO s2=0,2**3-1
        IF (ghkcheck(s2,3,m)==1) THEN
        wg(:,s1*128+s2*16+s3)=tb-ta

        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,s1*128+s2*16+s3,1)=-HUGE(v(j,s1*128+s2*16+s3,1))
   	        ELSE
          	v(j,s1*128+s2*16+s3,1)=INVNORMCDF((ta(j)*(1-draws4(j,s1*128+s2*16+s3,1))+tb(j)*draws4(j,s1*128+s2*16+s3,1)),0.0d0,1.0d0)
            ENDIF
        END DO
        ENDIF
        ENDDO
        ENDIF
        ENDDO
        ENDIF

    ENDIF

ENDIF
END DO         ! Strategy loop


DO s1=0,2**2-1       ! Strategy loop
IF (ghkcheck(s1,2,m)==1) THEN
DO s2=0,2**3-1
IF (ghkcheck(s2,3,m)==1) THEN
DO s3=0,2**4-1
IF (ghkcheck(s3,4,m)==1) THEN

    IF (s2==0) THEN
        utiltemp=-utilsub(k-1)
    	DO j=1, ns
        tb(j)=NORMCDF((utiltemp-DOT_PRODUCT(lchol(2,1:1),v(j,s1*128+s2*16+s3,1:1)))/lchol(2,2),0.0d0,1.0d0)
    	END DO         ! Simulation loop
        wg(:,s1*128+s2*16+s3)=wg(:,s1*128+s2*16+s3)*tb

        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,s1*128+s2*16+s3,2)=-HUGE(v(j,s1*128+s2*16+s3,2))
   	        ELSE
!            PRINT *,'k,s1,s2',k,s1,s2
   	        v(j,s1*128+s2*16+s3,2)=INVNORMCDF((tb(j)*draws4(j,s1*128+s2*16+s3,2)),0.0d0,1.0d0)
            ENDIF
        END DO
    
    ELSE IF (s2==7) THEN
        utiltemp=-utilsub(k-1)
    	DO j=1, ns
        ta(j)=NORMCDF((utiltemp-DOT_PRODUCT(lchol(2,1:1),v(j,s1*128+s2*16+s3,1:1)))/lchol(2,2),0.0d0,1.0d0)
    	END DO         ! Simulation loop
        wg(:,s1*128+s2*16+s3)=wg(:,s1*128+s2*16+s3)*(1.0d0-ta)

        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,s1*128+s2*16+s3,2)=HUGE(v(j,s1*128+s2*16+s3,2))
            ELSE
            	v(j,s1*128+s2*16+s3,2)=INVNORMCDF((ta(j)*(1-draws4(j,s1*128+s2*16+s3,2))+draws4(j,s1*128+s2*16+s3,2)),0.0d0,1.0d0)
	        ENDIF
        END DO


    ELSE IF (s2==1) THEN       ! s2==1
        utiltemp=-utilsub(k-1)

    	DO j=1, ns
        ta(j)=NORMCDF((utiltemp-DOT_PRODUCT(lchol(2,1:1),v(j,s1*128+s2*16+s3,1:1)))/lchol(2,2),0.0d0,1.0d0)
    	END DO         ! Simulation loop

        utiltemp=-utilsub(k-1)
    	DO j=1, ns
        tb(j)=NORMCDF((utiltemp-DOT_PRODUCT(lchol(2,1:1),v(j,s1*128+s2*16+s3,1:1)))/lchol(2,2),0.0d0,1.0d0)
    	END DO         ! Simulation loop


        DO j=1, ns
            IF (tb(j)-ta(j) > 0.0d0) THEN
            wg(j,s1*128+s2*16+s3)=wg(j,s1*128+s2*16+s3)*(tb(j)-ta(j))
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,s1*128+s2*16+s3,2)=HUGE(v(j,s1*128+s2*16+s3,2))
            ELSE IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,s1*128+s2*16+s3,2)=-HUGE(v(j,s1*128+s2*16+s3,2))
   	        ELSE
          	    v(j,s1*128+s2*16+s3,2)=INVNORMCDF((ta(j)*(1-draws4(j,s1*128+s2*16+s3,2))+tb(j)*draws4(j,s1*128+s2*16+s3,2)),0.0d0,1.0d0)
            ENDIF
            ELSE
                wg(j,s1*128+s2*16+s3)=0
            ENDIF
        END DO

    ELSE IF (s2==6) THEN       ! s2==6
        utiltemp=-utilsub(k-1)

    	DO j=1, ns
        ta(j)=NORMCDF((utiltemp-DOT_PRODUCT(lchol(2,1:1),v(j,s1*128+s2*16+s3,1:1)))/lchol(2,2),0.0d0,1.0d0)
    	END DO         ! Simulation loop

        utiltemp=-utilsub(k-1)
    	DO j=1, ns
        tb(j)=NORMCDF((utiltemp-DOT_PRODUCT(lchol(2,1:1),v(j,s1*128+s2*16+s3,1:1)))/lchol(2,2),0.0d0,1.0d0)
    	END DO         ! Simulation loop


        DO j=1, ns
            IF (tb(j)-ta(j) > 0.0d0) THEN
            wg(j,s1*128+s2*16+s3)=wg(j,s1*128+s2*16+s3)*(tb(j)-ta(j))
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,s1*128+s2*16+s3,2)=HUGE(v(j,s1*128+s2*16+s3,2))
            ELSE IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,s1*128+s2*16+s3,2)=-HUGE(v(j,s1*128+s2*16+s3,2))
   	        ELSE
          	    v(j,s1*128+s2*16+s3,2)=INVNORMCDF((ta(j)*(1-draws4(j,s1*128+s2*16+s3,2))+tb(j)*draws4(j,s1*128+s2*16+s3,2)),0.0d0,1.0d0)
            ENDIF
            ELSE
                wg(j,s1*128+s2*16+s3)=0
            ENDIF
        END DO

    ELSE

    	wg(:,s1*128+s2*16+s3)=0.0d0
    	v(:,s1*128+s2*16+s3,2)=0.0d0
        
    ENDIF


ELSE
    wg(:,s1*128+s2*16+s3)=0.0d0
ENDIF
END DO         ! Strategy loop
ELSE
    DO s3=0,2**4-1
    wg(:,s1*128+s2*16+s3)=0.0d0
    END DO
ENDIF
END DO         ! Strategy loop
ELSE
    DO s2=0,2**3-1
    DO s3=0,2**4-1
    wg(:,s1*128+s2*16+s3)=0.0d0
    END DO
    END DO
ENDIF
END DO         ! Strategy loop



DO s1=0,2**2-1       ! Strategy loop
IF (ghkcheck(s1,2,m)==1) THEN
DO s2=0,2**3-1
IF (ghkcheck(s2,3,m)==1) THEN
DO s3=0,2**4-1
IF (ghkcheck(s3,4,m)==1) THEN

    IF (s1==0) THEN
        utiltemp=-utilsub(k-2)
    	DO j=1, ns
        tb(j)=NORMCDF((utiltemp-DOT_PRODUCT(lchol(3,1:2),v(j,s1*128+s2*16+s3,1:2)))/lchol(3,3),0.0d0,1.0d0)
    	END DO         ! Simulation loop
        wg(:,s1*128+s2*16+s3)=wg(:,s1*128+s2*16+s3)*tb

        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,s1*128+s2*16+s3,3)=-HUGE(v(j,s1*128+s2*16+s3,3))
   	        ELSE
!            PRINT *,'k,s1,s2',k,s1,s2
   	        v(j,s1*128+s2*16+s3,3)=INVNORMCDF((tb(j)*draws4(j,s1*128+s2*16+s3,3)),0.0d0,1.0d0)
            ENDIF
        END DO
    
    ELSE IF (s1==3) THEN
        utiltemp=-utilsub(k-2)
    	DO j=1, ns
        ta(j)=NORMCDF((utiltemp-DOT_PRODUCT(lchol(3,1:2),v(j,s1*128+s2*16+s3,1:2)))/lchol(3,3),0.0d0,1.0d0)
    	END DO         ! Simulation loop
        wg(:,s1*128+s2*16+s3)=wg(:,s1*128+s2*16+s3)*(1.0d0-ta)

        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,s1*128+s2*16+s3,3)=HUGE(v(j,s1*128+s2*16+s3,3))
            ELSE
            	v(j,s1*128+s2*16+s3,3)=INVNORMCDF((ta(j)*(1-draws4(j,s1*128+s2*16+s3,3))+draws4(j,s1*128+s2*16+s3,3)),0.0d0,1.0d0)
	        ENDIF
        END DO


    ELSE IF (s1==1) THEN       ! s1==1
        utiltemp=-utilsub(k-2)

    	DO j=1, ns
        ta(j)=NORMCDF((utiltemp-DOT_PRODUCT(lchol(3,1:2),v(j,s1*128+s2*16+s3,1:2)))/lchol(3,3),0.0d0,1.0d0)
    	END DO         ! Simulation loop

        utiltemp=-utilsub(k-2)
    	DO j=1, ns
        tb(j)=NORMCDF((utiltemp-DOT_PRODUCT(lchol(3,1:2),v(j,s1*128+s2*16+s3,1:2)))/lchol(3,3),0.0d0,1.0d0)
    	END DO         ! Simulation loop


        DO j=1, ns
            IF (tb(j)-ta(j) > 0.0d0) THEN
            wg(j,s1*128+s2*16+s3)=wg(j,s1*128+s2*16+s3)*(tb(j)-ta(j))
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,s1*128+s2*16+s3,3)=HUGE(v(j,s1*128+s2*16+s3,3))
            ELSE IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,s1*128+s2*16+s3,3)=-HUGE(v(j,s1*128+s2*16+s3,3))
   	        ELSE
          	    v(j,s1*128+s2*16+s3,3)=INVNORMCDF((ta(j)*(1-draws4(j,s1*128+s2*16+s3,3))+tb(j)*draws4(j,s1*128+s2*16+s3,3)),0.0d0,1.0d0)
            ENDIF
            ELSE
                wg(j,s1*128+s2*16+s3)=0
            ENDIF
        END DO

    ELSE IF (s1==2) THEN       ! s1==2
        utiltemp=-utilsub(k-2)

    	DO j=1, ns
        ta(j)=NORMCDF((utiltemp-DOT_PRODUCT(lchol(3,1:2),v(j,s1*128+s2*16+s3,1:2)))/lchol(3,3),0.0d0,1.0d0)
    	END DO         ! Simulation loop

        utiltemp=-utilsub(k-2)
    	DO j=1, ns
        tb(j)=NORMCDF((utiltemp-DOT_PRODUCT(lchol(3,1:2),v(j,s1*128+s2*16+s3,1:2)))/lchol(3,3),0.0d0,1.0d0)
    	END DO         ! Simulation loop


        DO j=1, ns
            IF (tb(j)-ta(j) > 0.0d0) THEN
            wg(j,s1*128+s2*16+s3)=wg(j,s1*128+s2*16+s3)*(tb(j)-ta(j))
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,s1*128+s2*16+s3,3)=HUGE(v(j,s1*128+s2*16+s3,3))
            ELSE IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,s1*128+s2*16+s3,3)=-HUGE(v(j,s1*128+s2*16+s3,3))
   	        ELSE
          	    v(j,s1*128+s2*16+s3,3)=INVNORMCDF((ta(j)*(1-draws4(j,s1*128+s2*16+s3,3))+tb(j)*draws4(j,s1*128+s2*16+s3,3)),0.0d0,1.0d0)
            ENDIF
            ELSE
                wg(j,s1*128+s2*16+s3)=0
            ENDIF
        END DO

    ELSE

    	wg(:,s1*128+s2*16+s3)=0.0d0
    	v(:,s1*128+s2*16+s3,3)=0.0d0
        
    ENDIF


ELSE
    wg(:,s1*128+s2*16+s3)=0.0d0
ENDIF
END DO         ! Strategy loop
ELSE
    DO s3=0,2**4-1
    wg(:,s1*128+s2*16+s3)=0.0d0
    END DO
ENDIF
END DO         ! Strategy loop
ELSE
    DO s2=0,2**3-1
    DO s3=0,2**4-1
    wg(:,s1*128+s2*16+s3)=0.0d0
    END DO
    END DO
ENDIF
END DO         ! Strategy loop




! player 1

DO s1=0,2**2-1       ! Strategy loop
IF (ghkcheck(s1,2,m)==1) THEN
DO s2=0,2**3-1       ! Strategy loop
IF (ghkcheck(s2,3,m)==1) THEN
DO s3=0,2**4-1       ! Strategy loop
IF (ghkcheck(s3,4,m)==1 .and. wg(1,s1*128+s2*16+s3)>0.0d0) THEN

	IF (IBITS(near(m),3,1)==1) THEN
    utiltemp=utilsub(k-3)
	DO j=1, ns
    ta(j)=NORMCDF((-utiltemp-DOT_PRODUCT(lchol(nchild(m),1:nchild(m)-1),v(j,s1*128+s2*16+s3,1:nchild(m)-1)))/lchol(nchild(m),nchild(m)),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:,s1*128+s2*16+s3)=wg(:,s1*128+s2*16+s3)*(1.0-ta)
	ELSE
    utiltemp=utilsub(k-3)
	DO j=1, ns
    tb(j)=NORMCDF((-utiltemp-DOT_PRODUCT(lchol(nchild(m),1:nchild(m)-1),v(j,s1*128+s2*16+s3,1:nchild(m)-1)))/lchol(nchild(m),nchild(m)),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:,s1*128+s2*16+s3)=wg(:,s1*128+s2*16+s3)*tb
    ENDIF

ENDIF
END DO         ! Strategy loop
ENDIF
END DO         ! Strategy loop
ENDIF
END DO         ! Strategy loop


DO s1=0,2**2-1       ! Strategy loop
DO s2=0,2**3-1
DO s3=0,2**4-1
ghkval=ghkval+SUM(wg(:,s1*128+s2*16+s3))
END DO
END DO
END DO         ! Strategy loop
ghkval=ghkval/ns







CASE DEFAULT
    PRINT *,'nchild(m): ', nchild(m)
    PRINT *,'error'
    PAUSE
END SELECT




IF (ghkval<=0.0) THEN
!	PRINT *,'Error found!'
	ghkval=TINY(ghkval)
!	STOP
ENDIF

END FUNCTION GHK


FUNCTION ghkcheck(s,n,m)
USE Globvar
IMPLICIT NONE
INTEGER, INTENT(IN) :: m     ! mkt id
INTEGER, INTENT(IN) :: n      ! 
INTEGER, INTENT(IN) :: s      ! player n's strategy
INTEGER :: i, nup
INTEGER :: ghkcheck

ghkcheck=0
IF (n>=2) THEN
nup=nearcounti(IBITS(near(m),nchild(m)-n+1,n-1),n-1)
    IF (IBITS(s,nup,1)==1 .and. IBITS(near(m),nchild(m)-n,1)==1) THEN
        ghkcheck=1
    ELSEIF (IBITS(s,nup,1)==0 .and. IBITS(near(m),nchild(m)-n,1)==0) THEN
        ghkcheck=1
    END IF
ELSE
	PRINT *,'Error found!'
	PRINT *,'player #: ', n
	STOP
ENDIF


END FUNCTION ghkcheck



FUNCTION lcholbase(m,param,n) RESULT(lchol)
USE Globvar
USE MATRIX
IMPLICIT NONE
INTEGER, INTENT(IN) :: m			! family id
INTEGER, INTENT(IN) :: n 
REAL(8), INTENT(IN) :: param(n)
REAL(8) :: lchol(nchild(m),nchild(m)), omega(nchild(m),nchild(m))
INTEGER :: i

omega=0.0000d0
DO i = 1, nchild(m)
omega(i,i)=1.0000d0
END DO

lchol=CHOL(omega)

END FUNCTION lcholbase




FUNCTION nearcount(ent,n)
USE Globvar
IMPLICIT NONE
INTEGER, INTENT(IN) :: ent    ! family configuration
INTEGER, INTENT(IN) :: n      ! nchild
INTEGER :: i
REAL(8) :: nearcount

nearcount=0.0d0
DO i=0, n-1
    nearcount=nearcount+IBITS(ent,i,1)
END DO

END FUNCTION nearcount


FUNCTION nearcounti(ent,n)
USE Globvar
IMPLICIT NONE
INTEGER, INTENT(IN) :: ent    ! family configuration
INTEGER, INTENT(IN) :: n      ! nchild
INTEGER :: i
INTEGER :: nearcounti

nearcounti=0
IF (n>=0) THEN
DO i=0, n-1
    nearcounti=nearcounti+IBITS(ent,i,1)
END DO
END IF

END FUNCTION nearcounti




FUNCTION var_cov(param,n)
USE Globvar
! This function computes the VCov matrix of the estimates
! By using the numerical gradient of a function using forward differences
IMPLICIT NONE
INTEGER, INTENT(IN) :: n
REAL(8), INTENT(IN) :: param(n)

REAL(8) :: var_cov(n,n)
REAL(8) :: a(n,n)
REAL(8) :: f0			! Initial value of lnmlobj.
REAL(8) :: grdd(n),dh(n), arg(n,n)
REAL(8) :: ax0(n),xdh(n)
REAL(8) :: dax0(n),one
REAL(8) :: indgrad(n)
INTEGER :: i, j, k

! Set up the gradient calculation
one  = 1.0D0	! This parameter affects the length of dh
grdd = 0.0D0
	! Computation of stepsize (dh) for gradient
ax0  = ABS(param)
dax0 = 1.0D0
WHERE (param.ne.0.0D0) dax0 = param/ax0
dh = (1.0D-6)*(1.0D-2)*one*dax0
WHERE (ax0.gt.(1.0D-2)*one) dh = (1.0D-8)*ax0*dax0
xdh = param+dh
arg = SPREAD(param,DIM=2,NCOPIES=n)

!print *,param
!print *,'dh: ',dh
!print *,xdh
!stop

! Calculate VCOV
a=0.0d0

!$OMP PARALLEL DO SCHEDULE(STATIC) PRIVATE(arg,f0,grdd,indgrad)
DO i = 1, nfam

    arg = SPREAD(param,DIM=2,NCOPIES=n)

	utilsub(famindex(i)-nchild(i)+1:famindex(i))=MATMUL(param(1:ntheta1),x(:,famindex(i)-nchild(i)+1:famindex(i)))

	f0=DLOG(GHK(i,param,n))
		DO j = 1, n
			arg(j,j)=xdh(j)

			! Preparation for the likelihood calculation
			utilsub(famindex(i)-nchild(i)+1:famindex(i))=MATMUL(arg(1:ntheta1,j),x(:,famindex(i)-nchild(i)+1:famindex(i)))
			grdd(j)=DLOG(GHK(i,arg(:,j),n))
		END DO
!print *,'f0: ',f0
!print *,'grdd: ',grdd
!print *,'diff: ',grdd-f0
	indgrad= (grdd-f0)/dh
!print *,indgrad
	a=a+outerprod(indgrad,indgrad)
!print *,outerprod(indgrad,indgrad)
END DO
!$OMP END PARALLEL DO


DO i=1,ntheta
DO j=1,ntheta
	print *,'a(i,j):  ',i,j, a(i,j)
END DO
END DO

print *,'inverting var cov matrix'
var_cov=MATINV(a)

END FUNCTION var_cov






FUNCTION FAMSOLVE(param,n)
	! This function calculates the GHK simulator for each market*year
USE Globvar
USE PROBABILITY
IMPLICIT NONE
INTEGER, INTENT(IN) :: n 
REAL(8), INTENT(IN) :: param(n)

REAL(8) :: FAMSOLVE(0:2**nchildmax-1,5)

INTEGER :: i, j, k, l, m, ienta, ientb
INTEGER :: s0, s1, s2, s3, s4
! Specify the array sizes based on the number of players (i.e. possible number of strategy combinations)
REAL(8) :: wg(ns,0:sm4-1)
REAL(8) :: ta(ns), tb(ns)
REAL(8) :: v(ns,0:sm4-1,nchildmax-1)   ! GHK simulation draws constructed from draws()
REAL(8) :: utiltemp, temp
REAL(8) :: ghkval(ns,0:sm4*2-1)		! The calculated GHK value
REAL(8) :: configprob(0:15)
REAL(8) :: lchol(nchildmax,nchildmax)
INTEGER :: predconfig      ! predicted location configuration



utilsub(1:nobs)=MATMUL(param(1:ntheta1),x(:,1:nobs))

FAMSOLVE=0

DO m=1,nfam

k=famindex(m)
wg=0.0d0
v=0.0d0
ghkval=0.0d0
configprob=0.0d0

SELECT CASE(nchild(m))

CASE(1)

    FAMSOLVE(1,1)=FAMSOLVE(1,1)+1.0d0-NORMCDF(-utilsub(k),0.0d0,1.0d0)
    FAMSOLVE(0,1)=FAMSOLVE(0,1)+      NORMCDF(-utilsub(k),0.0d0,1.0d0)

    configprob(1)=1.0d0-NORMCDF(-utilsub(k),0.0d0,1.0d0)
    configprob(0)=NORMCDF(-utilsub(k),0.0d0,1.0d0)


CASE(2)

lchol(1:2,1:2)=lcholbase(m,param,n)

! s1==0
    tb=NORMCDF(-utilsub(k),0.0d0,1.0d0)
    wg(:,0)=tb

    DO j=1, ns
       	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
    	    v(j,0,1)=-HUGE(v(j,0,1))
        ELSE
        v(j,0,1)=INVNORMCDF((tb(j)*draws2(j,0)),0.0d0,1.0d0)
        ENDIF
    END DO

! s1==3
    ta=NORMCDF((-utilsub(k)),0.0d0,1.0d0)

    wg(:,3)=1.0-ta

    DO j=1, ns
   	    IF (ta(j)>=0.99999999d0) THEN 
!         PRINT *,'Warning: a is too large.'
 	    v(j,3,1)=HUGE(v(j,3,1))
        ELSE
        v(j,3,1)=INVNORMCDF((ta(j)*(1-draws2(j,3))+draws2(j,3)),0.0d0,1.0d0)
        ENDIF
    END DO

! s1==1
    ta=NORMCDF((-utilsub(k)),0.0d0,1.0d0)
    tb=NORMCDF((-utilsub(k)),0.0d0,1.0d0)
    IF (tb(1)-ta(1) > 0.0d0) THEN
    wg(:,1)=tb-ta

    DO j=1, ns
       	IF (tb(j)<=0.00000001d0) THEN 
    !   PRINT *,'Warning: b is too small.'
    	    v(j,1,1)=-HUGE(v(j,1,1))
        ELSE
          	v(j,1,1)=INVNORMCDF((ta(j)*(1-draws2(j,1))+tb(j)*draws2(j,1)),0.0d0,1.0d0)
        ENDIF
    END DO
    ENDIF

! s1==2
    ta=NORMCDF((-utilsub(k)),0.0d0,1.0d0)
    tb=NORMCDF((-utilsub(k)),0.0d0,1.0d0)
    IF (tb(1)-ta(1) > 0.0d0) THEN
    wg(:,2)=tb-ta

    DO j=1, ns
       	IF (tb(j)<=0.00000001d0) THEN 
    !	    PRINT *,'Warning: b is too small.'
    	    v(j,2,1)=-HUGE(v(j,2,1))
        ELSE
         	v(j,2,1)=INVNORMCDF((ta(j)*(1-draws2(j,2))+tb(j)*draws2(j,2)),0.0d0,1.0d0)
        ENDIF
    END DO
    ENDIF


! player 1

DO s1=0,2**2-1       ! Strategy loop
IF (SUM(wg(:,s1))>0.0d0) THEN

    utiltemp=utilsub(k-1)
	DO j=1, ns
    ta(j)=NORMCDF((-utiltemp-DOT_PRODUCT(lchol(nchild(m),1:nchild(m)-1),v(j,s1,1:nchild(m)-1)))/lchol(nchild(m),nchild(m)),0.0d0,1.0d0)
	END DO         ! Simulation loop
	ghkval(:,1*4+s1)=wg(:,s1)*(1.0-ta)

    utiltemp=utilsub(k-1)
	DO j=1, ns
    tb(j)=NORMCDF((-utiltemp-DOT_PRODUCT(lchol(nchild(m),1:nchild(m)-1),v(j,s1,1:nchild(m)-1)))/lchol(nchild(m),nchild(m)),0.0d0,1.0d0)
	END DO         ! Simulation loop
	ghkval(:,0*4+s1)=wg(:,s1)*tb

    FAMSOLVE(0*2+IBITS(s1,0,1),2)=FAMSOLVE(0*2+IBITS(s1,0,1),2)+SUM(ghkval(:,0*4+s1))/ns
    FAMSOLVE(1*2+IBITS(s1,1,1),2)=FAMSOLVE(1*2+IBITS(s1,1,1),2)+SUM(ghkval(:,1*4+s1))/ns

    configprob(0*2+IBITS(s1,0,1))=configprob(0*2+IBITS(s1,0,1))+SUM(ghkval(:,0*4+s1))/ns
    configprob(1*2+IBITS(s1,1,1))=configprob(1*2+IBITS(s1,1,1))+SUM(ghkval(:,1*4+s1))/ns


ENDIF
END DO         ! Strategy loop


CASE(3)

lchol(1:3,1:3)=lcholbase(m,param,n)

s2=0
        tb=NORMCDF(-utilsub(k),0.0d0,1.0d0)

        DO s1=0,2**2-1
        wg(:,s1*8+s2)=tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,s1*8+s2,1)=-HUGE(v(j,s1*8+s2,1))
   	        ELSE
   	            v(j,s1*8+s2,1)=INVNORMCDF((tb(j)*draws3(j,s1*8+s2,1)),0.0d0,1.0d0)
            ENDIF
        END DO
        END DO

s2=7

        ta=NORMCDF((-utilsub(k)),0.0d0,1.0d0)

        DO s1=0,2**2-1
        wg(:,s1*8+s2)=1.0-ta

        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,s1*8+s2,1)=HUGE(v(j,s1*8+s2,1))
            ELSE
            	v(j,s1*8+s2,1)=INVNORMCDF((ta(j)*(1-draws3(j,s1*8+s2,1))+draws3(j,s1*8+s2,1)),0.0d0,1.0d0)
	        ENDIF
        END DO
        END DO


s2=1

        ta=NORMCDF((-utilsub(k)),0.0d0,1.0d0)
        tb=NORMCDF((-utilsub(k)),0.0d0,1.0d0)
        IF (tb(1)-ta(1) > 0.0d0) THEN
        DO s1=0,2**2-1
        wg(:,s1*8+s2)=tb-ta

        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,s1*8+s2,1)=-HUGE(v(j,s1*8+s2,1))
   	        ELSE
          	v(j,s1*8+s2,1)=INVNORMCDF((ta(j)*(1-draws3(j,s1*8+s2,1))+tb(j)*draws3(j,s1*8+s2,1)),0.0d0,1.0d0)
            ENDIF
        END DO
        ENDDO
        ENDIF

s2=6

        ta=NORMCDF((-utilsub(k)),0.0d0,1.0d0)
        tb=NORMCDF((-utilsub(k)),0.0d0,1.0d0)
        IF (tb(1)-ta(1) > 0.0d0) THEN
        DO s1=0,2**2-1
        wg(:,s1*8+s2)=tb-ta

        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,s1*8+s2,1)=-HUGE(v(j,s1*8+s2,1))
   	        ELSE
          	v(j,s1*8+s2,1)=INVNORMCDF((ta(j)*(1-draws3(j,s1*8+s2,1))+tb(j)*draws3(j,s1*8+s2,1)),0.0d0,1.0d0)
            ENDIF
        END DO
        ENDDO
        ENDIF



DO s1=0,2**2-1       ! Strategy loop
DO s2=0,2**3-1

    IF (s1==0) THEN
        utiltemp=-utilsub(k-1)
    	DO j=1, ns
        tb(j)=NORMCDF((utiltemp-DOT_PRODUCT(lchol(2,1:1),v(j,s1*8+s2,1:1)))/lchol(2,2),0.0d0,1.0d0)
    	END DO         ! Simulation loop
        wg(:,s1*8+s2)=wg(:,s1*8+s2)*tb

        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,s1*8+s2,2)=-HUGE(v(j,s1*8+s2,2))
   	        ELSE
!            PRINT *,'k,s1,s2',k,s1,s2
   	        v(j,s1*8+s2,2)=INVNORMCDF((tb(j)*draws3(j,s1*8+s2,2)),0.0d0,1.0d0)
            ENDIF
        END DO

    ELSE IF (s1==3) THEN
        utiltemp=-utilsub(k-1)
    	DO j=1, ns
        ta(j)=NORMCDF((utiltemp-DOT_PRODUCT(lchol(2,1:1),v(j,s1*8+s2,1:1)))/lchol(2,2),0.0d0,1.0d0)
    	END DO         ! Simulation loop
        wg(:,s1*8+s2)=wg(:,s1*8+s2)*(1.0d0-ta)

        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,s1*8+s2,2)=HUGE(v(j,s1*8+s2,2))
            ELSE
            	v(j,s1*8+s2,2)=INVNORMCDF((ta(j)*(1-draws3(j,s1*8+s2,2))+draws3(j,s1*8+s2,2)),0.0d0,1.0d0)
	        ENDIF
        END DO


    ELSE IF (s1==1) THEN       ! s1==1
        utiltemp=-utilsub(k-1)

    	DO j=1, ns
        ta(j)=NORMCDF((utiltemp-DOT_PRODUCT(lchol(2,1:1),v(j,s1*8+s2,1:1)))/lchol(2,2),0.0d0,1.0d0)
    	END DO         ! Simulation loop

        utiltemp=-utilsub(k-1)
    	DO j=1, ns
        tb(j)=NORMCDF((utiltemp-DOT_PRODUCT(lchol(2,1:1),v(j,s1*8+s2,1:1)))/lchol(2,2),0.0d0,1.0d0)
    	END DO         ! Simulation loop


        DO j=1, ns
            IF (tb(j)-ta(j) > 0.0d0) THEN
            wg(j,s1*8+s2)=wg(j,s1*8+s2)*(tb(j)-ta(j))
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,s1*8+s2,2)=HUGE(v(j,s1*8+s2,2))
            ELSE IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,s1*8+s2,2)=-HUGE(v(j,s1*8+s2,2))
   	        ELSE
          	    v(j,s1*8+s2,2)=INVNORMCDF((ta(j)*(1-draws3(j,s1*8+s2,2))+tb(j)*draws3(j,s1*8+s2,2)),0.0d0,1.0d0)
            ENDIF
            ELSE
                wg(j,s1*8+s2)=0
            ENDIF
        END DO

    ELSE IF (s1==2) THEN       ! s1==2
        utiltemp=-utilsub(k-1)

    	DO j=1, ns
        ta(j)=NORMCDF((utiltemp-DOT_PRODUCT(lchol(2,1:1),v(j,s1*8+s2,1:1)))/lchol(2,2),0.0d0,1.0d0)
    	END DO         ! Simulation loop

        utiltemp=-utilsub(k-1)
    	DO j=1, ns
        tb(j)=NORMCDF((utiltemp-DOT_PRODUCT(lchol(2,1:1),v(j,s1*8+s2,1:1)))/lchol(2,2),0.0d0,1.0d0)
    	END DO         ! Simulation loop


        DO j=1, ns
            IF (tb(j)-ta(j) > 0.0d0) THEN
            wg(j,s1*8+s2)=wg(j,s1*8+s2)*(tb(j)-ta(j))
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,s1*8+s2,2)=HUGE(v(j,s1*8+s2,2))
            ELSE IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,s1*8+s2,2)=-HUGE(v(j,s1*8+s2,2))
   	        ELSE
          	    v(j,s1*8+s2,2)=INVNORMCDF((ta(j)*(1-draws3(j,s1*8+s2,2))+tb(j)*draws3(j,s1*8+s2,2)),0.0d0,1.0d0)
            ENDIF
            ELSE
                wg(j,s1*8+s2)=0
            ENDIF
        END DO

    ENDIF

END DO         ! Strategy loop
END DO         ! Strategy loop



! player 1

DO s1=0,2**2-1       ! Strategy loop
DO s2=0,2**3-1       ! Strategy loop
IF (SUM(wg(:,s1*8+s2))>0.0d0) THEN

    utiltemp=utilsub(k-2)
	DO j=1, ns
    ta(j)=NORMCDF((-utiltemp-DOT_PRODUCT(lchol(nchild(m),1:nchild(m)-1),v(j,s1*8+s2,1:nchild(m)-1)))/lchol(nchild(m),nchild(m)),0.0d0,1.0d0)
	END DO         ! Simulation loop
	ghkval(:,1*32+s1*8+s2)=wg(:,s1*8+s2)*(1.0-ta)


    utiltemp=utilsub(k-2)
	DO j=1, ns
    tb(j)=NORMCDF((-utiltemp-DOT_PRODUCT(lchol(nchild(m),1:nchild(m)-1),v(j,s1*8+s2,1:nchild(m)-1)))/lchol(nchild(m),nchild(m)),0.0d0,1.0d0)
	END DO         ! Simulation loop
	ghkval(:,0*32+s1*8+s2)=wg(:,s1*8+s2)*tb


    FAMSOLVE(0*4+IBITS(s1,0,1)*2+IBITS(s2,0+IBITS(s1,0,1),1),3)=FAMSOLVE(0*4+IBITS(s1,0,1)*2+IBITS(s2,0+IBITS(s1,0,1),1),3)+SUM(ghkval(:,0*32+s1*8+s2))/ns
    FAMSOLVE(1*4+IBITS(s1,1,1)*2+IBITS(s2,1+IBITS(s1,1,1),1),3)=FAMSOLVE(1*4+IBITS(s1,1,1)*2+IBITS(s2,1+IBITS(s1,1,1),1),3)+SUM(ghkval(:,1*32+s1*8+s2))/ns

    configprob(0*4+IBITS(s1,0,1)*2+IBITS(s2,0+IBITS(s1,0,1),1))=configprob(0*4+IBITS(s1,0,1)*2+IBITS(s2,0+IBITS(s1,0,1),1))+SUM(ghkval(:,0*32+s1*8+s2))/ns
    configprob(1*4+IBITS(s1,1,1)*2+IBITS(s2,1+IBITS(s1,1,1),1))=configprob(1*4+IBITS(s1,1,1)*2+IBITS(s2,1+IBITS(s1,1,1),1))+SUM(ghkval(:,1*32+s1*8+s2))/ns


ENDIF
END DO         ! Strategy loop
END DO         ! Strategy loop



CASE(4)

lchol(1:4,1:4)=lcholbase(m,param,n)


s3=0
        tb=NORMCDF(-utilsub(k),0.0d0,1.0d0)

        DO s1=0,2**2-1
        DO s2=0,2**3-1
        wg(:,s1*128+s2*16+s3)=tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,s1*128+s2*16+s3,1)=-HUGE(v(j,s1*128+s2*16+s3,1))
   	        ELSE
   	            v(j,s1*128+s2*16+s3,1)=INVNORMCDF((tb(j)*draws4(j,s1*128+s2*16+s3,1)),0.0d0,1.0d0)
            ENDIF
        END DO
        END DO
        END DO

s3=15

        ta=NORMCDF((-utilsub(k)),0.0d0,1.0d0)

        DO s1=0,2**2-1
        DO s2=0,2**3-1
        wg(:,s1*128+s2*16+s3)=1.0-ta

        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,s1*128+s2*16+s3,1)=HUGE(v(j,s1*128+s2*16+s3,1))
            ELSE
            	v(j,s1*128+s2*16+s3,1)=INVNORMCDF((ta(j)*(1-draws4(j,s1*128+s2*16+s3,1))+draws4(j,s1*128+s2*16+s3,1)),0.0d0,1.0d0)
	        ENDIF
        END DO
        END DO
        END DO


s3=1

        ta=NORMCDF((-utilsub(k)),0.0d0,1.0d0)
        tb=NORMCDF((-utilsub(k)),0.0d0,1.0d0)
        IF (tb(1)-ta(1) > 0.0d0) THEN
        DO s1=0,2**2-1
        DO s2=0,2**3-1
        wg(:,s1*128+s2*16+s3)=tb-ta

        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,s1*128+s2*16+s3,1)=-HUGE(v(j,s1*128+s2*16+s3,1))
   	        ELSE
          	v(j,s1*128+s2*16+s3,1)=INVNORMCDF((ta(j)*(1-draws4(j,s1*128+s2*16+s3,1))+tb(j)*draws4(j,s1*128+s2*16+s3,1)),0.0d0,1.0d0)
            ENDIF
        END DO
        ENDDO
        ENDDO
        ENDIF

s3=14

        ta=NORMCDF((-utilsub(k)),0.0d0,1.0d0)
        tb=NORMCDF((-utilsub(k)),0.0d0,1.0d0)
        IF (tb(1)-ta(1) > 0.0d0) THEN
        DO s1=0,2**2-1
        DO s2=0,2**3-1
        wg(:,s1*128+s2*16+s3)=tb-ta

        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,s1*128+s2*16+s3,1)=-HUGE(v(j,s1*128+s2*16+s3,1))
   	        ELSE
          	v(j,s1*128+s2*16+s3,1)=INVNORMCDF((ta(j)*(1-draws4(j,s1*128+s2*16+s3,1))+tb(j)*draws4(j,s1*128+s2*16+s3,1)),0.0d0,1.0d0)
            ENDIF
        END DO
        ENDDO
        ENDDO
        ENDIF



DO s1=0,2**2-1       ! Strategy loop
DO s2=0,2**3-1
DO s3=0,2**4-1

    IF (s2==0) THEN
        utiltemp=-utilsub(k-1)
    	DO j=1, ns
        tb(j)=NORMCDF((utiltemp-DOT_PRODUCT(lchol(2,1:1),v(j,s1*128+s2*16+s3,1:1)))/lchol(2,2),0.0d0,1.0d0)
    	END DO         ! Simulation loop
        wg(:,s1*128+s2*16+s3)=wg(:,s1*128+s2*16+s3)*tb

        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,s1*128+s2*16+s3,2)=-HUGE(v(j,s1*128+s2*16+s3,2))
   	        ELSE
!            PRINT *,'k,s1,s2',k,s1,s2
   	        v(j,s1*128+s2*16+s3,2)=INVNORMCDF((tb(j)*draws4(j,s1*128+s2*16+s3,2)),0.0d0,1.0d0)
            ENDIF
        END DO

    ELSE IF (s2==7) THEN
        utiltemp=-utilsub(k-1)
    	DO j=1, ns
        ta(j)=NORMCDF((utiltemp-DOT_PRODUCT(lchol(2,1:1),v(j,s1*128+s2*16+s3,1:1)))/lchol(2,2),0.0d0,1.0d0)
    	END DO         ! Simulation loop
        wg(:,s1*128+s2*16+s3)=wg(:,s1*128+s2*16+s3)*(1.0d0-ta)

        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,s1*128+s2*16+s3,2)=HUGE(v(j,s1*128+s2*16+s3,2))
            ELSE
            	v(j,s1*128+s2*16+s3,2)=INVNORMCDF((ta(j)*(1-draws4(j,s1*128+s2*16+s3,2))+draws4(j,s1*128+s2*16+s3,2)),0.0d0,1.0d0)
	        ENDIF
        END DO


    ELSE IF (s2==1) THEN       ! s2==1
        utiltemp=-utilsub(k-1)

    	DO j=1, ns
        ta(j)=NORMCDF((utiltemp-DOT_PRODUCT(lchol(2,1:1),v(j,s1*128+s2*16+s3,1:1)))/lchol(2,2),0.0d0,1.0d0)
    	END DO         ! Simulation loop

        utiltemp=-utilsub(k-1)
    	DO j=1, ns
        tb(j)=NORMCDF((utiltemp-DOT_PRODUCT(lchol(2,1:1),v(j,s1*128+s2*16+s3,1:1)))/lchol(2,2),0.0d0,1.0d0)
    	END DO         ! Simulation loop


        DO j=1, ns
            IF (tb(j)-ta(j) > 0.0d0) THEN
            wg(j,s1*128+s2*16+s3)=wg(j,s1*128+s2*16+s3)*(tb(j)-ta(j))
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,s1*128+s2*16+s3,2)=HUGE(v(j,s1*128+s2*16+s3,2))
            ELSE IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,s1*128+s2*16+s3,2)=-HUGE(v(j,s1*128+s2*16+s3,2))
   	        ELSE
          	    v(j,s1*128+s2*16+s3,2)=INVNORMCDF((ta(j)*(1-draws4(j,s1*128+s2*16+s3,2))+tb(j)*draws4(j,s1*128+s2*16+s3,2)),0.0d0,1.0d0)
            ENDIF
            ELSE
                wg(j,s1*128+s2*16+s3)=0
            ENDIF
        END DO

    ELSE IF (s2==6) THEN       ! s2==6
        utiltemp=-utilsub(k-1)

    	DO j=1, ns
        ta(j)=NORMCDF((utiltemp-DOT_PRODUCT(lchol(2,1:1),v(j,s1*128+s2*16+s3,1:1)))/lchol(2,2),0.0d0,1.0d0)
    	END DO         ! Simulation loop

        utiltemp=-utilsub(k-1)
    	DO j=1, ns
        tb(j)=NORMCDF((utiltemp-DOT_PRODUCT(lchol(2,1:1),v(j,s1*128+s2*16+s3,1:1)))/lchol(2,2),0.0d0,1.0d0)
    	END DO         ! Simulation loop

        DO j=1, ns
            IF (tb(j)-ta(j) > 0.0d0) THEN
            wg(j,s1*128+s2*16+s3)=wg(j,s1*128+s2*16+s3)*(tb(j)-ta(j))
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,s1*128+s2*16+s3,2)=HUGE(v(j,s1*128+s2*16+s3,2))
            ELSE IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,s1*128+s2*16+s3,2)=-HUGE(v(j,s1*128+s2*16+s3,2))
   	        ELSE
          	    v(j,s1*128+s2*16+s3,2)=INVNORMCDF((ta(j)*(1-draws4(j,s1*128+s2*16+s3,2))+tb(j)*draws4(j,s1*128+s2*16+s3,2)),0.0d0,1.0d0)
            ENDIF
            ELSE
                wg(j,s1*128+s2*16+s3)=0
            ENDIF
        END DO

    ELSE
        wg(:,s1*128+s2*16+s3)=0
    ENDIF

END DO         ! Strategy loop
END DO         ! Strategy loop
END DO         ! Strategy loop





DO s1=0,2**2-1       ! Strategy loop
DO s2=0,2**3-1
DO s3=0,2**4-1

    IF (s1==0) THEN
        utiltemp=-utilsub(k-2)
    	DO j=1, ns
        tb(j)=NORMCDF((utiltemp-DOT_PRODUCT(lchol(3,1:2),v(j,s1*128+s2*16+s3,1:2)))/lchol(3,3),0.0d0,1.0d0)
    	END DO         ! Simulation loop
        wg(:,s1*128+s2*16+s3)=wg(:,s1*128+s2*16+s3)*tb

        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,s1*128+s2*16+s3,3)=-HUGE(v(j,s1*128+s2*16+s3,3))
   	        ELSE
!            PRINT *,'k,s1,s2',k,s1,s2
   	        v(j,s1*128+s2*16+s3,3)=INVNORMCDF((tb(j)*draws4(j,s1*128+s2*16+s3,3)),0.0d0,1.0d0)
            ENDIF
        END DO

    ELSE IF (s1==3) THEN
        utiltemp=-utilsub(k-2)
    	DO j=1, ns
        ta(j)=NORMCDF((utiltemp-DOT_PRODUCT(lchol(3,1:2),v(j,s1*128+s2*16+s3,1:2)))/lchol(3,3),0.0d0,1.0d0)
    	END DO         ! Simulation loop
        wg(:,s1*128+s2*16+s3)=wg(:,s1*128+s2*16+s3)*(1.0d0-ta)

        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,s1*128+s2*16+s3,3)=HUGE(v(j,s1*128+s2*16+s3,3))
            ELSE
            	v(j,s1*128+s2*16+s3,3)=INVNORMCDF((ta(j)*(1-draws4(j,s1*128+s2*16+s3,3))+draws4(j,s1*128+s2*16+s3,3)),0.0d0,1.0d0)
	        ENDIF
        END DO


    ELSE IF (s1==1) THEN       ! s1==1
        utiltemp=-utilsub(k-2)

    	DO j=1, ns
        ta(j)=NORMCDF((utiltemp-DOT_PRODUCT(lchol(3,1:2),v(j,s1*128+s2*16+s3,1:2)))/lchol(3,3),0.0d0,1.0d0)
    	END DO         ! Simulation loop

        utiltemp=-utilsub(k-2)
    	DO j=1, ns
        tb(j)=NORMCDF((utiltemp-DOT_PRODUCT(lchol(3,1:2),v(j,s1*128+s2*16+s3,1:2)))/lchol(3,3),0.0d0,1.0d0)
    	END DO         ! Simulation loop


        DO j=1, ns
            IF (tb(j)-ta(j) > 0.0d0) THEN
            wg(j,s1*128+s2*16+s3)=wg(j,s1*128+s2*16+s3)*(tb(j)-ta(j))
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,s1*128+s2*16+s3,3)=HUGE(v(j,s1*128+s2*16+s3,3))
            ELSE IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,s1*128+s2*16+s3,3)=-HUGE(v(j,s1*128+s2*16+s3,3))
   	        ELSE
          	    v(j,s1*128+s2*16+s3,3)=INVNORMCDF((ta(j)*(1-draws4(j,s1*128+s2*16+s3,3))+tb(j)*draws4(j,s1*128+s2*16+s3,3)),0.0d0,1.0d0)
            ENDIF
            ELSE
                wg(j,s1*128+s2*16+s3)=0
            ENDIF
        END DO

    ELSE IF (s1==2) THEN       ! s1==2
        utiltemp=-utilsub(k-2)

    	DO j=1, ns
        ta(j)=NORMCDF((utiltemp-DOT_PRODUCT(lchol(3,1:2),v(j,s1*128+s2*16+s3,1:2)))/lchol(3,3),0.0d0,1.0d0)
    	END DO         ! Simulation loop

        utiltemp=-utilsub(k-2)
    	DO j=1, ns
        tb(j)=NORMCDF((utiltemp-DOT_PRODUCT(lchol(3,1:2),v(j,s1*128+s2*16+s3,1:2)))/lchol(3,3),0.0d0,1.0d0)
    	END DO         ! Simulation loop


        DO j=1, ns
            IF (tb(j)-ta(j) > 0.0d0) THEN
            wg(j,s1*128+s2*16+s3)=wg(j,s1*128+s2*16+s3)*(tb(j)-ta(j))
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,s1*128+s2*16+s3,3)=HUGE(v(j,s1*128+s2*16+s3,3))
            ELSE IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,s1*128+s2*16+s3,3)=-HUGE(v(j,s1*128+s2*16+s3,3))
   	        ELSE
          	    v(j,s1*128+s2*16+s3,3)=INVNORMCDF((ta(j)*(1-draws4(j,s1*128+s2*16+s3,3))+tb(j)*draws4(j,s1*128+s2*16+s3,3)),0.0d0,1.0d0)
            ENDIF
            ELSE
                wg(j,s1*128+s2*16+s3)=0
            ENDIF
        END DO

    ENDIF

END DO         ! Strategy loop
END DO         ! Strategy loop
END DO         ! Strategy loop




! player 1

DO s1=0,2**2-1       ! Strategy loop
DO s2=0,2**3-1       ! Strategy loop
DO s3=0,2**4-1       ! Strategy loop
IF (SUM(wg(:,s1*128+s2*16+s3))>0.0d0) THEN

    utiltemp=utilsub(k-3)
	DO j=1, ns
    ta(j)=NORMCDF((-utiltemp-DOT_PRODUCT(lchol(nchild(m),1:nchild(m)-1),v(j,s1*128+s2*16+s3,1:nchild(m)-1)))/lchol(nchild(m),nchild(m)),0.0d0,1.0d0)
	END DO         ! Simulation loop
	ghkval(:,1*512+s1*128+s2*16+s3)=wg(:,s1*128+s2*16+s3)*(1.0-ta)


    utiltemp=utilsub(k-3)
	DO j=1, ns
    tb(j)=NORMCDF((-utiltemp-DOT_PRODUCT(lchol(nchild(m),1:nchild(m)-1),v(j,s1*128+s2*16+s3,1:nchild(m)-1)))/lchol(nchild(m),nchild(m)),0.0d0,1.0d0)
	END DO         ! Simulation loop
	ghkval(:,0*512+s1*128+s2*16+s3)=wg(:,s1*128+s2*16+s3)*tb


    FAMSOLVE(0*8+IBITS(s1,0,1)*4+IBITS(s2,0+IBITS(s1,0,1),1)*2+IBITS(s3,0+IBITS(s1,0,1)+IBITS(s2,0+IBITS(s1,0,1),1),1),4)=FAMSOLVE(0*8+IBITS(s1,0,1)*4+IBITS(s2,0+IBITS(s1,0,1),1)*2+IBITS(s3,0+IBITS(s1,0,1)+IBITS(s2,0+IBITS(s1,0,1),1),1),4)+SUM(ghkval(:,0*512+s1*128+s2*16+s3))/ns
    FAMSOLVE(1*8+IBITS(s1,1,1)*4+IBITS(s2,1+IBITS(s1,1,1),1)*2+IBITS(s3,1+IBITS(s1,1,1)+IBITS(s2,1+IBITS(s1,1,1),1),1),4)=FAMSOLVE(1*8+IBITS(s1,1,1)*4+IBITS(s2,1+IBITS(s1,1,1),1)*2+IBITS(s3,1+IBITS(s1,1,1)+IBITS(s2,1+IBITS(s1,1,1),1),1),4)+SUM(ghkval(:,1*512+s1*128+s2*16+s3))/ns

    configprob(0*8+IBITS(s1,0,1)*4+IBITS(s2,0+IBITS(s1,0,1),1)*2+IBITS(s3,0+IBITS(s1,0,1)+IBITS(s2,0+IBITS(s1,0,1),1),1))=configprob(0*8+IBITS(s1,0,1)*4+IBITS(s2,0+IBITS(s1,0,1),1)*2+IBITS(s3,0+IBITS(s1,0,1)+IBITS(s2,0+IBITS(s1,0,1),1),1))+SUM(ghkval(:,0*512+s1*128+s2*16+s3))/ns
    configprob(1*8+IBITS(s1,1,1)*4+IBITS(s2,1+IBITS(s1,1,1),1)*2+IBITS(s3,1+IBITS(s1,1,1)+IBITS(s2,1+IBITS(s1,1,1),1),1))=configprob(1*8+IBITS(s1,1,1)*4+IBITS(s2,1+IBITS(s1,1,1),1)*2+IBITS(s3,1+IBITS(s1,1,1)+IBITS(s2,1+IBITS(s1,1,1),1),1))+SUM(ghkval(:,1*512+s1*128+s2*16+s3))/ns


ENDIF
END DO         ! Strategy loop
END DO         ! Strategy loop
END DO         ! Strategy loop




CASE DEFAULT
    PRINT *,'nchild(m): ', nchild(m)
    PRINT *,'error'
    PAUSE
END SELECT


temp=0
DO i=0,15
    IF (configprob(i)>temp) THEN
        temp=configprob(i)
        predconfig=i
    ENDIF
ENDDO
DO l=1,nchild(m)
    predobs(k-nchild(m)+l)=IBITS(predconfig,nchild(m)-l,1)
ENDDO


END DO  ! nfam



END FUNCTION FAMSOLVE











END MODULE Object






PROGRAM MAIN
USE Globvar
USE INIT
USE MINIMIZATION
USE Object
IMPLICIT NONE

REAL(8) :: resultmlv, resultmlse
REAL(8) :: gtol
REAL(4) :: time0, iniadj
INTEGER :: i, j, k, l, m, n, io
INTEGER :: s1, s2, s3, s4, s5, nup
CHARACTER(12) :: horz(3)
CHARACTER(80) :: flname2
CHARACTER(20) :: fmtname           ! To accomodate different number of parameters
CHARACTER(50) :: inivalnote
CHARACTER(60) :: setupnote

REAL(8), ALLOCATABLE :: initheta(:)
REAL(8), ALLOCATABLE :: vcov(:,:)
REAL(8), ALLOCATABLE :: se(:)
REAL(8), ALLOCATABLE :: tval(:)
CHARACTER(16), ALLOCATABLE :: vert(:)
REAL(8), ALLOCATABLE :: pred(:,:)

horz(1)='        coef' ; horz(2)='        s.e.' ; horz(3)=' asy t-val'

! Read the setting file
    flname2='setting.prn'
    OPEN(1,file=(flname2))
    READ(1,fmt='(A60)') setupnote
    READ(1,fmt=*) ntheta1
    READ(1,fmt=*) ntheta
    READ(1,fmt=*) nobs
    READ(1,fmt=*) nfam
    READ(1,fmt=*) ns
ALLOCATE (initheta(ntheta))
    READ(1,fmt=*) initheta
ALLOCATE (vert(ntheta))
    vert(1)='cons      '
    READ(1,fmt=*) vert(2:ntheta)
    CLOSE(1)


ALLOCATE (theta(ntheta))
ALLOCATE (vcov(ntheta,ntheta))
ALLOCATE (se(ntheta))
ALLOCATE (tval(ntheta))
ALLOCATE (famid(nobs))
ALLOCATE (famindex(nfam))
ALLOCATE (nchild(nfam))
ALLOCATE (x(ntheta1,nobs))
ALLOCATE (y(nobs))
ALLOCATE (near(nfam))
ALLOCATE (utilsub(nobs))
ALLOCATE (predobs(nobs))
ALLOCATE (pred(0:2**nchildmax-1,5))


! Set gradient tolerance
gtol=0.000001d0

time0 = SECNDS(0.0)

CALL GETTIM(i, j, k, l)
CALL GETDAT(l, m, n)
flname2=CHAR(ICHAR('0')+(m/10))//CHAR(ICHAR('0')+(m-m/10*10))//CHAR(ICHAR('0')+(n/10))//CHAR(ICHAR('0')+(n-n/10*10))//CHAR(ICHAR('0')+(i/10))//CHAR(ICHAR('0')+(i-i/10*10))//CHAR(ICHAR('0')+(j/10))//CHAR(ICHAR('0')+(j-j/10*10))//CHAR(ICHAR('0')+(k/10))//CHAR(ICHAR('0')+(k-k/10*10))//'.prn'

OPEN(4,file=(flname2),STATUS='REPLACE')
WRITE(4,fmt='(A20,4x,A60)') 'Estimation Setting: ', setupnote
WRITE(4,fmt=*) '# of families     :  ', nfam
WRITE(4,fmt=*) '# of observations:  ', nobs
WRITE(4,fmt=*) '# of simuln draws  :  ', ns
WRITE(4,fmt=*) ' '
WRITE(4,fmt=*) 'Initial Values: '
DO i=1,ntheta
!WRITE(4,fmt='(F16.10)') initheta(i)
END DO
WRITE(4,fmt=*) ' '


theta=initheta
numeval=0

! Load the data set
CALL READDATA(nobs)



ALLOCATE (draws2(ns,0:sm2-1))
ALLOCATE (draws3(ns,0:sm3-1,2))
ALLOCATE (draws4(ns,0:sm4-1,3))
CALL RANDOM_SEED(PUT=(/10000/))
CALL RANDOM_NUMBER (draws2)
CALL RANDOM_NUMBER (draws3)
CALL RANDOM_NUMBER (draws4)
draws2(ns/2+1:ns,:)=1-draws2(1:ns/2,:)        ! This line makes antithetics
draws3(ns/2+1:ns,:,:)=1-draws3(1:ns/2,:,:)        ! This line makes antithetics
draws4(ns/2+1:ns,:,:)=1-draws4(1:ns/2,:,:)        ! This line makes antithetics



! the following lines compute the estimates using a Quasi-Newton method
! with a numerical gradient



CALL dfpmin(theta,ntheta,gtol,lnmlobj)
objsaved=lnmlobj(theta,ntheta,0)
objsaved=-objsaved


print *,'check4-1! - 1st round Minimization Finished!'
print *,'objsaved : ',objsaved

print *,'Calculating variance-covariance matrix...'

vcov=var_cov(theta, ntheta)

print *,'check4-2! - vcov obtained!'
!	print *,vcov


DO i=1,ntheta
	print *,'vcov(i,i):  ',i, vcov(i,i)
END DO

DO i=1,ntheta
	se(i)=DSQRT(vcov(i,i))
	tval(i)=theta(i)/se(i)
END DO




WRITE(4,fmt=*) '# of parameters:  ', ntheta
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(10x,3A16)') (horz(i),i=1,3)
WRITE(4,fmt=*) ' '
DO i=1,ntheta
	WRITE(4,fmt='(A16,3F16.10)') vert(i), theta(i), se(i), tval(i)
ENDDO

pred=0
WRITE(4,fmt=*) ' '
WRITE(4,fmt=*) 'Location patterns: Data'
WRITE(4,fmt='(A8,8A8)') '  config',' 1 child','     (%)',' 2 child','     (%)',' 3 child','     (%)',' 4 child','     (%)'
DO i=0,2**nchildmax-1
    DO j=1,nfam
        IF (near(j)==i) THEN
            pred(i,nchild(j))=pred(i,nchild(j))+1
        ENDIF
    ENDDO
ENDDO
DO i=0,2**nchildmax-1
	WRITE(4,fmt='(4X,4I1,4(F8.1,F8.2))') IBITS(i,3,1),IBITS(i,2,1),IBITS(i,1,1),IBITS(i,0,1), (pred(i,k),(pred(i,k)/SUM(pred(:,k))*100.0d0),k=1,4)
ENDDO
WRITE(4,fmt='(A8,4(F8.1,F8.2))') '   total', ((SUM(pred(:,j)),100.0),j=1,4)

pred=0
WRITE(4,fmt=*) ' '
WRITE(4,fmt=*) 'Location patterns: Predicted'
WRITE(4,fmt='(A8,8A8)') '  config',' 1 child','     (%)',' 2 child','     (%)',' 3 child','     (%)',' 4 child','     (%)'
    pred(:,:)=FAMSOLVE(theta,ntheta)
DO i=0,2**nchildmax-1
	WRITE(4,fmt='(4X,4I1,4(F8.1,F8.2))') IBITS(i,3,1),IBITS(i,2,1),IBITS(i,1,1),IBITS(i,0,1), (pred(i,k),(pred(i,k)/SUM(pred(:,k))*100.0d0),k=1,4)
ENDDO
WRITE(4,fmt='(A8,4(F8.1,F8.2))') '   total', ((SUM(pred(:,j)),100.0),j=1,4)




WRITE(4,fmt=*) ' '
n=0
DO m=1,nfam
DO j=1,nchild(m)
    IF (predobs(famindex(m)-nchild(m)+j)==IBITS(near(m),nchild(m)-j,1)) n=n+1
ENDDO
ENDDO
WRITE(4,fmt=*) '% of observations correctly predicted:  ', n*1.0d0/nobs*100
n=0
DO m=1,nfam
    k=0
    DO j=1,nchild(m)
        IF (predobs(famindex(m)-nchild(m)+j)==IBITS(near(m),nchild(m)-j,1)) k=k+1
    ENDDO
    IF (k==nchild(m)) n=n+1
ENDDO
WRITE(4,fmt=*) '% of families correctly predicted:  ', n*1.0d0/nfam*100
n=0; l=0
DO m=1,nfam
    IF (nchild(m)==1) THEN
        IF (predobs(famindex(m))==IBITS(near(m),0,1)) n=n+1
    l=l+1
    ENDIF
ENDDO
WRITE(4,fmt=*) '% of 1 child families correctly predicted:  ', n*1.0d0/l*100
n=0; l=0
DO m=1,nfam
    IF (nchild(m)==2) THEN
    k=0
    DO j=1,nchild(m)
        IF (predobs(famindex(m)-nchild(m)+j)==IBITS(near(m),nchild(m)-j,1)) k=k+1
    ENDDO
    IF (k==nchild(m)) n=n+1
    l=l+1
    ENDIF
ENDDO
WRITE(4,fmt=*) '% of 2 child families correctly predicted:  ', n*1.0d0/l*100
n=0; l=0
DO m=1,nfam
    IF (nchild(m)==3) THEN
    k=0
    DO j=1,nchild(m)
        IF (predobs(famindex(m)-nchild(m)+j)==IBITS(near(m),nchild(m)-j,1)) k=k+1
    ENDDO
    IF (k==nchild(m)) n=n+1
    l=l+1
    ENDIF
ENDDO
WRITE(4,fmt=*) '% of 3 child families correctly predicted:  ', n*1.0d0/l*100
n=0; l=0
DO m=1,nfam
    IF (nchild(m)==4) THEN
    k=0
    DO j=1,nchild(m)
        IF (predobs(famindex(m)-nchild(m)+j)==IBITS(near(m),nchild(m)-j,1)) k=k+1
    ENDDO
    IF (k==nchild(m)) n=n+1
    l=l+1
    ENDIF
ENDDO
WRITE(4,fmt=*) '% of 4 child families correctly predicted:  ', n*1.0d0/l*100

WRITE(4,fmt=*) ' '
WRITE(4,fmt=*) 'ML objective:  ', objsaved
WRITE(4,fmt=*) ' '
WRITE(4,fmt=*) '# of objective function evaluations:  ', numeval
WRITE(4,fmt=*) ' '
WRITE(4,fmt=*) 'run time(seconds):  ', SECNDS(time0)
WRITE(4,fmt=*) ' '





CLOSE(4)

END PROGRAM MAIN
