MODULE Globvar
IMPLICIT NONE
! This module declares the variables that are going to be used throughout the program.
! To avoid stack overflow error --> Stack Reserve: 50,000,000

INTEGER, PARAMETER :: nchildmax=4  ! max number of children
INTEGER, PARAMETER :: sm2=2**(1+1)  ! necessary amount for strategy patterns
INTEGER, PARAMETER :: sm3=2**(1+1)*2**(2+1)  ! necessary amount for strategy patterns
INTEGER, PARAMETER :: sm4=2**(1+1)*2**(2+1)*2**(3+1)  ! necessary amount for strategy patterns
	! The 4th child in a 4-child family has 2**(3+1) strategies, because 
	! the number of possible elder siblings in the upstream from 0 to 3, i.e. 4.

INTEGER :: nobs
INTEGER :: nfam
INTEGER :: ns              ! number of simulation draws
REAL(8) :: objsaved
INTEGER :: numeval
INTEGER :: ntheta
INTEGER :: nthetab     ! # of parameters in beta term
INTEGER :: nthetaa     ! # of parameters in alpha term
INTEGER :: nthetag     ! # of parameters in gamma term
INTEGER :: nthetar     ! # of parameters in rho term
INTEGER, ALLOCATABLE :: famid(:)   ! famid(nobs)
INTEGER, ALLOCATABLE :: nchild(:), famindex(:)   ! nchild(nfam), famindex(nfam)
REAL(8), ALLOCATABLE :: theta(:) ! theta(ntheta)
REAL(8), ALLOCATABLE :: xb(:,:)   ! xb(nthetab,nobs)
REAL(8), ALLOCATABLE :: xa(:,:)   ! xa(nthetaa,nobs)
REAL(8), ALLOCATABLE :: xg(:,:)   ! xg(nthetag,nobs)
INTEGER :: sa, sg		! actual size of xa and xg
REAL(8), ALLOCATABLE :: xr(:,:,:)   ! xr(nthetar,nchildmax,nobs) xr(:,i,:)
INTEGER, ALLOCATABLE :: y(:)     ! y(nobs)
INTEGER, ALLOCATABLE :: near(:)   ! near(nfam)
INTEGER, ALLOCATABLE :: predobs(:)  !nearpred(nobs)


REAL(8), ALLOCATABLE :: draws2seed(:,:,:)    ! nchild=2. draws(ns,necessary amount,2)
REAL(8), ALLOCATABLE :: draws3seed(:,:,:)    ! nchild=3. draws(ns,necessary amount,3)
REAL(8), ALLOCATABLE :: draws4seed(:,:,:)    ! nchild=4. draws(ns,necessary amount,4)

CHARACTER(10) :: path                          ! Specify the data directory. To be read from setting.prn

END MODULE Globvar



MODULE INIT
	USE Globvar
	IMPLICIT NONE

CONTAINS

	SUBROUTINE READDATA(nobs)
		USE Globvar
		IMPLICIT NONE
		INTEGER, INTENT(IN) :: nobs
		INTEGER :: i,j,k,m,io

        CHARACTER(60) :: readfile_x      ! File names to be read

readfile_x='data.raw'

xb=0.0d0; y=0.0d0
xb(1,:)=1.00000000d0; xa(1,:)=1.00000000d0
IF (nthetag>=1) xg(1,:)=1.00000000d0
xr(1,:,:)=1.00000000d0

print *,'Reading Data ... '

	OPEN(1,file=(readfile_x),status='old')

    IF (nthetaa>3 .and. nthetar>1 .and. nthetag>2) THEN
        DO i = 1, nobs
	    READ(1,fmt=*,iostat=io) y(i), famid(i), xb(2:,i), xa(2:,i), xg(2:,i), xr(2:nthetar,1,i), xr(2:nthetar,2,i), xr(2:nthetar,3,i), xr(2:nthetar,4,i)
        IF (io<0) EXIT
        END DO
    ELSE IF (nthetaa>3 .and. nthetar>1) THEN
        DO i = 1, nobs
	    READ(1,fmt=*,iostat=io) y(i), famid(i), xb(2:,i), xa(2:,i), xr(2:nthetar,1,i), xr(2:nthetar,2,i), xr(2:nthetar,3,i), xr(2:nthetar,4,i)
        IF (io<0) EXIT
        END DO
    ELSE IF (nthetaa>1) THEN
        DO i = 1, nobs
	    READ(1,fmt=*,iostat=io) y(i), famid(i), xb(2:,i), xa(2:,i)
        IF (io<0) EXIT
        END DO
    ELSE
        DO i = 1, nobs
	    READ(1,fmt=*,iostat=io) y(i), famid(i), xb(2:,i)
        IF (io<0) EXIT
        END DO
    ENDIF
    CLOSE(1)


nchild=0
j=1
nchild(1)=1
DO i=2,nobs
	IF (famid(i)/=famid(i-1)) THEN
		j=j+1
	ENDIF
		nchild(j)=nchild(j)+1
ENDDO

famindex(1)=nchild(1)
DO i=2,nfam
	famindex(i)=famindex(i-1)+nchild(i)
ENDDO


DO m=1,nfam
near(m)=0
DO i=1,nchild(m)
near(m)=near(m)*2
near(m)=near(m)+y(famindex(m)-nchild(m)+i)
END DO
END DO



print *,'Reading Data ... Finished '

	END SUBROUTINE READDATA


END MODULE INIT

MODULE Object
	USE Globvar
	USE MATRIX
    USE PROBABILITY
	USE GHK
	IMPLICIT NONE

CONTAINS


FUNCTION lnmlobj(param,n,gradflag)	! This function computes the ML objective function
USE Globvar
IMPLICIT NONE
INTEGER, INTENT(IN) :: n 
INTEGER, INTENT(IN) :: gradflag
REAL(8), INTENT(IN) :: param(n)
REAL(8) :: mlobj
REAL(8) :: lnmlobj
REAL(8) :: famprob(nfam)			! The family-level likelihood
INTEGER :: i, j, k, m
REAL(4) :: tms
REAL(8) :: draws2(ns,0:sm2-1,2), draws2_2(ns,0:sm2-1,2), draws2_3(ns,0:sm2-1,2)
REAL(8) :: draws3(ns,0:sm3-1,3), draws3_2(ns,0:sm3-1,3), draws3_3(ns,0:sm3-1,3)
REAL(8) :: draws4(ns,0:sm4-1,4), draws4_2(ns,0:sm4-1,4), draws4_3(ns,0:sm4-1,4)

tms = SECNDS(0.0)

numeval=numeval+1

print ('(a16,I8)'),'mle evaluation: ',numeval


! This section is to elaborate the random draws (without this, every household uses the same set of 200 simulation draws)
! This is supposed to help convergence, though it doesn't affect the result much.
draws2=draws2seed+0.13
WHERE(draws2>1.0d0) draws2=draws2-1.0d0
draws3=draws3seed+0.13
WHERE(draws3>1.0d0) draws3=draws3-1.0d0
draws4=draws4seed+0.13
WHERE(draws4>1.0d0) draws4=draws4-1.0d0
    draws2=CSHIFT(draws2,1,2)
    draws3=CSHIFT(draws3,7,2)
    draws4=CSHIFT(draws4,9,2)
draws2_2=draws2+0.13
WHERE(draws2_2>1.0d0) draws2_2=draws2_2-1.0d0
draws3_2=draws3+0.13
WHERE(draws3_2>1.0d0) draws3_2=draws3_2-1.0d0
draws4_2=draws4+0.13
WHERE(draws4_2>1.0d0) draws4_2=draws4_2-1.0d0
    draws2_2=CSHIFT(draws2_2,1,2)
    draws3_2=CSHIFT(draws3_2,7,2)
    draws4_2=CSHIFT(draws4_2,9,2)
draws2_3=draws2_2+0.13
WHERE(draws2_3>1.0d0) draws2_3=draws2_3-1.0d0
draws3_3=draws3_2+0.13
WHERE(draws3_3>1.0d0) draws3_3=draws3_3-1.0d0
draws4_3=draws4_2+0.13
WHERE(draws4_3>1.0d0) draws4_3=draws4_3-1.0d0
    draws2_3=CSHIFT(draws2_3,1,2)
    draws3_3=CSHIFT(draws3_3,7,2)
    draws4_3=CSHIFT(draws4_3,9,2)


!$OMP PARALLEL DO SCHEDULE(DYNAMIC)
DO m = 1, nfam

!WRITE(4,fmt=*) m, nchild(m)

SELECT CASE(nchild(m))
CASE(1)
	famprob(m)=DLOG(GHK1(m,param,n))
CASE(2)
    IF (MOD(m,20)<5) THEN
	famprob(m)=DLOG(SUM(GHK2(m,param,n,ns,draws2seed,lcholbase(m,param,n),nthetab, nthetaa, nthetag, xb(:,famindex(m)-1:famindex(m)), xa(:,famindex(m)-1:famindex(m)), xg(:,famindex(m)-1:famindex(m)), sa, sg, near(m),0)))
	ELSEIF (MOD(m,20)<10) THEN
	famprob(m)=DLOG(SUM(GHK2(m,param,n,ns,draws2,lcholbase(m,param,n),nthetab, nthetaa, nthetag, xb(:,famindex(m)-1:famindex(m)), xa(:,famindex(m)-1:famindex(m)), xg(:,famindex(m)-1:famindex(m)), sa, sg, near(m),0)))
	ELSEIF (MOD(m,20)<15) THEN
	famprob(m)=DLOG(SUM(GHK2(m,param,n,ns,draws2_2,lcholbase(m,param,n),nthetab, nthetaa, nthetag, xb(:,famindex(m)-1:famindex(m)), xa(:,famindex(m)-1:famindex(m)), xg(:,famindex(m)-1:famindex(m)), sa, sg, near(m),0)))
	ELSE
	famprob(m)=DLOG(SUM(GHK2(m,param,n,ns,draws2_3,lcholbase(m,param,n),nthetab, nthetaa, nthetag, xb(:,famindex(m)-1:famindex(m)), xa(:,famindex(m)-1:famindex(m)), xg(:,famindex(m)-1:famindex(m)), sa, sg, near(m),0)))
	ENDIF
CASE(3)
    IF (MOD(m,20)<5) THEN
	famprob(m)=DLOG(SUM(GHK3(m,param,n,ns,draws3seed,lcholbase(m,param,n),nthetab, nthetaa, nthetag, xb(:,famindex(m)-2:famindex(m)), xa(:,famindex(m)-2:famindex(m)), xg(:,famindex(m)-2:famindex(m)), sa, sg, near(m),0)))
	ELSEIF (MOD(m,20)<10) THEN
	famprob(m)=DLOG(SUM(GHK3(m,param,n,ns,draws3,lcholbase(m,param,n),nthetab, nthetaa, nthetag, xb(:,famindex(m)-2:famindex(m)), xa(:,famindex(m)-2:famindex(m)), xg(:,famindex(m)-2:famindex(m)), sa, sg, near(m),0)))
	ELSEIF (MOD(m,20)<15) THEN
	famprob(m)=DLOG(SUM(GHK3(m,param,n,ns,draws3_2,lcholbase(m,param,n),nthetab, nthetaa, nthetag, xb(:,famindex(m)-2:famindex(m)), xa(:,famindex(m)-2:famindex(m)), xg(:,famindex(m)-2:famindex(m)), sa, sg, near(m),0)))
	ELSE
	famprob(m)=DLOG(SUM(GHK3(m,param,n,ns,draws3_3,lcholbase(m,param,n),nthetab, nthetaa, nthetag, xb(:,famindex(m)-2:famindex(m)), xa(:,famindex(m)-2:famindex(m)), xg(:,famindex(m)-2:famindex(m)), sa, sg, near(m),0)))
	ENDIF
CASE(4)

	IF (MOD(m,20)<5) THEN
	famprob(m)=DLOG(SUM(GHK4(m,param,n,ns,draws4seed,lcholbase(m,param,n),nthetab, nthetaa, nthetag, xb(:,famindex(m)-3:famindex(m)), xa(:,famindex(m)-3:famindex(m)), xg(:,famindex(m)-3:famindex(m)), sa, sg, near(m),0)))
	ELSEIF (MOD(m,20)<10) THEN
	famprob(m)=DLOG(SUM(GHK4(m,param,n,ns,draws4    ,lcholbase(m,param,n),nthetab, nthetaa, nthetag, xb(:,famindex(m)-3:famindex(m)), xa(:,famindex(m)-3:famindex(m)), xg(:,famindex(m)-3:famindex(m)), sa, sg, near(m),0)))
	ELSEIF (MOD(m,20)<15) THEN
	famprob(m)=DLOG(SUM(GHK4(m,param,n,ns,draws4_2  ,lcholbase(m,param,n),nthetab, nthetaa, nthetag, xb(:,famindex(m)-3:famindex(m)), xa(:,famindex(m)-3:famindex(m)), xg(:,famindex(m)-3:famindex(m)), sa, sg, near(m),0)))
	ELSE
	famprob(m)=DLOG(SUM(GHK4(m,param,n,ns,draws4_3  ,lcholbase(m,param,n),nthetab, nthetaa, nthetag, xb(:,famindex(m)-3:famindex(m)), xa(:,famindex(m)-3:famindex(m)), xg(:,famindex(m)-3:famindex(m)), sa, sg, near(m),0)))
	
	ENDIF
CASE DEFAULT
    PRINT *,'nchild(m): ', nchild(m)
    PRINT *,'error'
    PAUSE
END SELECT

END DO
!$OMP END PARALLEL DO



lnmlobj=(0-1)*SUM(famprob)


print *,'ln ML objective: ',-lnmlobj

print *,'Time spent for Likelihood evaluation: ',SECNDS(tms)


END FUNCTION lnmlobj






FUNCTION var_cov(param,n)
USE Globvar
! This function computes the VCov matrix of the estimates
! By using the numerical gradient of a function using forward differences
IMPLICIT NONE
INTEGER, INTENT(IN) :: n
REAL(8), INTENT(IN) :: param(n)
REAL(8) :: var_cov(n,n)

REAL(8) :: a(n,n)
REAL(8) :: f0			! Initial value of lnmlobj.
REAL(8) :: grdd(n),dh(n), arg(n,n)
REAL(8) :: ax0(n),xdh(n)
REAL(8) :: dax0(n),one
REAL(8) :: indgrad(n)
INTEGER :: i, j, k

! Set up the gradient calculation
one  = 1.0D0	! This parameter affects the length of dh
grdd = 0.0D0
	! Computation of stepsize (dh) for gradient
ax0  = ABS(param)
dax0 = 1.0D0
WHERE (param.ne.0.0D0) dax0 = param/ax0
dh = (1.0D-6)*(1.0D-2)*one*dax0
WHERE (ax0.gt.(1.0D-2)*one) dh = (1.0D-8)*ax0*dax0
xdh = param+dh

!print *,param
!print *,'dh: ',dh
!print *,xdh
!stop

! Calculate VCOV
a=0.0d0

!!$OMP PARALLEL DO SCHEDULE(STATIC) PRIVATE(f0,grdd,arg,indgrad) 
	! For this part, parallezation does not work because f0, j, arg, etc have inter-dependence.
DO i = 1, nfam

arg = SPREAD(param,DIM=2,NCOPIES=n)

SELECT CASE(nchild(i))
CASE(1)
	f0=DLOG(GHK1(i,param,n))
CASE(2)
	f0=DLOG(SUM(GHK2(i,param,n,ns,draws2seed,lcholbase(i,param,n),nthetab, nthetaa, nthetag, xb(:,famindex(i)-1:famindex(i)), xa(:,famindex(i)-1:famindex(i)), xg(:,famindex(i)-1:famindex(i)), sa, sg, near(i),0)))
CASE(3)
	f0=DLOG(SUM(GHK3(i,param,n,ns,draws3seed,lcholbase(i,param,n),nthetab, nthetaa, nthetag, xb(:,famindex(i)-2:famindex(i)), xa(:,famindex(i)-2:famindex(i)), xg(:,famindex(i)-2:famindex(i)), sa, sg, near(i),0)))
CASE(4)
	f0=DLOG(SUM(GHK4(i,param,n,ns,draws4seed,lcholbase(i,param,n),nthetab, nthetaa, nthetag, xb(:,famindex(i)-3:famindex(i)), xa(:,famindex(i)-3:famindex(i)), xg(:,famindex(i)-3:famindex(i)), sa, sg, near(i),0)))
CASE DEFAULT
    PRINT *,'error'
    PAUSE
END SELECT

		DO j = 1, n
			arg(j,j)=xdh(j)

        SELECT CASE(nchild(i))
        CASE(1)
	        grdd(j)=LOG(GHK1(i,arg(:,j),n))
        CASE(2)
			grdd(j)=LOG(SUM(GHK2(i,arg(:,j),n,ns,draws2seed,lcholbase(i,arg(:,j),n),nthetab, nthetaa, nthetag, xb(:,famindex(i)-1:famindex(i)), xa(:,famindex(i)-1:famindex(i)), xg(:,famindex(i)-1:famindex(i)), sa, sg, near(i),0)))
        CASE(3)
			grdd(j)=LOG(SUM(GHK3(i,arg(:,j),n,ns,draws3seed,lcholbase(i,arg(:,j),n),nthetab, nthetaa, nthetag, xb(:,famindex(i)-2:famindex(i)), xa(:,famindex(i)-2:famindex(i)), xg(:,famindex(i)-2:famindex(i)), sa, sg, near(i),0)))
        CASE(4)
			grdd(j)=LOG(SUM(GHK4(i,arg(:,j),n,ns,draws4seed,lcholbase(i,arg(:,j),n),nthetab, nthetaa, nthetag, xb(:,famindex(i)-3:famindex(i)), xa(:,famindex(i)-3:famindex(i)), xg(:,famindex(i)-3:famindex(i)), sa, sg, near(i),0)))
        CASE DEFAULT
            PRINT *,'error'
            PAUSE
        END SELECT


		END DO
!print *,'f0: ',f0
!print *,'grdd: ',grdd
!print *,'diff: ',grdd-f0
	indgrad= (grdd-f0)/dh
!print *,indgrad
	a=a+outerprod(indgrad,indgrad)
!print *,outerprod(indgrad,indgrad)
END DO
!!$OMP END PARALLEL DO

print *,'check'


DO j=1,ntheta
DO i=1,ntheta
	print *,'a(i,j):  ',i,j, a(i,j)
END DO
END DO

print *,'inverting var cov matrix'
var_cov=MATINV(a)

END FUNCTION var_cov




FUNCTION FAMSOLVE(param,n)
	! This function calculates simulated GHK probabilities for all the location
	! patterns by the number of children to evaluate the goodness-of-fit.
	! A standard way to do this would be to use a conventional frequency simulator.
	! But here it is based on GHK Monte Carlo integration, i.e. make simulation 
	! fraws for each possible location pattern (actually each SPNE that rationalizes
	! the location pattern) to compute its probability, and repeat this for every
	! location pattern.
USE Globvar
USE PROBABILITY
IMPLICIT NONE
INTEGER, INTENT(IN) :: n 
REAL(8), INTENT(IN) :: param(n)

REAL(8) :: FAMSOLVE(0:2**nchildmax-1,4)
REAL(8) :: wgv(0:15)

INTEGER :: i, j, k, l, m 
INTEGER :: s0, s1, s2, s3, s4, stmp
! Specify the array sizes based on the number of players (i.e. possible number of strategy combinations)
REAL(8) :: wg(ns,0:sm4-1)
REAL(8) :: ta(ns), tb(ns)
REAL(8) :: v(ns,0:sm4-1,nchildmax-1)   ! GHK simulation draws constructed from draws()
REAL(8) :: utiltemp, temp
REAL(8) :: ghkval(ns,0:sm4*2-1)		! The calculated GHK value
REAL(8) :: configprob(0:15)
REAL(8) :: utilsub(nchildmax), alpha(nchildmax,3),gamma(nchildmax,2)		! 1 ... oldest child
REAL(8) :: lchol(nchildmax,nchildmax)
INTEGER :: predconfig      ! predicted location configuration




FAMSOLVE=0.0d0

DO m=1,nfam

k=famindex(m)
wgv=0.0d0
wg=0.0d0
v=0.0d0
ghkval=0.0d0
configprob=0.0d0

SELECT CASE(nchild(m))

CASE(1)

utilsub(1:1)=MATMUL(param(1:nthetab),xb(:,k:k))
    IF(nthetaa>=3) THEN
    alpha(1:1,1)  =MATMUL(param(nthetab+1:nthetab+nthetaa-2),xa(:,k:k))
    ELSEIF(nthetaa==1) THEN
    alpha(1:1,1)  =param(nthetab+nthetaa)
    ELSE
    alpha(1:1,1) = 0.0d0
    ENDIF



    FAMSOLVE(1,1)=FAMSOLVE(1,1)+1.0d0-NORMCDF(-utilsub(1)-alpha(1,1),0.0d0,1.0d0)
    FAMSOLVE(0,1)=FAMSOLVE(0,1)+      NORMCDF(-utilsub(1)-alpha(1,1),0.0d0,1.0d0)

    configprob(1)=1.0d0-NORMCDF(-utilsub(1)-alpha(1,1),0.0d0,1.0d0)
    configprob(0)=NORMCDF(-utilsub(1)-alpha(1,1),0.0d0,1.0d0)


CASE(2)

wgv(0:3)=SUM(SUM(GHK2(m,param,n,ns,draws2seed,lcholbase(m,param,n),nthetab, nthetaa, nthetag, xb(:,famindex(m)-1:famindex(m)), xa(:,famindex(m)-1:famindex(m)), xg(:,famindex(m)-1:famindex(m)), sa, sg, near(m),1),3),2)


FAMSOLVE(0:3,2)=FAMSOLVE(0:3,2)+wgv(0:3)
configprob(0:3)=wgv(0:3)


CASE(3)

wgv(0:7)=SUM(SUM(GHK3(m,param,n,ns,draws3seed,lcholbase(m,param,n),nthetab, nthetaa, nthetag, xb(:,famindex(m)-2:famindex(m)), xa(:,famindex(m)-2:famindex(m)), xg(:,famindex(m)-2:famindex(m)), sa, sg, near(m),1),3),2)

FAMSOLVE(0:7,3)=FAMSOLVE(0:7,3)+wgv(0:7)
configprob(0:7)=wgv(0:7)

CASE(4)

wgv(0:15)=SUM(SUM(GHK4(m,param,n,ns,draws4seed,lcholbase(m,param,n),nthetab, nthetaa, nthetag, xb(:,famindex(m)-3:famindex(m)), xa(:,famindex(m)-3:famindex(m)), xg(:,famindex(m)-3:famindex(m)), sa, sg, near(m),1),3),2)


FAMSOLVE(0:15,4)=FAMSOLVE(0:15,4)+wgv(0:15)
configprob(0:15)=wgv(0:15)

CASE DEFAULT
    PRINT *,'nchild(m): ', nchild(m)
    PRINT *,'error'
    PAUSE
END SELECT


temp=0.0d0
DO i=0,15
    IF (configprob(i)>temp) THEN
        temp=configprob(i)
        predconfig=i
    ENDIF
ENDDO
DO l=1,nchild(m)
    predobs(k-nchild(m)+l)=IBITS(predconfig,nchild(m)-l,1)
ENDDO


END DO  ! nfam



END FUNCTION FAMSOLVE









END MODULE Object









PROGRAM MAIN
USE Globvar
USE INIT
USE MINIMIZATION
USE Object
IMPLICIT NONE

REAL(8) :: resultmlv, resultmlse
REAL(8) :: gtol
REAL(4) :: time0, iniadj
INTEGER :: i, j, k, l, m, n, io
INTEGER :: s1, s2, s3, s4, nup
CHARACTER(12) :: horz(3)
CHARACTER(20) :: flname
CHARACTER(80) :: flname2
CHARACTER(20) :: fmtname           ! To accomodate different number of parameters
CHARACTER(50) :: inivalnote
CHARACTER(60) :: setupnote

REAL(8), ALLOCATABLE :: initheta(:)
REAL(8), ALLOCATABLE :: vcov(:,:)
REAL(8), ALLOCATABLE :: se(:)
REAL(8), ALLOCATABLE :: tval(:)
CHARACTER(16), ALLOCATABLE :: vert(:)
REAL(8), ALLOCATABLE :: pred(:,:)

horz(1)='        coef' ; horz(2)='        s.e.' ; horz(3)=' asy t-val'

! Read the setting file
    flname2='setting.prn'
    OPEN(1,file=(flname2))
    READ(1,fmt='(A60)') setupnote
    READ(1,fmt=*) nthetab
    READ(1,fmt=*) nthetaa
    READ(1,fmt=*) nthetag
    READ(1,fmt=*) nthetar
    READ(1,fmt=*) ntheta
    READ(1,fmt=*) nobs
    READ(1,fmt=*) nfam
    READ(1,fmt=*) ns
ALLOCATE (initheta(ntheta))
    READ(1,fmt=*) initheta
ALLOCATE (vert(ntheta))
    vert(1)='cons      '
    READ(1,fmt=*) vert(2:ntheta)
    CLOSE(1)




ALLOCATE (theta(ntheta))
ALLOCATE (vcov(ntheta,ntheta))
ALLOCATE (se(ntheta))
ALLOCATE (tval(ntheta))
ALLOCATE (famid(nobs))
ALLOCATE (famindex(nfam))
ALLOCATE (nchild(nfam))
ALLOCATE (y(nobs))
ALLOCATE (xb(nthetab,nobs))
IF (nthetaa>=3) THEN
	sa=nthetaa-2
	ALLOCATE (xa(nthetaa-2,nobs))
ELSEIF (nthetaa==1) THEN
	sa=1
	ALLOCATE (xa(1,nobs))
ENDIF
IF (nthetag>=2) THEN
	sg=nthetag-1
	ALLOCATE (xg(nthetag-1,nobs))
ELSEIF (nthetag==1) THEN
	sg=1
	ALLOCATE (xg(1,nobs))
ENDIF
ALLOCATE (xr(nthetar,nchildmax,nobs))
ALLOCATE (near(nfam))
ALLOCATE (predobs(nobs))
ALLOCATE (pred(0:2**nchildmax-1,4))


! Set gradient tolerance
gtol=0.000001d0

time0 = SECNDS(0.0)

CALL GETTIM(i, j, k, l)
CALL GETDAT(l, m, n)
flname2=CHAR(ICHAR('0')+(m/10))//CHAR(ICHAR('0')+(m-m/10*10))//CHAR(ICHAR('0')+(n/10))//CHAR(ICHAR('0')+(n-n/10*10))//CHAR(ICHAR('0')+(i/10))//CHAR(ICHAR('0')+(i-i/10*10))//CHAR(ICHAR('0')+(j/10))//CHAR(ICHAR('0')+(j-j/10*10))//CHAR(ICHAR('0')+(k/10))//CHAR(ICHAR('0')+(k-k/10*10))//'.prn'

OPEN(4,file=(flname2),STATUS='REPLACE')
WRITE(4,fmt='(A20,4x,A60)') 'File Name: ', flname2
WRITE(4,fmt='(A20,4x,A60)') 'Estimation Setting: ', setupnote
WRITE(4,fmt=*) '# of families     :  ', nfam
WRITE(4,fmt=*) '# of observations:  ', nobs
WRITE(4,fmt=*) '# of simuln draws  :  ', ns
WRITE(4,fmt=*) ' '

theta=initheta
numeval=0

! Load the data set
CALL READDATA(nobs)


ALLOCATE (draws2seed(ns,0:sm2-1,2))
ALLOCATE (draws3seed(ns,0:sm3-1,3))
ALLOCATE (draws4seed(ns,0:sm4-1,4))
CALL RANDOM_SEED(PUT=(/3000/))
CALL RANDOM_NUMBER (draws2seed)
CALL RANDOM_NUMBER (draws3seed)
CALL RANDOM_NUMBER (draws4seed)
draws2seed(ns/2+1:ns,:,:)=1-draws2seed(1:ns/2,:,:)        ! This line makes antithetics
draws3seed(ns/2+1:ns,:,:)=1-draws3seed(1:ns/2,:,:)        ! This line makes antithetics
draws4seed(ns/2+1:ns,:,:)=1-draws4seed(1:ns/2,:,:)        ! This line makes antithetics


! the following lines compute the estimates using a Quasi-Newton method
! with a numerical gradient

CALL dfpmin(theta,ntheta,gtol,lnmlobj)
objsaved=lnmlobj(theta,ntheta,0)
objsaved=-objsaved


print *,'check4-1! - 1st round Minimization Finished!'
print *,'objsaved : ',objsaved

print *,'Calculating variance-covariance matrix...'

vcov=var_cov(theta, ntheta)

print *,'check4-2! - vcov obtained!'
!	print *,vcov


DO i=1,ntheta
	print *,'vcov(i,i):  ',i, vcov(i,i)
END DO

DO i=1,ntheta
	se(i)=DSQRT(vcov(i,i))
	tval(i)=theta(i)/se(i)
END DO




WRITE(4,fmt=*) '# of parameters:  ', ntheta
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(10x,3A16)') (horz(i),i=1,3)
WRITE(4,fmt=*) ' '
DO i=1,ntheta
	WRITE(4,fmt='(A16,3F16.10)') vert(i), theta(i), se(i), tval(i)
ENDDO

pred=0
WRITE(4,fmt=*) ' '
WRITE(4,fmt=*) 'Location patterns: Data'
WRITE(4,fmt='(A8,10A8)') '  config',' 1 child','     (%)',' 2 child','     (%)',' 3 child','     (%)',' 4 child','     (%)'
DO i=0,2**nchildmax-1
    DO j=1,nfam
        IF (near(j)==i) THEN
            pred(i,nchild(j))=pred(i,nchild(j))+1
        ENDIF
    ENDDO
ENDDO
DO i=0,2**nchildmax-1
	WRITE(4,fmt='(4X,4I1,4(F8.1,F8.2))') IBITS(i,3,1),IBITS(i,2,1),IBITS(i,1,1),IBITS(i,0,1), (pred(i,k),(pred(i,k)/SUM(pred(:,k))*100.0d0),k=1,4)
ENDDO
WRITE(4,fmt='(A8,4(F8.1,F8.2))') '   total', ((SUM(pred(:,j)),100.0),j=1,4)

pred=0
WRITE(4,fmt=*) ' '
WRITE(4,fmt=*) 'Location patterns: Predicted'
WRITE(4,fmt='(A8,10A8)') '  config',' 1 child','     (%)',' 2 child','     (%)',' 3 child','     (%)',' 4 child','     (%)'
    pred(:,:)=FAMSOLVE(theta,ntheta)
DO i=0,2**nchildmax-1
	WRITE(4,fmt='(4X,4I1,4(F8.1,F8.2))') IBITS(i,3,1),IBITS(i,2,1),IBITS(i,1,1),IBITS(i,0,1), (pred(i,k),(pred(i,k)/SUM(pred(:,k))*100.0d0),k=1,4)
ENDDO
WRITE(4,fmt='(A8,4(F8.1,F8.2))') '   total', ((SUM(pred(:,j)),100.0),j=1,4)



WRITE(4,fmt=*) ' '
n=0
DO m=1,nfam
DO j=1,nchild(m)
    IF (predobs(famindex(m)-nchild(m)+j)==IBITS(near(m),nchild(m)-j,1)) n=n+1
ENDDO
ENDDO
WRITE(4,fmt=*) '% of observations correctly predicted:  ', n*1.0d0/nobs*100
n=0
DO m=1,nfam
    k=0
    DO j=1,nchild(m)
        IF (predobs(famindex(m)-nchild(m)+j)==IBITS(near(m),nchild(m)-j,1)) k=k+1
    ENDDO
    IF (k==nchild(m)) n=n+1
ENDDO
WRITE(4,fmt=*) '% of families correctly predicted:  ', n*1.0d0/nfam*100
n=0; l=0
DO m=1,nfam
    IF (nchild(m)==1) THEN
        IF (predobs(famindex(m))==IBITS(near(m),0,1)) n=n+1
    l=l+1
    ENDIF
ENDDO
WRITE(4,fmt=*) '% of 1 child families correctly predicted:  ', n*1.0d0/l*100
n=0; l=0
DO m=1,nfam
    IF (nchild(m)==2) THEN
    k=0
    DO j=1,nchild(m)
        IF (predobs(famindex(m)-nchild(m)+j)==IBITS(near(m),nchild(m)-j,1)) k=k+1
    ENDDO
    IF (k==nchild(m)) n=n+1
    l=l+1
    ENDIF
ENDDO
WRITE(4,fmt=*) '% of 2 child families correctly predicted:  ', n*1.0d0/l*100
n=0; l=0
DO m=1,nfam
    IF (nchild(m)==3) THEN
    k=0
    DO j=1,nchild(m)
        IF (predobs(famindex(m)-nchild(m)+j)==IBITS(near(m),nchild(m)-j,1)) k=k+1
    ENDDO
    IF (k==nchild(m)) n=n+1
    l=l+1
    ENDIF
ENDDO
WRITE(4,fmt=*) '% of 3 child families correctly predicted:  ', n*1.0d0/l*100
n=0; l=0
DO m=1,nfam
    IF (nchild(m)==4) THEN
    k=0
    DO j=1,nchild(m)
        IF (predobs(famindex(m)-nchild(m)+j)==IBITS(near(m),nchild(m)-j,1)) k=k+1
    ENDDO
    IF (k==nchild(m)) n=n+1
    l=l+1
    ENDIF
ENDDO
WRITE(4,fmt=*) '% of 4 child families correctly predicted:  ', n*1.0d0/l*100

WRITE(4,fmt=*) ' '
WRITE(4,fmt=*) ' '
WRITE(4,fmt=*) 'ML objective:  ', objsaved
WRITE(4,fmt=*) '# of objective function evaluations:  ', numeval
WRITE(4,fmt=*) ' '
WRITE(4,fmt=*) 'run time(seconds):  ', SECNDS(time0)
WRITE(4,fmt=*) ' '
WRITE(4,fmt=*) ' '
WRITE(4,fmt=*) 'Initial Values: '
DO i=1,ntheta
WRITE(4,fmt='(F16.10)') initheta(i)
END DO
WRITE(4,fmt=*) ' '

CLOSE(4)

! estimated parameters
OPEN(1,file='param1.out',STATUS='REPLACE')
OPEN(2,file='param2.out',STATUS='REPLACE')
flname='('//CHAR((nthetab)/100+48)//CHAR(MOD(nthetab,100)/10+48)//CHAR(MOD(nthetab,10)+48)//'F20.14)'
WRITE(1,fmt=flname) theta(1:nthetab)
IF (ntheta>nthetab) THEN
flname='('//CHAR((ntheta-nthetab)/100+48)//CHAR(MOD(ntheta-nthetab,100)/10+48)//CHAR(MOD(ntheta-nthetab,10)+48)//'F20.14)'
WRITE(2,fmt=flname) theta(nthetab+1:ntheta)
ENDIF
CLOSE(1)
CLOSE(2)

END PROGRAM MAIN
