! This function conducts a GHK type Monte Carlo integration for each family for given decision rule.
! There are three main functions: GHK2, GHK3, GHK4.
	! GHK2, GHK3, GHK4 are for two-child, three-child, and four-child famlies.
MODULE GHK
	USE Globvar
	IMPLICIT NONE

CONTAINS


FUNCTION GHK2(param,n,ns,draws,lchol,a,near0) RESULT(ghkval)
USE PROBABILITY
IMPLICIT NONE
INTEGER, INTENT(IN) :: n, ns
REAL(8), INTENT(IN) :: param(n)
REAL(8), INTENT(IN) :: draws(ns,2)
REAL(8), INTENT(IN) :: lchol(2,2)
REAL(8), INTENT(IN) :: a(2)			! a(1) ... decision rule of the oldest child. Choose near if a(1)+e>0.
INTEGER, INTENT(IN) :: near0
REAL(8) :: ghkval
REAL(8) :: wg(ns)
REAL(8) :: v(ns,2:2)   ! GHK simulation draws constructed from draws()
INTEGER :: i, j, k, l
REAL(8) :: ta(ns), tb(ns)
REAL(8) :: temp

	wg=0.000d0
	v=0.000d0

	! 2nd child
	IF (IBITS(near0,0,1)==1) THEN	! 2nd child near
		ta=NORMCDF(-a(2),0.0d0,1.0d0)
		wg(:)=1.0-ta
		DO j=1, ns
			IF (ta(j)>=0.99999999d0) THEN 
	!	        PRINT *,'Warning: a is too large.'
				v(j,2)=HUGE(v(j,2))
			ELSE
        		v(j,2)=INVNORMCDF((ta(j)*(1-draws(j,2))+draws(j,2)),0.0d0,1.0d0)
			ENDIF
		END DO
	ELSE							! 2nd child far
		tb=NORMCDF(-a(2),0.0d0,1.0d0)
		wg(:)=tb
		DO j=1, ns
		   	IF (tb(j)<=0.00000001d0) THEN 
		!		PRINT *,'Warning: b is too small.'
				v(j,2)=-HUGE(v(j,2))
			ELSE
        		v(j,2)=INVNORMCDF((tb(j)*draws(j,2)),0.0d0,1.0d0)
			ENDIF
		END DO
	ENDIF
	
	! 1st child
	temp=1.0d0/lchol(2,2)
	IF (IBITS(near0,1,1)==1) THEN	! 1st child near
		DO j=1, ns
			ta(j)=NORMCDF((-a(1)-lchol(2,1)*v(j,2))*temp,0.0d0,1.0d0)
		END DO         ! Simulation loop
		wg(:)=wg(:)*(1.0-ta)
	ELSE							! 1st child far
		DO j=1, ns
			tb(j)=NORMCDF((-a(1)-lchol(2,1)*v(j,2))*temp,0.0d0,1.0d0)
		END DO         ! Simulation loop
		wg(:)=wg(:)*tb
	ENDIF
	
	ghkval=SUM(wg(:))/ns
	
END FUNCTION GHK2


FUNCTION GHK3(param,n,ns,draws,lchol,a,near0) RESULT(ghkval)
USE PROBABILITY
IMPLICIT NONE
INTEGER, INTENT(IN) :: n, ns
REAL(8), INTENT(IN) :: param(n)
REAL(8), INTENT(IN) :: draws(ns,3)
REAL(8), INTENT(IN) :: lchol(3,3)
REAL(8), INTENT(IN) :: a(3)
INTEGER, INTENT(IN) :: near0
REAL(8) :: ghkval
REAL(8) :: wg(ns)
REAL(8) :: v(ns,2:3)   ! GHK simulation draws constructed from draws()
INTEGER :: i, j, k, l
REAL(8) :: ta(ns), tb(ns)
REAL(8) :: temp

	wg=0.000d0
	v=0.000d0

	! 3rd child
	IF (IBITS(near0,0,1)==1) THEN	! 3rd child near
		ta=NORMCDF(-a(3),0.0d0,1.0d0)
		wg(:)=1.0d0-ta
		DO j=1, ns
			IF (ta(j)>=0.99999999d0) THEN 
	!	        PRINT *,'Warning: a is too large.'
				v(j,3)=HUGE(v(j,3))
			ELSE
        		v(j,3)=INVNORMCDF((ta(j)*(1-draws(j,3))+draws(j,3)),0.0d0,1.0d0)
			ENDIF
		END DO
	ELSE							! 3rd child far
		tb=NORMCDF(-a(3),0.0d0,1.0d0)
		wg(:)=tb
		DO j=1, ns
		   	IF (tb(j)<=0.00000001d0) THEN 
		!		PRINT *,'Warning: b is too small.'
				v(j,3)=-HUGE(v(j,3))
			ELSE
        		v(j,3)=INVNORMCDF((tb(j)*draws(j,3)),0.0d0,1.0d0)
			ENDIF
		END DO
	ENDIF
	
	! 2nd child
	temp=1.0d0/lchol(2,2)
	IF (IBITS(near0,1,1)==1) THEN	! 2nd child near
		DO j=1, ns
			ta(j)=NORMCDF((-a(2)-DOT_PRODUCT(lchol(2,1:1),v(j,3:3)))*temp,0.0d0,1.0d0)
			IF (ta(j)>=0.99999999d0) THEN 
	!	        PRINT *,'Warning: a is too large.'
				v(j,2)=HUGE(v(j,2))
			ELSE
        		v(j,2)=INVNORMCDF((ta(j)*(1-draws(j,2))+draws(j,2)),0.0d0,1.0d0)
			ENDIF
		END DO
		wg(:)=wg(:)*(1.0d0-ta(:))
	ELSE							! 2nd child far
		DO j=1, ns
			tb(j)=NORMCDF((-a(2)-DOT_PRODUCT(lchol(2,1:1),v(j,3:3)))*temp,0.0d0,1.0d0)
			IF (tb(j)<=0.00000001d0) THEN 
		!		PRINT *,'Warning: b is too small.'
				v(j,2)=-HUGE(v(j,2))
			ELSE
        		v(j,2)=INVNORMCDF((tb(j)*draws(j,2)),0.0d0,1.0d0)
			ENDIF
		ENDDO
		wg(:)=wg(:)*tb(:)
	ENDIF
	
	! 1st child
	temp=1.0d0/lchol(3,3)
	IF (IBITS(near0,2,1)==1) THEN	! 1st child near
		DO j=1, ns
			ta(j)=NORMCDF((-a(1)-DOT_PRODUCT(lchol(3,1:2),v(j,3:2:-1)))*temp,0.0d0,1.0d0)
		END DO         ! Simulation loop
		wg(:)=wg(:)*(1.0d0-ta)
	ELSE							! 1st child far
		DO j=1, ns
			tb(j)=NORMCDF((-a(1)-DOT_PRODUCT(lchol(3,1:2),v(j,3:2:-1)))*temp,0.0d0,1.0d0)
		END DO         ! Simulation loop
		wg(:)=wg(:)*tb
	ENDIF
	
	ghkval=SUM(wg(:))/ns
	
END FUNCTION GHK3




FUNCTION GHK4(param,n,ns,draws,lchol,a,near0) RESULT(ghkval)
USE PROBABILITY
IMPLICIT NONE
INTEGER, INTENT(IN) :: n, ns
REAL(8), INTENT(IN) :: param(n)
REAL(8), INTENT(IN) :: draws(ns,4)
REAL(8), INTENT(IN) :: lchol(4,4)
REAL(8), INTENT(IN) :: a(4)
INTEGER, INTENT(IN) :: near0
REAL(8) :: ghkval
REAL(8) :: wg(ns)
REAL(8) :: v(ns,2:4)   ! GHK simulation draws constructed from draws()
INTEGER :: i, j, k, l
REAL(8) :: ta(ns), tb(ns)
REAL(8) :: temp

	wg=0.000d0
	v=0.000d0

	! 4th child
	IF (IBITS(near0,0,1)==1) THEN	! 4th child near
		ta=NORMCDF(-a(4),0.0d0,1.0d0)
		wg(:)=1.0d0-ta
		DO j=1, ns
			IF (ta(j)>=0.99999999d0) THEN 
	!	        PRINT *,'Warning: a is too large.'
				v(j,4)=HUGE(v(j,4))
			ELSE
        		v(j,4)=INVNORMCDF((ta(j)*(1-draws(j,4))+draws(j,4)),0.0d0,1.0d0)
			ENDIF
		END DO
	ELSE							! 4th child far
		tb=NORMCDF(-a(4),0.0d0,1.0d0)
		wg(:)=tb
		DO j=1, ns
		   	IF (tb(j)<=0.00000001d0) THEN 
		!		PRINT *,'Warning: b is too small.'
				v(j,4)=-HUGE(v(j,4))
			ELSE
        		v(j,4)=INVNORMCDF((tb(j)*draws(j,4)),0.0d0,1.0d0)
			ENDIF
		END DO
	ENDIF
	
	! 3rd child
	temp=1.0d0/lchol(2,2)
	IF (IBITS(near0,1,1)==1) THEN	! 3rd child near
		DO j=1, ns
			ta(j)=NORMCDF((-a(3)-DOT_PRODUCT(lchol(2,1:1),v(j,4:4)))*temp,0.0d0,1.0d0)
			IF (ta(j)>=0.99999999d0) THEN 
	!	        PRINT *,'Warning: a is too large.'
				v(j,3)=HUGE(v(j,3))
			ELSE
        		v(j,3)=INVNORMCDF((ta(j)*(1-draws(j,3))+draws(j,3)),0.0d0,1.0d0)
			ENDIF
		END DO
		wg(:)=wg(:)*(1.0d0-ta(:))
	ELSE							! 3rd child far
		DO j=1, ns
			tb(j)=NORMCDF((-a(3)-DOT_PRODUCT(lchol(2,1:1),v(j,4:4)))*temp,0.0d0,1.0d0)
			IF (tb(j)<=0.00000001d0) THEN 
		!		PRINT *,'Warning: b is too small.'
				v(j,3)=-HUGE(v(j,3))
			ELSE
        		v(j,3)=INVNORMCDF((tb(j)*draws(j,3)),0.0d0,1.0d0)
			ENDIF
		ENDDO
		wg(:)=wg(:)*tb(:)
	ENDIF
	
	! 2nd child
	temp=1.0d0/lchol(3,3)
	IF (IBITS(near0,2,1)==1) THEN	! 2nd child near
		DO j=1, ns
			ta(j)=NORMCDF((-a(2)-DOT_PRODUCT(lchol(3,1:2),v(j,4:3:-1)))*temp,0.0d0,1.0d0)
			IF (ta(j)>=0.99999999d0) THEN 
	!	        PRINT *,'Warning: a is too large.'
				v(j,2)=HUGE(v(j,2))
			ELSE
        		v(j,2)=INVNORMCDF((ta(j)*(1-draws(j,2))+draws(j,2)),0.0d0,1.0d0)
			ENDIF
		END DO
		wg(:)=wg(:)*(1.0d0-ta(:))
	ELSE							! 2nd child far
		DO j=1, ns
			tb(j)=NORMCDF((-a(2)-DOT_PRODUCT(lchol(3,1:2),v(j,4:3:-1)))*temp,0.0d0,1.0d0)
			IF (tb(j)<=0.00000001d0) THEN 
		!		PRINT *,'Warning: b is too small.'
				v(j,2)=-HUGE(v(j,2))
			ELSE
        		v(j,2)=INVNORMCDF((tb(j)*draws(j,2)),0.0d0,1.0d0)
			ENDIF
		ENDDO
		wg(:)=wg(:)*tb(:)
	ENDIF
	
	! 1st child
	temp=1.0d0/lchol(4,4)
	IF (IBITS(near0,3,1)==1) THEN	! 1st child near
		DO j=1, ns
			ta(j)=NORMCDF((-a(1)-DOT_PRODUCT(lchol(4,1:3),v(j,4:2:-1)))*temp,0.0d0,1.0d0)
		END DO         ! Simulation loop
		wg(:)=wg(:)*(1.0d0-ta)
	ELSE							! 1st child far
		DO j=1, ns
			tb(j)=NORMCDF((-a(1)-DOT_PRODUCT(lchol(4,1:3),v(j,4:2:-1)))*temp,0.0d0,1.0d0)
		END DO         ! Simulation loop
		wg(:)=wg(:)*tb
	ENDIF
	
	ghkval=SUM(wg(:))/ns
	
END FUNCTION GHK4




END MODULE GHK
