MODULE Globvar
IMPLICIT NONE
! This module declares the variables that are going to be used throughout the program.
! To avoid stack overflow error --> Stack Reserve: 50,000,000

INTEGER, PARAMETER :: nchildmax=4  ! max number of children
INTEGER, PARAMETER :: sm2=2**(1+1)  ! necessary amount for strategy patterns
INTEGER, PARAMETER :: sm3=2**(1+1)*2**(2+1)  ! necessary amount for strategy patterns
INTEGER, PARAMETER :: sm4=2**(1+1)*2**(2+1)*2**(3+1)  ! necessary amount for strategy patterns
INTEGER :: nobs
INTEGER :: nfam
INTEGER :: ns              ! number of simulation draws
REAL(8) :: objsaved
INTEGER :: numeval
INTEGER :: ntheta
INTEGER :: nthetab     ! # of parameters in beta term
INTEGER :: nthetaa     ! # of parameters in alpha term
INTEGER :: nthetag     ! # of parameters in gamma term
INTEGER :: nthetar     ! # of parameters in rho term
INTEGER, ALLOCATABLE :: famid(:)   ! famid(nobs)
INTEGER, ALLOCATABLE :: nchild(:), famindex(:)   ! nchild(nfam), famindex(nfam)
REAL(8), ALLOCATABLE :: theta(:) ! theta(ntheta)
REAL(8), ALLOCATABLE :: xb(:,:)   ! xb(nobs,nthetab)
REAL(8), ALLOCATABLE :: xa(:,:)   ! xa(nobs,nthetaa)
REAL(8), ALLOCATABLE :: xg(:,:)   ! xg(nobs,nthetag)
REAL(8), ALLOCATABLE :: xr(:,:,:)   ! xr(nobs,nthetar,nchildmax)
INTEGER, ALLOCATABLE :: y(:)     ! y(nobs)
INTEGER, ALLOCATABLE :: near(:)   ! near(nfam)
INTEGER, ALLOCATABLE :: predobs(:)  !nearpred(nobs)

REAL(8), ALLOCATABLE :: draws2seed(:,:)    ! nchild=2. draws(ns,2:2)
REAL(8), ALLOCATABLE :: draws3seed(:,:)    ! nchild=3. draws(ns,2:3)
REAL(8), ALLOCATABLE :: draws4seed(:,:)    ! nchild=4. draws(ns,2:4)

CHARACTER(10) :: path                          ! Specify the data directory. To be read from setting.prn

END MODULE Globvar






MODULE INIT
	USE Globvar
	IMPLICIT NONE

CONTAINS

	SUBROUTINE READDATA(nobs)
		USE Globvar
		IMPLICIT NONE
		INTEGER, INTENT(IN) :: nobs
		INTEGER :: i,j,k,m,io

        CHARACTER(60) :: readfile_x      ! File names to be read

readfile_x='data.raw'

xb=0.0d0; y=0.0d0
xb(:,1)=1.00000000d0; xa(:,1)=1.00000000d0
IF (nthetag>=1) xg(:,1)=1.00000000d0
xr(:,1,:)=1.00000000d0

print *,'Reading Data ... '

	OPEN(1,file=(readfile_x),status='old')

    IF (nthetaa>3 .and. nthetar>1 .and. nthetag>2) THEN
        DO i = 1, nobs
	    READ(1,fmt=*,iostat=io) y(i), famid(i), xb(i,2:), xa(i,2:), xg(i,2:), xr(i,2:nthetar,1), xr(i,2:nthetar,2), xr(i,2:nthetar,3), xr(i,2:nthetar,4)
        IF (io<0) EXIT
        END DO
    ELSE IF (nthetaa>3 .and. nthetar>1) THEN
        DO i = 1, nobs
	    READ(1,fmt=*,iostat=io) y(i), famid(i), xb(i,2:), xa(i,2:), xr(i,2:nthetar,1), xr(i,2:nthetar,2), xr(i,2:nthetar,3), xr(i,2:nthetar,4)
        IF (io<0) EXIT
        END DO
    ELSE IF (nthetaa>1) THEN
        DO i = 1, nobs
	    READ(1,fmt=*,iostat=io) y(i), famid(i), xb(i,2:), xa(i,2:)
        IF (io<0) EXIT
        END DO
    ELSE
        DO i = 1, nobs
	    READ(1,fmt=*,iostat=io) y(i), famid(i), xb(i,2:)
        IF (io<0) EXIT
        END DO
    ENDIF
    CLOSE(1)


nchild=0
j=1
nchild(1)=1
DO i=2,nobs
	IF (famid(i)/=famid(i-1)) THEN
		j=j+1
	ENDIF
		nchild(j)=nchild(j)+1
ENDDO

famindex(1)=nchild(1)
DO i=2,nfam
	famindex(i)=famindex(i-1)+nchild(i)
ENDDO


DO m=1,nfam
near(m)=0
DO i=1,nchild(m)
near(m)=near(m)*2
near(m)=near(m)+y(famindex(m)-nchild(m)+i)
END DO
END DO



print *,'Reading Data ... Finished '

	END SUBROUTINE READDATA


END MODULE INIT

MODULE Object
	USE Globvar
	USE MATRIX
    USE PROBABILITY
	IMPLICIT NONE

CONTAINS


FUNCTION lnmlobj(param,n,gradflag)	! This function computes the ML objective function
USE Globvar
IMPLICIT NONE
INTEGER, INTENT(IN) :: n 
INTEGER, INTENT(IN) :: gradflag
REAL(8), INTENT(IN) :: param(n)
REAL(8) :: mlobj
REAL(8) :: lnmlobj
REAL(8) :: famprob(nfam)			! The family-level likelihood
INTEGER :: i, j, k, m
REAL(4) :: tms
REAL(8) :: draws2(ns,2:2)
REAL(8) :: draws3(ns,2:3)
REAL(8) :: draws4(ns,2:4)
REAL(8) :: draws2_2(ns,2:2)
REAL(8) :: draws3_2(ns,2:3)
REAL(8) :: draws4_2(ns,2:4)
REAL(8) :: draws2_3(ns,2:2)
REAL(8) :: draws3_3(ns,2:3)
REAL(8) :: draws4_3(ns,2:4)

tms = SECNDS(0.0)

numeval=numeval+1

print ('(a16,I8)'),'mle evaluation: ',numeval


! This section is to elaborate the random draws (without this, every household uses the same set of 200 simulation draws)
! This is supposed to help convergence, though it doesn't affect the result much.
draws2=draws2seed+0.13
WHERE(draws2>1.0d0) draws2=draws2-1.0d0
draws3=draws3seed+0.13
WHERE(draws3>1.0d0) draws3=draws3-1.0d0
draws4=draws4seed+0.13
WHERE(draws4>1.0d0) draws4=draws4-1.0d0
    draws2=CSHIFT(draws2,13)
    draws3=CSHIFT(draws3,13)
    draws4=CSHIFT(draws4,13)
draws2_2=draws2+0.13
WHERE(draws2_2>1.0d0) draws2_2=draws2_2-1.0d0
draws3_2=draws3+0.13
WHERE(draws3_2>1.0d0) draws3_2=draws3_2-1.0d0
draws4_2=draws4+0.13
WHERE(draws4_2>1.0d0) draws4_2=draws4_2-1.0d0
    draws2_2=CSHIFT(draws2_2,13)
    draws3_2=CSHIFT(draws3_2,13)
    draws4_2=CSHIFT(draws4_2,13)
draws2_3=draws2_2+0.13
WHERE(draws2_3>1.0d0) draws2_3=draws2_3-1.0d0
draws3_3=draws3_2+0.13
WHERE(draws3_3>1.0d0) draws3_3=draws3_3-1.0d0
draws4_3=draws4_2+0.13
WHERE(draws4_3>1.0d0) draws4_3=draws4_3-1.0d0
    draws2_3=CSHIFT(draws2_3,13)
    draws3_3=CSHIFT(draws3_3,13)
    draws4_3=CSHIFT(draws4_3,13)


!$OMP PARALLEL DO SCHEDULE(DYNAMIC)
DO m = 1, nfam

!WRITE(4,fmt=*) m, nchild(m)

SELECT CASE(nchild(m))
CASE(1)
	famprob(m)=DLOG(GHK1(m,param,n))
CASE(2)
    IF (MOD(m,20)<5) THEN
	famprob(m)=DLOG(GHK2(m,param,n,draws2seed))
	ELSEIF (MOD(m,20)<10) THEN
	famprob(m)=DLOG(GHK2(m,param,n,draws2))
	ELSEIF (MOD(m,20)<15) THEN
	famprob(m)=DLOG(GHK2(m,param,n,draws2_2))
	ELSE
	famprob(m)=DLOG(GHK2(m,param,n,draws2_3))
	ENDIF
CASE(3)
    IF (MOD(m,20)<5) THEN
	famprob(m)=DLOG(GHK3(m,param,n,draws3seed))
	ELSEIF (MOD(m,20)<10) THEN
	famprob(m)=DLOG(GHK3(m,param,n,draws3))
	ELSEIF (MOD(m,20)<15) THEN
	famprob(m)=DLOG(GHK3(m,param,n,draws3_2))
	ELSE
	famprob(m)=DLOG(GHK3(m,param,n,draws3_3))
	ENDIF
CASE(4)
    IF (MOD(m,20)<5) THEN
	famprob(m)=DLOG(GHK4(m,param,n,draws4seed))
	ELSEIF (MOD(m,20)<10) THEN
	famprob(m)=DLOG(GHK4(m,param,n,draws4))
	ELSEIF (MOD(m,20)<15) THEN
	famprob(m)=DLOG(GHK4(m,param,n,draws4_2))
	ELSE
	famprob(m)=DLOG(GHK4(m,param,n,draws4_3))
	ENDIF
CASE DEFAULT
    PRINT *,'nchild(m): ', nchild(m)
    PRINT *,'error'
    PAUSE
END SELECT

END DO
!$OMP END PARALLEL DO



lnmlobj=(0-1)*SUM(famprob)


print *,'ln ML objective: ',-lnmlobj
!IF (gradflag==0) objsaved=mlobj

print *,'Time spent for Likelihood evaluation: ',SECNDS(tms)

END FUNCTION lnmlobj





FUNCTION GHK1(m,param,n) RESULT(ghkval)
	! This function calculates the GHK simulator for each market*year
USE Globvar
USE PROBABILITY
IMPLICIT NONE
INTEGER, INTENT(IN) :: m			! family id
INTEGER, INTENT(IN) :: n 
REAL(8), INTENT(IN) :: param(n)
INTEGER :: i, j, k, l
REAL(8) :: ghkval		! The calculated GHK value
REAL(8) :: utilsub, alpha

k=famindex(m)

! PRINT *, 'GHK1-in'

	utilsub=DOT_PRODUCT(xb(k,:),param(1:nthetab))
    IF(nthetaa>=3) THEN
    alpha  = EXP( DOT_PRODUCT(xa(k,:),param(nthetab+1:nthetab+nthetaa-2)) )
    ELSEIF(nthetaa==1) THEN
    alpha  =param(nthetab+nthetaa)
    ELSE
    alpha = 0.0d0
    ENDIF

	IF (IBITS(near(m),0,1)==1) THEN
    ghkval=1.0d0-NORMCDF(-utilsub-alpha,0.0d0,1.0d0)
	ELSE
    ghkval=NORMCDF(-utilsub-alpha,0.0d0,1.0d0)
    ENDIF


IF (ghkval<=0.0) THEN
!	PRINT *,'Error found!'
!	PRINT *,'GHK Value: ', ghkval
!	PRINT *,'nmkt: ', nmkt
	ghkval=TINY(ghkval)
!	STOP
ENDIF

! PRINT *, 'GHK1-out'

END FUNCTION GHK1




FUNCTION GHK2(m,param,n,draws) RESULT(ghkval)
	! This function calculates the GHK simulator for each market*year
USE Globvar
USE PROBABILITY
IMPLICIT NONE
INTEGER, INTENT(IN) :: m			! family id
INTEGER, INTENT(IN) :: n 
REAL(8), INTENT(IN) :: param(n)
REAL(8), INTENT(IN) :: draws(ns,2:2)
INTEGER :: i, j, k, l
! Specify the array sizes based on the number of players (i.e. possible number of strategy combinations)
REAL(8) :: wg(ns)
REAL(8) :: ta(ns), tb(ns)
REAL(8) :: v(ns,nchild(m)-1)   ! GHK simulation draws constructed from draws(). v(ns,1) ... the youngest child
REAL(8) :: futil(1:nchild(m)*nchild(m)-1)  ! futil(1) ... family total utility when only last child lives near
REAL(8) :: ghkval		! The calculated GHK value
REAL(8) :: utilsub(2), alpha(2,3), gamma(2,2)		! 1 ... oldest child
REAL(8) :: lchol(2,2)



wg=0.000d0
v=0.000d0
ghkval=0.000d0

k=famindex(m)

!PRINT *, 'GHK2-in'

	utilsub(1:2)=MATMUL(xb(k-1:k,:),param(1:nthetab))
	IF (nthetaa>=3) THEN
    alpha(1:2,1) = MATMUL(xa(k-1:k,:),param(nthetab+1:nthetab+nthetaa-2))
    alpha(1:2,2) = param(nthetab+nthetaa-1)
    alpha(1:2,3) = param(nthetab+nthetaa)
	ELSEIF (nthetaa==1) THEN
	alpha(1:2,1)=param(nthetab+nthetaa)
	alpha(1:2,2:3)=0.0d0
	ELSE
	alpha=0.0d0
	ENDIF
	IF (nthetag>=2) THEN
	gamma(1:2,1) =MATMUL(xg(k-1:k,:),param(nthetab+nthetaa+1:nthetab+nthetaa+nthetag-1))
	gamma(1:2,2) =param(nthetab+nthetaa+nthetag)
	ELSEIF (nthetag==1) THEN
	gamma(1:2,1)=param(nthetab+nthetaa+nthetag)
	gamma(1:2,2)=0.0d0
	ELSE
	gamma=0.0d0
	ENDIF

lchol=lcholbase(m,param,n)

futil(1)=utilsub(2)+EXP(alpha(1,1))+EXP(alpha(2,1))
futil(2)=utilsub(1)+EXP(alpha(1,1))+EXP(alpha(2,1))
futil(3)=utilsub(1)+utilsub(2) + EXP( alpha(1,1)+alpha(1,2) ) + EXP( alpha(2,1)+alpha(2,2) ) + gamma(1,1)+gamma(2,1)

SELECT CASE(near(m))
CASE(0)
    ! condition: 0 > u1, u2, u3
    ! condition: 0 > futil1 + e1    ==> e1 < - futil1
    ! condition: 0 > futil2 + e2    ==> e2 < - futil2
    ! condition: 0 > futil3 + e1+e2 ==> e2 < - futil3 - e1

    tb=NORMCDF(-futil(1),0.0d0,1.0d0)
    wg(:)=tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,1)=-HUGE(v(j,1))
   	        ELSE
   	        v(j,1)=INVNORMCDF((tb(j)*draws(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( &
           MIN(-futil(2),-futil(3)-v(j,1)) &
           -DOT_PRODUCT(lchol(nchild(m),1:nchild(m)-1),v(j,1:1)) &
        ) &
        /lchol(nchild(m),nchild(m)),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb

CASE(1)
    ! condition: u1 > 0, u2, u3
    ! condition: futil1 + e1 > 0              ==> e1 > - futil1
    ! condition: futil1 + e1 > futil2 + e2    ==> e2 < futil1 - futil2 + e1
    ! condition: futil1 + e1 > futil3 + e1+e2 ==> e2 < futil1 - futil3

    ta=NORMCDF(-futil(1),0.0d0,1.0d0)
    wg(:)=1.0-ta
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,1)=HUGE(v(j,1))
            ELSE
            	v(j,1)=INVNORMCDF((ta(j)*(1-draws(j,2))+draws(j,2)),0.0d0,1.0d0)
	        ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( &
           MIN(futil(1)-futil(3),futil(1)-futil(2)+v(j,1)) &
           -DOT_PRODUCT(lchol(nchild(m),1:nchild(m)-1),v(j,1:1)) &
        ) &
        /lchol(nchild(m),nchild(m)),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb


CASE(2)
    ! condition: u2 > 0, u1, u3
    ! condition: futil2 + e2 > 0              ==> e2 > - futil2
    ! condition: futil2 + e2 > futil1 + e1    ==> e2 > futil1 - futil2 + e1
    ! condition: futil2 + e2 > futil3 + e1+e2 ==> e1 < futil2 - futil3

    tb=NORMCDF(futil(2)-futil(3),0.0d0,1.0d0)
    wg(:)=tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,1)=-HUGE(v(j,1))
   	        ELSE
   	        v(j,1)=INVNORMCDF((tb(j)*draws(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( &
           MAX(-futil(2),futil(1)-futil(2)+v(j,1)) &
           -DOT_PRODUCT(lchol(nchild(m),1:nchild(m)-1),v(j,1:1)) &
        ) &
        /lchol(nchild(m),nchild(m)),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)

CASE(3)

    ! condition: u3 > 0, u1, u2
    ! condition: futil3 + e1+e2 > 0           ==> e2 > - futil2 - e1
    ! condition: futil3 + e1+e2 > futil1 + e1 ==> e2 > futil1 - futil3
    ! condition: futil3 + e1+e2 > futil2 + e2 ==> e1 > futil2 - futil3

    ta=NORMCDF(futil(2)-futil(3),0.0d0,1.0d0)
    wg(:)=1.0-ta

        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,1)=HUGE(v(j,1))
            ELSE
            	v(j,1)=INVNORMCDF((ta(j)*(1-draws(j,2))+draws(j,2)),0.0d0,1.0d0)
	        ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( &
           MAX(-futil(2)-v(j,1),futil(1)-futil(3)) &
           -DOT_PRODUCT(lchol(nchild(m),1:nchild(m)-1),v(j,1:1)) &
        ) &
        /lchol(nchild(m),nchild(m)),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)



CASE DEFAULT
    PRINT *,'nchild(m): ', nchild(m)
    PRINT *,'error'
    PAUSE
END SELECT

ghkval=SUM(wg(:))/ns

IF (ghkval<=0.0) THEN
!	PRINT *,'Error found!'
!	PRINT *,'GHK Value: ', ghkval
!	PRINT *,'nmkt: ', nmkt
	ghkval=TINY(ghkval)
!	STOP
ENDIF

! PRINT *, 'GHK2-out'

END FUNCTION GHK2











FUNCTION GHK3(m,param,n,draws) RESULT(ghkval)
	! This function calculates the GHK simulator for each market*year
USE Globvar
USE PROBABILITY
IMPLICIT NONE
INTEGER, INTENT(IN) :: m			! family id
INTEGER, INTENT(IN) :: n 
REAL(8), INTENT(IN) :: param(n)
REAL(8), INTENT(IN) :: draws(ns,2:3)
INTEGER :: i, j, k, l
INTEGER :: s1, s2, s3, s4
! Specify the array sizes based on the number of players (i.e. possible number of strategy combinations)
REAL(8) :: wg(ns)
REAL(8) :: ta(ns), tb(ns)
REAL(8) :: v(ns,nchild(m)-1)   ! GHK simulation draws constructed from draws(). v(ns,1) ... the youngest child
REAL(8) :: ghkval		! The calculated GHK value
REAL(8) :: futil(1:nchild(m)*nchild(m)-1)  ! futil(1) ... family total utility when only last child lives near
REAL(8) :: utilsub(3), alpha(3,3), gamma(3,2)		! 1 ... oldest child
REAL(8) :: lchol(3,3)


wg=0.000d0
v=0.000d0
ghkval=0.000d0

k=famindex(m)

!PRINT *, 'GHK3-in'


utilsub(1:3)=MATMUL(xb(k-2:k,:),param(1:nthetab))
	IF (nthetaa>=3) THEN
    alpha(1:3,1)  =MATMUL(xa(k-2:k,:),param(nthetab+1:nthetab+nthetaa-2))
    alpha(1:3,2)=param(nthetab+nthetaa-1)
    alpha(1:3,3)=param(nthetab+nthetaa)
	ELSEIF (nthetaa==1) THEN
	alpha(1:3,1)=param(nthetab+nthetaa)
	alpha(1:3,2:3)=0.0d0
	ELSE
	alpha=0.0d0
	ENDIF
	IF (nthetag>=2) THEN
	gamma(1:3,1) =MATMUL(xg(k-2:k,:),param(nthetab+nthetaa+1:nthetab+nthetaa+nthetag-1))
	gamma(1:3,2) =param(nthetab+nthetaa+nthetag)
	ELSEIF (nthetag==1) THEN
	gamma(1:3,1)=param(nthetab+nthetaa+nthetag)
	gamma(1:3,2)=0.0d0
	ELSE
	gamma=0.0d0
	ENDIF


lchol=lcholbase(m,param,n)

futil(1)=utilsub(3) + EXP( alpha(1,1) ) + EXP( alpha(2,1) ) + EXP( alpha(3,1) )
futil(2)=utilsub(2) + EXP( alpha(1,1) ) + EXP( alpha(2,1) ) + EXP( alpha(3,1) )
futil(3)=utilsub(2)+utilsub(3) + EXP( alpha(1,1)+alpha(1,2) ) + EXP( alpha(2,1)+alpha(2,2) ) + EXP( alpha(3,1)+alpha(3,2) ) + gamma(2,1)+gamma(3,1)
futil(4)=utilsub(1) + EXP( alpha(1,1) ) + EXP( alpha(2,1) ) + EXP( alpha(3,1) )
futil(5)=utilsub(1)+utilsub(3) + EXP( alpha(1,1)+alpha(1,2) ) + EXP( alpha(2,1)+alpha(2,2) ) + EXP( alpha(3,1)+alpha(3,2) ) + gamma(1,1)+gamma(3,1)
futil(6)=utilsub(1)+utilsub(2) + EXP( alpha(1,1)+alpha(1,2) ) + EXP( alpha(2,1)+alpha(2,2) ) + EXP( alpha(3,1)+alpha(3,2) ) + gamma(1,1)+gamma(2,1)
futil(7)=utilsub(1)+utilsub(2)+utilsub(3) &
	+ EXP( alpha(1,1)+alpha(1,2)+alpha(1,3) ) &
	+ EXP( alpha(2,1)+alpha(2,2)+alpha(2,3) ) &
	+ EXP( alpha(3,1)+alpha(3,2)+alpha(3,3) ) &
	+ gamma(1,1)*(1.0d0+gamma(1,2))+gamma(2,1)*(1.0d0+gamma(2,2))+gamma(3,1)*(1.0d0+gamma(3,2))

SELECT CASE(near(m))
CASE(0)
    ! condition: 0 > u1, u2, u3, u4, u5, u6, u7
    ! condition: 0 > futil1 + e1    ==> e1 < - futil1
    ! condition: 0 > futil2 + e2    ==> e2 < - futil2
    ! condition: 0 > futil3 + e1+e2 ==> e2 < - futil3 - e1
    ! condition: 0 > futil4 + e3    ==> e3 < - futil4
    ! condition: 0 > futil5 + e1+e3 ==> e3 < - futil5 - e1
    ! condition: 0 > futil6 + e2+e3 ==> e3 < - futil6 - e2
    ! condition: 0 > futil7 + e1+e2+e3 ==> e3 < - futil7 - e2 - e1

    tb=NORMCDF(-futil(1),0.0d0,1.0d0)
    wg(:)=tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,1)=-HUGE(v(j,1))
   	        ELSE
   	        v(j,1)=INVNORMCDF((tb(j)*draws(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
        	-futil(2), &
        	-futil(3)-v(j,1) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,2)=-HUGE(v(j,2))
   	        ELSE
   	        v(j,2)=INVNORMCDF((tb(j)*draws(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
        	-futil(4), &
        	-futil(5)-v(j,1), &
        	-futil(6)-v(j,2), &
        	-futil(7)-v(j,1)-v(j,2) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb

CASE(1)
    ! condition: u1 > 0, u2, u3, u4, u5, u6, u7
    ! condition: futil1 + e1 > 0					==> e1 > - futil1
    ! condition: futil1 + e1 > futil2 + e2			==> e2 < futil1 + e1 - futil2
    ! condition: futil1 + e1 > futil3 + e1+e2		==> e2 < futil1 - futil3
    ! condition: futil1 + e1 > futil4 + e3			==> e3 < futil1 + e1 - futil4
    ! condition: futil1 + e1 > futil5 + e1+e3		==> e3 < futil1 - futil5
    ! condition: futil1 + e1 > futil6 + e2+e3		==> e3 < futil1 + e1 - futil6 - e2
    ! condition: futil1 + e1 > futil7 + e1+e2+e3	==> e3 < futil1 - futil7 - e2

    ta=NORMCDF(-futil(1),0.0d0,1.0d0)
    wg(:)=1.0-ta
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,1)=HUGE(v(j,1))
            ELSE
            	v(j,1)=INVNORMCDF((ta(j)*(1-draws(j,3))+draws(j,3)),0.0d0,1.0d0)
	        ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
        	futil(1)+v(j,1)-futil(2), &
        	futil(1)-futil(3) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,2)=-HUGE(v(j,2))
   	        ELSE
   	        v(j,2)=INVNORMCDF((tb(j)*draws(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
        	futil(1)+v(j,1)-futil(4), &
        	futil(1)-futil(5), &
        	futil(1)+v(j,1)-futil(6)-v(j,2), &
        	futil(1)-futil(7)-v(j,2) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb

CASE(2)
    ! condition: u2 > 0, u1, u3, u4, u5, u6, u7
    ! condition: futil2 + e2 > 0					==> e2 > - futil2
    ! condition: futil2 + e2 > futil1 + e1			==> e2 > futil1 + e1 - futil2
    ! condition: futil2 + e2 > futil3 + e1+e2		==> e1 < futil2- futil3
    ! condition: futil2 + e2 > futil4 + e3			==> e3 < futil2 + e2 - futil4
    ! condition: futil2 + e2 > futil5 + e1+e3		==> e3 < futil2 + e2 - futil5 - e1
    ! condition: futil2 + e2 > futil6 + e2+e3		==> e3 < futil2 - futil6
    ! condition: futil2 + e2 > futil7 + e1+e2+e3	==> e3 < futil2 - futil7 - e1

    tb=NORMCDF(futil(2)-futil(3),0.0d0,1.0d0)
    wg(:)=tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,1)=-HUGE(v(j,1))
   	        ELSE
   	        v(j,1)=INVNORMCDF((tb(j)*draws(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	-futil(2), &
        	futil(1)+v(j,1)-futil(2) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)
        DO j=1, ns
           	IF (ta(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,2)=-HUGE(v(j,2))
   	        ELSE
            	v(j,2)=INVNORMCDF((ta(j)*(1-draws(j,2))+draws(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
        	futil(2)+v(j,2)-futil(4), &
        	futil(2)+v(j,2)-futil(5)-v(j,1), &
        	futil(2)-futil(6), &
        	futil(2)-futil(7)-v(j,1) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb


CASE(3)
    ! condition: u3 > 0, u1, u2, u4, u5, u6, u7
    ! condition: futil3 + e1+e2 > 0					==> e2 > - futil3 - e1
    ! condition: futil3 + e1+e2 > futil1 + e1		==> e2 > futil1 - futil3
    ! condition: futil3 + e1+e2 > futil2 + e2		==> e1 > futil2 - futil3
    ! condition: futil3 + e1+e2 > futil4 + e3		==> e3 < futil3 + e1+e2 - futil4
    ! condition: futil3 + e1+e2 > futil5 + e1+e3	==> e3 < futil3 + e2 - futil5
    ! condition: futil3 + e1+e2 > futil6 + e2+e3	==> e3 < futil3 + e1 - futil6
    ! condition: futil3 + e1+e2 > futil7 + e1+e2+e3	==> e3 < futil3 - futil7

    ta=NORMCDF(futil(2)-futil(3),0.0d0,1.0d0)
    wg(:)=1.0-ta
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,1)=HUGE(v(j,1))
            ELSE
            	v(j,1)=INVNORMCDF((ta(j)*(1-draws(j,3))+draws(j,3)),0.0d0,1.0d0)
	        ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MIN( &
        	-futil(3)-v(j,1), &
        	futil(1)-futil(3) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,2)=-HUGE(v(j,2))
   	        ELSE
            	v(j,2)=INVNORMCDF((ta(j)*(1-draws(j,2))+draws(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
        	futil(3)+v(j,1)+v(j,2)-futil(4), &
        	futil(3)+v(j,2)-futil(5), &
        	futil(3)+v(j,1)-futil(6), &
        	futil(3)-futil(7) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb


CASE(4)
    ! condition: u4 > 0, u1, u2, u3, u5, u6, u7
    ! condition: futil4 + e3 > 0					==> e3 > - futil4
    ! condition: futil4 + e3 > futil1 + e1			==> e3 > futil1 + e1 - futil4
    ! condition: futil4 + e3 > futil2 + e2			==> e3 > futil2 + e2 - futil4
    ! condition: futil4 + e3 > futil3 + e1+e2		==> e3 > futil3 + e1+e2 - futil4
    ! condition: futil4 + e3 > futil5 + e1+e3		==> e1 < futil4 - futil5
    ! condition: futil4 + e3 > futil6 + e2+e3		==> e2 < futil4 - futil6
    ! condition: futil4 + e3 > futil7 + e1+e2+e3	==> e2 < futil4 - futil7 - e1

    tb=NORMCDF(futil(4)-futil(5),0.0d0,1.0d0)
    wg(:)=tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,1)=-HUGE(v(j,1))
   	        ELSE
   	        v(j,1)=INVNORMCDF((tb(j)*draws(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
        	futil(4)-futil(6), &
        	futil(4)-futil(7)-v(j,1) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,2)=-HUGE(v(j,2))
   	        ELSE
   	        v(j,2)=INVNORMCDF((tb(j)*draws(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	-futil(4), &
        	futil(1)+v(j,1)-futil(4), &
        	futil(2)+v(j,2)-futil(4), &
        	futil(3)+v(j,1)+v(j,2)-futil(4) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)


CASE(5)
    ! condition: u5 > 0, u1, u2, u3, u4, u6, u7
    ! condition: futil5 + e1+e3 > 0					==> e3 > - futil5 - e1
    ! condition: futil5 + e1+e3 > futil1 + e1		==> e3 > futil1 - futil5
    ! condition: futil5 + e1+e3 > futil2 + e2		==> e3 > futil2 + e2 - futil5 - e1
    ! condition: futil5 + e1+e3 > futil3 + e1+e2	==> e3 > futil3 + e2 - futil5
    ! condition: futil5 + e1+e3 > futil4 + e3		==> e1 > futil4 - futil5
    ! condition: futil5 + e1+e3 > futil6 + e2+e3	==> e2 < futil5 + e1 - futil6
    ! condition: futil5 + e1+e3 > futil7 + e1+e2+e3	==> e2 < futil5 - futil7

    ta=NORMCDF(futil(4)-futil(5),0.0d0,1.0d0)
    wg(:)=1.0-ta
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,1)=HUGE(v(j,1))
            ELSE
            	v(j,1)=INVNORMCDF((ta(j)*(1-draws(j,3))+draws(j,3)),0.0d0,1.0d0)
	        ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
        	futil(5)+v(j,1)-futil(6), &
        	futil(5)-futil(7) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,2)=-HUGE(v(j,2))
   	        ELSE
   	        v(j,2)=INVNORMCDF((tb(j)*draws(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	-futil(5)-v(j,1), &
        	futil(1)-futil(5), &
        	futil(2)+v(j,2)-futil(5)-v(j,1), &
        	futil(3)+v(j,2)-futil(5) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)



CASE(6)
    ! condition: u6 > 0, u1, u2, u3, u4, u5, u7
    ! condition: futil6 + e2+e3 > 0					==> e3 > - futil6 - e2
    ! condition: futil6 + e2+e3 > futil1 + e1		==> e3 > futil1 + e1 - futil6 - e2
    ! condition: futil6 + e2+e3 > futil2 + e2		==> e3 > futil2 - futil6
    ! condition: futil6 + e2+e3 > futil3 + e1+e2	==> e3 > futil3 + e1 - futil6
    ! condition: futil6 + e2+e3 > futil4 + e3		==> e2 > futil4 - futil6
    ! condition: futil6 + e2+e3 > futil5 + e1+e3	==> e2 > futil5 + e1 - futil6
    ! condition: futil6 + e2+e3 > futil7 + e1+e2+e3	==> e1 < futil6 - futil7

    tb=NORMCDF(futil(6)-futil(7),0.0d0,1.0d0)
    wg(:)=tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,1)=-HUGE(v(j,1))
   	        ELSE
   	        v(j,1)=INVNORMCDF((tb(j)*draws(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	-futil(4)-futil(6), &
        	futil(5)+v(j,1)-futil(6) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,2)=-HUGE(v(j,2))
   	        ELSE
   	        v(j,2)=INVNORMCDF((ta(j)*(1-draws(j,2))+draws(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	-futil(6)-v(j,2), &
        	futil(1)+v(j,1)-futil(6)-v(j,2), &
        	futil(2)-futil(6), &
        	futil(3)+v(j,1)-futil(6) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)


CASE(7)
    ! condition: u7 > 0, u1, u2, u3, u4, u5, u6
    ! condition: futil7 + e1+e2+e3 > 0				==> e3 > - futil7 - e1 - e2
    ! condition: futil7 + e1+e2+e3 > futil1 + e1	==> e3 > futil1 - futil7 - e2
    ! condition: futil7 + e1+e2+e3 > futil2 + e2	==> e3 > futil2 - futil7 - e1
    ! condition: futil7 + e1+e2+e3 > futil3 + e1+e2	==> e3 > futil3 - futil7
    ! condition: futil7 + e1+e2+e3 > futil4 + e3	==> e2 > futil4 - futil7 - e1
    ! condition: futil7 + e1+e2+e3 > futil5 + e1+e3	==> e2 > futil5 - futil7
    ! condition: futil7 + e1+e2+e3 > futil6 + e2+e3	==> e1 > futil6 - futil7

    ta=NORMCDF(futil(6)-futil(7),0.0d0,1.0d0)
    wg(:)=1.0-ta
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,1)=HUGE(v(j,1))
            ELSE
            	v(j,1)=INVNORMCDF((ta(j)*(1-draws(j,3))+draws(j,3)),0.0d0,1.0d0)
	        ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	futil(4)-futil(7)-v(j,1), &
        	futil(5)-futil(7) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,2)=-HUGE(v(j,2))
   	        ELSE
   	        v(j,2)=INVNORMCDF((ta(j)*(1-draws(j,2))+draws(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	-futil(7)-v(j,2)-v(j,1), &
        	futil(1)-futil(7)-v(j,2), &
        	futil(2)-futil(7)-v(j,1), &
        	futil(3)-futil(7) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)


CASE DEFAULT
    PRINT *,'nchild(m): ', nchild(m)
    PRINT *,'error'
    PAUSE
END SELECT

ghkval=SUM(wg(:))/ns




!	PRINT *,'mkt: ', m
!	PRINT *,'enter: ',enter(m)
!	PRINT *,'GHK Value: ', ghkval

IF (ghkval<=0.0) THEN
!	PRINT *,'Error found!'
!	PRINT *,'GHK Value: ', ghkval
!	PRINT *,'nmkt: ', nmkt
	ghkval=TINY(ghkval)
!	STOP
ENDIF

! PRINT *, 'GHK-out'

END FUNCTION GHK3







FUNCTION GHK4(m,param,n,draws) RESULT(ghkval)
	! This function calculates the GHK simulator for each market*year
USE Globvar
USE PROBABILITY
IMPLICIT NONE
INTEGER, INTENT(IN) :: m			! family id
INTEGER, INTENT(IN) :: n 
REAL(8), INTENT(IN) :: param(n)
REAL(8), INTENT(IN) :: draws(ns,2:4)
INTEGER :: i, j, k, l
REAL(8) :: wg(ns)
REAL(8) :: ta(ns), tb(ns)
REAL(8) :: v(ns,nchild(m)-1)   ! GHK simulation draws constructed from draws(). v(ns,1) ... the youngest child
REAL(8) :: ghkval		! The calculated GHK value
REAL(8) :: futil(1:nchild(m)*nchild(m)-1)  ! futil(1) ... family total utility when only last child lives near
REAL(8) :: utilsub(4), alpha(4,3), gamma(4,2)		! 1 ... oldest child
REAL(8) :: lchol(4,4)

!REAL(8) :: test1, test2


wg=0.000d0
v=0.000d0
ghkval=0.000d0

k=famindex(m)

!PRINT *, 'GHK4-in'

utilsub(1:4)=MATMUL(xb(k-3:k,:),param(1:nthetab))
	IF (nthetaa>=3) THEN
    alpha(1:4,1)  =MATMUL(xa(k-3:k,:),param(nthetab+1:nthetab+nthetaa-2))
    alpha(1:4,2)=param(nthetab+nthetaa-1)
    alpha(1:4,3)=param(nthetab+nthetaa)
	ELSEIF (nthetaa==1) THEN
	alpha(1:4,1)=param(nthetab+nthetaa)
	alpha(1:4,2:3)=0.0d0
	ELSE
	alpha=0.0d0
	ENDIF
	IF (nthetag>=2) THEN
	gamma(1:4,1) =MATMUL(xg(k-3:k,:),param(nthetab+nthetaa+1:nthetab+nthetaa+nthetag-1))
	gamma(1:4,2) =param(nthetab+nthetaa+nthetag)
	ELSEIF (nthetag==1) THEN
	gamma(1:4,1)=param(nthetab+nthetaa+nthetag)
	gamma(1:4,2)=0.0d0
	ELSE
	gamma=0.0d0
	ENDIF


lchol=lcholbase(m,param,n)

futil(1)=utilsub(4) + EXP( alpha(1,1) ) + EXP( alpha(2,1) ) + EXP( alpha(3,1) ) + EXP( alpha(4,1) )
futil(2)=utilsub(3) + EXP( alpha(1,1) ) + EXP( alpha(2,1) ) + EXP( alpha(3,1) ) + EXP( alpha(4,1) )
futil(3)=utilsub(3)+utilsub(4) &
	+ EXP( alpha(1,1)+alpha(1,2) ) &
	+ EXP( alpha(2,1)+alpha(2,2) ) &
	+ EXP( alpha(3,1)+alpha(3,2) ) &
	+ EXP( alpha(4,1)+alpha(4,2) ) &
	+gamma(3,1)+gamma(4,1)
futil(4)=utilsub(2) + EXP( alpha(1,1) ) + EXP( alpha(2,1) ) + EXP( alpha(3,1) ) + EXP( alpha(4,1) )
futil(5)=utilsub(2)+utilsub(4) &
	+ EXP( alpha(1,1)+alpha(1,2) ) &
	+ EXP( alpha(2,1)+alpha(2,2) ) &
	+ EXP( alpha(3,1)+alpha(3,2) ) &
	+ EXP( alpha(4,1)+alpha(4,2) ) &
	+gamma(2,1)+gamma(4,1)
futil(6)=utilsub(2)+utilsub(3) &
	+ EXP( alpha(1,1)+alpha(1,2) ) &
	+ EXP( alpha(2,1)+alpha(2,2) ) &
	+ EXP( alpha(3,1)+alpha(3,2) ) &
	+ EXP( alpha(4,1)+alpha(4,2) ) &
	+gamma(2,1)+gamma(3,1)
futil(7)=utilsub(2)+utilsub(3)+utilsub(4) &
	+ EXP( alpha(1,1)+alpha(1,2)+alpha(1,3) ) &
	+ EXP( alpha(2,1)+alpha(2,2)+alpha(2,3) ) &
	+ EXP( alpha(3,1)+alpha(3,2)+alpha(3,3) ) &
	+ EXP( alpha(4,1)+alpha(4,2)+alpha(4,3) ) &
	+gamma(2,1)*(1.0d0+gamma(2,2))+gamma(3,1)*(1.0d0+gamma(3,2))+gamma(4,1)*(1.0d0+gamma(4,2))
futil(8)=utilsub(1) + EXP( alpha(1,1) ) + EXP( alpha(2,1) ) + EXP( alpha(3,1) ) + EXP( alpha(4,1) )
futil(9)=utilsub(1)+utilsub(4) &
	+ EXP( alpha(1,1)+alpha(1,2) ) &
	+ EXP( alpha(2,1)+alpha(2,2) ) &
	+ EXP( alpha(3,1)+alpha(3,2) ) &
	+ EXP( alpha(4,1)+alpha(4,2) ) &
	+gamma(1,1)+gamma(4,1)
futil(10)=utilsub(1)+utilsub(3) &
	+ EXP( alpha(1,1)+alpha(1,2) ) &
	+ EXP( alpha(2,1)+alpha(2,2) ) &
	+ EXP( alpha(3,1)+alpha(3,2) ) &
	+ EXP( alpha(4,1)+alpha(4,2) ) &
	+gamma(1,1)+gamma(3,1)
futil(11)=utilsub(1)+utilsub(3)+utilsub(4) &
	+ EXP( alpha(1,1)+alpha(1,2)+alpha(1,3) ) &
	+ EXP( alpha(2,1)+alpha(2,2)+alpha(2,3) ) &
	+ EXP( alpha(3,1)+alpha(3,2)+alpha(3,3) ) &
	+ EXP( alpha(4,1)+alpha(4,2)+alpha(4,3) ) &
	+gamma(1,1)*(1.0d0+gamma(1,2))+gamma(3,1)*(1.0d0+gamma(3,2))+gamma(4,1)*(1.0d0+gamma(4,2))
futil(12)=utilsub(1)+utilsub(2) &
	+ EXP( alpha(1,1)+alpha(1,2) ) &
	+ EXP( alpha(2,1)+alpha(2,2) ) &
	+ EXP( alpha(3,1)+alpha(3,2) ) &
	+ EXP( alpha(4,1)+alpha(4,2) ) &
	+gamma(1,1)+gamma(2,1)
futil(13)=utilsub(1)+utilsub(2)+utilsub(4) &
	+ EXP( alpha(1,1)+alpha(1,2)+alpha(1,3) ) &
	+ EXP( alpha(2,1)+alpha(2,2)+alpha(2,3) ) &
	+ EXP( alpha(3,1)+alpha(3,2)+alpha(3,3) ) &
	+ EXP( alpha(4,1)+alpha(4,2)+alpha(4,3) ) &
	+gamma(1,1)*(1.0d0+gamma(1,2))+gamma(2,1)*(1.0d0+gamma(2,2))+gamma(4,1)*(1.0d0+gamma(4,2))
futil(14)=utilsub(1)+utilsub(2)+utilsub(3) &
	+ EXP( alpha(1,1)+alpha(1,2)+alpha(1,3) ) &
	+ EXP( alpha(2,1)+alpha(2,2)+alpha(2,3) ) &
	+ EXP( alpha(3,1)+alpha(3,2)+alpha(3,3) ) &
	+ EXP( alpha(4,1)+alpha(4,2)+alpha(4,3) ) &
	+gamma(1,1)*(1.0d0+gamma(1,2))+gamma(2,1)*(1.0d0+gamma(2,2))+gamma(3,1)*(1.0d0+gamma(3,2))
futil(15)=utilsub(1)+utilsub(2)+utilsub(3)+utilsub(4) &
	+ EXP( alpha(1,1)+alpha(1,2)+alpha(1,3) ) &
	+ EXP( alpha(2,1)+alpha(2,2)+alpha(2,3) ) &
	+ EXP( alpha(3,1)+alpha(3,2)+alpha(3,3) ) &
	+ EXP( alpha(4,1)+alpha(4,2)+alpha(4,3) ) &
	+gamma(1,1)*(1.0d0+gamma(1,2)*2.0d0)+gamma(2,1)*(1.0d0+gamma(2,2)*2.0d0)+gamma(3,1)*(1.0d0+gamma(3,2)*2.0d0)+gamma(4,1)*(1.0d0+gamma(4,2)*2.0d0)

SELECT CASE(near(m))

CASE(0)

    ! condition: 0 > u1 - u15
    ! condition: 0 > futil1 + e1			==> e1 < - futil1
    ! condition: 0 > futil2 + e2			==> e2 < - futil2
    ! condition: 0 > futil3 + e1+e2			==> e2 < - futil3 - e1
    ! condition: 0 > futil4 + e3			==> e3 < - futil4
    ! condition: 0 > futil5 + e1+e3			==> e3 < - futil5 - e1
    ! condition: 0 > futil6 + e2+e3			==> e3 < - futil6 - e2
    ! condition: 0 > futil7 + e1+e2+e3		==> e3 < - futil7 - e2 - e1
    ! condition: 0 > futil8 + e4			==> e4 < - futil8
    ! condition: 0 > futil9 + e1+e4			==> e4 < - futil9 - e1
    ! condition: 0 > futil10 + e2+e4		==> e4 < - futil10 - e2
    ! condition: 0 > futil11 +e1+e2+e4		==> e4 < - futil11 - e1 - e2
    ! condition: 0 > futil12 +e3+e4			==> e4 < - futil12 - e3
    ! condition: 0 > futil13 +e1+e3+e4		==> e4 < - futil13 - e1 - e3
    ! condition: 0 > futil14 +e2+e3+e4		==> e4 < - futil14 - e2 - e3
    ! condition: 0 > futil15 +e1+e2+e3+e4	==> e4 < - futil15 - e1 - e2 - e3

	tb=NORMCDF(-futil(1),0.0d0,1.0d0)
    wg(:)=tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,1)=-HUGE(v(j,1))
   	        ELSE
   	        v(j,1)=INVNORMCDF((tb(j)*draws(j,4)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
        	-futil(2), &
        	-futil(3)-v(j,1) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,2)=-HUGE(v(j,2))
   	        ELSE
   	        v(j,2)=INVNORMCDF((tb(j)*draws(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
			-futil(4), &
        	-futil(5)-v(j,1), &
        	-futil(6)-v(j,2), &
        	-futil(7)-v(j,1)-v(j,2) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,3)=-HUGE(v(j,3))
   	        ELSE
   	        v(j,3)=INVNORMCDF((tb(j)*draws(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
			-futil(8), &
        	-futil(9)-v(j,1), &
        	-futil(10)-v(j,2), &
        	-futil(11)-v(j,1)-v(j,2), &
        	-futil(12)-v(j,3), &
        	-futil(13)-v(j,1)-v(j,3), &
        	-futil(14)-v(j,2)-v(j,3), &
        	-futil(15)-v(j,1)-v(j,2)-v(j,3) &
	        ) &
           -DOT_PRODUCT(lchol(4,1:3),v(j,1:3)) &
        ) &
        /lchol(4,4),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb

CASE(1)

    ! condition: u1 > 0, u2 - u15
    ! condition: util1 + e1 > 0						==> e1 > - futil1
    ! condition: util1 + e1 > futil2 + e2			==> e2 < futil1 + e1 - futil2
    ! condition: util1 + e1 > futil3 + e1+e2		==> e2 < futil1 - futil3
    ! condition: util1 + e1 > futil4 + e3			==> e3 < futil1 + e1 - futil4
    ! condition: util1 + e1 > futil5 + e1+e3		==> e3 < futil1 - futil5
    ! condition: util1 + e1 > futil6 + e2+e3		==> e3 < futil1 + e1 - futil6 - e2
    ! condition: util1 + e1 > futil7 + e1+e2+e3		==> e3 < futil1 - futil7 - e2
    ! condition: util1 + e1 > futil8 + e4			==> e4 < futil1 + e1 - futil8
    ! condition: util1 + e1 > futil9 + e1+e4		==> e4 < futil1 - futil9
    ! condition: util1 + e1 > futil10 + e2+e4		==> e4 < futil1 + e1 - futil10 - e2
    ! condition: util1 + e1 > futil11 +e1+e2+e4		==> e4 < futil1 - futil11 - e2
    ! condition: util1 + e1 > futil12 +e3+e4		==> e4 < futil1 + e1 - futil12 - e3
    ! condition: util1 + e1 > futil13 +e1+e3+e4		==> e4 < futil1 - futil13 - e3
    ! condition: util1 + e1 > futil14 +e2+e3+e4		==> e4 < futil1 + e1 - futil14 - e2 - e3
    ! condition: util1 + e1 > futil15 +e1+e2+e3+e4	==> e4 < futil1 - futil15 - e2 - e3

    ta=NORMCDF(-futil(1),0.0d0,1.0d0)
    wg(:)=1.0-ta
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,1)=HUGE(v(j,1))
            ELSE
            	v(j,1)=INVNORMCDF((ta(j)*(1-draws(j,4))+draws(j,4)),0.0d0,1.0d0)
	        ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
        	futil(1)+v(j,1)-futil(2), &
        	futil(1)-futil(3) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,2)=-HUGE(v(j,2))
   	        ELSE
   	        v(j,2)=INVNORMCDF((tb(j)*draws(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
        	futil(1)+v(j,1)-futil(4), &
        	futil(1)-futil(5), &
        	futil(1)+v(j,1)-futil(6)-v(j,2), &
        	futil(1)-futil(7)-v(j,2) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,3)=-HUGE(v(j,3))
   	        ELSE
   	        v(j,3)=INVNORMCDF((tb(j)*draws(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
			futil(1)+v(j,1)-futil(8), &
        	futil(1)-futil(9), &
        	futil(1)+v(j,1)-futil(10)-v(j,2), &
        	futil(1)-futil(11)-v(j,2), &
        	futil(1)+v(j,1)-futil(12)-v(j,3), &
        	futil(1)-futil(13)-v(j,3), &
        	futil(1)+v(j,1)-futil(14)-v(j,2)-v(j,3), &
        	futil(1)-futil(15)-v(j,2)-v(j,3) &
	        ) &
           -DOT_PRODUCT(lchol(4,1:3),v(j,1:3)) &
        ) &
        /lchol(4,4),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb

CASE(2)

    ! condition: u2 > 0, u1, u3 - u15
    ! condition: futil2 + e2 > 0					==> e2 > - futil2
    ! condition: futil2 + e2 > futil1 + e1			==> e2 > futil1 + e1 - futil2
    ! condition: futil2 + e2 > futil3 + e1+e2		==> e1 < futil2 - futil3
    ! condition: futil2 + e2 > futil4 + e3			==> e3 < futil2 + e2 - futil4
    ! condition: futil2 + e2 > futil5 + e1+e3		==> e3 < futil2 + e2 - futil5 - e1
    ! condition: futil2 + e2 > futil6 + e2+e3		==> e3 < futil2 - futil6
    ! condition: futil2 + e2 > futil7 + e1+e2+e3	==> e3 < futil2 - futil7 - e1
    ! condition: futil2 + e2 > futil8 + e4			==> e4 < futil2 + e2 - futil8
    ! condition: futil2 + e2 > futil9 + e1+e4		==> e4 < futil2 + e2 - futil9 - e1
    ! condition: futil2 + e2 > futil10 + e2+e4		==> e4 < futil2 - futil10
    ! condition: futil2 + e2 > futil11 +e1+e2+e4	==> e4 < futil2 - futil11 - e1
    ! condition: futil2 + e2 > futil12 +e3+e4		==> e4 < futil2 + e2 - futil12 - e3
    ! condition: futil2 + e2 > futil13 +e1+e3+e4	==> e4 < futil2 + e2 - futil13 - e1 - e3
    ! condition: futil2 + e2 > futil14 +e2+e3+e4	==> e4 < futil2 + e2 - futil14 - e3
    ! condition: futil2 + e2 > futil15 +e1+e2+e3+e4	==> e4 < futil2 - futil15 - e1 - e3

	tb=NORMCDF(futil(2)-futil(3),0.0d0,1.0d0)
    wg(:)=tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,1)=-HUGE(v(j,1))
   	        ELSE
   	        v(j,1)=INVNORMCDF((tb(j)*draws(j,4)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	-futil(2), &
        	futil(1)+v(j,1)-futil(2) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
        	    v(j,2)=HUGE(v(j,2))
   	        ELSE
           	v(j,2)=INVNORMCDF((ta(j)*(1-draws(j,3))+draws(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
        	futil(2)+v(j,2)-futil(4), &
        	futil(2)+v(j,2)-futil(5)-v(j,1), &
        	futil(2)-futil(6), &
        	futil(2)-futil(7)-v(j,1) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,3)=-HUGE(v(j,3))
   	        ELSE
   	        v(j,3)=INVNORMCDF((tb(j)*draws(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
			futil(2)+v(j,2)-futil(8), &
        	futil(2)+v(j,2)-futil(9)-v(j,1), &
        	futil(2)-futil(10), &
        	futil(2)-futil(11)-v(j,1), &
        	futil(2)+v(j,2)-futil(12)-v(j,3), &
        	futil(2)+v(j,2)-futil(13)-v(j,1)-v(j,3), &
        	futil(2)-futil(14)-v(j,3), &
        	futil(2)-futil(15)-v(j,1)-v(j,3) &
	        ) &
           -DOT_PRODUCT(lchol(4,1:3),v(j,1:3)) &
        ) &
        /lchol(4,4),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb

CASE(3)

    ! condition: u3 > 0, u1, u2, u4 - u15
    ! condition: futil3 + e1+e2 > 0						==> e2 > - futil3 - e1
    ! condition: futil3 + e1+e2 > futil1 + e1			==> e2 > futil1 - futil3
    ! condition: futil3 + e1+e2 > futil2 + e2			==> e1 > futil2 - futil3
    ! condition: futil3 + e1+e2 > futil4 + e3			==> e3 < futil3 + e1+e2 - futil4
    ! condition: futil3 + e1+e2 > futil5 + e1+e3		==> e3 < futil3 + e2 - futil5
    ! condition: futil3 + e1+e2 > futil6 + e2+e3		==> e3 < futil3 + e1 - futil6
    ! condition: futil3 + e1+e2 > futil7 + e1+e2+e3		==> e3 < futil3 - futil7
    ! condition: futil3 + e1+e2 > futil8 + e4			==> e4 < futil3 + e1+e2 - futil8
    ! condition: futil3 + e1+e2 > futil9 + e1+e4		==> e4 < futil3 + e2 - futil9
    ! condition: futil3 + e1+e2 > futil10 + e2+e4		==> e4 < futil3 + e1 - futil10
    ! condition: futil3 + e1+e2 > futil11 +e1+e2+e4		==> e4 < futil3 - futil11
    ! condition: futil3 + e1+e2 > futil12 +e3+e4		==> e4 < futil3 + e1+e2 - futil12 - e3
    ! condition: futil3 + e1+e2 > futil13 +e1+e3+e4		==> e4 < futil3 + e2 - futil13 - e3
    ! condition: futil3 + e1+e2 > futil14 +e2+e3+e4		==> e4 < futil3 + e1 - futil14 - e3
    ! condition: futil3 + e1+e2 > futil15 +e1+e2+e3+e4	==> e4 < futil3 - futil15 - e3

    ta=NORMCDF(futil(2)-futil(3),0.0d0,1.0d0)
    wg(:)=1.0-ta
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,1)=HUGE(v(j,1))
            ELSE
            	v(j,1)=INVNORMCDF((ta(j)*(1-draws(j,4))+draws(j,4)),0.0d0,1.0d0)
	        ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	-futil(3)-v(j,1), &
        	futil(1)-futil(3) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
        	    v(j,2)=HUGE(v(j,2))
   	        ELSE
           	v(j,2)=INVNORMCDF((ta(j)*(1-draws(j,3))+draws(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
        	futil(3)+v(j,1)+v(j,2)-futil(4), &
        	futil(3)+v(j,2)-futil(5), &
        	futil(3)+v(j,1)-futil(6), &
        	futil(3)-futil(7) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,3)=-HUGE(v(j,3))
   	        ELSE
   	        v(j,3)=INVNORMCDF((tb(j)*draws(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
			futil(3)+v(j,1)+v(j,2)-futil(8), &
        	futil(3)+v(j,2)-futil(9), &
        	futil(3)+v(j,1)-futil(10), &
        	futil(3)-futil(11), &
        	futil(3)+v(j,1)+v(j,2)-futil(12)-v(j,3), &
        	futil(3)+v(j,2)-futil(13)-v(j,3), &
        	futil(3)+v(j,1)-futil(14)-v(j,3), &
        	futil(3)-futil(15)-v(j,3) &
	        ) &
           -DOT_PRODUCT(lchol(4,1:3),v(j,1:3)) &
        ) &
        /lchol(4,4),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb

CASE(4)

    ! condition: u4 > 0, u1 - u3, u5 - u15
    ! condition: futil4 + e3 > 0					==> e3 > - futil4
    ! condition: futil4 + e3 > futil1 + e1			==> e3 > futil1 + e1 - futil4
    ! condition: futil4 + e3 > futil2 + e2			==> e3 > futil2 + e2 - futil4
    ! condition: futil4 + e3 > futil3 + e1+e2		==> e3 > futil3 + e1+e2 - futil4
    ! condition: futil4 + e3 > futil5 + e1+e3		==> e1 < futil4 - futil5
    ! condition: futil4 + e3 > futil6 + e2+e3		==> e2 < futil4 - futil6
    ! condition: futil4 + e3 > futil7 + e1+e2+e3	==> e2 < futil4 - futil7 - e1
    ! condition: futil4 + e3 > futil8 + e4			==> e4 < futil4 + e3 - futil8
    ! condition: futil4 + e3 > futil9 + e1+e4		==> e4 < futil4 + e3 - futil9 - e1
    ! condition: futil4 + e3 > futil10 + e2+e4		==> e4 < futil4 + e3 - futil10 - e2
    ! condition: futil4 + e3 > futil11 +e1+e2+e4	==> e4 < futil4 + e3 - futil11 - e1 - e2
    ! condition: futil4 + e3 > futil12 +e3+e4		==> e4 < futil4 - futil12
    ! condition: futil4 + e3 > futil13 +e1+e3+e4	==> e4 < futil4 - futil13 - e1
    ! condition: futil4 + e3 > futil14 +e2+e3+e4	==> e4 < futil4 - futil14 - e2
    ! condition: futil4 + e3 > futil15 +e1+e2+e3+e4	==> e4 < futil4 - futil15 - e1 - e2

	tb=NORMCDF(futil(4)-futil(5),0.0d0,1.0d0)
    wg(:)=tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,1)=-HUGE(v(j,1))
   	        ELSE
   	        v(j,1)=INVNORMCDF((tb(j)*draws(j,4)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
        	futil(4)-futil(7)-v(j,1), &
        	futil(4)-futil(6) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,2)=-HUGE(v(j,2))
   	        ELSE
   	        v(j,2)=INVNORMCDF((tb(j)*draws(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	-futil(4), &
        	futil(1)+v(j,1)-futil(4), &
        	futil(2)+v(j,2)-futil(4), &
        	futil(3)+v(j,1)+v(j,2)-futil(4) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
        	    v(j,3)=HUGE(v(j,3))
   	        ELSE
           	v(j,3)=INVNORMCDF((ta(j)*(1-draws(j,2))+draws(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
			futil(4)+v(j,3)-futil(8), &
        	futil(4)+v(j,3)-futil(9)-v(j,1), &
        	futil(4)+v(j,3)-futil(10)-v(j,2), &
        	futil(4)+v(j,3)-futil(11)-v(j,1)-v(j,2), &
        	futil(4)-futil(12), &
        	futil(4)-futil(13)-v(j,1), &
        	futil(4)-futil(14)-v(j,2), &
        	futil(4)-futil(15)-v(j,1)-v(j,2) &
	        ) &
           -DOT_PRODUCT(lchol(4,1:3),v(j,1:3)) &
        ) &
        /lchol(4,4),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb

CASE(5)

    ! condition: u5 > 0, u1 - u4, u6 - u15
    ! condition: futil5 + e1+e3 > 0						==> e3 > - futil5 - e1
    ! condition: futil5 + e1+e3 > futil1 + e1			==> e3 > futil1 - futil5
    ! condition: futil5 + e1+e3 > futil2 + e2			==> e3 > futil2 + e2 - futil5 - e1
    ! condition: futil5 + e1+e3 > futil3 + e1+e2		==> e3 > futil3 + e2 - futil5
    ! condition: futil5 + e1+e3 > futil4 + e3			==> e1 > futil4 - futil5
    ! condition: futil5 + e1+e3 > futil6 + e2+e3		==> e2 < futil5 + e1 - futil6
    ! condition: futil5 + e1+e3 > futil7 + e1+e2+e3		==> e2 < futil5 - futil7
    ! condition: futil5 + e1+e3 > futil8 + e4			==> e4 < futil5 + e1+e3 - futil8
    ! condition: futil5 + e1+e3 > futil9 + e1+e4		==> e4 < futil5 + e3 - futil9
    ! condition: futil5 + e1+e3 > futil10 + e2+e4		==> e4 < futil5 + e1+e3 - futil10 - e2
    ! condition: futil5 + e1+e3 > futil11 +e1+e2+e4		==> e4 < futil5 + e3 - futil11 - e2
    ! condition: futil5 + e1+e3 > futil12 +e3+e4		==> e4 < futil5 + e1 - futil12
    ! condition: futil5 + e1+e3 > futil13 +e1+e3+e4		==> e4 < futil5 - futil13
    ! condition: futil5 + e1+e3 > futil14 +e2+e3+e4		==> e4 < futil5 + e1 - futil14 - e2
    ! condition: futil5 + e1+e3 > futil15 +e1+e2+e3+e4	==> e4 < futil5 - futil15 - e2

    ta=NORMCDF(futil(4)-futil(5),0.0d0,1.0d0)
    wg(:)=1.0-ta
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,1)=HUGE(v(j,1))
            ELSE
            	v(j,1)=INVNORMCDF((ta(j)*(1-draws(j,4))+draws(j,4)),0.0d0,1.0d0)
	        ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
        	futil(5)+v(j,1)-futil(6), &
        	futil(5)-futil(7) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,2)=-HUGE(v(j,2))
   	        ELSE
   	        v(j,2)=INVNORMCDF((tb(j)*draws(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	-futil(5)-v(j,1), &
        	futil(1)-futil(5), &
        	futil(2)+v(j,2)-futil(5)-v(j,1), &
        	futil(3)+v(j,2)-futil(5) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
        	    v(j,3)=HUGE(v(j,3))
   	        ELSE
           	v(j,3)=INVNORMCDF((ta(j)*(1-draws(j,2))+draws(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
			futil(5)+v(j,1)+v(j,3)-futil(8), &
        	futil(5)+v(j,3)-futil(9), &
        	futil(5)+v(j,1)+v(j,3)-futil(10)-v(j,2), &
        	futil(5)+v(j,3)-futil(11)-v(j,2), &
        	futil(5)+v(j,1)-futil(12), &
        	futil(5)-futil(13), &
        	futil(5)+v(j,1)-futil(14)-v(j,2), &
        	futil(5)-futil(15)-v(j,2) &
	        ) &
           -DOT_PRODUCT(lchol(4,1:3),v(j,1:3)) &
        ) &
        /lchol(4,4),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb

CASE(6)

    ! condition: u6 > 0, u1 - u5, u7 - u15
    ! condition: futil6 + e2+e3 > 0						==> e3 > - futil6 - e2
    ! condition: futil6 + e2+e3 > futil1 + e1			==> e3 > futil1 + e1 - futil6 - e2
    ! condition: futil6 + e2+e3 > futil2 + e2			==> e3 > futil2 - futil6
    ! condition: futil6 + e2+e3 > futil3 + e1+e2		==> e3 > futil3 + e1 - futil6
    ! condition: futil6 + e2+e3 > futil4 + e3			==> e2 > futil4 - futil6
    ! condition: futil6 + e2+e3 > futil5 + e1+e3		==> e2 > futil5 + e1 - futil6
    ! condition: futil6 + e2+e3 > futil7 + e1+e2+e3		==> e1 < futil6 - futil7
    ! condition: futil6 + e2+e3 > futil8 + e4			==> e4 < futil6 + e2+e3 - futil8
    ! condition: futil6 + e2+e3 > futil9 + e1+e4		==> e4 < futil6 + e2+e3 - futil9 - e1
    ! condition: futil6 + e2+e3 > futil10 + e2+e4		==> e4 < futil6 + e3 - futil10
    ! condition: futil6 + e2+e3 > futil11 +e1+e2+e4		==> e4 < futil6 + e3 - futil11 - e1
    ! condition: futil6 + e2+e3 > futil12 +e3+e4		==> e4 < futil6 + e2 - futil12
    ! condition: futil6 + e2+e3 > futil13 +e1+e3+e4		==> e4 < futil6 + e2 - futil13 - e1
    ! condition: futil6 + e2+e3 > futil14 +e2+e3+e4		==> e4 < futil6 - futil14
    ! condition: futil6 + e2+e3 > futil15 +e1+e2+e3+e4	==> e4 < futil6 - futil15 - e1

	tb=NORMCDF(futil(6)-futil(7),0.0d0,1.0d0)
    wg(:)=tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,1)=-HUGE(v(j,1))
   	        ELSE
   	        v(j,1)=INVNORMCDF((tb(j)*draws(j,4)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	futil(4)-futil(6), &
        	futil(5)+v(j,1)-futil(6) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
        	    v(j,2)=HUGE(v(j,2))
   	        ELSE
           	v(j,2)=INVNORMCDF((ta(j)*(1-draws(j,3))+draws(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	-futil(6)-v(j,2), &
        	futil(1)+v(j,1)-futil(6)-v(j,2), &
        	futil(2)-futil(6), &
        	futil(3)+v(j,1)-futil(6) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
        	    v(j,3)=HUGE(v(j,3))
   	        ELSE
           	v(j,3)=INVNORMCDF((ta(j)*(1-draws(j,2))+draws(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
			futil(6)+v(j,2)+v(j,3)-futil(8), &
        	futil(6)+v(j,2)+v(j,3)-futil(9)-v(j,1), &
        	futil(6)+v(j,3)-futil(10), &
        	futil(6)+v(j,3)-futil(11)-v(j,1), &
        	futil(6)+v(j,2)-futil(12), &
        	futil(6)+v(j,2)-futil(13)-v(j,1), &
        	futil(6)-futil(14), &
        	futil(6)-futil(15)-v(j,1) &
	        ) &
           -DOT_PRODUCT(lchol(4,1:3),v(j,1:3)) &
        ) &
        /lchol(4,4),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb

CASE(7)

    ! condition: u7 > 0, u1 - u6, u8 - u15
    ! condition: futil7 + e1+e2+e3 > 0						==> e3 > - futil7 - e1 - e2
    ! condition: futil7 + e1+e2+e3 > futil1 + e1			==> e3 > futil1 - futil7 - e2
    ! condition: futil7 + e1+e2+e3 > futil2 + e2			==> e3 > futil2 - futil7 - e1
    ! condition: futil7 + e1+e2+e3 > futil3 + e1+e2			==> e3 > futil3 - futil7
    ! condition: futil7 + e1+e2+e3 > futil4 + e3			==> e2 > futil4 - futil7 - e1
    ! condition: futil7 + e1+e2+e3 > futil5 + e1+e3			==> e2 > futil5 - futil7
    ! condition: futil7 + e1+e2+e3 > futil6 + e2+e3			==> e1 > futil6 - futil7
    ! condition: futil7 + e1+e2+e3 > futil8 + e4			==> e4 < futil7 + e1+e2+e3 - futil8
    ! condition: futil7 + e1+e2+e3 > futil9 + e1+e4			==> e4 < futil7 + e2+e3 - futil9
    ! condition: futil7 + e1+e2+e3 > futil10 + e2+e4		==> e4 < futil7 + e1+e3 - futil10
    ! condition: futil7 + e1+e2+e3 > futil11 +e1+e2+e4		==> e4 < futil7 + e3 - futil11
    ! condition: futil7 + e1+e2+e3 > futil12 +e3+e4			==> e4 < futil7 + e1+e2 - futil12
    ! condition: futil7 + e1+e2+e3 > futil13 +e1+e3+e4		==> e4 < futil7 + e2 - futil13
    ! condition: futil7 + e1+e2+e3 > futil14 +e2+e3+e4		==> e4 < futil7 + e1 - futil14
    ! condition: futil7 + e1+e2+e3 > futil15 +e1+e2+e3+e4	==> e4 < futil7 - futil15

    ta=NORMCDF(futil(6)-futil(7),0.0d0,1.0d0)
    wg(:)=1.0-ta
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,1)=HUGE(v(j,1))
            ELSE
            	v(j,1)=INVNORMCDF((ta(j)*(1-draws(j,4))+draws(j,4)),0.0d0,1.0d0)
	        ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	futil(4)-futil(7)-v(j,1), &
        	futil(5)-futil(7) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
        	    v(j,2)=HUGE(v(j,2))
   	        ELSE
           	v(j,2)=INVNORMCDF((ta(j)*(1-draws(j,3))+draws(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	-futil(7)-v(j,1)-v(j,2), &
        	futil(1)-futil(7)-v(j,2), &
        	futil(2)-futil(7)-v(j,1), &
        	futil(3)-futil(7) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
        	    v(j,3)=HUGE(v(j,3))
   	        ELSE
           	v(j,3)=INVNORMCDF((ta(j)*(1-draws(j,2))+draws(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
			futil(7)+v(j,1)+v(j,2)+v(j,3)-futil(8), &
        	futil(7)+v(j,2)+v(j,3)-futil(9), &
        	futil(7)+v(j,1)+v(j,3)-futil(10), &
        	futil(7)+v(j,3)-futil(11), &
        	futil(7)+v(j,1)+v(j,2)-futil(12), &
        	futil(7)+v(j,2)-futil(13), &
        	futil(7)+v(j,1)-futil(14), &
        	futil(7)-futil(15) &
	        ) &
           -DOT_PRODUCT(lchol(4,1:3),v(j,1:3)) &
        ) &
        /lchol(4,4),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb

CASE(8)

    ! condition: u8 > 0, u1 - u7, u9 - u15
    ! condition: futil8 + e4 > 0					==> e4 > - futil8
    ! condition: futil8 + e4 > futil1 + e1			==> e4 > futil1 + e1 - futil8
    ! condition: futil8 + e4 > futil2 + e2			==> e4 > futil2 + e2 - futil8
    ! condition: futil8 + e4 > futil3 + e1+e2		==> e4 > futil3 + e1+e2 - futil8
    ! condition: futil8 + e4 > futil4 + e3			==> e4 > futil4 + e3 - futil8
    ! condition: futil8 + e4 > futil5 + e1+e3		==> e4 > futil5 + e1+e3 - futil8
    ! condition: futil8 + e4 > futil6 + e2+e3		==> e4 > futil6 + e2+e3 - futil8
    ! condition: futil8 + e4 > futil7 + e1+e2+e3	==> e4 > futil7 + e1+e2+e3 - futil8
    ! condition: futil8 + e4 > futil9 + e1+e4		==> e1 < futil8 - futil9
    ! condition: futil8 + e4 > futil10 + e2+e4		==> e2 < futil8 - futil10
    ! condition: futil8 + e4 > futil11 +e1+e2+e4	==> e2 < futil8 - futil11 - e1
    ! condition: futil8 + e4 > futil12 +e3+e4		==> e3 < futil8 - futil12
    ! condition: futil8 + e4 > futil13 +e1+e3+e4	==> e3 < futil8 - futil13 - e1
    ! condition: futil8 + e4 > futil14 +e2+e3+e4	==> e3 < futil8 - futil14 - e2
    ! condition: futil8 + e4 > futil15 +e1+e2+e3+e4	==> e3 < futil8 - futil15 - e1 - e2

	tb=NORMCDF(futil(8)-futil(9),0.0d0,1.0d0)
    wg(:)=tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,1)=-HUGE(v(j,1))
   	        ELSE
   	        v(j,1)=INVNORMCDF((tb(j)*draws(j,4)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
        	futil(8)-futil(11)-v(j,1), &
        	futil(8)-futil(10) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,2)=-HUGE(v(j,2))
   	        ELSE
   	        v(j,2)=INVNORMCDF((tb(j)*draws(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
			futil(8)-futil(12), &
        	futil(8)-futil(13)-v(j,1), &
        	futil(8)-futil(14)-v(j,2), &
        	futil(8)-futil(15)-v(j,1)-v(j,2) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,3)=-HUGE(v(j,3))
   	        ELSE
   	        v(j,3)=INVNORMCDF((tb(j)*draws(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
			-futil(8), &
        	futil(1)+v(j,1)-futil(8), &
        	futil(2)+v(j,2)-futil(8), &
        	futil(3)+v(j,1)+v(j,2)-futil(8), &
        	futil(4)+v(j,3)-futil(8), &
        	futil(5)+v(j,1)+v(j,3)-futil(8), &
        	futil(6)+v(j,2)+v(j,3)-futil(8), &
        	futil(7)+v(j,1)+v(j,2)+v(j,3)-futil(8) &
	        ) &
           -DOT_PRODUCT(lchol(4,1:3),v(j,1:3)) &
        ) &
        /lchol(4,4),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)

CASE(9)

    ! condition: u9 > 0, u1 - u8, u10 - u15
    ! condition: futil9 + e1+e4 > 0						==> e4 > - futil9 - e1
    ! condition: futil9 + e1+e4 > futil1 + e1			==> e4 > futil1 - futil9
    ! condition: futil9 + e1+e4 > futil2 + e2			==> e4 > futil2 + e2 - futil9 - e1
    ! condition: futil9 + e1+e4 > futil3 + e1+e2		==> e4 > futil3 + e2 - futil9
    ! condition: futil9 + e1+e4 > futil4 + e3			==> e4 > futil4 + e3 - futil9 - e1
    ! condition: futil9 + e1+e4 > futil5 + e1+e3		==> e4 > futil5 + e3 - futil9
    ! condition: futil9 + e1+e4 > futil6 + e2+e3		==> e4 > futil6 + e2+e3 - futil9 - e1
    ! condition: futil9 + e1+e4 > futil7 + e1+e2+e3		==> e4 > futil7 + e2+e3 - futil9
    ! condition: futil9 + e1+e4 > futil8 + e4			==> e1 > futil8 - futil9
    ! condition: futil9 + e1+e4 > futil10 + e2+e4		==> e2 < futil9 + e1 - futil10
    ! condition: futil9 + e1+e4 > futil11 +e1+e2+e4		==> e2 < futil9 - futil11
    ! condition: futil9 + e1+e4 > futil12 +e3+e4		==> e3 < futil9 + e1 - futil12
    ! condition: futil9 + e1+e4 > futil13 +e1+e3+e4		==> e3 < futil9 - futil13
    ! condition: futil9 + e1+e4 > futil14 +e2+e3+e4		==> e3 < futil9 + e1 - futil14 - e2
    ! condition: futil9 + e1+e4 > futil15 +e1+e2+e3+e4	==> e3 < futil9 - futil15 - e2

    ta=NORMCDF(futil(8)-futil(9),0.0d0,1.0d0)
    wg(:)=1.0-ta
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,1)=HUGE(v(j,1))
            ELSE
            	v(j,1)=INVNORMCDF((ta(j)*(1-draws(j,4))+draws(j,4)),0.0d0,1.0d0)
	        ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
        	futil(9)+v(j,1)-futil(10), &
        	futil(9)-futil(11) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,2)=-HUGE(v(j,2))
   	        ELSE
   	        v(j,2)=INVNORMCDF((tb(j)*draws(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
			futil(9)+v(j,1)-futil(12), &
        	futil(9)-futil(13), &
        	futil(9)+v(j,1)-futil(14)-v(j,2), &
        	futil(9)-futil(15)-v(j,2) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,3)=-HUGE(v(j,3))
   	        ELSE
   	        v(j,3)=INVNORMCDF((tb(j)*draws(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
			-futil(9)-v(j,1), &
        	futil(1)-futil(9), &
        	futil(2)+v(j,2)-futil(9)-v(j,1), &
        	futil(3)+v(j,2)-futil(9), &
        	futil(4)+v(j,3)-futil(9)-v(j,1), &
        	futil(5)+v(j,3)-futil(9), &
        	futil(6)+v(j,2)+v(j,3)-futil(9)-v(j,1), &
        	futil(7)+v(j,2)+v(j,3)-futil(9) &
	        ) &
           -DOT_PRODUCT(lchol(4,1:3),v(j,1:3)) &
        ) &
        /lchol(4,4),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)

CASE(10)

    ! condition: u10 > 0, u1 - u9, u11 - u15
    ! condition: futil10 + e2+e4 > 0					==> e4 > - futil10 - e2
    ! condition: futil10 + e2+e4 > futil1 + e1			==> e4 > futil1 + e1 - futil10 - e2
    ! condition: futil10 + e2+e4 > futil2 + e2			==> e4 > futil2 - futil10
    ! condition: futil10 + e2+e4 > futil3 + e1+e2		==> e4 > futil3 + e1 - futil10
    ! condition: futil10 + e2+e4 > futil4 + e3			==> e4 > futil4 + e3 - futil10 - e2
    ! condition: futil10 + e2+e4 > futil5 + e1+e3		==> e4 > futil5 + e1+e3 - futil10 - e2
    ! condition: futil10 + e2+e4 > futil6 + e2+e3		==> e4 > futil6 + e3 - futil10
    ! condition: futil10 + e2+e4 > futil7 + e1+e2+e3	==> e4 > futil7 + e1+e3 - futil10
    ! condition: futil10 + e2+e4 > futil8 + e4			==> e2 > futil8 - futil10
    ! condition: futil10 + e2+e4 > futil9 + e1+e4		==> e2 > futil9 + e1 - futil10
    ! condition: futil10 + e2+e4 > futil11 +e1+e2+e4	==> e1 < futil10 - futil11
    ! condition: futil10 + e2+e4 > futil12 +e3+e4		==> e3 < futil10 + e2 - futil12
    ! condition: futil10 + e2+e4 > futil13 +e1+e3+e4	==> e3 < futil10 + e2 - futil13 - e1
    ! condition: futil10 + e2+e4 > futil14 +e2+e3+e4	==> e3 < futil10 - futil14
    ! condition: futil10 + e2+e4 > futil15 +e1+e2+e3+e4	==> e3 < futil10 - futil15 - e1

	tb=NORMCDF(futil(10)-futil(11),0.0d0,1.0d0)
    wg(:)=tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,1)=-HUGE(v(j,1))
   	        ELSE
   	        v(j,1)=INVNORMCDF((tb(j)*draws(j,4)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	futil(8)-futil(10), &
        	futil(9)+v(j,1)-futil(10) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
        	    v(j,2)=HUGE(v(j,2))
   	        ELSE
           	v(j,2)=INVNORMCDF((ta(j)*(1-draws(j,3))+draws(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
			futil(10)+v(j,2)-futil(12), &
        	futil(10)+v(j,2)-futil(13)-v(j,1), &
        	futil(10)-futil(14), &
        	futil(10)-futil(15)-v(j,1) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,3)=-HUGE(v(j,3))
   	        ELSE
   	        v(j,3)=INVNORMCDF((tb(j)*draws(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
			-futil(10)-v(j,2), &
        	futil(1)+v(j,1)-futil(10)-v(j,2), &
        	futil(2)-futil(10), &
        	futil(3)+v(j,1)-futil(10), &
        	futil(4)+v(j,3)-futil(10)-v(j,2), &
        	futil(5)+v(j,1)+v(j,3)-futil(10)-v(j,2), &
        	futil(6)+v(j,3)-futil(10), &
        	futil(7)+v(j,1)+v(j,3)-futil(10) &
	        ) &
           -DOT_PRODUCT(lchol(4,1:3),v(j,1:3)) &
        ) &
        /lchol(4,4),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)

CASE(11)

    ! condition: u11 > 0, u1 - u10, u12 - u15
    ! condition: futil11 + e1+e2+e4 > 0						==> e4 > - futil11 - e1 - e2
    ! condition: futil11 + e1+e2+e4 > futil1 + e1			==> e4 > futil1 - futil11 - e2
    ! condition: futil11 + e1+e2+e4 > futil2 + e2			==> e4 > futil2 - futil11 - e1
    ! condition: futil11 + e1+e2+e4 > futil3 + e1+e2		==> e4 > futil3 - futil11
    ! condition: futil11 + e1+e2+e4 > futil4 + e3			==> e4 > futil4 + e3 - futil11 - e1 - e2
    ! condition: futil11 + e1+e2+e4 > futil5 + e1+e3		==> e4 > futil5 + e3 - futil11 - e2
    ! condition: futil11 + e1+e2+e4 > futil6 + e2+e3		==> e4 > futil6 + e3 - futil11 - e1
    ! condition: futil11 + e1+e2+e4 > futil7 + e1+e2+e3		==> e4 > futil7 + e3 - futil11
    ! condition: futil11 + e1+e2+e4 > futil8 + e4			==> e2 > futil8 - futil11 - e1
    ! condition: futil11 + e1+e2+e4 > futil9 + e1+e4		==> e2 > futil9 - futil11
    ! condition: futil11 + e1+e2+e4 > futil10 +e2+e4		==> e1 > futil10 - futil11
    ! condition: futil11 + e1+e2+e4 > futil12 +e3+e4		==> e3 < futil11 + e1+e2 - futil12
    ! condition: futil11 + e1+e2+e4 > futil13 +e1+e3+e4		==> e3 < futil11 + e2 - futil13
    ! condition: futil11 + e1+e2+e4 > futil14 +e2+e3+e4		==> e3 < futil11 + e1 - futil14
    ! condition: futil11 + e1+e2+e4 > futil15 +e1+e2+e3+e4	==> e3 < futil11 - futil15

    ta=NORMCDF(futil(10)-futil(11),0.0d0,1.0d0)
    wg(:)=1.0-ta
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,1)=HUGE(v(j,1))
            ELSE
            	v(j,1)=INVNORMCDF((ta(j)*(1-draws(j,4))+draws(j,4)),0.0d0,1.0d0)
	        ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	futil(8)-futil(11)-v(j,1), &
        	futil(9)-futil(10) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
        	    v(j,2)=HUGE(v(j,2))
   	        ELSE
           	v(j,2)=INVNORMCDF((ta(j)*(1-draws(j,3))+draws(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
			futil(11)+v(j,1)+v(j,2)-futil(12), &
        	futil(11)+v(j,2)-futil(13), &
        	futil(11)+v(j,1)-futil(14), &
        	futil(11)-futil(15) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,3)=-HUGE(v(j,3))
   	        ELSE
   	        v(j,3)=INVNORMCDF((tb(j)*draws(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
			-futil(11)-v(j,1)-v(j,2), &
        	futil(1)-futil(11)-v(j,2), &
        	futil(2)-futil(11)-v(j,1), &
        	futil(3)-futil(11), &
        	futil(4)+v(j,3)-futil(11)-v(j,1)-v(j,2), &
        	futil(5)+v(j,3)-futil(11)-v(j,2), &
        	futil(6)+v(j,3)-futil(11)-v(j,1), &
        	futil(7)+v(j,3)-futil(11) &
	        ) &
           -DOT_PRODUCT(lchol(4,1:3),v(j,1:3)) &
        ) &
        /lchol(4,4),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)

CASE(12)

    ! condition: u12 > 0, u1 - u11, u13 - u15
    ! condition: futil12 + e3+e4 > 0					==> e4 > - futil12 - e3
    ! condition: futil12 + e3+e4 > futil1 + e1			==> e4 > futil1 + e1 - futil12 - e3
    ! condition: futil12 + e3+e4 > futil2 + e2			==> e4 > futil2 + e2 - futil12 - e3
    ! condition: futil12 + e3+e4 > futil3 + e1+e2		==> e4 > futil3 + e1+e2 - futil12 - e3
    ! condition: futil12 + e3+e4 > futil4 + e3			==> e4 > futil4 - futil12
    ! condition: futil12 + e3+e4 > futil5 + e1+e3		==> e4 > futil5 + e1 - futil12
    ! condition: futil12 + e3+e4 > futil6 + e2+e3		==> e4 > futil6 + e2 - futil12
    ! condition: futil12 + e3+e4 > futil7 + e1+e2+e3	==> e4 > futil7 + e1+e2 - futil12
    ! condition: futil12 + e3+e4 > futil8 + e4			==> e3 > futil8 - futil12
    ! condition: futil12 + e3+e4 > futil9 + e1+e4		==> e3 > futil9 + e1 - futil12
    ! condition: futil12 + e3+e4 > futil10 +e2+e4		==> e3 > futil10 + e2 - futil12
    ! condition: futil12 + e3+e4 > futil11 +e1+e2+e4	==> e3 > futil11 + e1+e2 - futil12
    ! condition: futil12 + e3+e4 > futil13 +e1+e3+e4	==> e1 < futil12 - futil13
    ! condition: futil12 + e3+e4 > futil14 +e2+e3+e4	==> e2 < futil12 - futil14
    ! condition: futil12 + e3+e4 > futil15 +e1+e2+e3+e4	==> e2 < futil12 - futil15 - e1

	tb=NORMCDF(futil(12)-futil(13),0.0d0,1.0d0)
    wg(:)=tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,1)=-HUGE(v(j,1))
   	        ELSE
   	        v(j,1)=INVNORMCDF((tb(j)*draws(j,4)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
        	futil(12)-futil(15)-v(j,1), &
        	futil(12)-futil(14) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,2)=-HUGE(v(j,2))
   	        ELSE
   	        v(j,2)=INVNORMCDF((tb(j)*draws(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	futil(8)-futil(12), &
        	futil(9)+v(j,1)-futil(12), &
        	futil(10)+v(j,2)-futil(12), &
        	futil(11)+v(j,1)+v(j,2)-futil(12) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
        	    v(j,3)=HUGE(v(j,3))
   	        ELSE
           	v(j,3)=INVNORMCDF((ta(j)*(1-draws(j,2))+draws(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
			-futil(12)-v(j,3), &
        	futil(1)+v(j,1)-futil(12)-v(j,3), &
        	futil(2)+v(j,2)-futil(12)-v(j,3), &
        	futil(3)+v(j,1)+v(j,2)-futil(12)-v(j,3), &
        	futil(4)-futil(12), &
        	futil(5)+v(j,1)-futil(12), &
        	futil(6)+v(j,2)-futil(12), &
        	futil(7)+v(j,1)+v(j,2)-futil(12) &
	        ) &
           -DOT_PRODUCT(lchol(4,1:3),v(j,1:3)) &
        ) &
        /lchol(4,4),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)

CASE(13)

    ! condition: u13 > 0, u1 - u12, u14 - u15
    ! condition: futil13 + e1+e3+e4 > 0						==> e4 > - futil13 - e1 - e3
    ! condition: futil13 + e1+e3+e4 > futil1 + e1			==> e4 > futil1 - futil13 - e3
    ! condition: futil13 + e1+e3+e4 > futil2 + e2			==> e4 > futil2 + e2 - futil13 - e1 - e3
    ! condition: futil13 + e1+e3+e4 > futil3 + e1+e2		==> e4 > futil3 + e2 - futil13 - e3
    ! condition: futil13 + e1+e3+e4 > futil4 + e3			==> e4 > futil4 - futil13 - e1
    ! condition: futil13 + e1+e3+e4 > futil5 + e1+e3		==> e4 > futil5 - futil13
    ! condition: futil13 + e1+e3+e4 > futil6 + e2+e3		==> e4 > futil6 + e2 - futil13 - e1
    ! condition: futil13 + e1+e3+e4 > futil7 + e1+e2+e3		==> e4 > futil7 + e2 - futil13
    ! condition: futil13 + e1+e3+e4 > futil8 + e4			==> e3 > futil8 - futil13 - e1
    ! condition: futil13 + e1+e3+e4 > futil9 + e1+e4		==> e3 > futil9 - futil13
    ! condition: futil13 + e1+e3+e4 > futil10 +e2+e4		==> e3 > futil10 + e2 - futil13 - e1
    ! condition: futil13 + e1+e3+e4 > futil11 +e1+e2+e4		==> e3 > futil11 + e2 - futil13
    ! condition: futil13 + e1+e3+e4 > futil12 +e3+e4		==> e1 > futil12 - futil13
    ! condition: futil13 + e1+e3+e4 > futil14 +e2+e3+e4		==> e2 < futil13 + e1 - futil14
    ! condition: futil13 + e1+e3+e4 > futil15 +e1+e2+e3+e4	==> e2 < futil13 - futil15

    ta=NORMCDF(futil(12)-futil(13),0.0d0,1.0d0)
    wg(:)=1.0-ta
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,1)=HUGE(v(j,1))
            ELSE
            	v(j,1)=INVNORMCDF((ta(j)*(1-draws(j,4))+draws(j,4)),0.0d0,1.0d0)
	        ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
        	futil(13)+v(j,1)-futil(14), &
        	futil(13)-futil(15) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,2)=-HUGE(v(j,2))
   	        ELSE
   	        v(j,2)=INVNORMCDF((tb(j)*draws(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	futil(8)-futil(13)-v(j,1), &
        	futil(9)-futil(13), &
        	futil(10)+v(j,2)-futil(13)-v(j,1), &
        	futil(11)+v(j,2)-futil(13) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
        	    v(j,3)=HUGE(v(j,3))
   	        ELSE
           	v(j,3)=INVNORMCDF((ta(j)*(1-draws(j,2))+draws(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
			-futil(13)-v(j,1)-v(j,3), &
        	futil(1)-futil(13)-v(j,3), &
        	futil(2)+v(j,2)-futil(13)-v(j,1)-v(j,3), &
        	futil(3)+v(j,2)-futil(13)-v(j,3), &
        	futil(4)-futil(13)-v(j,1), &
        	futil(5)-futil(13), &
        	futil(6)+v(j,2)-futil(13)-v(j,1), &
        	futil(7)+v(j,2)-futil(13) &
	        ) &
           -DOT_PRODUCT(lchol(4,1:3),v(j,1:3)) &
        ) &
        /lchol(4,4),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)

CASE(14)

    ! condition: u14 > 0, u1 - u13, u15
    ! condition: futil14 + e2+e3+e4 > 0						==> e4 > - futil14 - e2 - e3
    ! condition: futil14 + e2+e3+e4 > futil1 + e1			==> e4 > futil1 +e1 - futil14 - e2 - e3
    ! condition: futil14 + e2+e3+e4 > futil2 + e2			==> e4 > futil2 - futil14 - e3
    ! condition: futil14 + e2+e3+e4 > futil3 + e1+e2		==> e4 > futil3 + e1 - futil14 - e3
    ! condition: futil14 + e2+e3+e4 > futil4 + e3			==> e4 > futil4 - futil14 - e2
    ! condition: futil14 + e2+e3+e4 > futil5 + e1+e3		==> e4 > futil5 + e1 - futil14 - e2
    ! condition: futil14 + e2+e3+e4 > futil6 + e2+e3		==> e4 > futil6 - futil14
    ! condition: futil14 + e2+e3+e4 > futil7 + e1+e2+e3		==> e4 > futil7 + e1 - futil14
    ! condition: futil14 + e2+e3+e4 > futil8 + e4			==> e3 > futil8 - futil14 - e2
    ! condition: futil14 + e2+e3+e4 > futil9 + e1+e4		==> e3 > futil9 + e1 - futil14 - e2
    ! condition: futil14 + e2+e3+e4 > futil10 +e2+e4		==> e3 > futil10 - futil14
    ! condition: futil14 + e2+e3+e4 > futil11 +e1+e2+e4		==> e3 > futil11 + e1 - futil14
    ! condition: futil14 + e2+e3+e4 > futil12 +e3+e4		==> e2 > futil12 - futil14
    ! condition: futil14 + e2+e3+e4 > futil13 +e1+e3+e4		==> e2 > futil13 + e1 - futil14
    ! condition: futil14 + e2+e3+e4 > futil15 +e1+e2+e3+e4	==> e1 < futil14 - futil15

	tb=NORMCDF(futil(14)-futil(15),0.0d0,1.0d0)
    wg(:)=tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,1)=-HUGE(v(j,1))
   	        ELSE
   	        v(j,1)=INVNORMCDF((tb(j)*draws(j,4)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	futil(12)-futil(14), &
        	futil(13)+v(j,1)-futil(14) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
        	    v(j,2)=HUGE(v(j,2))
   	        ELSE
           	v(j,2)=INVNORMCDF((ta(j)*(1-draws(j,3))+draws(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	futil(8)-futil(14)-v(j,2), &
        	futil(9)+v(j,1)-futil(14)-v(j,2), &
        	futil(10)-futil(14), &
        	futil(11)+v(j,1)-futil(14) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
        	    v(j,3)=HUGE(v(j,3))
   	        ELSE
           	v(j,3)=INVNORMCDF((ta(j)*(1-draws(j,2))+draws(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
			-futil(14)-v(j,2)-v(j,3), &
        	futil(1)+v(j,1)-futil(14)-v(j,2)-v(j,3), &
        	futil(2)-futil(14)-v(j,3), &
        	futil(3)+v(j,1)-futil(14)-v(j,3), &
        	futil(4)-futil(14)-v(j,2), &
        	futil(5)+v(j,1)-futil(14)-v(j,2), &
        	futil(6)-futil(14), &
        	futil(7)+v(j,1)-futil(14) &
	        ) &
           -DOT_PRODUCT(lchol(4,1:3),v(j,1:3)) &
        ) &
        /lchol(4,4),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)




CASE(15)

    ! condition: u15 > 0, u1 - u14
    ! condition: futil15 + e1+e2+e3+e4 > 0					==> e4 > - futil15 - e1 - e2 - e3
    ! condition: futil15 + e1+e2+e3+e4 > futil1 + e1		==> e4 > futil1 - futil15 - e2 - e3
    ! condition: futil15 + e1+e2+e3+e4 > futil2 + e2		==> e4 > futil2 - futil15 - e1 - e3
    ! condition: futil15 + e1+e2+e3+e4 > futil3 + e1+e2		==> e4 > futil3 - futil15 - e3
    ! condition: futil15 + e1+e2+e3+e4 > futil4 + e3		==> e4 > futil4 - futil15 - e1 - e2
    ! condition: futil15 + e1+e2+e3+e4 > futil5 + e1+e3		==> e4 > futil5 - futil15 - e2
    ! condition: futil15 + e1+e2+e3+e4 > futil6 + e2+e3		==> e4 > futil6 - futil15 - e1
    ! condition: futil15 + e1+e2+e3+e4 > futil7 + e1+e2+e3	==> e4 > futil7 - futil15
    ! condition: futil15 + e1+e2+e3+e4 > futil8 + e4		==> e3 > futil8 - futil15 - e1 - e2
    ! condition: futil15 + e1+e2+e3+e4 > futil9 + e1+e4		==> e3 > futil9 - futil15 - e2
    ! condition: futil15 + e1+e2+e3+e4 > futil10 +e2+e4		==> e3 > futil10 - futil15 - e1
    ! condition: futil15 + e1+e2+e3+e4 > futil11 +e1+e2+e4	==> e3 > futil11 - futil15
    ! condition: futil15 + e1+e2+e3+e4 > futil12 +e3+e4		==> e2 > futil12 - futil15 - e1
    ! condition: futil15 + e1+e2+e3+e4 > futil13 +e1+e3+e4	==> e2 > futil13 - futil15
    ! condition: futil15 + e1+e2+e3+e4 > futil14 +e2+e3+e4	==> e1 > futil14 - futil15

    ta=NORMCDF(futil(14)-futil(15),0.0d0,1.0d0)
    wg(:)=1.0-ta
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,1)=HUGE(v(j,1))
            ELSE
            	v(j,1)=INVNORMCDF((ta(j)*(1-draws(j,4))+draws(j,4)),0.0d0,1.0d0)
	        ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	futil(12)-futil(15)-v(j,1), &
        	futil(13)-futil(15) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
        	    v(j,2)=HUGE(v(j,2))
   	        ELSE
           	v(j,2)=INVNORMCDF((ta(j)*(1-draws(j,3))+draws(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	futil(8)-futil(15)-v(j,1)-v(j,2), &
        	futil(9)-futil(15)-v(j,2), &
        	futil(10)-futil(15)-v(j,1), &
        	futil(11)-futil(15) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
        	    v(j,3)=HUGE(v(j,3))
   	        ELSE
           	v(j,3)=INVNORMCDF((ta(j)*(1-draws(j,2))+draws(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
			-futil(15)-v(j,1)-v(j,2)-v(j,3), &
        	futil(1)-futil(15)-v(j,2)-v(j,3), &
        	futil(2)-futil(15)-v(j,1)-v(j,3), &
        	futil(3)-futil(15)-v(j,3), &
        	futil(4)-futil(15)-v(j,1)-v(j,2), &
        	futil(5)-futil(15)-v(j,2), &
        	futil(6)-futil(15)-v(j,1), &
        	futil(7)-futil(15) &
	        ) &
           -DOT_PRODUCT(lchol(4,1:3),v(j,1:3)) &
        ) &
        /lchol(4,4),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)

CASE DEFAULT
!    PRINT *,'nchild(m): ', nchild(m)
!    PRINT *,'error'
!    PAUSE
END SELECT

ghkval=SUM(wg(:))/ns


IF (ghkval<=0.0) THEN
!	PRINT *,'Error found!'
!	PRINT *,'GHK Value: ', ghkval
!	PRINT *,'nmkt: ', nmkt
	ghkval=TINY(ghkval)
!	STOP
ENDIF

! PRINT *, 'GHK-out'

END FUNCTION GHK4






FUNCTION lcholbase(m,param,n) RESULT(lchol)
! VCOV matrix and Cholesky decomposition matrix
USE Globvar
USE MATRIX
IMPLICIT NONE
INTEGER, INTENT(IN) :: m			! family id
INTEGER, INTENT(IN) :: n 
REAL(8), INTENT(IN) :: param(n)
REAL(8) :: lchol(nchild(m),nchild(m)), omega(nchild(m),nchild(m))
INTEGER :: i

omega=0.0000d0
DO i = 1, nchild(m)
omega(i,i)=1.0000d0
END DO

SELECT CASE(nchild(m))
CASE(1)
omega(1,1)=1.0000d0
CASE(2)

omega(1,2)=DOT_PRODUCT(xr(famindex(m),:,1),param(nthetab+nthetaa+nthetag+1:ntheta))
    ! The following lines helps to contain omega values between -0.9 and 0.9 and get convergence
    IF (omega(1,2)<-0.8) omega(1,2)=-0.8d0-((DEXP(-omega(1,2)-0.8d0)/(1+DEXP(-omega(1,2)-0.8d0)))-0.5d0)*0.2d0
    IF (omega(1,2)> 0.8) omega(1,2)= 0.8d0+((DEXP( omega(1,2)-0.8d0)/(1+DEXP( omega(1,2)-0.8d0)))-0.5d0)*0.2d0
omega(2,1)=omega(1,2)

CASE(3)

omega(1,2)=DOT_PRODUCT(xr(famindex(m),:,2),param(nthetab+nthetaa+nthetag+1:ntheta))
    ! The following lines helps to contain omega values between -0.9 and 0.9 and get convergence
    IF (omega(1,2)<-0.8) omega(1,2)=-0.8d0-((DEXP(-omega(1,2)-0.8d0)/(1+DEXP(-omega(1,2)-0.8d0)))-0.5d0)*0.2d0
    IF (omega(1,2)> 0.8) omega(1,2)= 0.8d0+((DEXP( omega(1,2)-0.8d0)/(1+DEXP( omega(1,2)-0.8d0)))-0.5d0)*0.2d0
omega(2,1)=omega(1,2)

omega(1,3)=DOT_PRODUCT(xr(famindex(m),:,1),param(nthetab+nthetaa+nthetag+1:ntheta))
    IF (omega(1,3)<-0.8) omega(1,3)=-0.8d0-((DEXP(-omega(1,3)-0.8d0)/(1+DEXP(-omega(1,3)-0.8d0)))-0.5d0)*0.2d0
    IF (omega(1,3)> 0.8) omega(1,3)= 0.8d0+((DEXP( omega(1,3)-0.8d0)/(1+DEXP( omega(1,3)-0.8d0)))-0.5d0)*0.2d0
omega(3,1)=omega(1,3)

omega(2,3)=DOT_PRODUCT(xr(famindex(m)-1,:,1),param(nthetab+nthetaa+nthetag+1:ntheta))
    IF (omega(2,3)<-0.8) omega(2,3)=-0.8d0-((DEXP(-omega(2,3)-0.8d0)/(1+DEXP(-omega(2,3)-0.8d0)))-0.5d0)*0.2d0
    IF (omega(2,3)> 0.8) omega(2,3)= 0.8d0+((DEXP( omega(2,3)-0.8d0)/(1+DEXP( omega(2,3)-0.8d0)))-0.5d0)*0.2d0
omega(3,2)=omega(2,3)

CASE(4)

omega(1,2)=DOT_PRODUCT(xr(famindex(m),:,3),param(nthetab+nthetaa+nthetag+1:ntheta))
    ! The following lines helps to contain omega values between -0.9 and 0.9 and get convergence
    IF (omega(1,2)<-0.8) omega(1,2)=-0.8d0-((DEXP(-omega(1,2)-0.8d0)/(1+DEXP(-omega(1,2)-0.8d0)))-0.5d0)*0.2d0
    IF (omega(1,2)> 0.8) omega(1,2)= 0.8d0+((DEXP( omega(1,2)-0.8d0)/(1+DEXP( omega(1,2)-0.8d0)))-0.5d0)*0.2d0
omega(2,1)=omega(1,2)

omega(1,3)=DOT_PRODUCT(xr(famindex(m),:,2),param(nthetab+nthetaa+nthetag+1:ntheta))
    IF (omega(1,3)<-0.8) omega(1,3)=-0.8d0-((DEXP(-omega(1,3)-0.8d0)/(1+DEXP(-omega(1,3)-0.8d0)))-0.5d0)*0.2d0
    IF (omega(1,3)> 0.8) omega(1,3)= 0.8d0+((DEXP( omega(1,3)-0.8d0)/(1+DEXP( omega(1,3)-0.8d0)))-0.5d0)*0.2d0
omega(3,1)=omega(1,3)

omega(2,3)=DOT_PRODUCT(xr(famindex(m)-1,:,2),param(nthetab+nthetaa+nthetag+1:ntheta))
    IF (omega(2,3)<-0.8) omega(2,3)=-0.8d0-((DEXP(-omega(2,3)-0.8d0)/(1+DEXP(-omega(2,3)-0.8d0)))-0.5d0)*0.2d0
    IF (omega(2,3)> 0.8) omega(2,3)= 0.8d0+((DEXP( omega(2,3)-0.8d0)/(1+DEXP( omega(2,3)-0.8d0)))-0.5d0)*0.2d0
omega(3,2)=omega(2,3)

omega(1,4)=DOT_PRODUCT(xr(famindex(m),:,1),param(nthetab+nthetaa+nthetag+1:ntheta))
    IF (omega(1,4)<-0.8) omega(1,4)=-0.8d0-((DEXP(-omega(1,4)-0.8d0)/(1+DEXP(-omega(1,4)-0.8d0)))-0.5d0)*0.2d0
    IF (omega(1,4)> 0.8) omega(1,4)= 0.8d0+((DEXP( omega(1,4)-0.8d0)/(1+DEXP( omega(1,4)-0.8d0)))-0.5d0)*0.2d0
omega(4,1)=omega(1,4)

omega(2,4)=DOT_PRODUCT(xr(famindex(m)-1,:,1),param(nthetab+nthetaa+nthetag+1:ntheta))
    IF (omega(2,4)<-0.8) omega(2,4)=-0.8d0-((DEXP(-omega(2,4)-0.8d0)/(1+DEXP(-omega(2,4)-0.8d0)))-0.5d0)*0.2d0
    IF (omega(2,4)> 0.8) omega(2,4)= 0.8d0+((DEXP( omega(2,4)-0.8d0)/(1+DEXP( omega(2,4)-0.8d0)))-0.5d0)*0.2d0
omega(4,2)=omega(2,4)

omega(3,4)=DOT_PRODUCT(xr(famindex(m)-2,:,1),param(nthetab+nthetaa+nthetag+1:ntheta))
    IF (omega(3,4)<-0.8) omega(3,4)=-0.8d0-((DEXP(-omega(3,4)-0.8d0)/(1+DEXP(-omega(3,4)-0.8d0)))-0.5d0)*0.2d0
    IF (omega(3,4)> 0.8) omega(3,4)= 0.8d0+((DEXP( omega(3,4)-0.8d0)/(1+DEXP( omega(3,4)-0.8d0)))-0.5d0)*0.2d0
omega(4,3)=omega(3,4)

CASE DEFAULT
    PRINT *,'nchild(m): ', nchild(m)
    PRINT *,'error'
    PAUSE
END SELECT

lchol=CHOL(omega)

END FUNCTION lcholbase




FUNCTION nearcount(ent,n)
USE Globvar
IMPLICIT NONE
INTEGER, INTENT(IN) :: ent    ! family configuration
INTEGER, INTENT(IN) :: n      ! nchild
INTEGER :: i
REAL(8) :: nearcount

nearcount=0.0d0
DO i=0, n-1
    nearcount=nearcount+IBITS(ent,i,1)
END DO

END FUNCTION nearcount


FUNCTION nearcounti(ent,n)
USE Globvar
IMPLICIT NONE
INTEGER, INTENT(IN) :: ent    ! family configuration
INTEGER, INTENT(IN) :: n      ! nchild
INTEGER :: i
INTEGER :: nearcounti

nearcounti=0
IF (n>=0) THEN
DO i=0, n-1
    nearcounti=nearcounti+IBITS(ent,i,1)
END DO
END IF

END FUNCTION nearcounti




FUNCTION var_cov(param,n)
USE Globvar
! This function computes the VCov matrix of the estimates
! By using the numerical gradient of a function using forward differences
IMPLICIT NONE
INTEGER, INTENT(IN) :: n
REAL(8), INTENT(IN) :: param(n)

REAL(8) :: var_cov(n,n)
REAL(8) :: a(n,n)
REAL(8) :: f0			! Initial value of lnmlobj.
REAL(8) :: grdd(n),dh(n), arg(n,n)
REAL(8) :: ax0(n),xdh(n)
REAL(8) :: dax0(n),one
REAL(8) :: indgrad(n)
REAL(8) :: lchol(nchildmax,nchildmax)
INTEGER :: i, j, k

! Set up the gradient calculation
one  = 1.0D0	! This parameter affects the length of dh
grdd = 0.0D0
	! Computation of stepsize (dh) for gradient
ax0  = ABS(param)
dax0 = 1.0D0
WHERE (param.ne.0.0D0) dax0 = param/ax0
dh = (1.0D-6)*(1.0D-2)*one*dax0
WHERE (ax0.gt.(1.0D-2)*one) dh = (1.0D-8)*ax0*dax0
xdh = param+dh

! Calculate VCOV
a=0.0d0

!$OMP PARALLEL DO SCHEDULE(STATIC) PRIVATE(f0,grdd,arg,indgrad) 
DO i = 1, nfam

arg = SPREAD(param,DIM=2,NCOPIES=n)

SELECT CASE(nchild(i))
CASE(1)
	f0=DLOG(GHK1(i,param,n))
CASE(2)
	f0=DLOG(GHK2(i,param,n,draws2seed))
CASE(3)
	f0=DLOG(GHK3(i,param,n,draws3seed))
CASE(4)
	f0=DLOG(GHK4(i,param,n,draws4seed))
CASE DEFAULT
    PRINT *,'error'
    PAUSE
END SELECT

		DO j = 1, n
			arg(j,j)=xdh(j)

        SELECT CASE(nchild(i))
        CASE(1)
	        grdd(j)=DLOG(GHK1(i,arg(:,j),n))
        CASE(2)
	        grdd(j)=DLOG(GHK2(i,arg(:,j),n,draws2seed))
        CASE(3)
	        grdd(j)=DLOG(GHK3(i,arg(:,j),n,draws3seed))
        CASE(4)
	        grdd(j)=DLOG(GHK4(i,arg(:,j),n,draws4seed))
        CASE DEFAULT
            PRINT *,'error'
            PAUSE
        END SELECT


		END DO
!print *,'f0: ',f0
!print *,'grdd: ',grdd
!print *,'diff: ',grdd-f0
	indgrad= (grdd-f0)/dh
!print *,indgrad
	a=a+outerprod(indgrad,indgrad)
!print *,outerprod(indgrad,indgrad)
END DO
!$OMP END PARALLEL DO

print *,'check'


DO i=1,ntheta
DO j=1,ntheta
	print *,'a(i,j):  ',i,j, a(i,j)
END DO
END DO

print *,'inverting var cov matrix'
var_cov=MATINV(a)

END FUNCTION var_cov




FUNCTION FAMSOLVE(param,n)
	! This function calculates 
USE Globvar
USE PROBABILITY
IMPLICIT NONE
INTEGER, INTENT(IN) :: n 
REAL(8), INTENT(IN) :: param(n)

REAL(8) :: FAMSOLVE(0:2**nchildmax-1,4)

INTEGER :: i, j, k, l, m 
INTEGER :: s0, s1, s2, s3, s4
! Specify the array sizes based on the number of players (i.e. possible number of strategy combinations)
REAL(8) :: wg(ns)
REAL(8) :: ta(ns), tb(ns)
REAL(8) :: v(ns,nchildmax-1)   ! GHK simulation draws constructed from draws()
REAL(8) :: futil(1:nchildmax*nchildmax-1)  ! futil(1) ... family total utility when only last child lives near
REAL(8) :: configprob(0:15)
INTEGER :: predconfig      ! predicted location configuration
REAL(8) :: temp
REAL(8) :: utilsub(nchildmax), alpha(nchildmax,3),gamma(nchildmax,2)		! 1 ... oldest child
REAL(8) :: lchol(nchildmax,nchildmax)



FAMSOLVE=0

DO m=1,nfam

k=famindex(m)
wg=0.0d0
v=0.0d0
configprob=0.0d0

SELECT CASE(nchild(m))

CASE(1)

utilsub(1:1)=MATMUL(xb(k:k,:),param(1:nthetab))
    IF(nthetaa>=3) THEN
    alpha(1:1,1)  = EXP( MATMUL(xa(k:k,:),param(nthetab+1:nthetab+nthetaa-2)) )
    ELSEIF(nthetaa==1) THEN
    alpha(1:1,1)  =param(nthetab+nthetaa)
    ELSE
    alpha(1:1,1) = 0.0d0
    ENDIF

    FAMSOLVE(1,1)=FAMSOLVE(1,1)+1.0d0-NORMCDF(-utilsub(1)-alpha(1,1),0.0d0,1.0d0)
    FAMSOLVE(0,1)=FAMSOLVE(0,1)+      NORMCDF(-utilsub(1)-alpha(1,1),0.0d0,1.0d0)

    configprob(1)=1.0d0-NORMCDF(-utilsub(1)-alpha(1,1),0.0d0,1.0d0)
    configprob(0)=NORMCDF(-utilsub(1)-alpha(1,1),0.0d0,1.0d0)

CASE(2)

utilsub(1:2)=MATMUL(xb(k-1:k,:),param(1:nthetab))
	IF (nthetaa>=3) THEN
    alpha(1:2,1)  =MATMUL(xa(k-1:k,:),param(nthetab+1:nthetab+nthetaa-2))
    alpha(1:2,2)=param(nthetab+nthetaa-1)
    alpha(1:2,3)=param(nthetab+nthetaa)
	ELSEIF (nthetaa==1) THEN
	alpha(1:2,1)=param(nthetab+nthetaa)
	alpha(1:2,2:3)=0.0d0
	ELSE
	alpha=0.0d0
	ENDIF
	IF (nthetag>=2) THEN
	gamma(1:2,1) =MATMUL(xg(k-1:k,:),param(nthetab+nthetaa+1:nthetab+nthetaa+nthetag-1))
	gamma(1:2,2) =param(nthetab+nthetaa+nthetag)
	ELSEIF (nthetag==1) THEN
	gamma(1:2,1)=param(nthetab+nthetaa+nthetag)
	gamma(1:2,2)=0.0d0
	ELSE
	gamma=0.0d0
	ENDIF

lchol(1:2,1:2)=lcholbase(m,param,n)

futil(1)=utilsub(2)+EXP(alpha(1,1))+EXP(alpha(2,1))
futil(2)=utilsub(1)+EXP(alpha(1,1))+EXP(alpha(2,1))
futil(3)=utilsub(1)+utilsub(2) + EXP( alpha(1,1)+alpha(1,2) ) + EXP( alpha(2,1)+alpha(2,2) ) + gamma(1,1)+gamma(2,1)

! Configuration 0
    ! condition: 0 > u1, u2, u3
    ! condition: 0 > futil1 + e1    ==> e1 < - futil1
    ! condition: 0 > futil2 + e2    ==> e2 < - futil2
    ! condition: 0 > futil3 + e1+e2 ==> e2 < - futil3 - e1

    tb=NORMCDF(-futil(1),0.0d0,1.0d0)
    wg(:)=tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,1)=-HUGE(v(j,1))
   	        ELSE
   	        v(j,1)=INVNORMCDF((tb(j)*draws2seed(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( &
           MIN(-futil(2),-futil(3)-v(j,1)) &
           -DOT_PRODUCT(lchol(nchild(m),1:nchild(m)-1),v(j,1:1)) &
        ) &
        /lchol(nchild(m),nchild(m)),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb


    FAMSOLVE(0,2)=FAMSOLVE(0,2)+SUM(wg(:))/ns
    configprob(0)=SUM(wg(:))/ns


! Configuration 1
    ! condition: u1 > 0, u2, u3
    ! condition: futil1 + e1 > 0              ==> e1 > - futil1
    ! condition: futil1 + e1 > futil2 + e2    ==> e2 < futil1 - futil2 + e1
    ! condition: futil1 + e1 > futil3 + e1+e2 ==> e2 < futil1 - futil3

    ta=NORMCDF(-futil(1),0.0d0,1.0d0)
    wg(:)=1.0-ta
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,1)=HUGE(v(j,1))
            ELSE
            	v(j,1)=INVNORMCDF((ta(j)*(1-draws2seed(j,2))+draws2seed(j,2)),0.0d0,1.0d0)
	        ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( &
           MIN(futil(1)-futil(3),futil(1)-futil(2)+v(j,1)) &
           -DOT_PRODUCT(lchol(nchild(m),1:nchild(m)-1),v(j,1:1)) &
        ) &
        /lchol(nchild(m),nchild(m)),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb


    FAMSOLVE(1,2)=FAMSOLVE(1,2)+SUM(wg(:))/ns
    configprob(1)=SUM(wg(:))/ns


! Configuration 2
    ! condition: u2 > 0, u1, u3
    ! condition: futil2 + e2 > 0              ==> e2 > - futil2
    ! condition: futil2 + e2 > futil1 + e1    ==> e2 > futil1 - futil2 + e1
    ! condition: futil2 + e2 > futil3 + e1+e2 ==> e1 < futil2 - futil3

    tb=NORMCDF(futil(2)-futil(3),0.0d0,1.0d0)
    wg(:)=tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,1)=-HUGE(v(j,1))
   	        ELSE
   	        v(j,1)=INVNORMCDF((tb(j)*draws2seed(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( &
           MAX(-futil(2),futil(1)-futil(2)+v(j,1)) &
           -DOT_PRODUCT(lchol(nchild(m),1:nchild(m)-1),v(j,1:1)) &
        ) &
        /lchol(nchild(m),nchild(m)),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)


    FAMSOLVE(2,2)=FAMSOLVE(2,2)+SUM(wg(:))/ns
    configprob(2)=SUM(wg(:))/ns


! Configuration 3
    ! condition: u3 > 0, u1, u2
    ! condition: futil3 + e1+e2 > 0           ==> e2 > - futil2 - e1
    ! condition: futil3 + e1+e2 > futil1 + e1 ==> e2 > futil1 - futil3
    ! condition: futil3 + e1+e2 > futil2 + e2 ==> e1 > futil2 - futil3

    ta=NORMCDF(futil(2)-futil(3),0.0d0,1.0d0)
    wg(:)=1.0-ta

        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,1)=HUGE(v(j,1))
            ELSE
            	v(j,1)=INVNORMCDF((ta(j)*(1-draws2seed(j,2))+draws2seed(j,2)),0.0d0,1.0d0)
	        ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( &
           MAX(-futil(2)-v(j,1),futil(1)-futil(3)) &
           -DOT_PRODUCT(lchol(nchild(m),1:nchild(m)-1),v(j,1:1)) &
        ) &
        /lchol(nchild(m),nchild(m)),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)


    FAMSOLVE(3,2)=FAMSOLVE(3,2)+SUM(wg(:))/ns
    configprob(3)=SUM(wg(:))/ns




CASE(3)

utilsub(1:3)=MATMUL(xb(k-2:k,:),param(1:nthetab))
	IF (nthetaa>=3) THEN
    alpha(1:3,1)=MATMUL(xa(k-2:k,:),param(nthetab+1:nthetab+nthetaa-2))
    alpha(1:3,2)=param(nthetab+nthetaa-1)
    alpha(1:3,3)=param(nthetab+nthetaa)
	ELSEIF (nthetaa==1) THEN
	alpha(1:3,1)=param(nthetab+nthetaa)
	alpha(1:3,2:3)=0.0d0
	ELSE
	alpha=0.0d0
	ENDIF
	IF (nthetag>=2) THEN
	gamma(1:3,1) =MATMUL(xg(k-2:k,:),param(nthetab+nthetaa+1:nthetab+nthetaa+nthetag-1))
	gamma(1:3,2) =param(nthetab+nthetaa+nthetag)
	ELSEIF (nthetag==1) THEN
	gamma(1:3,1)=param(nthetab+nthetaa+nthetag)
	gamma(1:3,2)=0.0d0
	ELSE
	gamma=0.0d0
	ENDIF

lchol(1:3,1:3)=lcholbase(m,param,n)

futil(1)=utilsub(3) + EXP( alpha(1,1) ) + EXP( alpha(2,1) ) + EXP( alpha(3,1) )
futil(2)=utilsub(2) + EXP( alpha(1,1) ) + EXP( alpha(2,1) ) + EXP( alpha(3,1) )
futil(3)=utilsub(2)+utilsub(3) + EXP( alpha(1,1)+alpha(1,2) ) + EXP( alpha(2,1)+alpha(2,2) ) + EXP( alpha(3,1)+alpha(3,2) ) + gamma(2,1)+gamma(3,1)
futil(4)=utilsub(1) + EXP( alpha(1,1) ) + EXP( alpha(2,1) ) + EXP( alpha(3,1) )
futil(5)=utilsub(1)+utilsub(3) + EXP( alpha(1,1)+alpha(1,2) ) + EXP( alpha(2,1)+alpha(2,2) ) + EXP( alpha(3,1)+alpha(3,2) ) + gamma(1,1)+gamma(3,1)
futil(6)=utilsub(1)+utilsub(2) + EXP( alpha(1,1)+alpha(1,2) ) + EXP( alpha(2,1)+alpha(2,2) ) + EXP( alpha(3,1)+alpha(3,2) ) + gamma(1,1)+gamma(2,1)
futil(7)=utilsub(1)+utilsub(2)+utilsub(3) &
	+ EXP( alpha(1,1)+alpha(1,2)+alpha(1,3) ) &
	+ EXP( alpha(2,1)+alpha(2,2)+alpha(2,3) ) &
	+ EXP( alpha(3,1)+alpha(3,2)+alpha(3,3) ) &
	+ gamma(1,1)*(1.0d0+gamma(1,2))+gamma(2,1)*(1.0d0+gamma(2,2))+gamma(3,1)*(1.0d0+gamma(3,2))


! Configuration 0
    ! condition: 0 > u1, u2, u3, u4, u5, u6, u7
    ! condition: 0 > futil1 + e1    ==> e1 < - futil1
    ! condition: 0 > futil2 + e2    ==> e2 < - futil2
    ! condition: 0 > futil3 + e1+e2 ==> e2 < - futil3 - e1
    ! condition: 0 > futil4 + e3    ==> e3 < - futil4
    ! condition: 0 > futil5 + e1+e3 ==> e3 < - futil5 - e1
    ! condition: 0 > futil6 + e2+e3 ==> e3 < - futil6 - e2
    ! condition: 0 > futil7 + e1+e2+e3 ==> e3 < - futil7 - e2 - e1

    tb=NORMCDF(-futil(1),0.0d0,1.0d0)
    wg(:)=tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,1)=-HUGE(v(j,1))
   	        ELSE
   	        v(j,1)=INVNORMCDF((tb(j)*draws3seed(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
        	-futil(2), &
        	-futil(3)-v(j,1) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,2)=-HUGE(v(j,2))
   	        ELSE
   	        v(j,2)=INVNORMCDF((tb(j)*draws3seed(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
        	-futil(4), &
        	-futil(5)-v(j,1), &
        	-futil(6)-v(j,2), &
        	-futil(7)-v(j,1)-v(j,2) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb


    FAMSOLVE(0,3)=FAMSOLVE(0,3)+SUM(wg(:))/ns
    configprob(0)=SUM(wg(:))/ns


! Configuration 1
    ! condition: u1 > 0, u2, u3, u4, u5, u6, u7
    ! condition: futil1 + e1 > 0					==> e1 > - futil1
    ! condition: futil1 + e1 > futil2 + e2			==> e2 < futil1 + e1 - futil2
    ! condition: futil1 + e1 > futil3 + e1+e2		==> e2 < futil1 - futil3
    ! condition: futil1 + e1 > futil4 + e3			==> e3 < futil1 + e1 - futil4
    ! condition: futil1 + e1 > futil5 + e1+e3		==> e3 < futil1 - futil5
    ! condition: futil1 + e1 > futil6 + e2+e3		==> e3 < futil1 + e1 - futil6 - e2
    ! condition: futil1 + e1 > futil7 + e1+e2+e3	==> e3 < futil1 - futil7 - e2

    ta=NORMCDF(-futil(1),0.0d0,1.0d0)
    wg(:)=1.0-ta
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,1)=HUGE(v(j,1))
            ELSE
            	v(j,1)=INVNORMCDF((ta(j)*(1-draws3seed(j,3))+draws3seed(j,3)),0.0d0,1.0d0)
	        ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
        	futil(1)+v(j,1)-futil(2), &
        	futil(1)-futil(3) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,2)=-HUGE(v(j,2))
   	        ELSE
   	        v(j,2)=INVNORMCDF((tb(j)*draws3seed(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
        	futil(1)+v(j,1)-futil(4), &
        	futil(1)-futil(5), &
        	futil(1)+v(j,1)-futil(6)-v(j,2), &
        	futil(1)-futil(7)-v(j,2) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb


    FAMSOLVE(1,3)=FAMSOLVE(1,3)+SUM(wg(:))/ns
    configprob(1)=SUM(wg(:))/ns

! Configuration 2
    ! condition: u2 > 0, u1, u3, u4, u5, u6, u7
    ! condition: futil2 + e2 > 0					==> e2 > - futil2
    ! condition: futil2 + e2 > futil1 + e1			==> e2 > futil1 + e1 - futil2
    ! condition: futil2 + e2 > futil3 + e1+e2		==> e1 < futil2- futil3
    ! condition: futil2 + e2 > futil4 + e3			==> e3 < futil2 + e2 - futil4
    ! condition: futil2 + e2 > futil5 + e1+e3		==> e3 < futil2 + e2 - futil5 - e1
    ! condition: futil2 + e2 > futil6 + e2+e3		==> e3 < futil2 - futil6
    ! condition: futil2 + e2 > futil7 + e1+e2+e3	==> e3 < futil2 - futil7 - e1

    tb=NORMCDF(futil(2)-futil(3),0.0d0,1.0d0)
    wg(:)=tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,1)=-HUGE(v(j,1))
   	        ELSE
   	        v(j,1)=INVNORMCDF((tb(j)*draws3seed(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	-futil(2), &
        	futil(1)+v(j,1)-futil(2) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,2)=HUGE(v(j,2))
   	        ELSE
           	v(j,2)=INVNORMCDF((ta(j)*(1-draws3seed(j,2))+draws3seed(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
        	futil(2)+v(j,2)-futil(4), &
        	futil(2)+v(j,2)-futil(5)-v(j,1), &
        	futil(2)-futil(6), &
        	futil(2)-futil(7)-v(j,1) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb


    FAMSOLVE(2,3)=FAMSOLVE(2,3)+SUM(wg(:))/ns
    configprob(2)=SUM(wg(:))/ns

! Configuration 3
    ! condition: u3 > 0, u1, u2, u4, u5, u6, u7
    ! condition: futil3 + e1+e2 > 0					==> e2 > - futil3 - e1
    ! condition: futil3 + e1+e2 > futil1 + e1		==> e2 > futil1 - futil3
    ! condition: futil3 + e1+e2 > futil2 + e2		==> e1 > futil2 - futil3
    ! condition: futil3 + e1+e2 > futil4 + e3		==> e3 < futil3 + e1+e2 - futil4
    ! condition: futil3 + e1+e2 > futil5 + e1+e3	==> e3 < futil3 + e2 - futil5
    ! condition: futil3 + e1+e2 > futil6 + e2+e3	==> e3 < futil3 + e1 - futil6
    ! condition: futil3 + e1+e2 > futil7 + e1+e2+e3	==> e3 < futil3 - futil7

    ta=NORMCDF(futil(2)-futil(3),0.0d0,1.0d0)
    wg(:)=1.0-ta
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,1)=HUGE(v(j,1))
            ELSE
            	v(j,1)=INVNORMCDF((ta(j)*(1-draws3seed(j,3))+draws3seed(j,3)),0.0d0,1.0d0)
	        ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
        	futil(1)+v(j,1)-futil(2), &
        	futil(1)-futil(3) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,2)=-HUGE(v(j,2))
   	        ELSE
   	        v(j,2)=INVNORMCDF((tb(j)*draws3seed(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
        	futil(3)+v(j,1)+v(j,2)-futil(4), &
        	futil(3)+v(j,2)-futil(5), &
        	futil(3)+v(j,1)-futil(6), &
        	futil(3)-futil(7) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb


    FAMSOLVE(3,3)=FAMSOLVE(3,3)+SUM(wg(:))/ns
    configprob(3)=SUM(wg(:))/ns

! Configuration 4
    ! condition: u4 > 0, u1, u2, u3, u5, u6, u7
    ! condition: futil4 + e3 > 0					==> e3 > - futil4
    ! condition: futil4 + e3 > futil1 + e1			==> e3 > futil1 + e1 - futil4
    ! condition: futil4 + e3 > futil2 + e2			==> e3 > futil2 + e2 - futil4
    ! condition: futil4 + e3 > futil3 + e1+e2		==> e3 > futil3 + e1+e2 - futil4
    ! condition: futil4 + e3 > futil5 + e1+e3		==> e1 < futil4 - futil5
    ! condition: futil4 + e3 > futil6 + e2+e3		==> e2 < futil4 - futil6
    ! condition: futil4 + e3 > futil7 + e1+e2+e3	==> e2 < futil4 - futil7 - e1

    tb=NORMCDF(futil(4)-futil(5),0.0d0,1.0d0)
    wg(:)=tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,1)=-HUGE(v(j,1))
   	        ELSE
   	        v(j,1)=INVNORMCDF((tb(j)*draws3seed(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
        	futil(4)-futil(6), &
        	futil(4)-futil(7)-v(j,1) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,2)=-HUGE(v(j,2))
   	        ELSE
   	        v(j,2)=INVNORMCDF((tb(j)*draws3seed(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	-futil(4), &
        	futil(1)+v(j,1)-futil(4), &
        	futil(2)+v(j,2)-futil(4), &
        	futil(3)+v(j,1)+v(j,2)-futil(4) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)


    FAMSOLVE(4,3)=FAMSOLVE(4,3)+SUM(wg(:))/ns
    configprob(4)=SUM(wg(:))/ns

! Configuration 5
    ! condition: u5 > 0, u1, u2, u3, u4, u6, u7
    ! condition: futil5 + e1+e3 > 0					==> e3 > - futil5 - e1
    ! condition: futil5 + e1+e3 > futil1 + e1		==> e3 > futil1 - futil5
    ! condition: futil5 + e1+e3 > futil2 + e2		==> e3 > futil2 + e2 - futil5 - e1
    ! condition: futil5 + e1+e3 > futil3 + e1+e2	==> e3 > futil3 + e2 - futil5
    ! condition: futil5 + e1+e3 > futil4 + e3		==> e1 > futil4 - futil5
    ! condition: futil5 + e1+e3 > futil6 + e2+e3	==> e2 < futil5 + e1 - futil6
    ! condition: futil5 + e1+e3 > futil7 + e1+e2+e3	==> e2 < futil5 - futil7

    ta=NORMCDF(futil(4)-futil(5),0.0d0,1.0d0)
    wg(:)=1.0-ta
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,1)=HUGE(v(j,1))
            ELSE
            	v(j,1)=INVNORMCDF((ta(j)*(1-draws3seed(j,3))+draws3seed(j,3)),0.0d0,1.0d0)
	        ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
        	futil(5)+v(j,1)-futil(6), &
        	futil(5)-futil(7) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,2)=-HUGE(v(j,2))
   	        ELSE
   	        v(j,2)=INVNORMCDF((tb(j)*draws3seed(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	-futil(5)-v(j,1), &
        	futil(1)-futil(5), &
        	futil(2)+v(j,2)-futil(5)-v(j,1), &
        	futil(3)+v(j,2)-futil(5) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)


    FAMSOLVE(5,3)=FAMSOLVE(5,3)+SUM(wg(:))/ns
    configprob(5)=SUM(wg(:))/ns

! Configuration 6
    ! condition: u6 > 0, u1, u2, u3, u4, u5, u7
    ! condition: futil6 + e2+e3 > 0					==> e3 > - futil6 - e2
    ! condition: futil6 + e2+e3 > futil1 + e1		==> e3 > futil1 + e1 - futil6 - e2
    ! condition: futil6 + e2+e3 > futil2 + e2		==> e3 > futil2 - futil6
    ! condition: futil6 + e2+e3 > futil3 + e1+e2	==> e3 > futil3 + e1 - futil6
    ! condition: futil6 + e2+e3 > futil4 + e3		==> e2 > futil4 - futil6
    ! condition: futil6 + e2+e3 > futil5 + e1+e3	==> e2 > futil5 + e1 - futil6
    ! condition: futil6 + e2+e3 > futil7 + e1+e2+e3	==> e1 < futil6 - futil7

    tb=NORMCDF(futil(6)-futil(7),0.0d0,1.0d0)
    wg(:)=tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,1)=-HUGE(v(j,1))
   	        ELSE
   	        v(j,1)=INVNORMCDF((tb(j)*draws3seed(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	-futil(4)-futil(6), &
        	futil(5)+v(j,1)-futil(6) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,2)=HUGE(v(j,2))
   	        ELSE
           	v(j,2)=INVNORMCDF((ta(j)*(1-draws3seed(j,2))+draws3seed(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	-futil(6)-v(j,2), &
        	futil(1)+v(j,1)-futil(6)-v(j,2), &
        	futil(2)-futil(6), &
        	futil(3)+v(j,1)-futil(6) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)


    FAMSOLVE(6,3)=FAMSOLVE(6,3)+SUM(wg(:))/ns
    configprob(6)=SUM(wg(:))/ns

! Configuration 7
    ! condition: u7 > 0, u1, u2, u3, u4, u5, u6
    ! condition: futil7 + e1+e2+e3 > 0				==> e3 > - futil7 - e1 - e2
    ! condition: futil7 + e1+e2+e3 > futil1 + e1	==> e3 > futil1 - futil7 - e2
    ! condition: futil7 + e1+e2+e3 > futil2 + e2	==> e3 > futil2 - futil7 - e1
    ! condition: futil7 + e1+e2+e3 > futil3 + e1+e2	==> e3 > futil3 - futil7
    ! condition: futil7 + e1+e2+e3 > futil4 + e3	==> e2 > futil4 - futil7 - e1
    ! condition: futil7 + e1+e2+e3 > futil5 + e1+e3	==> e2 > futil5 - futil7
    ! condition: futil7 + e1+e2+e3 > futil6 + e2+e3	==> e1 > futil6 - futil7

    ta=NORMCDF(futil(6)-futil(7),0.0d0,1.0d0)
    wg(:)=1.0-ta
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,1)=HUGE(v(j,1))
            ELSE
            	v(j,1)=INVNORMCDF((ta(j)*(1-draws3seed(j,3))+draws3seed(j,3)),0.0d0,1.0d0)
	        ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	futil(4)-futil(7)-v(j,1), &
        	futil(5)-futil(7) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,2)=HUGE(v(j,2))
   	        ELSE
   	        v(j,2)=INVNORMCDF((ta(j)*(1-draws3seed(j,2))+draws3seed(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	-futil(7)-v(j,2)-v(j,1), &
        	futil(1)-futil(7)-v(j,2), &
        	futil(2)-futil(7)-v(j,1), &
        	futil(3)-futil(7) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)


    FAMSOLVE(7,3)=FAMSOLVE(7,3)+SUM(wg(:))/ns
    configprob(7)=SUM(wg(:))/ns



CASE(4)

utilsub(1:4)=MATMUL(xb(k-3:k,:),param(1:nthetab))
	IF (nthetaa>=3) THEN
    alpha(1:4,1)=MATMUL(xa(k-3:k,:),param(nthetab+1:nthetab+nthetaa-2))
    alpha(1:4,2)=param(nthetab+nthetaa-1)
    alpha(1:4,3)=param(nthetab+nthetaa)
	ELSEIF (nthetaa==1) THEN
	alpha(1:4,1)=param(nthetab+nthetaa)
	alpha(1:4,2:3)=0.0d0
	ELSE
	alpha=0.0d0
	ENDIF
	IF (nthetag>=2) THEN
	gamma(1:4,1) =MATMUL(xg(k-3:k,:),param(nthetab+nthetaa+1:nthetab+nthetaa+nthetag-1))
	gamma(1:4,2) =param(nthetab+nthetaa+nthetag)
	ELSEIF (nthetag==1) THEN
	gamma(1:4,1)=param(nthetab+nthetaa+nthetag)
	gamma(1:4,2)=0.0d0
	ELSE
	gamma=0.0d0
	ENDIF



lchol(1:4,1:4)=lcholbase(m,param,n)

futil(1)=utilsub(4) + EXP( alpha(1,1) ) + EXP( alpha(2,1) ) + EXP( alpha(3,1) ) + EXP( alpha(4,1) )
futil(2)=utilsub(3) + EXP( alpha(1,1) ) + EXP( alpha(2,1) ) + EXP( alpha(3,1) ) + EXP( alpha(4,1) )
futil(3)=utilsub(3)+utilsub(4) &
	+ EXP( alpha(1,1)+alpha(1,2) ) &
	+ EXP( alpha(2,1)+alpha(2,2) ) &
	+ EXP( alpha(3,1)+alpha(3,2) ) &
	+ EXP( alpha(4,1)+alpha(4,2) ) &
	+gamma(3,1)+gamma(4,1)
futil(4)=utilsub(2) + EXP( alpha(1,1) ) + EXP( alpha(2,1) ) + EXP( alpha(3,1) ) + EXP( alpha(4,1) )
futil(5)=utilsub(2)+utilsub(4) &
	+ EXP( alpha(1,1)+alpha(1,2) ) &
	+ EXP( alpha(2,1)+alpha(2,2) ) &
	+ EXP( alpha(3,1)+alpha(3,2) ) &
	+ EXP( alpha(4,1)+alpha(4,2) ) &
	+gamma(2,1)+gamma(4,1)
futil(6)=utilsub(2)+utilsub(3) &
	+ EXP( alpha(1,1)+alpha(1,2) ) &
	+ EXP( alpha(2,1)+alpha(2,2) ) &
	+ EXP( alpha(3,1)+alpha(3,2) ) &
	+ EXP( alpha(4,1)+alpha(4,2) ) &
	+gamma(2,1)+gamma(3,1)
futil(7)=utilsub(2)+utilsub(3)+utilsub(4) &
	+ EXP( alpha(1,1)+alpha(1,2)+alpha(1,3) ) &
	+ EXP( alpha(2,1)+alpha(2,2)+alpha(2,3) ) &
	+ EXP( alpha(3,1)+alpha(3,2)+alpha(3,3) ) &
	+ EXP( alpha(4,1)+alpha(4,2)+alpha(4,3) ) &
	+gamma(2,1)*(1.0d0+gamma(2,2))+gamma(3,1)*(1.0d0+gamma(3,2))+gamma(4,1)*(1.0d0+gamma(4,2))
futil(8)=utilsub(1) + EXP( alpha(1,1) ) + EXP( alpha(2,1) ) + EXP( alpha(3,1) ) + EXP( alpha(4,1) )
futil(9)=utilsub(1)+utilsub(4) &
	+ EXP( alpha(1,1)+alpha(1,2) ) &
	+ EXP( alpha(2,1)+alpha(2,2) ) &
	+ EXP( alpha(3,1)+alpha(3,2) ) &
	+ EXP( alpha(4,1)+alpha(4,2) ) &
	+gamma(1,1)+gamma(4,1)
futil(10)=utilsub(1)+utilsub(3) &
	+ EXP( alpha(1,1)+alpha(1,2) ) &
	+ EXP( alpha(2,1)+alpha(2,2) ) &
	+ EXP( alpha(3,1)+alpha(3,2) ) &
	+ EXP( alpha(4,1)+alpha(4,2) ) &
	+gamma(1,1)+gamma(3,1)
futil(11)=utilsub(1)+utilsub(3)+utilsub(4) &
	+ EXP( alpha(1,1)+alpha(1,2)+alpha(1,3) ) &
	+ EXP( alpha(2,1)+alpha(2,2)+alpha(2,3) ) &
	+ EXP( alpha(3,1)+alpha(3,2)+alpha(3,3) ) &
	+ EXP( alpha(4,1)+alpha(4,2)+alpha(4,3) ) &
	+gamma(1,1)*(1.0d0+gamma(1,2))+gamma(3,1)*(1.0d0+gamma(3,2))+gamma(4,1)*(1.0d0+gamma(4,2))
futil(12)=utilsub(1)+utilsub(2) &
	+ EXP( alpha(1,1)+alpha(1,2) ) &
	+ EXP( alpha(2,1)+alpha(2,2) ) &
	+ EXP( alpha(3,1)+alpha(3,2) ) &
	+ EXP( alpha(4,1)+alpha(4,2) ) &
	+gamma(1,1)+gamma(2,1)
futil(13)=utilsub(1)+utilsub(2)+utilsub(4) &
	+ EXP( alpha(1,1)+alpha(1,2)+alpha(1,3) ) &
	+ EXP( alpha(2,1)+alpha(2,2)+alpha(2,3) ) &
	+ EXP( alpha(3,1)+alpha(3,2)+alpha(3,3) ) &
	+ EXP( alpha(4,1)+alpha(4,2)+alpha(4,3) ) &
	+gamma(1,1)*(1.0d0+gamma(1,2))+gamma(2,1)*(1.0d0+gamma(2,2))+gamma(4,1)*(1.0d0+gamma(4,2))
futil(14)=utilsub(1)+utilsub(2)+utilsub(3) &
	+ EXP( alpha(1,1)+alpha(1,2)+alpha(1,3) ) &
	+ EXP( alpha(2,1)+alpha(2,2)+alpha(2,3) ) &
	+ EXP( alpha(3,1)+alpha(3,2)+alpha(3,3) ) &
	+ EXP( alpha(4,1)+alpha(4,2)+alpha(4,3) ) &
	+gamma(1,1)*(1.0d0+gamma(1,2))+gamma(2,1)*(1.0d0+gamma(2,2))+gamma(3,1)*(1.0d0+gamma(3,2))
futil(15)=utilsub(1)+utilsub(2)+utilsub(3)+utilsub(4) &
	+ EXP( alpha(1,1)+alpha(1,2)+alpha(1,3) ) &
	+ EXP( alpha(2,1)+alpha(2,2)+alpha(2,3) ) &
	+ EXP( alpha(3,1)+alpha(3,2)+alpha(3,3) ) &
	+ EXP( alpha(4,1)+alpha(4,2)+alpha(4,3) ) &
	+gamma(1,1)*(1.0d0+gamma(1,2)*2.0d0)+gamma(2,1)*(1.0d0+gamma(2,2)*2.0d0)+gamma(3,1)*(1.0d0+gamma(3,2)*2.0d0)+gamma(4,1)*(1.0d0+gamma(4,2)*2.0d0)

! Configuration 0
    ! condition: 0 > u1 - u15
    ! condition: 0 > futil1 + e1			==> e1 < - futil1
    ! condition: 0 > futil2 + e2			==> e2 < - futil2
    ! condition: 0 > futil3 + e1+e2			==> e2 < - futil3 - e1
    ! condition: 0 > futil4 + e3			==> e3 < - futil4
    ! condition: 0 > futil5 + e1+e3			==> e3 < - futil5 - e1
    ! condition: 0 > futil6 + e2+e3			==> e3 < - futil6 - e2
    ! condition: 0 > futil7 + e1+e2+e3		==> e3 < - futil7 - e2 - e1
    ! condition: 0 > futil8 + e4			==> e4 < - futil8
    ! condition: 0 > futil9 + e1+e4			==> e4 < - futil9 - e1
    ! condition: 0 > futil10 + e2+e4		==> e4 < - futil10 - e2
    ! condition: 0 > futil11 +e1+e2+e4		==> e4 < - futil11 - e1 - e2
    ! condition: 0 > futil12 +e3+e4			==> e4 < - futil12 - e3
    ! condition: 0 > futil13 +e1+e3+e4		==> e4 < - futil13 - e1 - e3
    ! condition: 0 > futil14 +e2+e3+e4		==> e4 < - futil14 - e2 - e3
    ! condition: 0 > futil15 +e1+e2+e3+e4	==> e4 < - futil15 - e1 - e2 - e3

	tb=NORMCDF(-futil(1),0.0d0,1.0d0)
    wg(:)=tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,1)=-HUGE(v(j,1))
   	        ELSE
   	        v(j,1)=INVNORMCDF((tb(j)*draws4seed(j,4)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
        	-futil(2), &
        	-futil(3)-v(j,1) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,2)=-HUGE(v(j,2))
   	        ELSE
   	        v(j,2)=INVNORMCDF((tb(j)*draws4seed(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
			-futil(4), &
        	-futil(5)-v(j,1), &
        	-futil(6)-v(j,2), &
        	-futil(7)-v(j,1)-v(j,2) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,3)=-HUGE(v(j,3))
   	        ELSE
   	        v(j,3)=INVNORMCDF((tb(j)*draws4seed(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
			-futil(8), &
        	-futil(9)-v(j,1), &
        	-futil(10)-v(j,2), &
        	-futil(11)-v(j,1)-v(j,2), &
        	-futil(12)-v(j,3), &
        	-futil(13)-v(j,1)-v(j,3), &
        	-futil(14)-v(j,2)-v(j,3), &
        	-futil(15)-v(j,1)-v(j,2)-v(j,3) &
	        ) &
           -DOT_PRODUCT(lchol(4,1:3),v(j,1:3)) &
        ) &
        /lchol(4,4),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb


    FAMSOLVE(0,3)=FAMSOLVE(0,3)+SUM(wg(:))/ns
    configprob(0)=SUM(wg(:))/ns

! Configuration 1
    ! condition: u1 > 0, u2 - u15
    ! condition: util1 + e1 > 0						==> e1 > - futil1
    ! condition: util1 + e1 > futil2 + e2			==> e2 < futil1 + e1 - futil2
    ! condition: util1 + e1 > futil3 + e1+e2		==> e2 < futil1 - futil3
    ! condition: util1 + e1 > futil4 + e3			==> e3 < futil1 + e1 - futil4
    ! condition: util1 + e1 > futil5 + e1+e3		==> e3 < futil1 - futil5
    ! condition: util1 + e1 > futil6 + e2+e3		==> e3 < futil1 + e1 - futil6 - e2
    ! condition: util1 + e1 > futil7 + e1+e2+e3		==> e3 < futil1 - futil7 - e2
    ! condition: util1 + e1 > futil8 + e4			==> e4 < futil1 + e1 - futil8
    ! condition: util1 + e1 > futil9 + e1+e4		==> e4 < futil1 - futil9
    ! condition: util1 + e1 > futil10 + e2+e4		==> e4 < futil1 + e1 - futil10 - e2
    ! condition: util1 + e1 > futil11 +e1+e2+e4		==> e4 < futil1 - futil11 - e2
    ! condition: util1 + e1 > futil12 +e3+e4		==> e4 < futil1 + e1 - futil12 - e3
    ! condition: util1 + e1 > futil13 +e1+e3+e4		==> e4 < futil1 - futil13 - e3
    ! condition: util1 + e1 > futil14 +e2+e3+e4		==> e4 < futil1 + e1 - futil14 - e2 - e3
    ! condition: util1 + e1 > futil15 +e1+e2+e3+e4	==> e4 < futil1 - futil15 - e2 - e3

    ta=NORMCDF(-futil(1),0.0d0,1.0d0)
    wg(:)=1.0-ta
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,1)=HUGE(v(j,1))
            ELSE
            	v(j,1)=INVNORMCDF((ta(j)*(1-draws4seed(j,4))+draws4seed(j,4)),0.0d0,1.0d0)
	        ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
        	futil(1)+v(j,1)-futil(2), &
        	futil(1)-futil(3) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,2)=-HUGE(v(j,2))
   	        ELSE
   	        v(j,2)=INVNORMCDF((tb(j)*draws4seed(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
        	futil(1)+v(j,1)-futil(4), &
        	futil(1)-futil(5), &
        	futil(1)+v(j,1)-futil(6)-v(j,2), &
        	futil(1)-futil(7)-v(j,2) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,3)=-HUGE(v(j,3))
   	        ELSE
   	        v(j,3)=INVNORMCDF((tb(j)*draws4seed(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
			futil(1)+v(j,1)-futil(8), &
        	futil(1)-futil(9), &
        	futil(1)+v(j,1)-futil(10)-v(j,2), &
        	futil(1)-futil(11)-v(j,2), &
        	futil(1)+v(j,1)-futil(12)-v(j,3), &
        	futil(1)-futil(13)-v(j,3), &
        	futil(1)+v(j,1)-futil(14)-v(j,2)-v(j,3), &
        	futil(1)-futil(15)-v(j,2)-v(j,3) &
	        ) &
           -DOT_PRODUCT(lchol(4,1:3),v(j,1:3)) &
        ) &
        /lchol(4,4),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb


    FAMSOLVE(1,3)=FAMSOLVE(1,3)+SUM(wg(:))/ns
    configprob(1)=SUM(wg(:))/ns

! Configuration 2
    ! condition: u2 > 0, u1, u3 - u15
    ! condition: futil2 + e2 > 0					==> e2 > - futil2
    ! condition: futil2 + e2 > futil1 + e1			==> e2 > futil1 + e1 - futil2
    ! condition: futil2 + e2 > futil3 + e1+e2		==> e1 < futil2 - futil3
    ! condition: futil2 + e2 > futil4 + e3			==> e3 < futil2 + e2 - futil4
    ! condition: futil2 + e2 > futil5 + e1+e3		==> e3 < futil2 + e2 - futil5 - e1
    ! condition: futil2 + e2 > futil6 + e2+e3		==> e3 < futil2 - futil6
    ! condition: futil2 + e2 > futil7 + e1+e2+e3	==> e3 < futil2 - futil7 - e1
    ! condition: futil2 + e2 > futil8 + e4			==> e4 < futil2 + e2 - futil8
    ! condition: futil2 + e2 > futil9 + e1+e4		==> e4 < futil2 + e2 - futil9 - e1
    ! condition: futil2 + e2 > futil10 + e2+e4		==> e4 < futil2 - futil10
    ! condition: futil2 + e2 > futil11 +e1+e2+e4	==> e4 < futil2 - futil11 - e1
    ! condition: futil2 + e2 > futil12 +e3+e4		==> e4 < futil2 + e2 - futil12 - e3
    ! condition: futil2 + e2 > futil13 +e1+e3+e4	==> e4 < futil2 + e2 - futil13 - e1 - e3
    ! condition: futil2 + e2 > futil14 +e2+e3+e4	==> e4 < futil2 + e2 - futil14 - e3
    ! condition: futil2 + e2 > futil15 +e1+e2+e3+e4	==> e4 < futil2 - futil15 - e1 - e3

	tb=NORMCDF(futil(2)-futil(3),0.0d0,1.0d0)
    wg(:)=tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,1)=-HUGE(v(j,1))
   	        ELSE
   	        v(j,1)=INVNORMCDF((tb(j)*draws4seed(j,4)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	-futil(2), &
        	futil(1)+v(j,1)-futil(2) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
        	    v(j,2)=HUGE(v(j,2))
   	        ELSE
           	v(j,2)=INVNORMCDF((ta(j)*(1-draws4seed(j,3))+draws4seed(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
        	futil(2)+v(j,2)-futil(4), &
        	futil(2)+v(j,2)-futil(5)-v(j,1), &
        	futil(2)-futil(6), &
        	futil(2)-futil(7)-v(j,1) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,3)=-HUGE(v(j,3))
   	        ELSE
   	        v(j,3)=INVNORMCDF((tb(j)*draws4seed(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
			futil(2)+v(j,2)-futil(8), &
        	futil(2)+v(j,2)-futil(9)-v(j,1), &
        	futil(2)-futil(10), &
        	futil(2)-futil(11)-v(j,1), &
        	futil(2)+v(j,2)-futil(12)-v(j,3), &
        	futil(2)+v(j,2)-futil(13)-v(j,1)-v(j,3), &
        	futil(2)-futil(14)-v(j,3), &
        	futil(2)-futil(15)-v(j,1)-v(j,3) &
	        ) &
           -DOT_PRODUCT(lchol(4,1:3),v(j,1:3)) &
        ) &
        /lchol(4,4),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb


    FAMSOLVE(2,3)=FAMSOLVE(2,3)+SUM(wg(:))/ns
    configprob(2)=SUM(wg(:))/ns

! Configuration 3
    ! condition: u3 > 0, u1, u2, u4 - u15
    ! condition: futil3 + e1+e2 > 0						==> e2 > - futil3 - e1
    ! condition: futil3 + e1+e2 > futil1 + e1			==> e2 > futil1 - futil3
    ! condition: futil3 + e1+e2 > futil2 + e2			==> e1 > futil2 - futil3
    ! condition: futil3 + e1+e2 > futil4 + e3			==> e3 < futil3 + e1+e2 - futil4
    ! condition: futil3 + e1+e2 > futil5 + e1+e3		==> e3 < futil3 + e2 - futil5
    ! condition: futil3 + e1+e2 > futil6 + e2+e3		==> e3 < futil3 + e1 - futil6
    ! condition: futil3 + e1+e2 > futil7 + e1+e2+e3		==> e3 < futil3 - futil7
    ! condition: futil3 + e1+e2 > futil8 + e4			==> e4 < futil3 + e1+e2 - futil8
    ! condition: futil3 + e1+e2 > futil9 + e1+e4		==> e4 < futil3 + e2 - futil9
    ! condition: futil3 + e1+e2 > futil10 + e2+e4		==> e4 < futil3 + e1 - futil10
    ! condition: futil3 + e1+e2 > futil11 +e1+e2+e4		==> e4 < futil3 - futil11
    ! condition: futil3 + e1+e2 > futil12 +e3+e4		==> e4 < futil3 + e1+e2 - futil12 - e3
    ! condition: futil3 + e1+e2 > futil13 +e1+e3+e4		==> e4 < futil3 + e2 - futil13 - e3
    ! condition: futil3 + e1+e2 > futil14 +e2+e3+e4		==> e4 < futil3 + e1 - futil14 - e3
    ! condition: futil3 + e1+e2 > futil15 +e1+e2+e3+e4	==> e4 < futil3 - futil15 - e3

    ta=NORMCDF(futil(2)-futil(3),0.0d0,1.0d0)
    wg(:)=1.0-ta
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,1)=HUGE(v(j,1))
            ELSE
            	v(j,1)=INVNORMCDF((ta(j)*(1-draws4seed(j,4))+draws4seed(j,4)),0.0d0,1.0d0)
	        ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	-futil(3)-v(j,1), &
        	futil(1)-futil(3) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
        	    v(j,2)=HUGE(v(j,2))
   	        ELSE
           	v(j,2)=INVNORMCDF((ta(j)*(1-draws4seed(j,3))+draws4seed(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
        	futil(3)+v(j,1)+v(j,2)-futil(4), &
        	futil(3)+v(j,2)-futil(5), &
        	futil(3)+v(j,1)-futil(6), &
        	futil(3)-futil(7) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,3)=-HUGE(v(j,3))
   	        ELSE
   	        v(j,3)=INVNORMCDF((tb(j)*draws4seed(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
			futil(3)+v(j,1)+v(j,2)-futil(8), &
        	futil(3)+v(j,2)-futil(9), &
        	futil(3)+v(j,1)-futil(10), &
        	futil(3)-futil(11), &
        	futil(3)+v(j,1)+v(j,2)-futil(12)-v(j,3), &
        	futil(3)+v(j,2)-futil(13)-v(j,3), &
        	futil(3)+v(j,1)-futil(14)-v(j,3), &
        	futil(3)-futil(15)-v(j,3) &
	        ) &
           -DOT_PRODUCT(lchol(4,1:3),v(j,1:3)) &
        ) &
        /lchol(4,4),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb


    FAMSOLVE(3,3)=FAMSOLVE(3,3)+SUM(wg(:))/ns
    configprob(3)=SUM(wg(:))/ns

! Configuration 4
    ! condition: u4 > 0, u1 - u3, u5 - u15
    ! condition: futil4 + e3 > 0					==> e3 > - futil4
    ! condition: futil4 + e3 > futil1 + e1			==> e3 > futil1 + e1 - futil4
    ! condition: futil4 + e3 > futil2 + e2			==> e3 > futil2 + e2 - futil4
    ! condition: futil4 + e3 > futil3 + e1+e2		==> e3 > futil3 + e1+e2 - futil4
    ! condition: futil4 + e3 > futil5 + e1+e3		==> e1 < futil4 - futil5
    ! condition: futil4 + e3 > futil6 + e2+e3		==> e2 < futil4 - futil6
    ! condition: futil4 + e3 > futil7 + e1+e2+e3	==> e2 < futil4 - futil7 - e1
    ! condition: futil4 + e3 > futil8 + e4			==> e4 < futil4 + e3 - futil8
    ! condition: futil4 + e3 > futil9 + e1+e4		==> e4 < futil4 + e3 - futil9 - e1
    ! condition: futil4 + e3 > futil10 + e2+e4		==> e4 < futil4 + e3 - futil10 - e2
    ! condition: futil4 + e3 > futil11 +e1+e2+e4	==> e4 < futil4 + e3 - futil11 - e1 - e2
    ! condition: futil4 + e3 > futil12 +e3+e4		==> e4 < futil4 - futil12
    ! condition: futil4 + e3 > futil13 +e1+e3+e4	==> e4 < futil4 - futil13 - e1
    ! condition: futil4 + e3 > futil14 +e2+e3+e4	==> e4 < futil4 - futil14 - e2
    ! condition: futil4 + e3 > futil15 +e1+e2+e3+e4	==> e4 < futil4 - futil15 - e1 - e2

	tb=NORMCDF(futil(4)-futil(5),0.0d0,1.0d0)
    wg(:)=tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,1)=-HUGE(v(j,1))
   	        ELSE
   	        v(j,1)=INVNORMCDF((tb(j)*draws4seed(j,4)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
        	futil(4)-futil(7)-v(j,1), &
        	futil(4)-futil(6) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,2)=-HUGE(v(j,2))
   	        ELSE
   	        v(j,2)=INVNORMCDF((tb(j)*draws4seed(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	-futil(4), &
        	futil(1)+v(j,1)-futil(4), &
        	futil(2)+v(j,2)-futil(4), &
        	futil(3)+v(j,1)+v(j,2)-futil(4) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
        	    v(j,3)=HUGE(v(j,3))
   	        ELSE
           	v(j,3)=INVNORMCDF((ta(j)*(1-draws4seed(j,2))+draws4seed(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
			futil(4)+v(j,3)-futil(8), &
        	futil(4)+v(j,3)-futil(9)-v(j,1), &
        	futil(4)+v(j,3)-futil(10)-v(j,2), &
        	futil(4)+v(j,3)-futil(11)-v(j,1)-v(j,2), &
        	futil(4)-futil(12), &
        	futil(4)-futil(13)-v(j,1), &
        	futil(4)-futil(14)-v(j,2), &
        	futil(4)-futil(15)-v(j,1)-v(j,2) &
	        ) &
           -DOT_PRODUCT(lchol(4,1:3),v(j,1:3)) &
        ) &
        /lchol(4,4),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb


    FAMSOLVE(4,3)=FAMSOLVE(4,3)+SUM(wg(:))/ns
    configprob(4)=SUM(wg(:))/ns

! Configuration 5
    ! condition: u5 > 0, u1 - u4, u6 - u15
    ! condition: futil5 + e1+e3 > 0						==> e3 > - futil5 - e1
    ! condition: futil5 + e1+e3 > futil1 + e1			==> e3 > futil1 - futil5
    ! condition: futil5 + e1+e3 > futil2 + e2			==> e3 > futil2 + e2 - futil5 - e1
    ! condition: futil5 + e1+e3 > futil3 + e1+e2		==> e3 > futil3 + e2 - futil5
    ! condition: futil5 + e1+e3 > futil4 + e3			==> e1 > futil4 - futil5
    ! condition: futil5 + e1+e3 > futil6 + e2+e3		==> e2 < futil5 + e1 - futil6
    ! condition: futil5 + e1+e3 > futil7 + e1+e2+e3		==> e2 < futil5 - futil7
    ! condition: futil5 + e1+e3 > futil8 + e4			==> e4 < futil5 + e1+e3 - futil8
    ! condition: futil5 + e1+e3 > futil9 + e1+e4		==> e4 < futil5 + e3 - futil9
    ! condition: futil5 + e1+e3 > futil10 + e2+e4		==> e4 < futil5 + e1+e3 - futil10 - e2
    ! condition: futil5 + e1+e3 > futil11 +e1+e2+e4		==> e4 < futil5 + e3 - futil11 - e2
    ! condition: futil5 + e1+e3 > futil12 +e3+e4		==> e4 < futil5 + e1 - futil12
    ! condition: futil5 + e1+e3 > futil13 +e1+e3+e4		==> e4 < futil5 - futil13
    ! condition: futil5 + e1+e3 > futil14 +e2+e3+e4		==> e4 < futil5 + e1 - futil14 - e2
    ! condition: futil5 + e1+e3 > futil15 +e1+e2+e3+e4	==> e4 < futil5 - futil15 - e2

    ta=NORMCDF(futil(4)-futil(5),0.0d0,1.0d0)
    wg(:)=1.0-ta
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,1)=HUGE(v(j,1))
            ELSE
            	v(j,1)=INVNORMCDF((ta(j)*(1-draws4seed(j,4))+draws4seed(j,4)),0.0d0,1.0d0)
	        ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
        	futil(5)+v(j,1)-futil(6), &
        	futil(5)-futil(7) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,2)=-HUGE(v(j,2))
   	        ELSE
   	        v(j,2)=INVNORMCDF((tb(j)*draws4seed(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	-futil(5)-v(j,1), &
        	futil(1)-futil(5), &
        	futil(2)+v(j,2)-futil(5)-v(j,1), &
        	futil(3)+v(j,2)-futil(5) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
        	    v(j,3)=HUGE(v(j,3))
   	        ELSE
           	v(j,3)=INVNORMCDF((ta(j)*(1-draws4seed(j,2))+draws4seed(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
			futil(5)+v(j,1)+v(j,3)-futil(8), &
        	futil(5)+v(j,3)-futil(9), &
        	futil(5)+v(j,1)+v(j,3)-futil(10)-v(j,2), &
        	futil(5)+v(j,3)-futil(11)-v(j,2), &
        	futil(5)+v(j,1)-futil(12), &
        	futil(5)-futil(13), &
        	futil(5)+v(j,1)-futil(14)-v(j,2), &
        	futil(5)-futil(15)-v(j,2) &
	        ) &
           -DOT_PRODUCT(lchol(4,1:3),v(j,1:3)) &
        ) &
        /lchol(4,4),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb


    FAMSOLVE(5,3)=FAMSOLVE(5,3)+SUM(wg(:))/ns
    configprob(5)=SUM(wg(:))/ns

! Configuration 6
    ! condition: u6 > 0, u1 - u5, u7 - u15
    ! condition: futil6 + e2+e3 > 0						==> e3 > - futil6 - e2
    ! condition: futil6 + e2+e3 > futil1 + e1			==> e3 > futil1 + e1 - futil6 - e2
    ! condition: futil6 + e2+e3 > futil2 + e2			==> e3 > futil2 - futil6
    ! condition: futil6 + e2+e3 > futil3 + e1+e2		==> e3 > futil3 + e1 - futil6
    ! condition: futil6 + e2+e3 > futil4 + e3			==> e2 > futil4 - futil6
    ! condition: futil6 + e2+e3 > futil5 + e1+e3		==> e2 > futil5 + e1 - futil6
    ! condition: futil6 + e2+e3 > futil7 + e1+e2+e3		==> e1 < futil6 - futil7
    ! condition: futil6 + e2+e3 > futil8 + e4			==> e4 < futil6 + e2+e3 - futil8
    ! condition: futil6 + e2+e3 > futil9 + e1+e4		==> e4 < futil6 + e2+e3 - futil9 - e1
    ! condition: futil6 + e2+e3 > futil10 + e2+e4		==> e4 < futil6 + e3 - futil10
    ! condition: futil6 + e2+e3 > futil11 +e1+e2+e4		==> e4 < futil6 + e3 - futil11 - e1
    ! condition: futil6 + e2+e3 > futil12 +e3+e4		==> e4 < futil6 + e2 - futil12
    ! condition: futil6 + e2+e3 > futil13 +e1+e3+e4		==> e4 < futil6 + e2 - futil13 - e1
    ! condition: futil6 + e2+e3 > futil14 +e2+e3+e4		==> e4 < futil6 - futil14
    ! condition: futil6 + e2+e3 > futil15 +e1+e2+e3+e4	==> e4 < futil6 - futil15 - e1

	tb=NORMCDF(futil(6)-futil(7),0.0d0,1.0d0)
    wg(:)=tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,1)=-HUGE(v(j,1))
   	        ELSE
   	        v(j,1)=INVNORMCDF((tb(j)*draws4seed(j,4)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	futil(4)-futil(6), &
        	futil(5)+v(j,1)-futil(6) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
        	    v(j,2)=HUGE(v(j,2))
   	        ELSE
           	v(j,2)=INVNORMCDF((ta(j)*(1-draws4seed(j,3))+draws4seed(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	-futil(6)-v(j,2), &
        	futil(1)+v(j,1)-futil(6)-v(j,2), &
        	futil(2)-futil(6), &
        	futil(3)+v(j,1)-futil(6) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
        	    v(j,3)=HUGE(v(j,3))
   	        ELSE
           	v(j,3)=INVNORMCDF((ta(j)*(1-draws4seed(j,2))+draws4seed(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
			futil(6)+v(j,2)+v(j,3)-futil(8), &
        	futil(6)+v(j,2)+v(j,3)-futil(9)-v(j,1), &
        	futil(6)+v(j,3)-futil(10), &
        	futil(6)+v(j,3)-futil(11)-v(j,1), &
        	futil(6)+v(j,2)-futil(12), &
        	futil(6)+v(j,2)-futil(13)-v(j,1), &
        	futil(6)-futil(14), &
        	futil(6)-futil(15)-v(j,1) &
	        ) &
           -DOT_PRODUCT(lchol(4,1:3),v(j,1:3)) &
        ) &
        /lchol(4,4),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb


    FAMSOLVE(6,3)=FAMSOLVE(6,3)+SUM(wg(:))/ns
    configprob(6)=SUM(wg(:))/ns

! Configuration 7
    ! condition: u7 > 0, u1 - u6, u8 - u15
    ! condition: futil7 + e1+e2+e3 > 0						==> e3 > - futil7 - e1 - e2
    ! condition: futil7 + e1+e2+e3 > futil1 + e1			==> e3 > futil1 - futil7 - e2
    ! condition: futil7 + e1+e2+e3 > futil2 + e2			==> e3 > futil2 - futil7 - e1
    ! condition: futil7 + e1+e2+e3 > futil3 + e1+e2			==> e3 > futil3 - futil7
    ! condition: futil7 + e1+e2+e3 > futil4 + e3			==> e2 > futil4 - futil7 - e1
    ! condition: futil7 + e1+e2+e3 > futil5 + e1+e3			==> e2 > futil5 - futil7
    ! condition: futil7 + e1+e2+e3 > futil6 + e2+e3			==> e1 > futil6 - futil7
    ! condition: futil7 + e1+e2+e3 > futil8 + e4			==> e4 < futil7 + e1+e2+e3 - futil8
    ! condition: futil7 + e1+e2+e3 > futil9 + e1+e4			==> e4 < futil7 + e2+e3 - futil9
    ! condition: futil7 + e1+e2+e3 > futil10 + e2+e4		==> e4 < futil7 + e1+e3 - futil10
    ! condition: futil7 + e1+e2+e3 > futil11 +e1+e2+e4		==> e4 < futil7 + e3 - futil11
    ! condition: futil7 + e1+e2+e3 > futil12 +e3+e4			==> e4 < futil7 + e1+e2 - futil12
    ! condition: futil7 + e1+e2+e3 > futil13 +e1+e3+e4		==> e4 < futil7 + e2 - futil13
    ! condition: futil7 + e1+e2+e3 > futil14 +e2+e3+e4		==> e4 < futil7 + e1 - futil14
    ! condition: futil7 + e1+e2+e3 > futil15 +e1+e2+e3+e4	==> e4 < futil7 - futil15

    ta=NORMCDF(futil(6)-futil(7),0.0d0,1.0d0)
    wg(:)=1.0-ta
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,1)=HUGE(v(j,1))
            ELSE
            	v(j,1)=INVNORMCDF((ta(j)*(1-draws4seed(j,4))+draws4seed(j,4)),0.0d0,1.0d0)
	        ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	futil(4)-futil(7)-v(j,1), &
        	futil(5)-futil(7) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
        	    v(j,2)=HUGE(v(j,2))
   	        ELSE
           	v(j,2)=INVNORMCDF((ta(j)*(1-draws4seed(j,3))+draws4seed(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	-futil(7)-v(j,1)-v(j,2), &
        	futil(1)-futil(7)-v(j,2), &
        	futil(2)-futil(7)-v(j,1), &
        	futil(3)-futil(7) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
        	    v(j,3)=HUGE(v(j,3))
   	        ELSE
           	v(j,3)=INVNORMCDF((ta(j)*(1-draws4seed(j,2))+draws4seed(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
			futil(7)+v(j,1)+v(j,2)+v(j,3)-futil(8), &
        	futil(7)+v(j,2)+v(j,3)-futil(9), &
        	futil(7)+v(j,1)+v(j,3)-futil(10), &
        	futil(7)+v(j,3)-futil(11), &
        	futil(7)+v(j,1)+v(j,2)-futil(12), &
        	futil(7)+v(j,2)-futil(13), &
        	futil(7)+v(j,1)-futil(14), &
        	futil(7)-futil(15) &
	        ) &
           -DOT_PRODUCT(lchol(4,1:3),v(j,1:3)) &
        ) &
        /lchol(4,4),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb


    FAMSOLVE(7,3)=FAMSOLVE(7,3)+SUM(wg(:))/ns
    configprob(7)=SUM(wg(:))/ns

! Configuration 8
    ! condition: u8 > 0, u1 - u7, u9 - u15
    ! condition: futil8 + e4 > 0					==> e4 > - futil8
    ! condition: futil8 + e4 > futil1 + e1			==> e4 > futil1 + e1 - futil8
    ! condition: futil8 + e4 > futil2 + e2			==> e4 > futil2 + e2 - futil8
    ! condition: futil8 + e4 > futil3 + e1+e2		==> e4 > futil3 + e1+e2 - futil8
    ! condition: futil8 + e4 > futil4 + e3			==> e4 > futil4 + e3 - futil8
    ! condition: futil8 + e4 > futil5 + e1+e3		==> e4 > futil5 + e1+e3 - futil8
    ! condition: futil8 + e4 > futil6 + e2+e3		==> e4 > futil6 + e2+e3 - futil8
    ! condition: futil8 + e4 > futil7 + e1+e2+e3	==> e4 > futil7 + e1+e2+e3 - futil8
    ! condition: futil8 + e4 > futil9 + e1+e4		==> e1 < futil8 - futil9
    ! condition: futil8 + e4 > futil10 + e2+e4		==> e2 < futil8 - futil10
    ! condition: futil8 + e4 > futil11 +e1+e2+e4	==> e2 < futil8 - futil11 - e1
    ! condition: futil8 + e4 > futil12 +e3+e4		==> e3 < futil8 - futil12
    ! condition: futil8 + e4 > futil13 +e1+e3+e4	==> e3 < futil8 - futil13 - e1
    ! condition: futil8 + e4 > futil14 +e2+e3+e4	==> e3 < futil8 - futil14 - e2
    ! condition: futil8 + e4 > futil15 +e1+e2+e3+e4	==> e3 < futil8 - futil15 - e1 - e2

	tb=NORMCDF(futil(8)-futil(9),0.0d0,1.0d0)
    wg(:)=tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,1)=-HUGE(v(j,1))
   	        ELSE
   	        v(j,1)=INVNORMCDF((tb(j)*draws4seed(j,4)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
        	futil(8)-futil(11)-v(j,1), &
        	futil(8)-futil(10) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,2)=-HUGE(v(j,2))
   	        ELSE
   	        v(j,2)=INVNORMCDF((tb(j)*draws4seed(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
			futil(8)-futil(12), &
        	futil(8)-futil(13)-v(j,1), &
        	futil(8)-futil(14)-v(j,2), &
        	futil(8)-futil(15)-v(j,1)-v(j,2) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,3)=-HUGE(v(j,3))
   	        ELSE
   	        v(j,3)=INVNORMCDF((tb(j)*draws4seed(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
			-futil(8), &
        	futil(1)+v(j,1)-futil(8), &
        	futil(2)+v(j,2)-futil(8), &
        	futil(3)+v(j,1)+v(j,2)-futil(8), &
        	futil(4)+v(j,3)-futil(8), &
        	futil(5)+v(j,1)+v(j,3)-futil(8), &
        	futil(6)+v(j,2)+v(j,3)-futil(8), &
        	futil(7)+v(j,1)+v(j,2)+v(j,3)-futil(8) &
	        ) &
           -DOT_PRODUCT(lchol(4,1:3),v(j,1:3)) &
        ) &
        /lchol(4,4),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)


    FAMSOLVE(8,3)=FAMSOLVE(8,3)+SUM(wg(:))/ns
    configprob(8)=SUM(wg(:))/ns

! Configuration 9
    ! condition: u9 > 0, u1 - u8, u10 - u15
    ! condition: futil9 + e1+e4 > 0						==> e4 > - futil9 - e1
    ! condition: futil9 + e1+e4 > futil1 + e1			==> e4 > futil1 - futil9
    ! condition: futil9 + e1+e4 > futil2 + e2			==> e4 > futil2 + e2 - futil9 - e1
    ! condition: futil9 + e1+e4 > futil3 + e1+e2		==> e4 > futil3 + e2 - futil9
    ! condition: futil9 + e1+e4 > futil4 + e3			==> e4 > futil4 + e3 - futil9 - e1
    ! condition: futil9 + e1+e4 > futil5 + e1+e3		==> e4 > futil5 + e3 - futil9
    ! condition: futil9 + e1+e4 > futil6 + e2+e3		==> e4 > futil6 + e2+e3 - futil9 - e1
    ! condition: futil9 + e1+e4 > futil7 + e1+e2+e3		==> e4 > futil7 + e2+e3 - futil9
    ! condition: futil9 + e1+e4 > futil8 + e4			==> e1 > futil8 - futil9
    ! condition: futil9 + e1+e4 > futil10 + e2+e4		==> e2 < futil9 + e1 - futil10
    ! condition: futil9 + e1+e4 > futil11 +e1+e2+e4		==> e2 < futil9 - futil11
    ! condition: futil9 + e1+e4 > futil12 +e3+e4		==> e3 < futil9 + e1 - futil12
    ! condition: futil9 + e1+e4 > futil13 +e1+e3+e4		==> e3 < futil9 - futil13
    ! condition: futil9 + e1+e4 > futil14 +e2+e3+e4		==> e3 < futil9 + e1 - futil14 - e2
    ! condition: futil9 + e1+e4 > futil15 +e1+e2+e3+e4	==> e3 < futil9 - futil15 - e2

    ta=NORMCDF(futil(8)-futil(9),0.0d0,1.0d0)
    wg(:)=1.0-ta
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,1)=HUGE(v(j,1))
            ELSE
            	v(j,1)=INVNORMCDF((ta(j)*(1-draws4seed(j,4))+draws4seed(j,4)),0.0d0,1.0d0)
	        ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
        	futil(9)+v(j,1)-futil(10), &
        	futil(9)-futil(11) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,2)=-HUGE(v(j,2))
   	        ELSE
   	        v(j,2)=INVNORMCDF((tb(j)*draws4seed(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
			futil(9)+v(j,1)-futil(12), &
        	futil(9)-futil(13), &
        	futil(9)+v(j,1)-futil(14)-v(j,2), &
        	futil(9)-futil(15)-v(j,2) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,3)=-HUGE(v(j,3))
   	        ELSE
   	        v(j,3)=INVNORMCDF((tb(j)*draws4seed(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
			-futil(9)-v(j,1), &
        	futil(1)-futil(9), &
        	futil(2)+v(j,2)-futil(9)-v(j,1), &
        	futil(3)+v(j,2)-futil(9), &
        	futil(4)+v(j,3)-futil(9)-v(j,1), &
        	futil(5)+v(j,3)-futil(9), &
        	futil(6)+v(j,2)+v(j,3)-futil(9)-v(j,1), &
        	futil(7)+v(j,2)+v(j,3)-futil(9) &
	        ) &
           -DOT_PRODUCT(lchol(4,1:3),v(j,1:3)) &
        ) &
        /lchol(4,4),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)


    FAMSOLVE(9,3)=FAMSOLVE(9,3)+SUM(wg(:))/ns
    configprob(9)=SUM(wg(:))/ns

! Configuration 10
    ! condition: u10 > 0, u1 - u9, u11 - u15
    ! condition: futil10 + e2+e4 > 0					==> e4 > - futil10 - e2
    ! condition: futil10 + e2+e4 > futil1 + e1			==> e4 > futil1 + e1 - futil10 - e2
    ! condition: futil10 + e2+e4 > futil2 + e2			==> e4 > futil2 - futil10
    ! condition: futil10 + e2+e4 > futil3 + e1+e2		==> e4 > futil3 + e1 - futil10
    ! condition: futil10 + e2+e4 > futil4 + e3			==> e4 > futil4 + e3 - futil10 - e2
    ! condition: futil10 + e2+e4 > futil5 + e1+e3		==> e4 > futil5 + e1+e3 - futil10 - e2
    ! condition: futil10 + e2+e4 > futil6 + e2+e3		==> e4 > futil6 + e3 - futil10
    ! condition: futil10 + e2+e4 > futil7 + e1+e2+e3	==> e4 > futil7 + e1+e3 - futil10
    ! condition: futil10 + e2+e4 > futil8 + e4			==> e2 > futil8 - futil10
    ! condition: futil10 + e2+e4 > futil9 + e1+e4		==> e2 > futil9 + e1 - futil10
    ! condition: futil10 + e2+e4 > futil11 +e1+e2+e4	==> e1 < futil10 - futil11
    ! condition: futil10 + e2+e4 > futil12 +e3+e4		==> e3 < futil10 + e2 - futil12
    ! condition: futil10 + e2+e4 > futil13 +e1+e3+e4	==> e3 < futil10 + e2 - futil13 - e1
    ! condition: futil10 + e2+e4 > futil14 +e2+e3+e4	==> e3 < futil10 - futil14
    ! condition: futil10 + e2+e4 > futil15 +e1+e2+e3+e4	==> e3 < futil10 - futil15 - e1

	tb=NORMCDF(futil(10)-futil(11),0.0d0,1.0d0)
    wg(:)=tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,1)=-HUGE(v(j,1))
   	        ELSE
   	        v(j,1)=INVNORMCDF((tb(j)*draws4seed(j,4)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	futil(8)-futil(10), &
        	futil(9)+v(j,1)-futil(10) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
        	    v(j,2)=HUGE(v(j,2))
   	        ELSE
           	v(j,2)=INVNORMCDF((ta(j)*(1-draws4seed(j,3))+draws4seed(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
			futil(10)+v(j,2)-futil(12), &
        	futil(10)+v(j,2)-futil(13)-v(j,1), &
        	futil(10)-futil(14), &
        	futil(10)-futil(15)-v(j,1) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,3)=-HUGE(v(j,3))
   	        ELSE
   	        v(j,3)=INVNORMCDF((tb(j)*draws4seed(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
			-futil(10)-v(j,2), &
        	futil(1)+v(j,1)-futil(10)-v(j,2), &
        	futil(2)-futil(10), &
        	futil(3)+v(j,1)-futil(10), &
        	futil(4)+v(j,3)-futil(10)-v(j,2), &
        	futil(5)+v(j,1)+v(j,3)-futil(10)-v(j,2), &
        	futil(6)+v(j,3)-futil(10), &
        	futil(7)+v(j,1)+v(j,3)-futil(10) &
	        ) &
           -DOT_PRODUCT(lchol(4,1:3),v(j,1:3)) &
        ) &
        /lchol(4,4),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)


    FAMSOLVE(10,3)=FAMSOLVE(10,3)+SUM(wg(:))/ns
    configprob(10)=SUM(wg(:))/ns

! Configuration 11
    ! condition: u11 > 0, u1 - u10, u12 - u15
    ! condition: futil11 + e1+e2+e4 > 0						==> e4 > - futil11 - e1 - e2
    ! condition: futil11 + e1+e2+e4 > futil1 + e1			==> e4 > futil1 - futil11 - e2
    ! condition: futil11 + e1+e2+e4 > futil2 + e2			==> e4 > futil2 - futil11 - e1
    ! condition: futil11 + e1+e2+e4 > futil3 + e1+e2		==> e4 > futil3 - futil11
    ! condition: futil11 + e1+e2+e4 > futil4 + e3			==> e4 > futil4 + e3 - futil11 - e1 - e2
    ! condition: futil11 + e1+e2+e4 > futil5 + e1+e3		==> e4 > futil5 + e3 - futil11 - e2
    ! condition: futil11 + e1+e2+e4 > futil6 + e2+e3		==> e4 > futil6 + e3 - futil11 - e1
    ! condition: futil11 + e1+e2+e4 > futil7 + e1+e2+e3		==> e4 > futil7 + e3 - futil11
    ! condition: futil11 + e1+e2+e4 > futil8 + e4			==> e2 > futil8 - futil11 - e1
    ! condition: futil11 + e1+e2+e4 > futil9 + e1+e4		==> e2 > futil9 - futil11
    ! condition: futil11 + e1+e2+e4 > futil10 +e2+e4		==> e1 > futil10 - futil11
    ! condition: futil11 + e1+e2+e4 > futil12 +e3+e4		==> e3 < futil11 + e1+e2 - futil12
    ! condition: futil11 + e1+e2+e4 > futil13 +e1+e3+e4		==> e3 < futil11 + e2 - futil13
    ! condition: futil11 + e1+e2+e4 > futil14 +e2+e3+e4		==> e3 < futil11 + e1 - futil14
    ! condition: futil11 + e1+e2+e4 > futil15 +e1+e2+e3+e4	==> e3 < futil11 - futil15

    ta=NORMCDF(futil(10)-futil(11),0.0d0,1.0d0)
    wg(:)=1.0-ta
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,1)=HUGE(v(j,1))
            ELSE
            	v(j,1)=INVNORMCDF((ta(j)*(1-draws4seed(j,4))+draws4seed(j,4)),0.0d0,1.0d0)
	        ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	futil(8)-futil(11)-v(j,1), &
        	futil(9)-futil(10) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
        	    v(j,2)=HUGE(v(j,2))
   	        ELSE
           	v(j,2)=INVNORMCDF((ta(j)*(1-draws4seed(j,3))+draws4seed(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
			futil(11)+v(j,1)+v(j,2)-futil(12), &
        	futil(11)+v(j,2)-futil(13), &
        	futil(11)+v(j,1)-futil(14), &
        	futil(11)-futil(15) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,3)=-HUGE(v(j,3))
   	        ELSE
   	        v(j,3)=INVNORMCDF((tb(j)*draws4seed(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
			-futil(11)-v(j,1)-v(j,2), &
        	futil(1)-futil(11)-v(j,2), &
        	futil(2)-futil(11)-v(j,1), &
        	futil(3)-futil(11), &
        	futil(4)+v(j,3)-futil(11)-v(j,1)-v(j,2), &
        	futil(5)+v(j,3)-futil(11)-v(j,2), &
        	futil(6)+v(j,3)-futil(11)-v(j,1), &
        	futil(7)+v(j,3)-futil(11) &
	        ) &
           -DOT_PRODUCT(lchol(4,1:3),v(j,1:3)) &
        ) &
        /lchol(4,4),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)


    FAMSOLVE(11,3)=FAMSOLVE(11,3)+SUM(wg(:))/ns
    configprob(11)=SUM(wg(:))/ns

! Configuration 12
    ! condition: u12 > 0, u1 - u11, u13 - u15
    ! condition: futil12 + e3+e4 > 0					==> e4 > - futil12 - e3
    ! condition: futil12 + e3+e4 > futil1 + e1			==> e4 > futil1 + e1 - futil12 - e3
    ! condition: futil12 + e3+e4 > futil2 + e2			==> e4 > futil2 + e2 - futil12 - e3
    ! condition: futil12 + e3+e4 > futil3 + e1+e2		==> e4 > futil3 + e1+e2 - futil12 - e3
    ! condition: futil12 + e3+e4 > futil4 + e3			==> e4 > futil4 - futil12
    ! condition: futil12 + e3+e4 > futil5 + e1+e3		==> e4 > futil5 + e1 - futil12
    ! condition: futil12 + e3+e4 > futil6 + e2+e3		==> e4 > futil6 + e2 - futil12
    ! condition: futil12 + e3+e4 > futil7 + e1+e2+e3	==> e4 > futil7 + e1+e2 - futil12
    ! condition: futil12 + e3+e4 > futil8 + e4			==> e3 > futil8 - futil12
    ! condition: futil12 + e3+e4 > futil9 + e1+e4		==> e3 > futil9 + e1 - futil12
    ! condition: futil12 + e3+e4 > futil10 +e2+e4		==> e3 > futil10 + e2 - futil12
    ! condition: futil12 + e3+e4 > futil11 +e1+e2+e4	==> e3 > futil11 + e1+e2 - futil12
    ! condition: futil12 + e3+e4 > futil13 +e1+e3+e4	==> e1 < futil12 - futil13
    ! condition: futil12 + e3+e4 > futil14 +e2+e3+e4	==> e2 < futil12 - futil14
    ! condition: futil12 + e3+e4 > futil15 +e1+e2+e3+e4	==> e2 < futil12 - futil15 - e1

	tb=NORMCDF(futil(12)-futil(13),0.0d0,1.0d0)
    wg(:)=tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,1)=-HUGE(v(j,1))
   	        ELSE
   	        v(j,1)=INVNORMCDF((tb(j)*draws4seed(j,4)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
        	futil(12)-futil(15)-v(j,1), &
        	futil(12)-futil(14) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,2)=-HUGE(v(j,2))
   	        ELSE
   	        v(j,2)=INVNORMCDF((tb(j)*draws4seed(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	futil(8)-futil(12), &
        	futil(9)+v(j,1)-futil(12), &
        	futil(10)+v(j,2)-futil(12), &
        	futil(11)+v(j,1)+v(j,2)-futil(12) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
        	    v(j,3)=HUGE(v(j,3))
   	        ELSE
           	v(j,3)=INVNORMCDF((ta(j)*(1-draws4seed(j,2))+draws4seed(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
			-futil(12)-v(j,3), &
        	futil(1)+v(j,1)-futil(12)-v(j,3), &
        	futil(2)+v(j,2)-futil(12)-v(j,3), &
        	futil(3)+v(j,1)+v(j,2)-futil(12)-v(j,3), &
        	futil(4)-futil(12), &
        	futil(5)+v(j,1)-futil(12), &
        	futil(6)+v(j,2)-futil(12), &
        	futil(7)+v(j,1)+v(j,2)-futil(12) &
	        ) &
           -DOT_PRODUCT(lchol(4,1:3),v(j,1:3)) &
        ) &
        /lchol(4,4),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)


    FAMSOLVE(12,3)=FAMSOLVE(12,3)+SUM(wg(:))/ns
    configprob(12)=SUM(wg(:))/ns

! Configuration 13
    ! condition: u13 > 0, u1 - u12, u14 - u15
    ! condition: futil13 + e1+e3+e4 > 0						==> e4 > - futil13 - e1 - e3
    ! condition: futil13 + e1+e3+e4 > futil1 + e1			==> e4 > futil1 - futil13 - e3
    ! condition: futil13 + e1+e3+e4 > futil2 + e2			==> e4 > futil2 + e2 - futil13 - e1 - e3
    ! condition: futil13 + e1+e3+e4 > futil3 + e1+e2		==> e4 > futil3 + e2 - futil13 - e3
    ! condition: futil13 + e1+e3+e4 > futil4 + e3			==> e4 > futil4 - futil13 - e1
    ! condition: futil13 + e1+e3+e4 > futil5 + e1+e3		==> e4 > futil5 - futil13
    ! condition: futil13 + e1+e3+e4 > futil6 + e2+e3		==> e4 > futil6 + e2 - futil13 - e1
    ! condition: futil13 + e1+e3+e4 > futil7 + e1+e2+e3		==> e4 > futil7 + e2 - futil13
    ! condition: futil13 + e1+e3+e4 > futil8 + e4			==> e3 > futil8 - futil13 - e1
    ! condition: futil13 + e1+e3+e4 > futil9 + e1+e4		==> e3 > futil9 - futil13
    ! condition: futil13 + e1+e3+e4 > futil10 +e2+e4		==> e3 > futil10 + e2 - futil13 - e1
    ! condition: futil13 + e1+e3+e4 > futil11 +e1+e2+e4		==> e3 > futil11 + e2 - futil13
    ! condition: futil13 + e1+e3+e4 > futil12 +e3+e4		==> e1 > futil12 - futil13
    ! condition: futil13 + e1+e3+e4 > futil14 +e2+e3+e4		==> e2 < futil13 + e1 - futil14
    ! condition: futil13 + e1+e3+e4 > futil15 +e1+e2+e3+e4	==> e2 < futil13 - futil15

    ta=NORMCDF(futil(12)-futil(13),0.0d0,1.0d0)
    wg(:)=1.0-ta
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,1)=HUGE(v(j,1))
            ELSE
            	v(j,1)=INVNORMCDF((ta(j)*(1-draws4seed(j,4))+draws4seed(j,4)),0.0d0,1.0d0)
	        ENDIF
        END DO

	DO j=1, ns
    tb(j)=NORMCDF( &
        ( MIN( &
        	futil(13)+v(j,1)-futil(14), &
        	futil(13)-futil(15) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,2)=-HUGE(v(j,2))
   	        ELSE
   	        v(j,2)=INVNORMCDF((tb(j)*draws4seed(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	futil(8)-futil(13)-v(j,1), &
        	futil(9)-futil(13), &
        	futil(10)+v(j,2)-futil(13)-v(j,1), &
        	futil(11)+v(j,2)-futil(13) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
        	    v(j,3)=HUGE(v(j,3))
   	        ELSE
           	v(j,3)=INVNORMCDF((ta(j)*(1-draws4seed(j,2))+draws4seed(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
			-futil(13)-v(j,1)-v(j,3), &
        	futil(1)-futil(13)-v(j,3), &
        	futil(2)+v(j,2)-futil(13)-v(j,1)-v(j,3), &
        	futil(3)+v(j,2)-futil(13)-v(j,3), &
        	futil(4)-futil(13)-v(j,1), &
        	futil(5)-futil(13), &
        	futil(6)+v(j,2)-futil(13)-v(j,1), &
        	futil(7)+v(j,2)-futil(13) &
	        ) &
           -DOT_PRODUCT(lchol(4,1:3),v(j,1:3)) &
        ) &
        /lchol(4,4),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)


    FAMSOLVE(13,3)=FAMSOLVE(13,3)+SUM(wg(:))/ns
    configprob(13)=SUM(wg(:))/ns

! Configuration 14
    ! condition: u14 > 0, u1 - u13, u15
    ! condition: futil14 + e2+e3+e4 > 0						==> e4 > - futil14 - e2 - e3
    ! condition: futil14 + e2+e3+e4 > futil1 + e1			==> e4 > futil1 +e1 - futil14 - e2 - e3
    ! condition: futil14 + e2+e3+e4 > futil2 + e2			==> e4 > futil2 - futil14 - e3
    ! condition: futil14 + e2+e3+e4 > futil3 + e1+e2		==> e4 > futil3 + e1 - futil14 - e3
    ! condition: futil14 + e2+e3+e4 > futil4 + e3			==> e4 > futil4 - futil14 - e2
    ! condition: futil14 + e2+e3+e4 > futil5 + e1+e3		==> e4 > futil5 + e1 - futil14 - e2
    ! condition: futil14 + e2+e3+e4 > futil6 + e2+e3		==> e4 > futil6 - futil14
    ! condition: futil14 + e2+e3+e4 > futil7 + e1+e2+e3		==> e4 > futil7 + e1 - futil14
    ! condition: futil14 + e2+e3+e4 > futil8 + e4			==> e3 > futil8 - futil14 - e2
    ! condition: futil14 + e2+e3+e4 > futil9 + e1+e4		==> e3 > futil9 + e1 - futil14 - e2
    ! condition: futil14 + e2+e3+e4 > futil10 +e2+e4		==> e3 > futil10 - futil14
    ! condition: futil14 + e2+e3+e4 > futil11 +e1+e2+e4		==> e3 > futil11 + e1 - futil14
    ! condition: futil14 + e2+e3+e4 > futil12 +e3+e4		==> e2 > futil12 - futil14
    ! condition: futil14 + e2+e3+e4 > futil13 +e1+e3+e4		==> e2 > futil13 + e1 - futil14
    ! condition: futil14 + e2+e3+e4 > futil15 +e1+e2+e3+e4	==> e1 < futil14 - futil15

	tb=NORMCDF(futil(14)-futil(15),0.0d0,1.0d0)
    wg(:)=tb
        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,1)=-HUGE(v(j,1))
   	        ELSE
   	        v(j,1)=INVNORMCDF((tb(j)*draws4seed(j,4)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	futil(12)-futil(14), &
        	futil(13)+v(j,1)-futil(14) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
        	    v(j,2)=HUGE(v(j,2))
   	        ELSE
           	v(j,2)=INVNORMCDF((ta(j)*(1-draws4seed(j,3))+draws4seed(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	futil(8)-futil(14)-v(j,2), &
        	futil(9)+v(j,1)-futil(14)-v(j,2), &
        	futil(10)-futil(14), &
        	futil(11)+v(j,1)-futil(14) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
        	    v(j,3)=HUGE(v(j,3))
   	        ELSE
           	v(j,3)=INVNORMCDF((ta(j)*(1-draws4seed(j,2))+draws4seed(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
			-futil(14)-v(j,2)-v(j,3), &
        	futil(1)+v(j,1)-futil(14)-v(j,2)-v(j,3), &
        	futil(2)-futil(14)-v(j,3), &
        	futil(3)+v(j,1)-futil(14)-v(j,3), &
        	futil(4)-futil(14)-v(j,2), &
        	futil(5)+v(j,1)-futil(14)-v(j,2), &
        	futil(6)-futil(14), &
        	futil(7)+v(j,1)-futil(14) &
	        ) &
           -DOT_PRODUCT(lchol(4,1:3),v(j,1:3)) &
        ) &
        /lchol(4,4),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)


    FAMSOLVE(14,3)=FAMSOLVE(14,3)+SUM(wg(:))/ns
    configprob(14)=SUM(wg(:))/ns

! Configuration 15
    ! condition: u15 > 0, u1 - u14
    ! condition: futil15 + e1+e2+e3+e4 > 0					==> e4 > - futil15 - e1 - e2 - e3
    ! condition: futil15 + e1+e2+e3+e4 > futil1 + e1		==> e4 > futil1 - futil15 - e2 - e3
    ! condition: futil15 + e1+e2+e3+e4 > futil2 + e2		==> e4 > futil2 - futil15 - e1 - e3
    ! condition: futil15 + e1+e2+e3+e4 > futil3 + e1+e2		==> e4 > futil3 - futil15 - e3
    ! condition: futil15 + e1+e2+e3+e4 > futil4 + e3		==> e4 > futil4 - futil15 - e1 - e2
    ! condition: futil15 + e1+e2+e3+e4 > futil5 + e1+e3		==> e4 > futil5 - futil15 - e2
    ! condition: futil15 + e1+e2+e3+e4 > futil6 + e2+e3		==> e4 > futil6 - futil15 - e1
    ! condition: futil15 + e1+e2+e3+e4 > futil7 + e1+e2+e3	==> e4 > futil7 - futil15
    ! condition: futil15 + e1+e2+e3+e4 > futil8 + e4		==> e3 > futil8 - futil15 - e1 - e2
    ! condition: futil15 + e1+e2+e3+e4 > futil9 + e1+e4		==> e3 > futil9 - futil15 - e2
    ! condition: futil15 + e1+e2+e3+e4 > futil10 +e2+e4		==> e3 > futil10 - futil15 - e1
    ! condition: futil15 + e1+e2+e3+e4 > futil11 +e1+e2+e4	==> e3 > futil11 - futil15
    ! condition: futil15 + e1+e2+e3+e4 > futil12 +e3+e4		==> e2 > futil12 - futil15 - e1
    ! condition: futil15 + e1+e2+e3+e4 > futil13 +e1+e3+e4	==> e2 > futil13 - futil15
    ! condition: futil15 + e1+e2+e3+e4 > futil14 +e2+e3+e4	==> e1 > futil14 - futil15

    ta=NORMCDF(futil(14)-futil(15),0.0d0,1.0d0)
    wg(:)=1.0-ta
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,1)=HUGE(v(j,1))
            ELSE
            	v(j,1)=INVNORMCDF((ta(j)*(1-draws4seed(j,4))+draws4seed(j,4)),0.0d0,1.0d0)
	        ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	futil(12)-futil(15)-v(j,1), &
        	futil(13)-futil(15) &
	        ) &
           -DOT_PRODUCT(lchol(2,1:1),v(j,1:1)) &
        ) &
        /lchol(2,2),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
        	    v(j,2)=HUGE(v(j,2))
   	        ELSE
           	v(j,2)=INVNORMCDF((ta(j)*(1-draws4seed(j,3))+draws4seed(j,3)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
        	futil(8)-futil(15)-v(j,1)-v(j,2), &
        	futil(9)-futil(15)-v(j,2), &
        	futil(10)-futil(15)-v(j,1), &
        	futil(11)-futil(15) &
	        ) &
           -DOT_PRODUCT(lchol(3,1:2),v(j,1:2)) &
        ) &
        /lchol(3,3),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)
        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
        	    v(j,3)=HUGE(v(j,3))
   	        ELSE
           	v(j,3)=INVNORMCDF((ta(j)*(1-draws4seed(j,2))+draws4seed(j,2)),0.0d0,1.0d0)
            ENDIF
        END DO

	DO j=1, ns
    ta(j)=NORMCDF( &
        ( MAX( &
			-futil(15)-v(j,1)-v(j,2)-v(j,3), &
        	futil(1)-futil(15)-v(j,2)-v(j,3), &
        	futil(2)-futil(15)-v(j,1)-v(j,3), &
        	futil(3)-futil(15)-v(j,3), &
        	futil(4)-futil(15)-v(j,1)-v(j,2), &
        	futil(5)-futil(15)-v(j,2), &
        	futil(6)-futil(15)-v(j,1), &
        	futil(7)-futil(15) &
	        ) &
           -DOT_PRODUCT(lchol(4,1:3),v(j,1:3)) &
        ) &
        /lchol(4,4),0.0d0,1.0d0)
	END DO         ! Simulation loop
	wg(:)=wg(:)*(1.0d0-ta)


    FAMSOLVE(15,3)=FAMSOLVE(15,3)+SUM(wg(:))/ns
    configprob(15)=SUM(wg(:))/ns



CASE DEFAULT
    PRINT *,'nchild(m): ', nchild(m)
    PRINT *,'error'
    PAUSE
END SELECT


temp=0
DO i=0,15
    IF (configprob(i)>temp) THEN
        temp=configprob(i)
        predconfig=i
    ENDIF
ENDDO
DO l=1,nchild(m)
    predobs(k-nchild(m)+l)=IBITS(predconfig,nchild(m)-l,1)
ENDDO



END DO  ! nfam



END FUNCTION FAMSOLVE









END MODULE Object









PROGRAM MAIN
USE Globvar
USE INIT
USE MINIMIZATION
USE Object
IMPLICIT NONE

REAL(8) :: resultmlv, resultmlse
REAL(8) :: gtol
REAL(4) :: time0, iniadj
INTEGER :: i, j, k, l, m, n, io
INTEGER :: s1, s2, s3, s4, nup
CHARACTER(12) :: horz(3)
CHARACTER(20) :: flname
CHARACTER(80) :: flname2
CHARACTER(20) :: fmtname           ! To accomodate different number of parameters
CHARACTER(50) :: inivalnote
CHARACTER(60) :: setupnote

REAL(8), ALLOCATABLE :: initheta(:)
REAL(8), ALLOCATABLE :: vcov(:,:)
REAL(8), ALLOCATABLE :: se(:)
REAL(8), ALLOCATABLE :: tval(:)
CHARACTER(16), ALLOCATABLE :: vert(:)
REAL(8), ALLOCATABLE :: pred(:,:)

horz(1)='        coef' ; horz(2)='        s.e.' ; horz(3)=' asy t-val'

! Read the setting file
    flname2='setting.prn'
    OPEN(1,file=(flname2))
    READ(1,fmt='(A60)') setupnote
    READ(1,fmt=*) nthetab
    READ(1,fmt=*) nthetaa
    READ(1,fmt=*) nthetag
    READ(1,fmt=*) nthetar
    READ(1,fmt=*) ntheta
    READ(1,fmt=*) nobs
    READ(1,fmt=*) nfam
    READ(1,fmt=*) ns
ALLOCATE (initheta(ntheta))
    READ(1,fmt=*) initheta
ALLOCATE (vert(ntheta))
    vert(1)='cons      '
    READ(1,fmt=*) vert(2:ntheta)
!    READ(1,fmt=*) path
    CLOSE(1)


ALLOCATE (theta(ntheta))
ALLOCATE (vcov(ntheta,ntheta))
ALLOCATE (se(ntheta))
ALLOCATE (tval(ntheta))
ALLOCATE (famid(nobs))
ALLOCATE (famindex(nfam))
ALLOCATE (nchild(nfam))
ALLOCATE (y(nobs))
ALLOCATE (xb(nobs,nthetab))
IF (nthetaa>=3) THEN
ALLOCATE (xa(nobs,nthetaa-2))
ELSEIF (nthetaa==1) THEN
ALLOCATE (xa(nobs,1))
ENDIF
IF (nthetag>=2) THEN
ALLOCATE (xg(nobs,nthetag-1))
ELSEIF (nthetag==1) THEN
ALLOCATE (xg(nobs,1))
ENDIF
ALLOCATE (xr(nobs,nthetar,nchildmax))
ALLOCATE (near(nfam))
ALLOCATE (predobs(nobs))
ALLOCATE (pred(0:2**nchildmax-1,5))


! Set gradient tolerance
gtol=0.000001d0

time0 = SECNDS(0.0)

CALL GETTIM(i, j, k, l)
CALL GETDAT(l, m, n)
flname2=CHAR(ICHAR('0')+(m/10))//CHAR(ICHAR('0')+(m-m/10*10))//CHAR(ICHAR('0')+(n/10))//CHAR(ICHAR('0')+(n-n/10*10))//CHAR(ICHAR('0')+(i/10))//CHAR(ICHAR('0')+(i-i/10*10))//CHAR(ICHAR('0')+(j/10))//CHAR(ICHAR('0')+(j-j/10*10))//CHAR(ICHAR('0')+(k/10))//CHAR(ICHAR('0')+(k-k/10*10))//'.prn'

OPEN(4,file=(flname2),STATUS='REPLACE')
WRITE(4,fmt='(A20,4x,A60)') 'File Name: ', flname2
WRITE(4,fmt='(A20,4x,A60)') 'Estimation Setting: ', setupnote
WRITE(4,fmt=*) '# of families     :  ', nfam
WRITE(4,fmt=*) '# of observations:  ', nobs
WRITE(4,fmt=*) '# of simuln draws  :  ', ns
WRITE(4,fmt=*) ' '

theta=initheta
numeval=0

! Load the data set
CALL READDATA(nobs)


ALLOCATE (draws2seed(ns,2:2))
ALLOCATE (draws3seed(ns,2:3))
ALLOCATE (draws4seed(ns,2:4))
CALL RANDOM_SEED(PUT=(/3000/))
CALL RANDOM_NUMBER (draws2seed)
CALL RANDOM_NUMBER (draws3seed)
CALL RANDOM_NUMBER (draws4seed)
draws2seed(ns/2+1:ns,:)=1-draws2seed(1:ns/2,:)        ! This line makes antithetics
draws3seed(ns/2+1:ns,:)=1-draws3seed(1:ns/2,:)        ! This line makes antithetics
draws4seed(ns/2+1:ns,:)=1-draws4seed(1:ns/2,:)        ! This line makes antithetics


! the following lines compute the estimates using a Quasi-Newton method
! with a numerical gradient

CALL dfpmin(theta,ntheta,gtol,lnmlobj)
objsaved=lnmlobj(theta,ntheta,0)
objsaved=-objsaved


print *,'check4-1! - 1st round Minimization Finished!'
print *,'objsaved : ',objsaved

print *,'Calculating variance-covariance matrix...'

vcov=var_cov(theta, ntheta)

print *,'check4-2! - vcov obtained!'
!	print *,vcov


DO i=1,ntheta
	print *,'vcov(i,i):  ',i, vcov(i,i)
END DO

DO i=1,ntheta
	se(i)=DSQRT(vcov(i,i))
	tval(i)=theta(i)/se(i)
END DO




WRITE(4,fmt=*) '# of parameters:  ', ntheta
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(10x,3A16)') (horz(i),i=1,3)
WRITE(4,fmt=*) ' '
DO i=1,ntheta
	WRITE(4,fmt='(A16,3F16.10)') vert(i), theta(i), se(i), tval(i)
ENDDO

pred=0
WRITE(4,fmt=*) ' '
WRITE(4,fmt=*) 'Location patterns: Data'
WRITE(4,fmt='(A8,10A8)') '  config',' 1 child','     (%)',' 2 child','     (%)',' 3 child','     (%)',' 4 child','     (%)'
DO i=0,2**nchildmax-1
    DO j=1,nfam
        IF (near(j)==i) THEN
            pred(i,nchild(j))=pred(i,nchild(j))+1
        ENDIF
    ENDDO
ENDDO
DO i=0,2**nchildmax-1
	WRITE(4,fmt='(4X,4I1,4(F8.1,F8.2))') IBITS(i,3,1),IBITS(i,2,1),IBITS(i,1,1),IBITS(i,0,1), (pred(i,k),(pred(i,k)/SUM(pred(:,k))*100.0d0),k=1,4)
ENDDO
WRITE(4,fmt='(A8,4(F8.1,F8.2))') '   total', ((SUM(pred(:,j)),100.0),j=1,4)

pred=0
WRITE(4,fmt=*) ' '
WRITE(4,fmt=*) 'Location patterns: Predicted'
WRITE(4,fmt='(A8,10A8)') '  config',' 1 child','     (%)',' 2 child','     (%)',' 3 child','     (%)',' 4 child','     (%)'
    pred(:,:)=FAMSOLVE(theta,ntheta)
DO i=0,2**nchildmax-1
	WRITE(4,fmt='(4X,4I1,4(F8.1,F8.2))') IBITS(i,3,1),IBITS(i,2,1),IBITS(i,1,1),IBITS(i,0,1), (pred(i,k),(pred(i,k)/SUM(pred(:,k))*100.0d0),k=1,4)
ENDDO
WRITE(4,fmt='(A8,4(F8.1,F8.2))') '   total', ((SUM(pred(:,j)),100.0),j=1,4)



WRITE(4,fmt=*) ' '
n=0
DO m=1,nfam
DO j=1,nchild(m)
    IF (predobs(famindex(m)-nchild(m)+j)==IBITS(near(m),nchild(m)-j,1)) n=n+1
ENDDO
ENDDO
WRITE(4,fmt=*) '% of observations correctly predicted:  ', n*1.0d0/nobs*100
n=0
DO m=1,nfam
    k=0
    DO j=1,nchild(m)
        IF (predobs(famindex(m)-nchild(m)+j)==IBITS(near(m),nchild(m)-j,1)) k=k+1
    ENDDO
    IF (k==nchild(m)) n=n+1
ENDDO
WRITE(4,fmt=*) '% of families correctly predicted:  ', n*1.0d0/nfam*100
n=0; l=0
DO m=1,nfam
    IF (nchild(m)==1) THEN
        IF (predobs(famindex(m))==IBITS(near(m),0,1)) n=n+1
    l=l+1
    ENDIF
ENDDO
WRITE(4,fmt=*) '% of 1 child families correctly predicted:  ', n*1.0d0/l*100
n=0; l=0
DO m=1,nfam
    IF (nchild(m)==2) THEN
    k=0
    DO j=1,nchild(m)
        IF (predobs(famindex(m)-nchild(m)+j)==IBITS(near(m),nchild(m)-j,1)) k=k+1
    ENDDO
    IF (k==nchild(m)) n=n+1
    l=l+1
    ENDIF
ENDDO
WRITE(4,fmt=*) '% of 2 child families correctly predicted:  ', n*1.0d0/l*100
n=0; l=0
DO m=1,nfam
    IF (nchild(m)==3) THEN
    k=0
    DO j=1,nchild(m)
        IF (predobs(famindex(m)-nchild(m)+j)==IBITS(near(m),nchild(m)-j,1)) k=k+1
    ENDDO
    IF (k==nchild(m)) n=n+1
    l=l+1
    ENDIF
ENDDO
WRITE(4,fmt=*) '% of 3 child families correctly predicted:  ', n*1.0d0/l*100
n=0; l=0
DO m=1,nfam
    IF (nchild(m)==4) THEN
    k=0
    DO j=1,nchild(m)
        IF (predobs(famindex(m)-nchild(m)+j)==IBITS(near(m),nchild(m)-j,1)) k=k+1
    ENDDO
    IF (k==nchild(m)) n=n+1
    l=l+1
    ENDIF
ENDDO
WRITE(4,fmt=*) '% of 4 child families correctly predicted:  ', n*1.0d0/l*100

WRITE(4,fmt=*) ' '
WRITE(4,fmt=*) ' '
WRITE(4,fmt=*) 'ML objective:  ', objsaved
WRITE(4,fmt=*) '# of objective function evaluations:  ', numeval
WRITE(4,fmt=*) ' '
WRITE(4,fmt=*) 'run time(seconds):  ', SECNDS(time0)
WRITE(4,fmt=*) ' '
WRITE(4,fmt=*) ' '
WRITE(4,fmt=*) 'Initial Values: '
DO i=1,ntheta
WRITE(4,fmt='(F16.10)') initheta(i)
END DO
WRITE(4,fmt=*) ' '

CLOSE(4)

! estimated parameters
OPEN(1,file='param1.out',STATUS='REPLACE')
OPEN(2,file='param2.out',STATUS='REPLACE')
flname='('//CHAR((nthetab)/100+48)//CHAR(MOD(nthetab,100)/10+48)//CHAR(MOD(nthetab,10)+48)//'F20.14)'
WRITE(1,fmt=flname) theta(1:nthetab)
IF (ntheta>nthetab) THEN
flname='('//CHAR((ntheta-nthetab)/100+48)//CHAR(MOD(ntheta-nthetab,100)/10+48)//CHAR(MOD(ntheta-nthetab,10)+48)//'F20.14)'
WRITE(2,fmt=flname) theta(nthetab+1:ntheta)
ENDIF
CLOSE(1)
CLOSE(2)

END PROGRAM MAIN
