MODULE Globvar
IMPLICIT NONE
! This module declares the variables that are going to be used throughout the program.
! To avoid stack overflow error --> Stack Reserve: 50,000,000

INTEGER, PARAMETER :: nchildmax=4  ! max number of children
INTEGER, PARAMETER :: sm2=2**(1+1)  ! necessary amount for strategy patterns
INTEGER, PARAMETER :: sm3=2**(1+1)*2**(2+1)  ! = 32 = necessary amount for strategy patterns
INTEGER, PARAMETER :: sm4=2**(1+1)*2**(2+1)*2**(3+1)  ! = 512 = necessary amount for strategy patterns
	! The 4th child in a 4-child family has 2**(3+1) strategies, because 
	! the number of possible elder siblings in the upstream from 0 to 3, i.e. 4.
INTEGER :: nobs
INTEGER :: nfam
INTEGER :: ns              ! number of simulation draws
INTEGER :: ntheta
INTEGER :: nthetab     ! # of parameters in beta term
INTEGER :: nthetaa     ! # of parameters in alpha term
INTEGER :: nthetag     ! # of parameters in gamma term
INTEGER :: nthetar     ! # of parameters in rho term
REAL(4) :: amp1         ! alpha multiplier for counterfactual simulation
REAL(4) :: amp2         ! gamma multiplier for counterfactual simulation
INTEGER, ALLOCATABLE :: famid(:)   ! famid(nobs)
INTEGER, ALLOCATABLE :: nchild(:), famindex(:)   ! nchild(nfam), famindex(nfam)
REAL(8), ALLOCATABLE :: theta(:) ! theta(ntheta)
REAL(8), ALLOCATABLE :: xb(:,:)   ! xb(nthetab,nobs)
REAL(8), ALLOCATABLE :: xa(:,:)   ! xa(nthetaa,nobs)
REAL(8), ALLOCATABLE :: xg(:,:)   ! xg(nthetag,nobs)
INTEGER :: sa, sg		! actual size of xa and xg
REAL(8), ALLOCATABLE :: xr(:,:,:)   ! xr(nthetar,nchildmax,nobs)
INTEGER, ALLOCATABLE :: y(:)     ! y(nobs)
INTEGER, ALLOCATABLE :: near(:)   ! near(nfam)
REAL(8), ALLOCATABLE :: near_cf(:,:)   ! near_cf(nfam,0:15) ... proportions of living arrangement types for counterfactual simulation
REAL(8), ALLOCATABLE :: utilxb(:) ! utilxb(nobs)

REAL(8), ALLOCATABLE :: draws2(:,:,:)		! nchild=2. draws(ns,necessary amount,2)
REAL(8), ALLOCATABLE :: draws3(:,:,:)		! nchild=3. draws(ns,necessary amount,3)
REAL(8), ALLOCATABLE :: draws4(:,:,:)		! nchild=4. draws(ns,necessary amount,4)
REAL(8), ALLOCATABLE :: draws2seed(:,:,:)	! nchild=2. draws(ns,necessary amount,2)
REAL(8), ALLOCATABLE :: draws3seed(:,:,:)	! nchild=3. draws(ns,necessary amount,3)
REAL(8), ALLOCATABLE :: draws4seed(:,:,:)	! nchild=4. draws(ns,necessary amount,4)

REAL(8), ALLOCATABLE :: exputil(:)   ! exputil(nobs) ... expected utility
REAL(8), ALLOCATABLE :: dst(:)   ! dst(nobs) ... whether a chidl is playing a daminant strategy (in a simultaneous move game sense)
                                    !!! This is not whether the child has a dominant strategy. A child who has a dominant strategy may not be playing it in SPNE.
REAL(8), ALLOCATABLE :: famutil(:) ! famutil(nfam)
REAL(8), ALLOCATABLE :: fbfamutil(:) !  fbfamutil(nfam)
REAL(8), ALLOCATABLE :: pbfamutil(:) !  pbfamutil(nfam)
REAL(8), ALLOCATABLE :: y_any_fb(:) !  y_any_fb(nfam) ... probability that a parent has anyone near in the fb outcome
REAL(8), ALLOCATABLE :: gtype1a(:,:,:,:) !  gtype1a(nfam,0:15,1:6,0:15) ... proportions of game type patterns (1-4 child families)
REAL(8), ALLOCATABLE :: gtype1b(:,:,:,:) !  gtype1b(nfam,0:15,0:15,0:15) ... proportions of game type patterns (1-4 child families)
REAL(8), ALLOCATABLE :: gtype1c(:,:,:,:,:,:) !  gtype1c(nfam,0:6,0:4,1:3,0:15,0:15) ... proportions of game type patterns (1-4 child families)


CHARACTER(10) :: path                          ! Specify the data directory. To be read from setting.prn

END MODULE Globvar






MODULE INIT
	USE Globvar
	IMPLICIT NONE

CONTAINS

	SUBROUTINE READDATA(nobs)
		USE Globvar
		IMPLICIT NONE
		INTEGER, INTENT(IN) :: nobs
		INTEGER :: i,j,k,m,io

        CHARACTER(60) :: readfile_x      ! File names to be read

readfile_x='data.raw'

xb=0.0d0; y=0.0d0
xb(1,:)=1.00000000d0
xa(1,:)=1.0000000000d0
xr(1,:,:)=1.00000000d0
IF (nthetag>=1) xg(1,:)=1.00000000d0

print *,'Reading Data ... '

	OPEN(1,file=(readfile_x),status='old')

    IF (nthetaa>3 .and. nthetar>1 .and. nthetag>2) THEN
        DO i = 1, nobs
	    READ(1,fmt=*,iostat=io) y(i), famid(i), xb(2:,i), xa(2:,i), xg(2:,i), xr(2:nthetar,1,i), xr(2:nthetar,2,i), xr(2:nthetar,3,i), xr(2:nthetar,4,i)
        IF (io<0) EXIT
        END DO
    ELSE IF (nthetaa>3 .and. nthetar>1) THEN
        DO i = 1, nobs
	    READ(1,fmt=*,iostat=io) y(i), famid(i), xb(2:,i), xa(2:,i), xr(2:nthetar,1,i), xr(2:nthetar,2,i), xr(2:nthetar,3,i), xr(2:nthetar,4,i)
        IF (io<0) EXIT
        END DO
    ELSE IF (nthetaa>1) THEN
        DO i = 1, nobs
	    READ(1,fmt=*,iostat=io) y(i), famid(i), xb(2:,i), xa(2:,i)
        IF (io<0) EXIT
        END DO
    ELSE
        DO i = 1, nobs
	    READ(1,fmt=*,iostat=io) y(i), famid(i), xb(2:,i)
        IF (io<0) EXIT
        END DO
    ENDIF
    CLOSE(1)

nchild=0
j=1
nchild(1)=1
DO i=2,nobs
	IF (famid(i)/=famid(i-1)) THEN
		j=j+1
	ENDIF
		nchild(j)=nchild(j)+1
ENDDO

famindex(1)=nchild(1)
DO i=2,nfam
	famindex(i)=famindex(i-1)+nchild(i)
ENDDO


DO m=1,nfam
near(m)=0
DO i=1,nchild(m)
near(m)=near(m)*2
near(m)=near(m)+y(famindex(m)-nchild(m)+i)
END DO
END DO



print *,'Reading Data ... Finished '

	END SUBROUTINE READDATA

END MODULE INIT

	
	
MODULE Object
	USE Globvar
	USE GHK
	USE MATRIX
    USE PROBABILITY
	IMPLICIT NONE

CONTAINS


	FUNCTION SIMULATION(param,n) RESULT(lnmlobj)	! This subroutine conducts simulations
		USE Globvar
		IMPLICIT NONE
		INTEGER, INTENT(IN) :: n 
		REAL(8), INTENT(IN) :: param(n)
		REAL(8) :: mlobj
		REAL(8) :: lnmlobj
		REAL(8) :: famprob(nfam)			! The family-level likelihood
		INTEGER :: i, j, k, m
		REAL(4) :: tms

		print *,'SUBROUTINE SIMULATION ... In '

		tms = SECNDS(0.0)

		exputil=0.0d0
		dst=0.0d0
		famutil=0.0d0
		fbfamutil=0.0d0
		pbfamutil=0.0d0
		y_any_fb=0.0d0
		gtype1a=0.0d0
		gtype1b=0.0d0
		gtype1c=0.0d0
		near_cf=0.0d0

		draws2=draws2seed
		draws3=draws3seed
		draws4=draws4seed

		DO m = 1, nfam

		IF (MOD(m,500)==0) THEN
			PRINT *,' ... current family # / nfam: ', m, '/',nfam
		ENDIF

		SELECT CASE(nchild(m))
		CASE(2)
			draws2=CSHIFT(draws2,1,2)
		CASE(3)
			draws3=CSHIFT(draws3,7,2)
		CASE(4)
			draws4=CSHIFT(draws4,9,2)
		END SELECT


		SELECT CASE(nchild(m))
		CASE(1)
			famprob(m)=DLOG(SIM1(m,param,n))
		CASE(2)
			famprob(m)=DLOG(SIM2(m,param,n))
		CASE(3)
			famprob(m)=DLOG(SIM3(m,param,n))
		CASE(4)
			famprob(m)=DLOG(SIM4(m,param,n))
		CASE DEFAULT
			PRINT *,'nchild(m): ', nchild(m)
			PRINT *,'error'
			PAUSE
		END SELECT


		END DO

		lnmlobj=(0-1)*SUM(famprob)
		print *,'ln ML objective: ',-lnmlobj
		print *,'Time spent for Simulation: ',SECNDS(tms)

	END FUNCTION SIMULATION


FUNCTION SIM1(m,param,n) RESULT(ghkval)
	! This function calculates the GHK simulator for each market*year
USE Globvar
USE PROBABILITY
IMPLICIT NONE
INTEGER, INTENT(IN) :: m			! market id
INTEGER, INTENT(IN) :: n 
REAL(8), INTENT(IN) :: param(n)
REAL(8) :: v(ns)   ! GHK simulation draws constructed from draws()
INTEGER :: i, j, k, l
REAL(8) :: ghkval		! The calculated GHK value
REAL(8) :: utilsub, alpha

ghkval=0.000d0

k=famindex(m)

! PRINT *, 'SIM1-in'

	utilsub=DOT_PRODUCT(param(1:nthetab),xb(:,k))
    IF(nthetaa>=3) THEN
    alpha  = EXP( DOT_PRODUCT(param(nthetab+1:nthetab+nthetaa-2),xa(:,k)) )
!    alpha  = DOT_PRODUCT(param(nthetab+1:nthetab+nthetaa-2),xa(:,k)) 
    ELSEIF(nthetaa==1) THEN
    alpha  =param(nthetab+nthetaa)
    ELSE
    alpha = 0.0d0
    ENDIF


	IF (IBITS(near(m),0,1)==1) THEN
    ghkval=1.0d0-NORMCDF(-utilsub-alpha,0.0d0,1.0d0)

    DO j=1, ns
        v(j)=INVNORMCDF((NORMCDF(-utilsub-alpha,0.0d0,1.0d0)*(1-draws2(j,1,1))+draws2(j,1,1)),0.0d0,1.0d0)
        IF (v(j)+utilsub+alpha*amp1 > 0.0d0) THEN
            near_cf(m,1)=near_cf(m,1)+1.0d0
            gtype1a(m,1,1,1)=gtype1a(m,1,1,1)+1.0d0
            gtype1b(m,1,1,1)=gtype1b(m,1,1,1)+1.0d0
            gtype1c(m,1,0,1,1,1)=gtype1c(m,1,0,1,1,1)+1.0d0
            exputil(k)=exputil(k)+(v(j)+utilsub+alpha*amp1)
            y_any_fb(m)=y_any_fb(m)+1.0d0
        ELSE
            near_cf(m,0)=near_cf(m,0)+1.0d0
            gtype1a(m,0,1,0)=gtype1a(m,0,1,0)+1.0d0
            gtype1b(m,1,1,0)=gtype1b(m,1,1,0)+1.0d0
            gtype1c(m,1,0,1,1,0)=gtype1c(m,1,0,1,1,0)+1.0d0
        ENDIF
    END DO

	ELSE
    ghkval=NORMCDF(-utilsub-alpha,0.0d0,1.0d0)
    DO j=1, ns
        v(j)=INVNORMCDF((NORMCDF(-utilsub-alpha,0.0d0,1.0d0)*draws2(j,1,1)),0.0d0,1.0d0)
        IF (v(j)+utilsub+alpha*amp1 > 0.0d0) THEN
            near_cf(m,1)=near_cf(m,1)+1.0d0
            gtype1a(m,1,1,1)=gtype1a(m,1,1,1)+1.0d0
            gtype1b(m,1,1,1)=gtype1b(m,1,1,1)+1.0d0
            gtype1c(m,1,0,1,1,1)=gtype1c(m,1,0,1,1,1)+1.0d0
            exputil(k)=exputil(k)+(v(j)+utilsub+alpha*amp1)
            y_any_fb(m)=y_any_fb(m)+1.0d0
        ELSE
            near_cf(m,0)=near_cf(m,0)+1.0d0
            gtype1a(m,0,1,0)=gtype1a(m,0,1,0)+1.0d0
            gtype1b(m,1,1,0)=gtype1b(m,1,1,0)+1.0d0
            gtype1c(m,1,0,1,1,0)=gtype1c(m,1,0,1,1,0)+1.0d0
        ENDIF
    END DO

    ENDIF

    exputil(k)=exputil(k)/ns
    famutil(m)=exputil(k)
    fbfamutil(m)=exputil(k)
    pbfamutil(m)=exputil(k)
    gtype1a(m,:,:,:)=gtype1a(m,:,:,:)/ns
    gtype1b(m,:,:,:)=gtype1b(m,:,:,:)/ns
    gtype1c(m,:,:,:,:,:)=gtype1c(m,:,:,:,:,:)/ns
    near_cf(m,:)=near_cf(m,:)/ns
    y_any_fb(m)=y_any_fb(m)/ns

dst(k)=1.0d0

END FUNCTION SIM1


FUNCTION SIM2(m,param,n) RESULT(ghkval)
	! This function calculates the GHK simulator for each market*year
USE Globvar
USE PROBABILITY
IMPLICIT NONE
INTEGER, INTENT(IN) :: m			! market id
INTEGER, INTENT(IN) :: n 
REAL(8), INTENT(IN) :: param(n)
INTEGER :: i, j, k, l
INTEGER :: s1, s2, s3, s4
REAL(8) :: wgv(ns,0:sm2-1,0:2)		! wgv(ns, # strategy combitnations, 1 + # child) ... wgv(ns,s,0) for wg.
REAL(8) :: wg(ns,0:sm2-1)
REAL(8) :: ta(ns), tb(ns)
REAL(8) :: v(ns,0:sm2-1,nchild(m))   ! GHK simulation draws constructed from draws()
REAL(8) :: gtemp1a(0:15,1:6,0:15)
REAL(8) :: gtemp1b(0:15,0:15,0:15)
REAL(8) :: gtemp1c(0:6,0:4,1:3,0:15,0:15)
REAL(8) :: nearcftemp(0:15)
REAL(8) :: utiltemp
REAL(8) :: utemp(nchild(m)),dstemp(nchild(m))
REAL(8) :: fbutemp, pbutemp, fbytemp
REAL(8) :: ghkval, ghkvals		! The calculated GHK value
REAL(8) :: utilsub(2), alpha(2,3), gamma(2,3)		! 1 ... oldest child
REAL(8) :: lchol(2,2)

wgv=0.000d0
wg=0.000d0
v=0.000d0
ghkval=0.000d0

k=famindex(m)

utemp=0.000d0
fbutemp=0.000d0 ; pbutemp=0.000d0 ; fbytemp=0.000d0

lchol=lcholbase(m,param,n)
wgv=GHK2(m,param,n,ns,draws2,lchol,nthetab, nthetaa, nthetag, xb(:,famindex(m)-1:famindex(m)), xa(:,famindex(m)-1:famindex(m)), xg(:,famindex(m)-1:famindex(m)), sa, sg, near(m),2)

	utilsub(1:2)=MATMUL(param(1:nthetab),xb(:,k-1:k))
	IF (nthetaa>=3) THEN
	alpha(1:2,1) = EXP( MATMUL(param(nthetab+1:nthetab+nthetaa-2),xa(:,k-1:k)) )
	alpha(1:2,2) = EXP( MATMUL(param(nthetab+1:nthetab+nthetaa-2),xa(:,k-1:k)) + param(nthetab+nthetaa-1) )
	alpha(1:2,3) = EXP( MATMUL(param(nthetab+1:nthetab+nthetaa-2),xa(:,k-1:k)) + param(nthetab+nthetaa-1) + param(nthetab+nthetaa) )
	ELSEIF (nthetaa==1) THEN
	alpha(1:2,1:3)=param(nthetab+nthetaa)
	ELSE
	alpha=0.0d0
	ENDIF
	IF (nthetag>=2) THEN
	gamma(1:2,1) =MATMUL(param(nthetab+nthetaa+1:nthetab+nthetaa+nthetag-1),xg(:,k-1:k))
	gamma(1:2,2) =MATMUL(param(nthetab+nthetaa+1:nthetab+nthetaa+nthetag-1),xg(:,k-1:k))*(1.0d0 + param(nthetab+nthetaa+nthetag))
	gamma(1:2,3) =MATMUL(param(nthetab+nthetaa+1:nthetab+nthetaa+nthetag-1),xg(:,k-1:k))*(1.0d0 + param(nthetab+nthetaa+nthetag)*2.0d0)
	ELSEIF (nthetag==1) THEN
	gamma(1:2,1:3)=param(nthetab+nthetaa+nthetag)
	ELSE
	gamma=0.0d0
	ENDIF


alpha(:,:)=alpha(:,:)*amp1
gamma(:,:)=gamma(:,:)*amp2

DO s1=0,2**2-1       ! Strategy loop

IF (ghkcheck(s1,2,m)==1 .and. wgv(1,s1,0)>0.0d0) THEN
    CALL geval(gtemp1a(:,:,:),gtemp1b(:,:,:),gtemp1c(:,:,:,:,:),nearcftemp(:),utemp,pbutemp,fbutemp,fbytemp,dstemp,m,wgv(:,s1,1:2),alpha,gamma,lchol)
ENDIF

ghkvals=SUM(wgv(:,s1,0))/ns
ghkval=ghkval+ghkvals
exputil(k-1:k)=exputil(k-1:k)+utemp(1:2)*ghkvals
dst(k-1:k)=dst(k-1:k)+dstemp(1:2)*ghkvals
fbfamutil(m)=fbfamutil(m)+fbutemp*ghkvals
pbfamutil(m)=pbfamutil(m)+pbutemp*ghkvals
y_any_fb(m)=y_any_fb(m)+fbytemp*ghkvals
gtype1a(m,:,:,:)=gtype1a(m,:,:,:)+gtemp1a(:,:,:)*ghkvals
gtype1b(m,:,:,:)=gtype1b(m,:,:,:)+gtemp1b(:,:,:)*ghkvals
gtype1c(m,:,:,:,:,:)=gtype1c(m,:,:,:,:,:)+gtemp1c(:,:,:,:,:)*ghkvals
near_cf(m,:)=near_cf(m,:)+nearcftemp(:)*ghkvals

END DO         ! Strategy loop

exputil(k-1:k)=exputil(k-1:k)/ghkval
dst(k-1:k)=dst(k-1:k)/ghkval
famutil(m)=SUM(exputil(k-1:k))
fbfamutil(m)=fbfamutil(m)/ghkval
pbfamutil(m)=pbfamutil(m)/ghkval
y_any_fb(m)=y_any_fb(m)/ghkval
gtype1a(m,:,:,:)  =gtype1a(m,:,:,:)/ghkval
gtype1b(m,:,:,:)  =gtype1b(m,:,:,:)/ghkval
gtype1c(m,:,:,:,:,:)=gtype1c(m,:,:,:,:,:)/ghkval
near_cf(m,:)=near_cf(m,:)/ghkval

END FUNCTION SIM2


FUNCTION SIM3(m,param,n) RESULT(ghkval)
	! This function calculates the GHK simulator for each market*year
USE Globvar
USE PROBABILITY
IMPLICIT NONE
INTEGER, INTENT(IN) :: m			! market id
INTEGER, INTENT(IN) :: n 
REAL(8), INTENT(IN) :: param(n)
INTEGER :: i, j, k, l
INTEGER :: s1, s2, s3, s4
REAL(8) :: wgv(ns,0:sm3-1,0:3)		! wgv(ns, # strategy combitnations, 1 + # child) ... wgv(ns,s,0) for wg.
REAL(8) :: wg(ns,0:sm3-1)
REAL(8) :: ta(ns), tb(ns)
REAL(8) :: v(ns,0:sm3-1,3)   ! GHK simulation draws constructed from draws()
REAL(8) :: gtemp1a(0:15,1:6,0:15)
REAL(8) :: gtemp1b(0:15,0:15,0:15)
REAL(8) :: gtemp1c(0:6,0:4,1:3,0:15,0:15)
REAL(8) :: nearcftemp(0:15)
REAL(8) :: utiltemp
REAL(8) :: utemp(3), dstemp(3)
REAL(8) :: fbutemp, pbutemp, fbytemp
REAL(8) :: ghkval, ghkvals		! The calculated GHK value
REAL(8) :: utilsub(3), alpha(3,3), gamma(3,3)		! 1 ... oldest child
REAL(8) :: lchol(3,3)

wgv=0.000d0
wg=0.000d0
v=0.000d0
ghkval=0.000d0

k=famindex(m)

utemp=0.000d0
fbutemp=0.000d0 ; pbutemp=0.000d0 ; fbytemp=0.000d0

lchol=lcholbase(m,param,n)
wgv=GHK3(m,param,n,ns,draws3,lchol,nthetab, nthetaa, nthetag, xb(:,famindex(m)-2:famindex(m)), xa(:,famindex(m)-2:famindex(m)), xg(:,famindex(m)-2:famindex(m)), sa, sg, near(m),2)

utilsub(1:3)=MATMUL(param(1:nthetab),xb(:,k-2:k))
	IF (nthetaa>=3) THEN
	alpha(1:3,1) = EXP( MATMUL(param(nthetab+1:nthetab+nthetaa-2),xa(:,k-2:k)) )
	alpha(1:3,2) = EXP( MATMUL(param(nthetab+1:nthetab+nthetaa-2),xa(:,k-2:k)) + param(nthetab+nthetaa-1) )
	alpha(1:3,3) = EXP( MATMUL(param(nthetab+1:nthetab+nthetaa-2),xa(:,k-2:k)) + param(nthetab+nthetaa-1) + param(nthetab+nthetaa) )
	ELSEIF (nthetaa==1) THEN
	alpha(1:3,1:3)=param(nthetab+nthetaa)
	ELSE
	alpha=0.0d0
	ENDIF
	IF (nthetag>=2) THEN
	gamma(1:3,1) =MATMUL(param(nthetab+nthetaa+1:nthetab+nthetaa+nthetag-1),xg(:,k-2:k))
	gamma(1:3,2) =MATMUL(param(nthetab+nthetaa+1:nthetab+nthetaa+nthetag-1),xg(:,k-2:k))*(1.0d0 + param(nthetab+nthetaa+nthetag))
	gamma(1:3,3) =MATMUL(param(nthetab+nthetaa+1:nthetab+nthetaa+nthetag-1),xg(:,k-2:k))*(1.0d0 + param(nthetab+nthetaa+nthetag)*2.0d0)
	ELSEIF (nthetag==1) THEN
	gamma(1:3,1:3)=param(nthetab+nthetaa+nthetag)
	ELSE
	gamma=0.0d0
	ENDIF


alpha(:,:)=alpha(:,:)*amp1
gamma(:,:)=gamma(:,:)*amp2

DO s1=0,2**2-1       ! Strategy loop
DO s2=0,2**3-1

IF (ghkcheck(s1,2,m)==1 .and. ghkcheck(s2,3,m)==1 .and. wgv(1,s1*8+s2,0)>0.0d0) THEN
    CALL geval(gtemp1a(:,:,:),gtemp1b(:,:,:),gtemp1c(:,:,:,:,:),nearcftemp(:),utemp,pbutemp,fbutemp,fbytemp,dstemp,m,wgv(:,s1*8+s2,1:3),alpha,gamma,lchol)
ENDIF


ghkvals=SUM(wgv(:,s1*8+s2,0))/ns
ghkval=ghkval+ghkvals
exputil(k-2:k)=exputil(k-2:k)+utemp(1:3)*ghkvals
dst(k-2:k)=dst(k-2:k)+dstemp(1:3)*ghkvals
fbfamutil(m)=fbfamutil(m)+fbutemp*ghkvals
pbfamutil(m)=pbfamutil(m)+pbutemp*ghkvals
y_any_fb(m) = y_any_fb(m)+fbytemp*ghkvals
gtype1a(m,:,:,:)=gtype1a(m,:,:,:)+gtemp1a(:,:,:)*ghkvals
gtype1b(m,:,:,:)=gtype1b(m,:,:,:)+gtemp1b(:,:,:)*ghkvals
gtype1c(m,:,:,:,:,:)=gtype1c(m,:,:,:,:,:)+gtemp1c(:,:,:,:,:)*ghkvals
near_cf(m,:)=near_cf(m,:)+nearcftemp(:)*ghkvals
END DO
END DO         ! Strategy loop
exputil(k-2:k)=exputil(k-2:k)/ghkval
dst(k-2:k)=dst(k-2:k)/ghkval
famutil(m)=SUM(exputil(k-2:k))
fbfamutil(m)=fbfamutil(m)/ghkval
pbfamutil(m)=pbfamutil(m)/ghkval
y_any_fb(m)=y_any_fb(m)/ghkval
gtype1a(m,:,:,:)=gtype1a(m,:,:,:)/ghkval
gtype1b(m,:,:,:)=gtype1b(m,:,:,:)/ghkval
gtype1c(m,:,:,:,:,:)=gtype1c(m,:,:,:,:,:)/ghkval
near_cf(m,:)=near_cf(m,:)/ghkvals

END FUNCTION SIM3




FUNCTION SIM4(m,param,n) RESULT(ghkval)
	! This function calculates the GHK simulator for each market*year
USE Globvar
USE PROBABILITY
IMPLICIT NONE
INTEGER, INTENT(IN) :: m			! market id
INTEGER, INTENT(IN) :: n 
REAL(8), INTENT(IN) :: param(n)
INTEGER :: i, j, k, l
INTEGER :: s1, s2, s3, s4, stmp
REAL(8) :: wgv(ns,0:sm4-1,0:4)		! wgv(ns, # strategy combitnations, 1 + # child) ... wgv(ns,s,0) for wg.
REAL(8) :: wg(ns,0:sm4-1)
REAL(8) :: ta(ns), tb(ns)
REAL(8) :: v(ns,0:sm4-1,nchild(m))   ! GHK simulation draws constructed from draws()
REAL(8) :: gtemp1a(0:15,1:6,0:15)
REAL(8) :: gtemp1b(0:15,0:15,0:15)
REAL(8) :: gtemp1c(0:6,0:4,1:3,0:15,0:15)
REAL(8) :: nearcftemp(0:15)
REAL(8) :: utiltemp
REAL(8) :: utemp(nchild(m)), dstemp(nchild(m))
REAL(8) :: fbutemp, pbutemp, fbytemp
REAL(8) :: ghkval, ghkvals		! The calculated GHK value
REAL(8) :: utilsub(4), alpha(4,3), gamma(4,3)		! 1 ... oldest child
REAL(8) :: lchol(4,4)

wgv=0.000d0
wg=0.000d0
v=0.000d0
ghkval=0.000d0

k=famindex(m)

utemp=0.000d0
fbutemp=0.000d0 ; pbutemp=0.000d0 ; fbytemp=0.000d0

lchol=lcholbase(m,param,n)

wgv=GHK4(m,param,n,ns,draws4,lchol,nthetab, nthetaa, nthetag, xb(:,famindex(m)-3:famindex(m)), xa(:,famindex(m)-3:famindex(m)), xg(:,famindex(m)-3:famindex(m)), sa, sg, near(m),2)

utilsub(1:4)=MATMUL(param(1:nthetab),xb(:,k-3:k))
	IF (nthetaa>=3) THEN
	alpha(1:4,1) = EXP( MATMUL(param(nthetab+1:nthetab+nthetaa-2),xa(:,k-3:k)) )
	alpha(1:4,2) = EXP( MATMUL(param(nthetab+1:nthetab+nthetaa-2),xa(:,k-3:k)) + param(nthetab+nthetaa-1) )
	alpha(1:4,3) = EXP( MATMUL(param(nthetab+1:nthetab+nthetaa-2),xa(:,k-3:k)) + param(nthetab+nthetaa-1) + param(nthetab+nthetaa) )
	ELSEIF (nthetaa==1) THEN
	alpha(1:4,1:3)=param(nthetab+nthetaa)
	ELSE
	alpha=0.0d0
	ENDIF
	IF (nthetag>=2) THEN
	gamma(1:4,1) =MATMUL(param(nthetab+nthetaa+1:nthetab+nthetaa+nthetag-1),xg(:,k-3:k))
	gamma(1:4,2) =MATMUL(param(nthetab+nthetaa+1:nthetab+nthetaa+nthetag-1),xg(:,k-3:k))*(1.0d0 + param(nthetab+nthetaa+nthetag))
	gamma(1:4,3) =MATMUL(param(nthetab+nthetaa+1:nthetab+nthetaa+nthetag-1),xg(:,k-3:k))*(1.0d0 + param(nthetab+nthetaa+nthetag)*2.0d0)
	ELSEIF (nthetag==1) THEN
	gamma(1:4,1:3)=param(nthetab+nthetaa+nthetag)
	ELSE
	gamma=0.0d0
	ENDIF


alpha(:,:)=alpha(:,:)*amp1
gamma(:,:)=gamma(:,:)*amp2

DO s1=0,2**2-1       ! Strategy loop
DO s2=0,2**3-1
DO s3=0,2**4-1

IF (ghkcheck(s1,2,m)==1 .and. ghkcheck(s2,3,m)==1 .and. ghkcheck(s3,4,m)==1 .and. SUM(wgv(:,s1*128+s2*16+s3,0))>0.0d0) THEN
    CALL geval(gtemp1a(:,:,:),gtemp1b(:,:,:),gtemp1c(:,:,:,:,:),nearcftemp(:),utemp,pbutemp,fbutemp,fbytemp,dstemp,m,wgv(:,s1*128+s2*16+s3,1:4),alpha,gamma,lchol)
		! IN :: m, v(ns,nchild(m)), tmpalpha(nchild(m),3), tmpgamma(nchild(m),2), lchol(nchild(m),nchild(m))
ENDIF

ghkvals=SUM(wgv(:,s1*128+s2*16+s3,0))/ns
ghkval=ghkval+ghkvals
exputil(k-3:k)=exputil(k-3:k)+utemp(1:4)*ghkvals
dst(k-3:k)=dst(k-3:k)+dstemp(1:4)*ghkvals
fbfamutil(m)=fbfamutil(m)+fbutemp*ghkvals
pbfamutil(m)=pbfamutil(m)+pbutemp*ghkvals
y_any_fb(m) =y_any_fb(m) +fbytemp*ghkvals
gtype1a(m,:,:,:)=gtype1a(m,:,:,:)+gtemp1a(:,:,:)*ghkvals
gtype1b(m,:,:,:)=gtype1b(m,:,:,:)+gtemp1b(:,:,:)*ghkvals
gtype1c(m,:,:,:,:,:)=gtype1c(m,:,:,:,:,:)+gtemp1c(:,:,:,:,:)*ghkvals
near_cf(m,:)=near_cf(m,:)+nearcftemp(:)*ghkvals
END DO
END DO
END DO         ! Strategy loop

exputil(k-3:k)=exputil(k-3:k)/ghkval
dst(k-3:k)=dst(k-3:k)/ghkval
famutil(m)=SUM(exputil(k-3:k))
fbfamutil(m)=fbfamutil(m)/ghkval
pbfamutil(m)=pbfamutil(m)/ghkval
y_any_fb(m)=y_any_fb(m)/ghkval
gtype1a(m,:,:,:)=gtype1a(m,:,:,:)/ghkval
gtype1b(m,:,:,:)=gtype1b(m,:,:,:)/ghkval
gtype1c(m,:,:,:,:,:)=gtype1c(m,:,:,:,:,:)/ghkval
near_cf(m,:)=near_cf(m,:)/ghkvals

END FUNCTION SIM4



FUNCTION ghkcheck(s,n,m)
USE Globvar
IMPLICIT NONE
INTEGER, INTENT(IN) :: m     ! mkt id
INTEGER, INTENT(IN) :: n      ! child id
INTEGER, INTENT(IN) :: s      ! player n's strategy
INTEGER :: i, nup
INTEGER :: ghkcheck

ghkcheck=0
IF (n>=2) THEN
nup=nearcounti(IBITS(near(m),nchild(m)-n+1,n-1),n-1)
    IF (IBITS(s,nup,1)==1 .and. IBITS(near(m),nchild(m)-n,1)==1) THEN
        ghkcheck=1
    ELSEIF (IBITS(s,nup,1)==0 .and. IBITS(near(m),nchild(m)-n,1)==0) THEN
        ghkcheck=1
    END IF
ELSE
	PRINT *,'Error found!'
	PRINT *,'player #: ', n
	STOP
ENDIF


END FUNCTION ghkcheck



SUBROUTINE geval(gtp1a,gtp1b,gtp1c,nearcfval,uval,pbuval,fbuval,fbyval,dsval,m,v,tmpalpha,tmpgamma,lchol)        ! compute first best
USE Globvar
IMPLICIT NONE
INTEGER, INTENT(IN) :: m    	! family id
REAL(8), INTENT(IN) :: v(ns,nchild(m))	! error term. 1 ... oldest
REAL(8), INTENT(IN) :: tmpalpha(nchild(m),3)
REAL(8), INTENT(IN) :: tmpgamma(nchild(m),3)
REAL(8), INTENT(IN) :: lchol(nchild(m),nchild(m))
REAL(8), INTENT(OUT) :: gtp1a(0:15,1:6,0:15)
REAL(8), INTENT(OUT) :: gtp1b(0:15,0:15,0:15)
REAL(8), INTENT(OUT) :: gtp1c(0:6,0:4,1:3,0:15,0:15)
REAL(8), INTENT(OUT) :: nearcfval(0:15)
REAL(8), INTENT(OUT) :: uval(nchild(m))
REAL(8), INTENT(OUT) :: dsval(nchild(m))
REAL(8), INTENT(OUT) :: pbuval, fbuval, fbyval
INTEGER :: i, j, k, st, i_pareto, i_ne
INTEGER :: fbnear, pbnear
INTEGER :: g1check1, g1check2, g1check3, g1check4
INTEGER :: g1check5, g1check6, g1check7
REAL(8) :: error(ns,nchild(m)) ! error(:,1) ... oldest
INTEGER :: near_cf_s(ns)
REAL(8) :: uval_s(ns,nchild(m)), pbu_s(ns), fbu_s(ns)
REAL(8) :: uval_s_dev(ns,nchild(m))
INTEGER :: dsval_s(ns,nchild(m))
REAL(8) :: temp(nchild(m)), tempu1, tempu2
REAL(8) :: u1(0:1,0:1), u2(0:1,0:1)
INTEGER :: nearr, ne(0:1,0:1), neunique

uval=0.0d0
uval_s=0.0d0
uval_s_dev=0.0d0
dsval=0.0d0
pbuval=0.0d0
pbu_s=0.0d0
fbuval=0.0d0
fbyval=0.0d0
fbu_s=0.0d0
gtp1a=0.0d0
gtp1b=0.0d0
gtp1c=0.0d0
nearcfval=0.0d0

!!! Construct the error term !!!
DO j=1,nchild(m)    ! j=1 means oldest
    k=nchild(m)-j+1
    error(:,j)=MATMUL(v(:,nchild(m):j:-1),(lchol(k,1:k)))
END DO



DO i=1,ns
    g1check1=0 ; g1check2=0 ; g1check3=0 ; g1check4=0

    !!! Solve for SPNE if amp/=1.0
    IF (amp1==1.0d0 .and. amp1==2.0d0) THEN
        near_cf_s(i)=near(m)
    ELSE
        CALL NESOLVE(near_cf_s(i),m,error(i,:),utilxb(famindex(m)-nchild(m)+1:famindex(m)),tmpalpha,tmpgamma)
    ENDIF

    !!! Compute current (SPNE) utility values !!!
    DO j=1,nchild(m)    ! j=1 means oldest
        IF (IBITS(near_cf_s(i),nchild(m)-j,1)==1) THEN		! when near
            uval_s(i,j)=error(i,j)+utilxb(famindex(m)-nchild(m)+j)
            IF     (nearcounti(near_cf_s(i),nchild(m))==4) THEN
                uval_s(i,j)    =uval_s(i,j)+tmpalpha(j,3)
                uval_s(i,j)    =uval_s(i,j)+tmpgamma(j,3)
                uval_s_dev(i,j)=tmpalpha(j,3)
            ELSEIF (nearcounti(near_cf_s(i),nchild(m))==3) THEN
                uval_s(i,j)    =uval_s(i,j)+tmpalpha(j,3)
                uval_s(i,j)    =uval_s(i,j)+tmpgamma(j,2)
                uval_s_dev(i,j)=tmpalpha(j,2)
            ELSEIF (nearcounti(near_cf_s(i),nchild(m))==2) THEN
                uval_s(i,j)    =uval_s(i,j)+tmpalpha(j,2)
                uval_s(i,j)    =uval_s(i,j)+tmpgamma(j,1)
                uval_s_dev(i,j)=tmpalpha(j,1)
            ELSEIF (nearcounti(near_cf_s(i),nchild(m))==1) THEN
                uval_s(i,j)    =uval_s(i,j)+tmpalpha(j,1)
                uval_s_dev(i,j)=0.0d0
            ENDIF
        ELSE
            uval_s_dev(i,j)=error(i,j)+utilxb(famindex(m)-nchild(m)+j)
            IF     (nearcounti(near_cf_s(i),nchild(m))==3) THEN
                uval_s(i,j)    =tmpalpha(j,3)
                uval_s_dev(i,j)=uval_s_dev(i,j)+tmpalpha(j,3)
                uval_s_dev(i,j)=uval_s_dev(i,j)+tmpgamma(j,3)
            ELSEIF (nearcounti(near_cf_s(i),nchild(m))==2) THEN
                uval_s(i,j)    =tmpalpha(j,2)
                uval_s_dev(i,j)=uval_s_dev(i,j)+tmpalpha(j,3)
                uval_s_dev(i,j)=uval_s_dev(i,j)+tmpgamma(j,2)
            ELSEIF (nearcounti(near_cf_s(i),nchild(m))==1) THEN
                uval_s(i,j)    =tmpalpha(j,1)
                uval_s_dev(i,j)=uval_s_dev(i,j)+tmpalpha(j,2)
                uval_s_dev(i,j)=uval_s_dev(i,j)+tmpgamma(j,1)
            ELSEIF (nearcounti(near_cf_s(i),nchild(m))==0) THEN
                uval_s(i,j)    =0.0d0
                uval_s_dev(i,j)=uval_s_dev(i,j)+tmpalpha(j,1)
            ENDIF
        ENDIF
    END DO

    !!! Check whether SPNE (near) is a normal form NE.
    i_ne=1
    DO j=1,nchild(m)    ! j=1 means oldest
        IF (uval_s_dev(i,j)>uval_s(i,j)) THEN
            i_ne=0
        ENDIF
    ENDDO

    !!! Compute Pareto best !!!
    pbnear=near_cf_s(i)
    pbu_s(i)=SUM(uval_s(i,:))
    DO st=0,2**nchild(m)-1       ! Strategy loop
	    temp=0.0d0
	    i_pareto=1
	    DO j=1,nchild(m)
		    IF (IBITS(st,nchild(m)-j,1)==1) THEN	! when near
		        IF (nearcounti(st,nchild(m))==1) THEN
    		        temp(j)=error(i,j)+utilxb(famindex(m)-nchild(m)+j)+tmpalpha(j,1)
		        ELSEIF (nearcounti(st,nchild(m))==2) THEN
    		        temp(j)=error(i,j)+utilxb(famindex(m)-nchild(m)+j)+tmpalpha(j,2)+tmpgamma(j,1)
		        ELSEIF (nearcounti(st,nchild(m))==3) THEN
    		        temp(j)=error(i,j)+utilxb(famindex(m)-nchild(m)+j)+tmpalpha(j,3)+tmpgamma(j,2)
		        ELSEIF (nearcounti(st,nchild(m))==4) THEN
    		        temp(j)=error(i,j)+utilxb(famindex(m)-nchild(m)+j)+tmpalpha(j,3)+tmpgamma(j,3)
		        ENDIF
		    ELSE
		        IF (nearcounti(st,nchild(m))==0) THEN
    		        temp(j)=0.0d0
		        ELSEIF (nearcounti(st,nchild(m))==1) THEN
    		        temp(j)=tmpalpha(j,1)
		        ELSEIF (nearcounti(st,nchild(m))==2) THEN
    		        temp(j)=tmpalpha(j,2)
		        ELSEIF (nearcounti(st,nchild(m))>2) THEN
    		        temp(j)=tmpalpha(j,3)
		        ENDIF
		    ENDIF
		    IF (temp(j)<uval_s(i,j)) THEN
		        i_pareto=0
		    ENDIF
	    ENDDO
	    IF (SUM(temp(:))>pbu_s(i) .and. i_pareto==1) THEN
		    pbu_s(i)=SUM(temp(:))
		    pbnear=st
	    ENDIF
    ENDDO         ! Strategy loop
    IF (pbnear==near_cf_s(i)) THEN   !!! No Pareto improvement
        g1check5=2
        IF (i_ne==1) THEN
            g1check3=3
        ELSE
            g1check3=4
        ENDIF
    ELSE   !!! Pareto dominated
        g1check5=3
        IF (i_ne==1) THEN
            g1check3=5
        ELSE
            g1check3=6
        ENDIF
    ENDIF

    !!! Compute first best !!!
    fbnear=0
    DO st=0,2**nchild(m)-1       ! Strategy loop
	    temp=0.0d0
	    DO j=1,nchild(m)
		    IF (IBITS(st,nchild(m)-j,1)==1) THEN
		        IF (nearcounti(st,nchild(m))==1) THEN
    		        temp(j)=error(i,j)+utilxb(famindex(m)-nchild(m)+j)+tmpalpha(j,1)
		        ELSEIF (nearcounti(st,nchild(m))==2) THEN
    		        temp(j)=error(i,j)+utilxb(famindex(m)-nchild(m)+j)+tmpalpha(j,2)+tmpgamma(j,1)
		        ELSEIF (nearcounti(st,nchild(m))==3) THEN
    		        temp(j)=error(i,j)+utilxb(famindex(m)-nchild(m)+j)+tmpalpha(j,3)+tmpgamma(j,2)
		        ELSEIF (nearcounti(st,nchild(m))==4) THEN
    		        temp(j)=error(i,j)+utilxb(famindex(m)-nchild(m)+j)+tmpalpha(j,3)+tmpgamma(j,3)
		        ENDIF
		    ELSE
		        IF (nearcounti(st,nchild(m))==0) THEN
    		        temp(j)=0.0d0
		        ELSEIF (nearcounti(st,nchild(m))==1) THEN
    		        temp(j)=tmpalpha(j,1)
		        ELSEIF (nearcounti(st,nchild(m))==2) THEN
    		        temp(j)=tmpalpha(j,2)
		        ELSEIF (nearcounti(st,nchild(m))>2) THEN
    		        temp(j)=tmpalpha(j,3)
		        ENDIF
		    ENDIF
	    ENDDO
	    IF (SUM(temp(:))>fbu_s(i)) THEN
		    fbu_s(i)=SUM(temp(:))
		    fbnear=st
	    ENDIF
    ENDDO         ! Strategy loop
    IF (fbnear>0) THEN
        fbyval=fbyval+1.0d0
    ENDIF
    g1check4=fbnear
    IF (fbnear==near_cf_s(i) .and. pbnear==near_cf_s(i)) THEN
        g1check5=1
        IF (i_ne==1) THEN
            g1check3=1
        ELSE
            g1check3=2
        ENDIF
    ELSEIF (fbnear==near_cf_s(i) .and. pbnear/=near_cf_s(i)) THEN
        PRINT *,'Error: First best but Pareto dominated'
        PAUSE
    ENDIF


    !!! Check dominant strategies !!!
    DO j=1,nchild(m)
        dsval_s(i,j)=1                      !!! check if "near" is a dominant strategy
        DO st=0,2**nchild(m)-1       ! Strategy loop
            IF (IBITS(st,nchild(m)-j,1)==1) THEN
		        IF (nearcounti(st,nchild(m))==1) THEN
    		        tempu1=error(i,j)+utilxb(famindex(m)-nchild(m)+j)+tmpalpha(j,1)
    		        tempu2=0.0d0
		        ELSEIF (nearcounti(st,nchild(m))==2) THEN
    		        tempu1=error(i,j)+utilxb(famindex(m)-nchild(m)+j)+tmpalpha(j,2)+tmpgamma(j,1)
    		        tempu2=tmpalpha(j,1)
		        ELSEIF (nearcounti(st,nchild(m))==3) THEN
    		        tempu1=error(i,j)+utilxb(famindex(m)-nchild(m)+j)+tmpalpha(j,3)+tmpgamma(j,2)
    		        tempu2=tmpalpha(j,2)
		        ELSEIF (nearcounti(st,nchild(m))==4) THEN
    		        tempu1=error(i,j)+utilxb(famindex(m)-nchild(m)+j)+tmpalpha(j,3)+tmpgamma(j,3)
    		        tempu2=tmpalpha(j,3)
		        ENDIF
	            IF (tempu1<tempu2) THEN
	                dsval_s(i,j)=2
	            ENDIF
            ENDIF
        ENDDO
        IF (dsval_s(i,j)==2) THEN           !!! when near is not a dominant strategy, check if far is dominant
            dsval_s(i,j)=0
            DO st=0,2**nchild(m)-1       ! Strategy loop
                IF (IBITS(st,nchild(m)-j,1)==0) THEN
		            IF (nearcounti(st,nchild(m))==0) THEN
    		            tempu1=0.0d0
    		            tempu2=error(i,j)+utilxb(famindex(m)-nchild(m)+j)+tmpalpha(j,1)
		            ELSEIF (nearcounti(st,nchild(m))==1) THEN
    		            tempu1=tmpalpha(j,1)
    		            tempu2=error(i,j)+utilxb(famindex(m)-nchild(m)+j)+tmpalpha(j,2)+tmpgamma(j,1)
		            ELSEIF (nearcounti(st,nchild(m))==2) THEN
    		            tempu1=tmpalpha(j,2)
    		            tempu2=error(i,j)+utilxb(famindex(m)-nchild(m)+j)+tmpalpha(j,3)+tmpgamma(j,2)
		            ELSEIF (nearcounti(st,nchild(m))==3) THEN
    		            tempu1=tmpalpha(j,3)
    		            tempu2=error(i,j)+utilxb(famindex(m)-nchild(m)+j)+tmpalpha(j,3)+tmpgamma(j,3)
		            ENDIF
	                IF (tempu1<tempu2) THEN
	                    dsval_s(i,j)=2
	                ENDIF
                ENDIF
            ENDDO
        ENDIF

    ENDDO

    !!! Check dominant strategies
    g1check1=0 ;      g1check2=0
    DO j=1,nchild(m)
        IF (dsval_s(i,j)/=2) THEN
            g1check1=g1check1+2**(nchild(m)-j)
        ENDIF
        IF (dsval_s(i,j)==IBITS(near_cf_s(i),nchild(m)-j,1)) THEN
            g1check2=g1check2+2**(nchild(m)-j)
        ENDIF
    ENDDO

    !!! Count the number of children with increasing difference
    g1check6=0
    DO j=1,nchild(m)
        u1(0,0)=0.0d0
        u1(0,1)=tmpalpha(j,1)
        u1(1,0)=error(i,j)+utilxb(famindex(m)-nchild(m)+j)+tmpalpha(j,1)
        u1(1,1)=error(i,j)+utilxb(famindex(m)-nchild(m)+j)+tmpalpha(j,2)+tmpgamma(j,1)
        IF (u1(1,1)-u1(0,1) > u1(1,0)-u1(0,0) ) g1check6=g1check6+1
    ENDDO

    !!! Count the number of normal form NE
	g1check7=0
    DO st=0,2**nchild(m)-1       ! Strategy loop
	i_ne=1
    DO j=1,nchild(m)    ! j=1 means oldest
        IF (IBITS(st,nchild(m)-j,1)==1) THEN	! when near
            tempu1=error(i,j)+utilxb(famindex(m)-nchild(m)+j)
            IF     (nearcounti(st,nchild(m))==4) THEN
                tempu1    =tempu1+tmpalpha(j,3)
                tempu1    =tempu1+tmpgamma(j,3)
                tempu2=tmpalpha(j,3)
            ELSEIF (nearcounti(st,nchild(m))==3) THEN
                tempu1    =tempu1+tmpalpha(j,3)
                tempu1    =tempu1+tmpgamma(j,2)
                tempu2=tmpalpha(j,2)
            ELSEIF (nearcounti(st,nchild(m))==2) THEN
                tempu1    =tempu1+tmpalpha(j,2)
                tempu1    =tempu1+tmpgamma(j,1)
                tempu2=tmpalpha(j,1)
            ELSEIF (nearcounti(st,nchild(m))==1) THEN
                tempu1    =tempu1+tmpalpha(j,1)
                tempu2=0.0d0
            ENDIF
        ELSE
            tempu2=error(i,j)+utilxb(famindex(m)-nchild(m)+j)
            IF     (nearcounti(st,nchild(m))==3) THEN
                tempu1    =tmpalpha(j,2)
                tempu2=tempu2+tmpalpha(j,3)
                tempu2=tempu2+tmpgamma(j,3)
            ELSEIF (nearcounti(st,nchild(m))==2) THEN
                tempu1    =tmpalpha(j,2)
                tempu2=tempu2+tmpalpha(j,3)
                tempu2=tempu2+tmpgamma(j,2)
            ELSEIF (nearcounti(st,nchild(m))==1) THEN
                tempu1    =tmpalpha(j,1)
                tempu2=tempu2+tmpalpha(j,2)
                tempu2=tempu2+tmpgamma(j,1)
            ELSEIF (nearcounti(st,nchild(m))==0) THEN
                tempu1    =0.0d0
                tempu2=tempu2+tmpalpha(j,1)
            ENDIF
        ENDIF
	    IF (tempu1<tempu2) THEN
	        i_ne=0
	    ENDIF
    END DO		! Siblings loop
	    IF (i_ne==1) 	g1check7=g1check7+1
    END DO		! Strategy loop



    gtp1a(g1check4,g1check3,near_cf_s(i))=gtp1a(g1check4,g1check3,near_cf_s(i))+1.0d0
    gtp1b(g1check2,g1check1,near_cf_s(i))=gtp1b(g1check2,g1check1,near_cf_s(i))+1.0d0
    gtp1c(g1check7,g1check6,g1check5,g1check4,near_cf_s(i))=gtp1c(g1check7,g1check6,g1check5,g1check4,near_cf_s(i))+1.0d0
    nearcfval(near_cf_s(i))=nearcfval(near_cf_s(i))+1.0d0

    !!! re-code dsval_s
    DO j=1,nchild(m)
        IF (dsval_s(i,j)==IBITS(near_cf_s(i),nchild(m)-j,1)) THEN
            dsval_s(i,j)=1
        ELSE
            dsval_s(i,j)=0
        ENDIF
    ENDDO


ENDDO          ! Sim draw loop
uval=SUM(uval_s(:,:),1)/ns

dsval=SUM(dsval_s(:,:),1)*1.0d0/ns
pbuval=SUM(pbu_s(:))/ns
fbuval=SUM(fbu_s(:))/ns
fbyval=fbyval/ns
gtp1a=gtp1a/ns
gtp1b=gtp1b/ns
gtp1c=gtp1c/ns
nearcfval=nearcfval/ns

END SUBROUTINE geval





END MODULE Object











PROGRAM MAIN
USE Globvar
USE INIT
USE Object
IMPLICIT NONE

REAL(8) :: resultmlv, resultmlse
REAL(8) :: objsaved
REAL(4) :: time0, iniadj
REAL(8) :: gtable0(0:15,1:5), gtable1(0:15,1:5), gtable2(1:6,1:5), gtable3(0:15,1:5), gtable4(0:15,1:5), gtable5(0:15,1:5)      !! temporary variables to make tables
REAL(8) :: gtable6a(0:3,1:4), gtable6b(0:7,1:4), gtable6c(0:15,1:4)
REAL(8) :: gtable7a(0:3,1:4), gtable7b(0:7,1:4), gtable7c(0:15,1:4)
REAL(8) :: gtable8a(0:3,1:4), gtable8b(0:7,1:4), gtable8c(0:15,1:4)
REAL(8) :: gtable9a(0:2,1:4), gtable9b(0:3,1:5), gtable9c(0:4,1:6)
REAL(8) :: gtable10a(0:2,1:4), gtable10b(0:3,1:5), gtable10c(0:4,1:6)
REAL(8) :: gtable11a(0:2,1:4), gtable11b(0:3,1:5), gtable11c(0:4,1:6)
REAL(8) :: gtable12(0:4,1:5), gtable12a(0:4,1:5), gtable13(0:6,1:5), gtable13a(0:6,1:5)      !! temporary variables to make tables
REAL(8) :: gtable14a(0:2,1:4), gtable14b(0:3,1:5), gtable14c(0:4,1:6)
REAL(8) :: gtable15a(0:3,1:4), gtable15b(0:6,1:5), gtable15c(0:6,1:6)
REAL(8) :: gtable16a(0:2,1:4), gtable16b(0:3,1:5), gtable16c(0:4,1:6)
REAL(8) :: gtable17a(0:3,1:4), gtable17b(0:6,1:5), gtable17c(0:6,1:6)
REAL(8) :: gtable18a(0:3,1:4), gtable18b(0:6,1:5), gtable18c(0:6,1:6)
REAL(8) :: gtable19a(0:3,1:4), gtable19b(0:7,1:5), gtable19c(0:15,1:6)
REAL(8) :: gtable20a(1:3,1:4), gtable20b(1:3,1:5), gtable20c(1:3,1:6)
REAL(8) :: gtable21a(0:3,1:4), gtable21b(0:7,1:5), gtable21c(0:15,1:6)
REAL(8) :: gtable22a(0:2,1:4), gtable22b(0:3,1:5), gtable22c(0:4,1:6)
REAL(8) :: gtable23a(1:3,1:4), gtable23b(1:3,1:5), gtable23c(1:3,1:6)
REAL(8) :: gtable24a(0:3,1:4), gtable24b(0:6,1:5), gtable24c(0:6,1:6)

INTEGER :: i, j, k, l, m, n, io, nnear
INTEGER :: i0, i1, i2, i3, i4, i5, i6, i7
INTEGER :: gtct_1c, gtct_2c, gtct_3c, gtct_4c
REAL(8) :: gtct_2ce1, gtct_3ce1, gtct_4ce1
REAL(8) :: gtct_2ce2, gtct_3ce2, gtct_4ce2
REAL(8) :: gtct_2ce3, gtct_3ce3, gtct_4ce3
REAL(8) :: gtct_2cfb0, gtct_2cfb1, gtct_2cfb2         ! # of child with Near in FB in 2 Child Families
REAL(8) :: gtct_3cfb0, gtct_3cfb1, gtct_3cfb2, gtct_3cfb3         ! # of child with Near in FB in 3 Child Families
REAL(8) :: gtct_4cfb0, gtct_4cfb1, gtct_4cfb2, gtct_4cfb3, gtct_4cfb4         ! # of child with Near in FB in 4 Child Families
REAL(8) :: gtct_2cid0, gtct_2cid1, gtct_2cid2         ! # of child with Inc. Diff. in 2 Child Families
REAL(8) :: gtct_3cid0, gtct_3cid1, gtct_3cid2, gtct_3cid3         ! # of child with Inc. Diff. in 3 Child Families
REAL(8) :: gtct_4cid0, gtct_4cid1, gtct_4cid2, gtct_4cid3, gtct_4cid4         ! # of child with Inc. Diff. in 4 Child Families


CHARACTER(12) :: horz(3)
!CHARACTER(60) :: flname(10)        ! File names to be written
CHARACTER(20) :: flname
CHARACTER(80) :: flname2
CHARACTER(20) :: fmtname           ! To accomodate different number of parameters
CHARACTER(50) :: inivalnote
CHARACTER(60) :: setupnote

CHARACTER(16), ALLOCATABLE :: vert(:)

horz(1)='        coef' ; horz(2)='        s.e.' ; horz(3)=' asy t-val'

! Read the setting file
    flname2='setting.prn'
    OPEN(1,file=(flname2))
    READ(1,fmt='(A60)') setupnote
    READ(1,fmt=*) nthetab
    READ(1,fmt=*) nthetaa
    READ(1,fmt=*) nthetag
    READ(1,fmt=*) nthetar
    READ(1,fmt=*) ntheta
    READ(1,fmt=*) nobs
    READ(1,fmt=*) nfam
    READ(1,fmt=*) ns
ALLOCATE (vert(ntheta))
    vert(1)='cons      '
    READ(1,fmt=*) vert(2:ntheta)
!    READ(1,fmt=*) path
    READ(1,fmt=*) amp1
    READ(1,fmt=*) amp2
    CLOSE(1)

ALLOCATE (theta(ntheta))
ALLOCATE (famid(nobs))
ALLOCATE (famindex(nfam))
ALLOCATE (nchild(nfam))
ALLOCATE (y(nobs))
ALLOCATE (xb(nthetab,nobs))
IF (nthetaa>=3) THEN
	sa=nthetaa-2
	ALLOCATE (xa(nthetaa-2,nobs))
ELSEIF (nthetaa==1) THEN
	sa=1
	ALLOCATE (xa(1,nobs))
ENDIF
IF (nthetag>=2) THEN
	sg=nthetag-1
	ALLOCATE (xg(nthetag-1,nobs))
ELSEIF (nthetag==1) THEN
	sg=1
	ALLOCATE (xg(1,nobs))
ENDIF
ALLOCATE (xr(nthetar,nchildmax,nobs))
ALLOCATE (near(nfam))
ALLOCATE (near_cf(nfam,0:15))
ALLOCATE (utilxb(nobs))
ALLOCATE (exputil(nobs))
ALLOCATE (dst(nobs))
ALLOCATE (famutil(nfam))
ALLOCATE (fbfamutil(nfam))
ALLOCATE (pbfamutil(nfam))
ALLOCATE (y_any_fb(nfam))
ALLOCATE (gtype1a(nfam,0:15,1:6,0:15))
ALLOCATE (gtype1b(nfam,0:15,0:15,0:15))
ALLOCATE (gtype1c(nfam,0:6,0:4,1:3,0:15,0:15))

OPEN(1,file='param1.out')
    READ(1,fmt=*) theta(1:nthetab)
CLOSE(1)
IF (ntheta>nthetab) THEN
OPEN(1,file='param2.out')
    READ(1,fmt=*) theta(nthetab+1:ntheta)
CLOSE(1)
ENDIF


time0 = SECNDS(0.0)

CALL GETTIM(i, j, k, l)
CALL GETDAT(l, m, n)
flname2=CHAR(ICHAR('0')+(m/10))//CHAR(ICHAR('0')+(m-m/10*10))//CHAR(ICHAR('0')+(n/10))//CHAR(ICHAR('0')+(n-n/10*10))//CHAR(ICHAR('0')+(i/10))//CHAR(ICHAR('0')+(i-i/10*10))//CHAR(ICHAR('0')+(j/10))//CHAR(ICHAR('0')+(j-j/10*10))//CHAR(ICHAR('0')+(k/10))//CHAR(ICHAR('0')+(k-k/10*10))//'.prn'

OPEN(4,file=(flname2),STATUS='REPLACE')
WRITE(4,fmt='(A20,4x,A60)') 'Estimation Setting: ', setupnote
WRITE(4,fmt=*) '# of families     :  ', nfam
WRITE(4,fmt=*) '# of observations:  ', nobs
WRITE(4,fmt=*) '# of simuln draws  :  ', ns
WRITE(4,fmt=*) ' '
WRITE(4,fmt=*) 'Counterfactual simulation. Alpha multiplier  :  ', amp1
WRITE(4,fmt=*) 'Counterfactual simulation. Gamma multiplier  :  ', amp2
WRITE(4,fmt=*) ' '
WRITE(4,fmt=*) 'Output: Tables for # children 1-4  '
WRITE(4,fmt=*) ' '


! Load the data set
CALL READDATA(nobs)

ALLOCATE (draws2seed(ns,0:sm2-1,2))
ALLOCATE (draws3seed(ns,0:sm3-1,3))
ALLOCATE (draws4seed(ns,0:sm4-1,4))
ALLOCATE (draws2(ns,0:sm2-1,2))
ALLOCATE (draws3(ns,0:sm3-1,3))
ALLOCATE (draws4(ns,0:sm4-1,4))
CALL RANDOM_SEED(PUT=(/3000/))
CALL RANDOM_NUMBER (draws2seed)
CALL RANDOM_NUMBER (draws3seed)
CALL RANDOM_NUMBER (draws4seed)
draws2seed(ns/2+1:ns,:,:)=1-draws2seed(1:ns/2,:,:)        ! This line makes antithetics
draws3seed(ns/2+1:ns,:,:)=1-draws3seed(1:ns/2,:,:)        ! This line makes antithetics
draws4seed(ns/2+1:ns,:,:)=1-draws4seed(1:ns/2,:,:)        ! This line makes antithetics


utilxb(1:nobs)=MATMUL(theta(1:nthetab),xb(:,1:nobs))


objsaved=SIMULATION(theta,ntheta)

WRITE(4,fmt=*) ' '
WRITE(4,fmt=*) 'ML objective:  ', objsaved
WRITE(4,fmt=*) ' '
WRITE(4,fmt=*) 'run time(seconds):  ', SECNDS(time0)
WRITE(4,fmt=*) ' '


print *,'Making Tables ... '

!!! Make counts for game type analysis

gtct_1c=0 ; gtct_2c=0 ; gtct_3c=0 ; gtct_4c=0               ! # of Families with x Children
gtct_2ce1=0.0d0 ; gtct_3ce1=0.0d0 ; gtct_4ce1=0.0d0         ! # of x child Families with KH best outcomes
gtct_2ce2=0.0d0 ; gtct_3ce2=0.0d0 ; gtct_4ce2=0.0d0         ! # of x child Families with P.O. outcomes
gtct_2ce3=0.0d0 ; gtct_3ce3=0.0d0 ; gtct_4ce3=0.0d0         ! # of x child Families with P.D. outcomes

gtct_2cid0=0.0d0 ; gtct_2cid1=0.0d0 ; gtct_2cid2=0.0d0         ! # of child with Inc. Diff. in 2 Child Families
gtct_3cid0=0.0d0 ; gtct_3cid1=0.0d0 ; gtct_3cid2=0.0d0 ; gtct_3cid3=0.0d0         ! # of child with Inc. Diff. in 3 Child Families
gtct_4cid0=0.0d0 ; gtct_4cid1=0.0d0 ; gtct_4cid2=0.0d0 ; gtct_4cid3=0.0d0 ; gtct_4cid4=0.0d0         ! # of child with Inc. Diff. in 4 Child Families

gtct_2cfb0=0.0d0 ; gtct_2cfb1=0.0d0 ; gtct_2cfb2=0.0d0         ! # of child with Near in FB in 2 Child Families
gtct_3cfb0=0.0d0 ; gtct_3cfb1=0.0d0 ; gtct_3cfb2=0.0d0 ; gtct_3cfb3=0.0d0         ! # of child with Near in FB in 3 Child Families
gtct_4cfb0=0.0d0 ; gtct_4cfb1=0.0d0 ; gtct_4cfb2=0.0d0 ; gtct_4cfb3=0.0d0 ; gtct_4cfb4=0.0d0         ! # of child with Near in FB in 4 Child Families



DO m=1,nfam
    SELECT CASE(nchild(m))
    CASE(1)
        gtct_1c=gtct_1c+1
    CASE(2)
        gtct_2c=gtct_2c+1
        DO i0=0, 3
        DO i3=1, 6
        DO i4=0, 3
            IF     (i3==1 .or. i3==2) THEN
                gtct_2ce1=gtct_2ce1+gtype1a(m,i4,i3,i0)
            ELSEIF (i3==3 .or. i3==4) THEN
                gtct_2ce2=gtct_2ce2+gtype1a(m,i4,i3,i0)
            ELSEIF (i3==5 .or. i3==6) THEN
                gtct_2ce3=gtct_2ce3+gtype1a(m,i4,i3,i0)
            ENDIF
            IF     (i4==0) THEN
                gtct_2cfb0=gtct_2cfb0+gtype1a(m,i4,i3,i0)
            ELSEIF (i4==1 .or. i4==2) THEN
                gtct_2cfb1=gtct_2cfb1+gtype1a(m,i4,i3,i0)
            ELSEIF (i4==3) THEN
                gtct_2cfb2=gtct_2cfb2+gtype1a(m,i4,i3,i0)
            ENDIF
        ENDDO
        ENDDO
        ENDDO
        DO i0=0, 3
        DO i4=0, 3
        DO i5=1, 3
        DO i6=0, 2
        DO i7=0, 3
            IF     (i6==0) THEN
                gtct_2cid0=gtct_2cid0+gtype1c(m,i7,i6,i5,i4,i0)
            ELSEIF (i6==1) THEN
                gtct_2cid1=gtct_2cid1+gtype1c(m,i7,i6,i5,i4,i0)
            ELSEIF (i6==2) THEN
                gtct_2cid2=gtct_2cid2+gtype1c(m,i7,i6,i5,i4,i0)
            ENDIF
        ENDDO
        ENDDO
        ENDDO
        ENDDO
        ENDDO
    CASE(3)
        gtct_3c=gtct_3c+1
        DO i0=0, 7
        DO i3=1, 6
        DO i4=0, 7
            IF     (i3==1 .or. i3==2) THEN
                gtct_3ce1=gtct_3ce1+gtype1a(m,i4,i3,i0)
            ELSEIF (i3==3 .or. i3==4) THEN
                gtct_3ce2=gtct_3ce2+gtype1a(m,i4,i3,i0)
            ELSEIF (i3==5 .or. i3==6) THEN
                gtct_3ce3=gtct_3ce3+gtype1a(m,i4,i3,i0)
            ENDIF
            IF     (i4==0) THEN
                gtct_3cfb0=gtct_3cfb0+gtype1a(m,i4,i3,i0)
            ELSEIF (i4==1 .or. i4==2 .or. i4==4) THEN
                gtct_3cfb1=gtct_3cfb1+gtype1a(m,i4,i3,i0)
            ELSEIF (i4==3 .or. i4==5 .or. i4==6) THEN
                gtct_3cfb2=gtct_3cfb2+gtype1a(m,i4,i3,i0)
            ELSEIF (i4==7) THEN
                gtct_3cfb3=gtct_3cfb3+gtype1a(m,i4,i3,i0)
            ENDIF
        ENDDO
        ENDDO
        ENDDO
        DO i0=0, 7
        DO i4=0, 7
        DO i5=1, 3
        DO i6=0, 3
        DO i7=0, 6
            IF     (i6==0) THEN
                gtct_3cid0=gtct_3cid0+gtype1c(m,i7,i6,i5,i4,i0)
            ELSEIF (i6==1) THEN
                gtct_3cid1=gtct_3cid1+gtype1c(m,i7,i6,i5,i4,i0)
            ELSEIF (i6==2) THEN
                gtct_3cid2=gtct_3cid2+gtype1c(m,i7,i6,i5,i4,i0)
            ELSEIF (i6==3) THEN
                gtct_3cid3=gtct_3cid3+gtype1c(m,i7,i6,i5,i4,i0)
            ENDIF
        ENDDO
        ENDDO
        ENDDO
        ENDDO
        ENDDO
    CASE(4)
        gtct_4c=gtct_4c+1
        DO i0=0, 15
        DO i3=1, 6
        DO i4=0, 15
            IF     (i3==1 .or. i3==2) THEN
                gtct_4ce1=gtct_4ce1+gtype1a(m,i4,i3,i0)
            ELSEIF (i3==3 .or. i3==4) THEN
                gtct_4ce2=gtct_4ce2+gtype1a(m,i4,i3,i0)
            ELSEIF (i3==5 .or. i3==6) THEN
                gtct_4ce3=gtct_4ce3+gtype1a(m,i4,i3,i0)
            ENDIF
            IF     (i4==0) THEN
                gtct_4cfb0=gtct_4cfb0+gtype1a(m,i4,i3,i0)
            ELSEIF (i4==1 .or. i4==2 .or. i4==4 .or. i4==8) THEN
                gtct_4cfb1=gtct_4cfb1+gtype1a(m,i4,i3,i0)
            ELSEIF (i4==3 .or. i4==5 .or. i4==6 .or. i4==9 .or. i4==10 .or. i4==12) THEN
                gtct_4cfb2=gtct_4cfb2+gtype1a(m,i4,i3,i0)
            ELSEIF (i4==7 .or. i4==11 .or. i4==13 .or. i4==14) THEN
                gtct_4cfb3=gtct_4cfb3+gtype1a(m,i4,i3,i0)
            ELSEIF (i4==15) THEN
                gtct_4cfb4=gtct_4cfb4+gtype1a(m,i4,i3,i0)
            ENDIF
        ENDDO
        ENDDO
        ENDDO
        DO i0=0, 15
        DO i4=0, 15
        DO i5=1, 3
        DO i6=0, 4
        DO i7=0, 6
            IF     (i6==0) THEN
                gtct_4cid0=gtct_4cid0+gtype1c(m,i7,i6,i5,i4,i0)
            ELSEIF (i6==1) THEN
                gtct_4cid1=gtct_4cid1+gtype1c(m,i7,i6,i5,i4,i0)
            ELSEIF (i6==2) THEN
                gtct_4cid2=gtct_4cid2+gtype1c(m,i7,i6,i5,i4,i0)
            ELSEIF (i6==3) THEN
                gtct_4cid3=gtct_4cid3+gtype1c(m,i7,i6,i5,i4,i0)
            ELSEIF (i6==4) THEN
                gtct_4cid4=gtct_4cid4+gtype1c(m,i7,i6,i5,i4,i0)
            ENDIF
        ENDDO
        ENDDO
        ENDDO
        ENDDO
        ENDDO
    CASE DEFAULT
        PAUSE
    END SELECT
ENDDO





!!! Game type analysis (nchild 0-4)

gtable0=0.0d0 ; gtable1=0.0d0
gtable2=0.0d0 ; gtable3=0.0d0 ; gtable4=0.0d0 ; gtable5=0.0d0

DO m=1,nfam
    DO i0=0, 15
    DO i3=1, 6
    DO i4=0, 15
        gtable0(i0,1)=gtable0(i0,1)+gtype1a(m,i4,i3,i0)
        gtable2(i3,1)=gtable2(i3,1)+gtype1a(m,i4,i3,i0)
        gtable5(i4,1)=gtable5(i4,1)+gtype1a(m,i4,i3,i0)
    ENDDO
    ENDDO
    ENDDO

    DO i0=0, 15
    DO i1=0, 15
    DO i2=0, 15
        gtable3(i1,1)=gtable3(i1,1)+gtype1b(m,i2,i1,i0)
        gtable4(i2,1)=gtable4(i2,1)+gtype1b(m,i2,i1,i0)
    ENDDO
    ENDDO
    ENDDO

    SELECT CASE(nchild(m))
    CASE(1)
        DO i0=0, 15
        DO i3=1, 6
        DO i4=0, 15
            gtable0(i0,2)=gtable0(i0,2)+gtype1a(m,i4,i3,i0)
            gtable2(i3,2)=gtable2(i3,2)+gtype1a(m,i4,i3,i0)
            gtable5(i4,2)=gtable5(i4,2)+gtype1a(m,i4,i3,i0)
        ENDDO
        ENDDO
        ENDDO
        DO i0=0, 15
        DO i1=0, 15
        DO i2=0, 15
            gtable3(i1,2)=gtable3(i1,2)+gtype1b(m,i2,i1,i0)
            gtable4(i2,2)=gtable4(i2,2)+gtype1b(m,i2,i1,i0)
        ENDDO
        ENDDO
        ENDDO
    CASE(2)
        DO i0=0, 15
        DO i3=1, 6
        DO i4=0, 15
            gtable0(i0,3)=gtable0(i0,3)+gtype1a(m,i4,i3,i0)
            gtable2(i3,3)=gtable2(i3,3)+gtype1a(m,i4,i3,i0)
            gtable5(i4,3)=gtable5(i4,3)+gtype1a(m,i4,i3,i0)
        ENDDO
        ENDDO
        ENDDO
        DO i0=0, 15
        DO i1=0, 15
        DO i2=0, 15
            gtable3(i1,3)=gtable3(i1,3)+gtype1b(m,i2,i1,i0)
            gtable4(i2,3)=gtable4(i2,3)+gtype1b(m,i2,i1,i0)
        ENDDO
        ENDDO
        ENDDO
    CASE(3)
        DO i0=0, 15
        DO i3=1, 6
        DO i4=0, 15
            gtable0(i0,4)=gtable0(i0,4)+gtype1a(m,i4,i3,i0)
            gtable2(i3,4)=gtable2(i3,4)+gtype1a(m,i4,i3,i0)
            gtable5(i4,4)=gtable5(i4,4)+gtype1a(m,i4,i3,i0)
        ENDDO
        ENDDO
        ENDDO
        DO i0=0, 15
        DO i1=0, 15
        DO i2=0, 15
            gtable3(i1,4)=gtable3(i1,4)+gtype1b(m,i2,i1,i0)
            gtable4(i2,4)=gtable4(i2,4)+gtype1b(m,i2,i1,i0)
        ENDDO
        ENDDO
        ENDDO
    CASE(4)
        DO i0=0, 15
        DO i3=1, 6
        DO i4=0, 15
            gtable0(i0,5)=gtable0(i0,5)+gtype1a(m,i4,i3,i0)
            gtable2(i3,5)=gtable2(i3,5)+gtype1a(m,i4,i3,i0)
            gtable5(i4,5)=gtable5(i4,5)+gtype1a(m,i4,i3,i0)
        ENDDO
        ENDDO
        ENDDO
        DO i0=0, 15
        DO i1=0, 15
        DO i2=0, 15
            gtable3(i1,5)=gtable3(i1,5)+gtype1b(m,i2,i1,i0)
            gtable4(i2,5)=gtable4(i2,5)+gtype1b(m,i2,i1,i0)
        ENDDO
        ENDDO
        ENDDO
    CASE DEFAULT
        PAUSE
    END SELECT

ENDDO
    gtable1(:,1) =gtable0(:,1)/nfam ; gtable1(:,2)=gtable0(:,2)/gtct_1c ; gtable1(:,3)=gtable0(:,3)/gtct_2c ; gtable1(:,4)=gtable0(:,4)/gtct_3c ; gtable1(:,5)=gtable0(:,5)/gtct_4c
    gtable0(:,:) =gtable0(:,:)/nfam
    gtable2(:,1) =gtable2(:,1)/nfam ; gtable2(:,2)=gtable2(:,2)/gtct_1c ; gtable2(:,3)=gtable2(:,3)/gtct_2c ; gtable2(:,4)=gtable2(:,4)/gtct_3c ; gtable2(:,5)=gtable2(:,5)/gtct_4c
    gtable3(:,1) =gtable3(:,1)/nfam ; gtable3(:,2)=gtable3(:,2)/gtct_1c ; gtable3(:,3)=gtable3(:,3)/gtct_2c ; gtable3(:,4)=gtable3(:,4)/gtct_3c ; gtable3(:,5)=gtable3(:,5)/gtct_4c
    gtable4(:,1) =gtable4(:,1)/nfam ; gtable4(:,2)=gtable4(:,2)/gtct_1c ; gtable4(:,3)=gtable4(:,3)/gtct_2c ; gtable4(:,4)=gtable4(:,4)/gtct_3c ; gtable4(:,5)=gtable4(:,5)/gtct_4c
    gtable5(:,1) =gtable5(:,1)/nfam ; gtable5(:,2)=gtable5(:,2)/gtct_1c ; gtable5(:,3)=gtable5(:,3)/gtct_2c ; gtable5(:,4)=gtable5(:,4)/gtct_3c ; gtable5(:,5)=gtable5(:,5)/gtct_4c

gtable6a=0.0d0 ; gtable6b=0.0d0 ; gtable6c=0.0d0
gtable7a=0.0d0 ; gtable7b=0.0d0 ; gtable7c=0.0d0
gtable8a=0.0d0 ; gtable8b=0.0d0 ; gtable8c=0.0d0
gtable9a=0.0d0 ; gtable9b=0.0d0 ; gtable9c=0.0d0
gtable10a=0.0d0 ; gtable10b=0.0d0 ; gtable10c=0.0d0
gtable11a=0.0d0 ; gtable11b=0.0d0 ; gtable11c=0.0d0

DO m=1,nfam

    SELECT CASE(nchild(m))
    CASE(1)
    CASE(2)

	    DO i0=0, 3
	    DO i3=1, 6
	    DO i4=0, 3
		        gtable7a(i0,1)=gtable7a(i0,1)+gtype1a(m,i4,i3,i0)
		        gtable8a(i4,1)=gtable8a(i4,1)+gtype1a(m,i4,i3,i0)
	        IF     (i3==1 .or. i3==2) THEN
		        gtable7a(i0,2)=gtable7a(i0,2)+gtype1a(m,i4,i3,i0)
		        gtable8a(i4,2)=gtable8a(i4,2)+gtype1a(m,i4,i3,i0)
	        ELSEIF (i3==3 .or. i3==4) THEN
		        gtable7a(i0,3)=gtable7a(i0,3)+gtype1a(m,i4,i3,i0)
		        gtable8a(i4,3)=gtable8a(i4,3)+gtype1a(m,i4,i3,i0)
	        ELSEIF (i3==5 .or. i3==6) THEN
		        gtable7a(i0,4)=gtable7a(i0,4)+gtype1a(m,i4,i3,i0)
		        gtable8a(i4,4)=gtable8a(i4,4)+gtype1a(m,i4,i3,i0)
	        ENDIF
            IF (i4==0) THEN
                gtable11a(0,1) =gtable11a(0,1) +gtype1a(m,i4,i3,i0)
                IF (i0==0) THEN
                gtable11a(0,2) =gtable11a(0,2) +gtype1a(m,i4,i3,i0)
                ELSEIF (i0==1 .or. i0==2) THEN
                gtable11a(0,3) =gtable11a(0,3) +gtype1a(m,i4,i3,i0)
                ELSEIF (i0==3) THEN
                gtable11a(0,4) =gtable11a(0,4) +gtype1a(m,i4,i3,i0)
                ENDIF
            ELSEIF (i4==1 .or. i4==2) THEN
                gtable11a(1,1) =gtable11a(1,1) +gtype1a(m,i4,i3,i0)
                IF (i0==0) THEN
                gtable11a(1,2) =gtable11a(1,2) +gtype1a(m,i4,i3,i0)
                ELSEIF (i0==1 .or. i0==2) THEN
                gtable11a(1,3) =gtable11a(1,3) +gtype1a(m,i4,i3,i0)
                ELSEIF (i0==3) THEN
                gtable11a(1,4) =gtable11a(1,4) +gtype1a(m,i4,i3,i0)
                ENDIF
            ELSEIF (i4==3) THEN
                gtable11a(2,1) =gtable11a(2,1) +gtype1a(m,i4,i3,i0)
                IF (i0==0) THEN
                gtable11a(2,2) =gtable11a(2,2) +gtype1a(m,i4,i3,i0)
                ELSEIF (i0==1 .or. i0==2) THEN
                gtable11a(2,3) =gtable11a(2,3) +gtype1a(m,i4,i3,i0)
                ELSEIF (i0==3) THEN
                gtable11a(2,4) =gtable11a(2,4) +gtype1a(m,i4,i3,i0)
                ENDIF
            ENDIF
	    ENDDO
	    ENDDO
	    ENDDO
        DO i0=0, 3
        DO i1=0, 3
        DO i2=0, 3
            IF (i1==0) THEN
                gtable9a(0,1) =gtable9a(0,1) +gtype1b(m,i2,i1,i0)
                IF (i0==0) THEN
                gtable9a(0,2) =gtable9a(0,2) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==1 .or. i0==2) THEN
                gtable9a(0,3) =gtable9a(0,3) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==3) THEN
                gtable9a(0,4) =gtable9a(0,4) +gtype1b(m,i2,i1,i0)
                ENDIF
            ELSEIF (i1==1 .or. i1==2) THEN
                gtable9a(1,1) =gtable9a(1,1) +gtype1b(m,i2,i1,i0)
                IF (i0==0) THEN
                gtable9a(1,2) =gtable9a(1,2) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==1 .or. i0==2) THEN
                gtable9a(1,3) =gtable9a(1,3) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==3) THEN
                gtable9a(1,4) =gtable9a(1,4) +gtype1b(m,i2,i1,i0)
                ENDIF
            ELSEIF (i1==3) THEN
                gtable9a(2,1) =gtable9a(2,1) +gtype1b(m,i2,i1,i0)
                IF (i0==0) THEN
                gtable9a(2,2) =gtable9a(2,2) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==1 .or. i0==2) THEN
                gtable9a(2,3) =gtable9a(2,3) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==3) THEN
                gtable9a(2,4) =gtable9a(2,4) +gtype1b(m,i2,i1,i0)
                ENDIF
            ENDIF
            IF (i2==0) THEN
                gtable10a(0,1) =gtable10a(0,1) +gtype1b(m,i2,i1,i0)
                IF (i0==0) THEN
                gtable10a(0,2) =gtable10a(0,2) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==1 .or. i0==2) THEN
                gtable10a(0,3) =gtable10a(0,3) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==3) THEN
                gtable10a(0,4) =gtable10a(0,4) +gtype1b(m,i2,i1,i0)
                ENDIF
            ELSEIF (i2==1 .or. i2==2) THEN
                gtable10a(1,1) =gtable10a(1,1) +gtype1b(m,i2,i1,i0)
                IF (i0==0) THEN
                gtable10a(1,2) =gtable10a(1,2) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==1 .or. i0==2) THEN
                gtable10a(1,3) =gtable10a(1,3) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==3) THEN
                gtable10a(1,4) =gtable10a(1,4) +gtype1b(m,i2,i1,i0)
                ENDIF
            ELSEIF (i2==3) THEN
                gtable10a(2,1) =gtable10a(2,1) +gtype1b(m,i2,i1,i0)
                IF (i0==0) THEN
                gtable10a(2,2) =gtable10a(2,2) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==1 .or. i0==2) THEN
                gtable10a(2,3) =gtable10a(2,3) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==3) THEN
                gtable10a(2,4) =gtable10a(2,4) +gtype1b(m,i2,i1,i0)
                ENDIF
            ENDIF
        ENDDO
        ENDDO
        ENDDO

    CASE(3)

	    DO i0=0, 7
	    DO i3=1, 6
	    DO i4=0, 7
		        gtable7b(i0,1)=gtable7b(i0,1)+gtype1a(m,i4,i3,i0)
		        gtable8b(i4,1)=gtable8b(i4,1)+gtype1a(m,i4,i3,i0)
	        IF     (i3==1 .or. i3==2) THEN
		        gtable7b(i0,2)=gtable7b(i0,2)+gtype1a(m,i4,i3,i0)
		        gtable8b(i4,2)=gtable8b(i4,2)+gtype1a(m,i4,i3,i0)
	        ELSEIF (i3==3 .or. i3==4) THEN
		        gtable7b(i0,3)=gtable7b(i0,3)+gtype1a(m,i4,i3,i0)
		        gtable8b(i4,3)=gtable8b(i4,3)+gtype1a(m,i4,i3,i0)
	        ELSEIF (i3==5 .or. i3==6) THEN
		        gtable7b(i0,4)=gtable7b(i0,4)+gtype1a(m,i4,i3,i0)
		        gtable8b(i4,4)=gtable8b(i4,4)+gtype1a(m,i4,i3,i0)
	        ENDIF
            IF (i4==0) THEN
                gtable11b(0,1) =gtable11b(0,1) +gtype1a(m,i4,i3,i0)
                IF (i0==0) THEN
                gtable11b(0,2) =gtable11b(0,2) +gtype1a(m,i4,i3,i0)
                ELSEIF (i0==1 .or. i0==2 .or. i0==4) THEN
                gtable11b(0,3) =gtable11b(0,3) +gtype1a(m,i4,i3,i0)
                ELSEIF (i0==3 .or. i0==5 .or. i0==6) THEN
                gtable11b(0,4) =gtable11b(0,4) +gtype1a(m,i4,i3,i0)
                ELSEIF (i0==7) THEN
                gtable11b(0,5) =gtable11b(0,5) +gtype1a(m,i4,i3,i0)
                ENDIF
            ELSEIF (i4==1 .or. i4==2 .or. i4==4) THEN
                gtable11b(1,1) =gtable11b(1,1) +gtype1a(m,i4,i3,i0)
                IF (i0==0) THEN
                gtable11b(1,2) =gtable11b(1,2) +gtype1a(m,i4,i3,i0)
                ELSEIF (i0==1 .or. i0==2 .or. i0==4) THEN
                gtable11b(1,3) =gtable11b(1,3) +gtype1a(m,i4,i3,i0)
                ELSEIF (i0==3 .or. i0==5 .or. i0==6) THEN
                gtable11b(1,4) =gtable11b(1,4) +gtype1a(m,i4,i3,i0)
                ELSEIF (i0==7) THEN
                gtable11b(1,5) =gtable11b(1,5) +gtype1a(m,i4,i3,i0)
                ENDIF
            ELSEIF (i4==3 .or. i4==5 .or. i4==6) THEN
                gtable11b(2,1) =gtable11b(2,1) +gtype1a(m,i4,i3,i0)
                IF (i0==0) THEN
                gtable11b(2,2) =gtable11b(2,2) +gtype1a(m,i4,i3,i0)
                ELSEIF (i0==1 .or. i0==2 .or. i0==4) THEN
                gtable11b(2,3) =gtable11b(2,3) +gtype1a(m,i4,i3,i0)
                ELSEIF (i0==3 .or. i0==5 .or. i0==6) THEN
                gtable11b(2,4) =gtable11b(2,4) +gtype1a(m,i4,i3,i0)
                ELSEIF (i0==7) THEN
                gtable11b(2,5) =gtable11b(2,5) +gtype1a(m,i4,i3,i0)
                ENDIF
            ELSEIF (i4==7) THEN
                gtable11b(3,1) =gtable11b(3,1) +gtype1a(m,i4,i3,i0)
                IF (i0==0) THEN
                gtable11b(3,2) =gtable11b(3,2) +gtype1a(m,i4,i3,i0)
                ELSEIF (i0==1 .or. i0==2 .or. i0==4) THEN
                gtable11b(3,3) =gtable11b(3,3) +gtype1a(m,i4,i3,i0)
                ELSEIF (i0==3 .or. i0==5 .or. i0==6) THEN
                gtable11b(3,4) =gtable11b(3,4) +gtype1a(m,i4,i3,i0)
                ELSEIF (i0==7) THEN
                gtable11b(3,5) =gtable11b(3,5) +gtype1a(m,i4,i3,i0)
                ENDIF
            ENDIF
	    ENDDO
	    ENDDO
	    ENDDO
        DO i0=0, 7
        DO i1=0, 7
        DO i2=0, 7
            IF (i1==0) THEN
                gtable9b(0,1) =gtable9b(0,1) +gtype1b(m,i2,i1,i0)
                IF (i0==0) THEN
                gtable9b(0,2) =gtable9b(0,2) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==1 .or. i0==2 .or. i0==4) THEN
                gtable9b(0,3) =gtable9b(0,3) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==3 .or. i0==5 .or. i0==6) THEN
                gtable9b(0,4) =gtable9b(0,4) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==7) THEN
                gtable9b(0,5) =gtable9b(0,5) +gtype1b(m,i2,i1,i0)
                ENDIF
            ELSEIF (i1==1 .or. i1==2 .or. i1==4) THEN
                gtable9b(1,1) =gtable9b(1,1) +gtype1b(m,i2,i1,i0)
                IF (i0==0) THEN
                gtable9b(1,2) =gtable9b(1,2) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==1 .or. i0==2 .or. i0==4) THEN
                gtable9b(1,3) =gtable9b(1,3) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==3 .or. i0==5 .or. i0==6) THEN
                gtable9b(1,4) =gtable9b(1,4) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==7) THEN
                gtable9b(1,5) =gtable9b(1,5) +gtype1b(m,i2,i1,i0)
                ENDIF
            ELSEIF (i1==3 .or. i1==5 .or. i1==6) THEN
                gtable9b(2,1) =gtable9b(2,1) +gtype1b(m,i2,i1,i0)
                IF (i0==0) THEN
                gtable9b(2,2) =gtable9b(2,2) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==1 .or. i0==2 .or. i0==4) THEN
                gtable9b(2,3) =gtable9b(2,3) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==3 .or. i0==5 .or. i0==6) THEN
                gtable9b(2,4) =gtable9b(2,4) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==7) THEN
                gtable9b(2,5) =gtable9b(2,5) +gtype1b(m,i2,i1,i0)
                ENDIF
            ELSEIF (i1==7) THEN
                gtable9b(3,1) =gtable9b(3,1) +gtype1b(m,i2,i1,i0)
                IF (i0==0) THEN
                gtable9b(3,2) =gtable9b(3,2) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==1 .or. i0==2 .or. i0==4) THEN
                gtable9b(3,3) =gtable9b(3,3) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==3 .or. i0==5 .or. i0==6) THEN
                gtable9b(3,4) =gtable9b(3,4) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==7) THEN
                gtable9b(3,5) =gtable9b(3,5) +gtype1b(m,i2,i1,i0)
                ENDIF
            ENDIF
            IF (i2==0) THEN
                gtable10b(0,1) =gtable10b(0,1) +gtype1b(m,i2,i1,i0)
                IF (i0==0) THEN
                gtable10b(0,2) =gtable10b(0,2) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==1 .or. i0==2 .or. i0==4) THEN
                gtable10b(0,3) =gtable10b(0,3) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==3 .or. i0==5 .or. i0==6) THEN
                gtable10b(0,4) =gtable10b(0,4) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==7) THEN
                gtable10b(0,5) =gtable10b(0,5) +gtype1b(m,i2,i1,i0)
                ENDIF
            ELSEIF (i2==1 .or. i2==2 .or. i2==4) THEN
                gtable10b(1,1) =gtable10b(1,1) +gtype1b(m,i2,i1,i0)
                IF (i0==0) THEN
                gtable10b(1,2) =gtable10b(1,2) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==1 .or. i0==2 .or. i0==4) THEN
                gtable10b(1,3) =gtable10b(1,3) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==3 .or. i0==5 .or. i0==6) THEN
                gtable10b(1,4) =gtable10b(1,4) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==7) THEN
                gtable10b(1,5) =gtable10b(1,5) +gtype1b(m,i2,i1,i0)
                ENDIF
            ELSEIF (i2==3 .or. i2==5 .or. i2==6) THEN
                gtable10b(2,1) =gtable10b(2,1) +gtype1b(m,i2,i1,i0)
                IF (i0==0) THEN
                gtable10b(2,2) =gtable10b(2,2) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==1 .or. i0==2 .or. i0==4) THEN
                gtable10b(2,3) =gtable10b(2,3) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==3 .or. i0==5 .or. i0==6) THEN
                gtable10b(2,4) =gtable10b(2,4) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==7) THEN
                gtable10b(2,5) =gtable10b(2,5) +gtype1b(m,i2,i1,i0)
                ENDIF
            ELSEIF (i2==7) THEN
                gtable10b(3,1) =gtable10b(3,1) +gtype1b(m,i2,i1,i0)
                IF (i0==0) THEN
                gtable10b(3,2) =gtable10b(3,2) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==1 .or. i0==2 .or. i0==4) THEN
                gtable10b(3,3) =gtable10b(3,3) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==3 .or. i0==5 .or. i0==6) THEN
                gtable10b(3,4) =gtable10b(3,4) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==7) THEN
                gtable10b(3,5) =gtable10b(3,5) +gtype1b(m,i2,i1,i0)
                ENDIF
            ENDIF
        ENDDO
        ENDDO
        ENDDO

    CASE(4)

	    DO i0=0, 15
	    DO i3=1, 6
	    DO i4=0, 15
		        gtable7c(i0,1)=gtable7c(i0,1)+gtype1a(m,i4,i3,i0)
		        gtable8c(i4,1)=gtable8c(i4,1)+gtype1a(m,i4,i3,i0)
	        IF     (i3==1 .or. i3==2) THEN
		        gtable7c(i0,2)=gtable7c(i0,2)+gtype1a(m,i4,i3,i0)
		        gtable8c(i4,2)=gtable8c(i4,2)+gtype1a(m,i4,i3,i0)
	        ELSEIF (i3==3 .or. i3==4) THEN
		        gtable7c(i0,3)=gtable7c(i0,3)+gtype1a(m,i4,i3,i0)
		        gtable8c(i4,3)=gtable8c(i4,3)+gtype1a(m,i4,i3,i0)
	        ELSEIF (i3==5 .or. i3==6) THEN
		        gtable7c(i0,4)=gtable7c(i0,4)+gtype1a(m,i4,i3,i0)
		        gtable8c(i4,4)=gtable8c(i4,4)+gtype1a(m,i4,i3,i0)
	        ENDIF
            IF (i4==0) THEN
                gtable11c(0,1) =gtable11c(0,1) +gtype1a(m,i4,i3,i0)
                IF (i0==0) THEN
                gtable11c(0,2) =gtable11c(0,2) +gtype1a(m,i4,i3,i0)
                ELSEIF (i0==1 .or. i0==2 .or. i0==4 .or. i0==8) THEN
                gtable11c(0,3) =gtable11c(0,3) +gtype1a(m,i4,i3,i0)
                ELSEIF (i0==3 .or. i0==5 .or. i0==6 .or. i0==9 .or. i0==10 .or. i0==12) THEN
                gtable11c(0,4) =gtable11c(0,4) +gtype1a(m,i4,i3,i0)
                ELSEIF (i0==7 .or. i0==11 .or. i0==13 .or. i0==14) THEN
                gtable11c(0,5) =gtable11c(0,5) +gtype1a(m,i4,i3,i0)
                ELSEIF (i0==15) THEN
                gtable11c(0,6) =gtable11c(0,6) +gtype1a(m,i4,i3,i0)
                ENDIF
            ELSEIF (i4==1 .or. i4==2 .or. i4==4 .or. i4==8) THEN
                gtable11c(1,1) =gtable11c(1,1) +gtype1a(m,i4,i3,i0)
                IF (i0==0) THEN
                gtable11c(1,2) =gtable11c(1,2) +gtype1a(m,i4,i3,i0)
                ELSEIF (i0==1 .or. i0==2 .or. i0==4 .or. i0==8) THEN
                gtable11c(1,3) =gtable11c(1,3) +gtype1a(m,i4,i3,i0)
                ELSEIF (i0==3 .or. i0==5 .or. i0==6 .or. i0==9 .or. i0==10 .or. i0==12) THEN
                gtable11c(1,4) =gtable11c(1,4) +gtype1a(m,i4,i3,i0)
                ELSEIF (i0==7 .or. i0==11 .or. i0==13 .or. i0==14) THEN
                gtable11c(1,5) =gtable11c(1,5) +gtype1a(m,i4,i3,i0)
                ELSEIF (i0==15) THEN
                gtable11c(1,6) =gtable11c(1,6) +gtype1a(m,i4,i3,i0)
                ENDIF
            ELSEIF (i4==3 .or. i4==5 .or. i4==6 .or. i4==9 .or. i4==10 .or. i4==12) THEN
                gtable11c(2,1) =gtable11c(2,1) +gtype1a(m,i4,i3,i0)
                IF (i0==0) THEN
                gtable11c(2,2) =gtable11c(2,2) +gtype1a(m,i4,i3,i0)
                ELSEIF (i0==1 .or. i0==2 .or. i0==4 .or. i0==8) THEN
                gtable11c(2,3) =gtable11c(2,3) +gtype1a(m,i4,i3,i0)
                ELSEIF (i0==3 .or. i0==5 .or. i0==6 .or. i0==9 .or. i0==10 .or. i0==12) THEN
                gtable11c(2,4) =gtable11c(2,4) +gtype1a(m,i4,i3,i0)
                ELSEIF (i0==7 .or. i0==11 .or. i0==13 .or. i0==14) THEN
                gtable11c(2,5) =gtable11c(2,5) +gtype1a(m,i4,i3,i0)
                ELSEIF (i0==15) THEN
                gtable11c(2,6) =gtable11c(2,6) +gtype1a(m,i4,i3,i0)
                ENDIF
            ELSEIF (i4==7 .or. i4==11 .or. i4==13 .or. i4==14) THEN
                gtable11c(3,1) =gtable11c(3,1) +gtype1a(m,i4,i3,i0)
                IF (i0==0) THEN
                gtable11c(3,2) =gtable11c(3,2) +gtype1a(m,i4,i3,i0)
                ELSEIF (i0==1 .or. i0==2 .or. i0==4 .or. i0==8) THEN
                gtable11c(3,3) =gtable11c(3,3) +gtype1a(m,i4,i3,i0)
                ELSEIF (i0==3 .or. i0==5 .or. i0==6 .or. i0==9 .or. i0==10 .or. i0==12) THEN
                gtable11c(3,4) =gtable11c(3,4) +gtype1a(m,i4,i3,i0)
                ELSEIF (i0==7 .or. i0==11 .or. i0==13 .or. i0==14) THEN
                gtable11c(3,5) =gtable11c(3,5) +gtype1a(m,i4,i3,i0)
                ELSEIF (i0==15) THEN
                gtable11c(3,6) =gtable11c(3,6) +gtype1a(m,i4,i3,i0)
                ENDIF
            ELSEIF (i4==15) THEN
                gtable11c(4,1) =gtable11c(4,1) +gtype1a(m,i4,i3,i0)
                IF (i0==0) THEN
                gtable11c(4,2) =gtable11c(4,2) +gtype1a(m,i4,i3,i0)
                ELSEIF (i0==1 .or. i0==2 .or. i0==4 .or. i0==8) THEN
                gtable11c(4,3) =gtable11c(4,3) +gtype1a(m,i4,i3,i0)
                ELSEIF (i0==3 .or. i0==5 .or. i0==6 .or. i0==9 .or. i0==10 .or. i0==12) THEN
                gtable11c(4,4) =gtable11c(4,4) +gtype1a(m,i4,i3,i0)
                ELSEIF (i0==7 .or. i0==11 .or. i0==13 .or. i0==14) THEN
                gtable11c(4,5) =gtable11c(4,5) +gtype1a(m,i4,i3,i0)
                ELSEIF (i0==15) THEN
                gtable11c(4,6) =gtable11c(4,6) +gtype1a(m,i4,i3,i0)
                ENDIF
            ENDIF
	    ENDDO
	    ENDDO
	    ENDDO
        DO i0=0, 15
        DO i1=0, 15
        DO i2=0, 15
            IF (i1==0) THEN
                gtable9c(0,1) =gtable9c(0,1) +gtype1b(m,i2,i1,i0)
                IF (i0==0) THEN
                gtable9c(0,2) =gtable9c(0,2) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==1 .or. i0==2 .or. i0==4 .or. i0==8) THEN
                gtable9c(0,3) =gtable9c(0,3) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==3 .or. i0==5 .or. i0==6 .or. i0==9 .or. i0==10 .or. i0==12) THEN
                gtable9c(0,4) =gtable9c(0,4) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==7 .or. i0==11 .or. i0==13 .or. i0==14) THEN
                gtable9c(0,5) =gtable9c(0,5) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==15) THEN
                gtable9c(0,6) =gtable9c(0,6) +gtype1b(m,i2,i1,i0)
                ENDIF
            ELSEIF (i1==1 .or. i1==2 .or. i1==4 .or. i1==8) THEN
                gtable9c(1,1) =gtable9c(1,1) +gtype1b(m,i2,i1,i0)
                IF (i0==0) THEN
                gtable9c(1,2) =gtable9c(1,2) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==1 .or. i0==2 .or. i0==4 .or. i0==8) THEN
                gtable9c(1,3) =gtable9c(1,3) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==3 .or. i0==5 .or. i0==6 .or. i0==9 .or. i0==10 .or. i0==12) THEN
                gtable9c(1,4) =gtable9c(1,4) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==7 .or. i0==11 .or. i0==13 .or. i0==14) THEN
                gtable9c(1,5) =gtable9c(1,5) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==15) THEN
                gtable9c(1,6) =gtable9c(1,6) +gtype1b(m,i2,i1,i0)
                ENDIF
            ELSEIF (i1==3 .or. i1==5 .or. i1==6 .or. i1==9 .or. i1==10 .or. i1==12) THEN
                gtable9c(2,1) =gtable9c(2,1) +gtype1b(m,i2,i1,i0)
                IF (i0==0) THEN
                gtable9c(2,2) =gtable9c(2,2) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==1 .or. i0==2 .or. i0==4 .or. i0==8) THEN
                gtable9c(2,3) =gtable9c(2,3) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==3 .or. i0==5 .or. i0==6 .or. i0==9 .or. i0==10 .or. i0==12) THEN
                gtable9c(2,4) =gtable9c(2,4) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==7 .or. i0==11 .or. i0==13 .or. i0==14) THEN
                gtable9c(2,5) =gtable9c(2,5) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==15) THEN
                gtable9c(2,6) =gtable9c(2,6) +gtype1b(m,i2,i1,i0)
                ENDIF
            ELSEIF (i1==7 .or. i1==11 .or. i1==13 .or. i1==14) THEN
                gtable9c(3,1) =gtable9c(3,1) +gtype1b(m,i2,i1,i0)
                IF (i0==0) THEN
                gtable9c(3,2) =gtable9c(3,2) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==1 .or. i0==2 .or. i0==4 .or. i0==8) THEN
                gtable9c(3,3) =gtable9c(3,3) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==3 .or. i0==5 .or. i0==6 .or. i0==9 .or. i0==10 .or. i0==12) THEN
                gtable9c(3,4) =gtable9c(3,4) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==7 .or. i0==11 .or. i0==13 .or. i0==14) THEN
                gtable9c(3,5) =gtable9c(3,5) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==15) THEN
                gtable9c(3,6) =gtable9c(3,6) +gtype1b(m,i2,i1,i0)
                ENDIF
            ELSEIF (i1==15) THEN
                gtable9c(4,1) =gtable9c(4,1) +gtype1b(m,i2,i1,i0)
                IF (i0==0) THEN
                gtable9c(4,2) =gtable9c(4,2) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==1 .or. i0==2 .or. i0==4 .or. i0==8) THEN
                gtable9c(4,3) =gtable9c(4,3) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==3 .or. i0==5 .or. i0==6 .or. i0==9 .or. i0==10 .or. i0==12) THEN
                gtable9c(4,4) =gtable9c(4,4) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==7 .or. i0==11 .or. i0==13 .or. i0==14) THEN
                gtable9c(4,5) =gtable9c(4,5) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==15) THEN
                gtable9c(4,6) =gtable9c(4,6) +gtype1b(m,i2,i1,i0)
                ENDIF
            ENDIF
            IF (i2==0) THEN
                gtable10c(0,1) =gtable10c(0,1) +gtype1b(m,i2,i1,i0)
                IF (i0==0) THEN
                gtable10c(0,2) =gtable10c(0,2) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==1 .or. i0==2 .or. i0==4 .or. i0==8) THEN
                gtable10c(0,3) =gtable10c(0,3) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==3 .or. i0==5 .or. i0==6 .or. i0==9 .or. i0==10 .or. i0==12) THEN
                gtable10c(0,4) =gtable10c(0,4) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==7 .or. i0==11 .or. i0==13 .or. i0==14) THEN
                gtable10c(0,5) =gtable10c(0,5) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==15) THEN
                gtable10c(0,6) =gtable10c(0,6) +gtype1b(m,i2,i1,i0)
                ENDIF
            ELSEIF (i2==1 .or. i2==2 .or. i2==4 .or. i2==8) THEN
                gtable10c(1,1) =gtable10c(1,1) +gtype1b(m,i2,i1,i0)
                IF (i0==0) THEN
                gtable10c(1,2) =gtable10c(1,2) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==1 .or. i0==2 .or. i0==4 .or. i0==8) THEN
                gtable10c(1,3) =gtable10c(1,3) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==3 .or. i0==5 .or. i0==6 .or. i0==9 .or. i0==10 .or. i0==12) THEN
                gtable10c(1,4) =gtable10c(1,4) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==7 .or. i0==11 .or. i0==13 .or. i0==14) THEN
                gtable10c(1,5) =gtable10c(1,5) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==15) THEN
                gtable10c(1,6) =gtable10c(1,6) +gtype1b(m,i2,i1,i0)
                ENDIF
            ELSEIF (i2==3 .or. i2==5 .or. i2==6 .or. i2==9 .or. i2==10 .or. i2==12) THEN
                gtable10c(2,1) =gtable10c(2,1) +gtype1b(m,i2,i1,i0)
                IF (i0==0) THEN
                gtable10c(2,2) =gtable10c(2,2) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==1 .or. i0==2 .or. i0==4 .or. i0==8) THEN
                gtable10c(2,3) =gtable10c(2,3) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==3 .or. i0==5 .or. i0==6 .or. i0==9 .or. i0==10 .or. i0==12) THEN
                gtable10c(2,4) =gtable10c(2,4) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==7 .or. i0==11 .or. i0==13 .or. i0==14) THEN
                gtable10c(2,5) =gtable10c(2,5) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==15) THEN
                gtable10c(2,6) =gtable10c(2,6) +gtype1b(m,i2,i1,i0)
                ENDIF
            ELSEIF (i2==7 .or. i2==11 .or. i2==13 .or. i2==14) THEN
                gtable10c(3,1) =gtable10c(3,1) +gtype1b(m,i2,i1,i0)
                IF (i0==0) THEN
                gtable10c(3,2) =gtable10c(3,2) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==1 .or. i0==2 .or. i0==4 .or. i0==8) THEN
                gtable10c(3,3) =gtable10c(3,3) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==3 .or. i0==5 .or. i0==6 .or. i0==9 .or. i0==10 .or. i0==12) THEN
                gtable10c(3,4) =gtable10c(3,4) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==7 .or. i0==11 .or. i0==13 .or. i0==14) THEN
                gtable10c(3,5) =gtable10c(3,5) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==15) THEN
                gtable10c(3,6) =gtable10c(3,6) +gtype1b(m,i2,i1,i0)
                ENDIF
            ELSEIF (i2==15) THEN
                gtable10c(4,1) =gtable10c(4,1) +gtype1b(m,i2,i1,i0)
                IF (i0==0) THEN
                gtable10c(4,2) =gtable10c(4,2) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==1 .or. i0==2 .or. i0==4 .or. i0==8) THEN
                gtable10c(4,3) =gtable10c(4,3) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==3 .or. i0==5 .or. i0==6 .or. i0==9 .or. i0==10 .or. i0==12) THEN
                gtable10c(4,4) =gtable10c(4,4) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==7 .or. i0==11 .or. i0==13 .or. i0==14) THEN
                gtable10c(4,5) =gtable10c(4,5) +gtype1b(m,i2,i1,i0)
                ELSEIF (i0==15) THEN
                gtable10c(4,6) =gtable10c(4,6) +gtype1b(m,i2,i1,i0)
                ENDIF
            ENDIF
        ENDDO
        ENDDO
        ENDDO

    CASE DEFAULT
        PAUSE
    END SELECT

ENDDO

gtable7a(:,1) =gtable7a(:,1)/gtct_2c ; gtable7a(:,2)=gtable7a(:,2)/gtct_2ce1 ; gtable7a(:,3)=gtable7a(:,3)/gtct_2ce2 ; gtable7a(:,4)=gtable7a(:,4)/gtct_2ce3
gtable7b(:,1) =gtable7b(:,1)/gtct_3c ; gtable7b(:,2)=gtable7b(:,2)/gtct_3ce1 ; gtable7b(:,3)=gtable7b(:,3)/gtct_3ce2 ; gtable7b(:,4)=gtable7b(:,4)/gtct_3ce3
gtable7c(:,1) =gtable7c(:,1)/gtct_4c ; gtable7c(:,2)=gtable7c(:,2)/gtct_4ce1 ; gtable7c(:,3)=gtable7c(:,3)/gtct_4ce2 ; gtable7c(:,4)=gtable7c(:,4)/gtct_4ce3

gtable8a(:,1) =gtable8a(:,1)/gtct_2c ; gtable8a(:,2)=gtable8a(:,2)/gtct_2ce1 ; gtable8a(:,3)=gtable8a(:,3)/gtct_2ce2 ; gtable8a(:,4)=gtable8a(:,4)/gtct_2ce3
gtable8b(:,1) =gtable8b(:,1)/gtct_3c ; gtable8b(:,2)=gtable8b(:,2)/gtct_3ce1 ; gtable8b(:,3)=gtable8b(:,3)/gtct_3ce2 ; gtable8b(:,4)=gtable8b(:,4)/gtct_3ce3
gtable8c(:,1) =gtable8c(:,1)/gtct_4c ; gtable8c(:,2)=gtable8c(:,2)/gtct_4ce1 ; gtable8c(:,3)=gtable8c(:,3)/gtct_4ce2 ; gtable8c(:,4)=gtable8c(:,4)/gtct_4ce3

gtable9a(:,:) =gtable9a(:,:)/gtct_2c
gtable9b(:,:) =gtable9b(:,:)/gtct_3c
gtable9c(:,:) =gtable9c(:,:)/gtct_4c

gtable10a(:,:) =gtable10a(:,:)/gtct_2c
gtable10b(:,:) =gtable10b(:,:)/gtct_3c
gtable10c(:,:) =gtable10c(:,:)/gtct_4c

gtable11a(:,:) =gtable11a(:,:)/gtct_2c
gtable11b(:,:) =gtable11b(:,:)/gtct_3c
gtable11c(:,:) =gtable11c(:,:)/gtct_4c



gtable12=0.0d0
gtable13=0.0d0
gtable14a=0.0d0 ; gtable14b=0.0d0 ; gtable14c=0.0d0
gtable15a=0.0d0 ; gtable15b=0.0d0 ; gtable15c=0.0d0
gtable16a=0.0d0 ; gtable16b=0.0d0 ; gtable16c=0.0d0
gtable17a=0.0d0 ; gtable17b=0.0d0 ; gtable17c=0.0d0
gtable18a=0.0d0 ; gtable18b=0.0d0 ; gtable18c=0.0d0
gtable19a=0.0d0 ; gtable19b=0.0d0 ; gtable19c=0.0d0
gtable20a=0.0d0 ; gtable20b=0.0d0 ; gtable20c=0.0d0
gtable21a=0.0d0 ; gtable21b=0.0d0 ; gtable21c=0.0d0
gtable22a=0.0d0 ; gtable22b=0.0d0 ; gtable22c=0.0d0
gtable23a=0.0d0 ; gtable23b=0.0d0 ; gtable23c=0.0d0
gtable24a=0.0d0 ; gtable24b=0.0d0 ; gtable24c=0.0d0



DO m=1,nfam
    DO i0=0, 15
    DO i4=0, 15
    DO i5=1, 3
    DO i6=0, 4
    DO i7=0, 6
        gtable12(i6,1)=gtable12(i6,1)+gtype1c(m,i7,i6,i5,i4,i0)
        gtable13(i7,1)=gtable13(i7,1)+gtype1c(m,i7,i6,i5,i4,i0)
    ENDDO
    ENDDO
    ENDDO
    ENDDO
    ENDDO

    SELECT CASE(nchild(m))
    CASE(1)
        DO i0=0, 15
        DO i4=0, 15
        DO i5=1, 3
        DO i6=0, 4
        DO i7=0, 6
            gtable12(i6,2)=gtable12(i6,2)+gtype1c(m,i7,i6,i5,i4,i0)
            gtable13(i7,2)=gtable13(i7,2)+gtype1c(m,i7,i6,i5,i4,i0)
        ENDDO
        ENDDO
        ENDDO
        ENDDO
        ENDDO
    CASE(2)
        DO i0=0, 15
        DO i4=0, 15
        DO i5=1, 3
        DO i6=0, 4
        DO i7=0, 6
            gtable12(i6,3)=gtable12(i6,3)+gtype1c(m,i7,i6,i5,i4,i0)
            gtable13(i7,3)=gtable13(i7,3)+gtype1c(m,i7,i6,i5,i4,i0)
        ENDDO
        ENDDO
        ENDDO
        ENDDO
        ENDDO
    CASE(3)
        DO i0=0, 15
        DO i4=0, 15
        DO i5=1, 3
        DO i6=0, 4
        DO i7=0, 6
            gtable12(i6,4)=gtable12(i6,4)+gtype1c(m,i7,i6,i5,i4,i0)
            gtable13(i7,4)=gtable13(i7,4)+gtype1c(m,i7,i6,i5,i4,i0)
        ENDDO
        ENDDO
        ENDDO
        ENDDO
        ENDDO
    CASE(4)
        DO i0=0, 15
        DO i4=0, 15
        DO i5=1, 3
        DO i6=0, 4
        DO i7=0, 6
            gtable12(i6,5)=gtable12(i6,5)+gtype1c(m,i7,i6,i5,i4,i0)
            gtable13(i7,5)=gtable13(i7,5)+gtype1c(m,i7,i6,i5,i4,i0)
        ENDDO
        ENDDO
        ENDDO
        ENDDO
        ENDDO
    CASE DEFAULT
        PAUSE
    END SELECT

ENDDO
    gtable12a(:,1) =gtable12(:,1)/nfam ; gtable12a(:,2)=gtable12(:,2)/gtct_1c ; gtable12a(:,3)=gtable12(:,3)/gtct_2c ; gtable12a(:,4)=gtable12(:,4)/gtct_3c ; gtable12a(:,5)=gtable12(:,5)/gtct_4c
    gtable12(:,:)=gtable12(:,:)/nfam
    gtable13a(:,1) =gtable13(:,1)/nfam ; gtable13a(:,2)=gtable13(:,2)/gtct_1c ; gtable13a(:,3)=gtable13(:,3)/gtct_2c ; gtable13a(:,4)=gtable13(:,4)/gtct_3c ; gtable13a(:,5)=gtable13(:,5)/gtct_4c
    gtable13(:,:)=gtable13(:,:)/nfam


DO m=1,nfam

    SELECT CASE(nchild(m))
    CASE(1)
    CASE(2)

        DO i0=0, 3
        DO i4=0, 3
        DO i5=1, 3
        DO i6=0, 2
        DO i7=0, 3
            gtable14a(i6,1) =gtable14a(i6,1) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable15a(i7,1) =gtable15a(i7,1) +gtype1c(m,i7,i6,i5,i4,i0)
            IF (i0==0) THEN
            gtable14a(i6,2) =gtable14a(i6,2) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable15a(i7,2) =gtable15a(i7,2) +gtype1c(m,i7,i6,i5,i4,i0)
            ELSEIF (i0==1 .or. i0==2) THEN
            gtable14a(i6,3) =gtable14a(i6,3) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable15a(i7,3) =gtable15a(i7,3) +gtype1c(m,i7,i6,i5,i4,i0)
            ELSEIF (i0==3) THEN
            gtable14a(i6,4) =gtable14a(i6,4) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable15a(i7,4) =gtable15a(i7,4) +gtype1c(m,i7,i6,i5,i4,i0)
            ENDIF

            gtable16a(i6,1) =gtable16a(i6,1) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable17a(i7,1) =gtable17a(i7,1) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable16a(i6,i5+1) =gtable16a(i6,i5+1) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable17a(i7,i5+1) =gtable17a(i7,i5+1) +gtype1c(m,i7,i6,i5,i4,i0)

            gtable18a(i7,1) =gtable18a(i7,1) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable19a(i0,1) =gtable19a(i0,1) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable18a(i7,i6+2) =gtable18a(i7,i6+2) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable19a(i0,i6+2) =gtable19a(i0,i6+2) +gtype1c(m,i7,i6,i5,i4,i0)

            gtable20a(i5,1) =gtable20a(i5,1) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable21a(i4,1) =gtable21a(i4,1) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable20a(i5,i6+2) =gtable20a(i5,i6+2) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable21a(i4,i6+2) =gtable21a(i4,i6+2) +gtype1c(m,i7,i6,i5,i4,i0)

            gtable22a(nearcounti(i0,nchild(m)),1) =gtable22a(nearcounti(i0,nchild(m)),1) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable22a(nearcounti(i0,nchild(m)),nearcounti(i4,nchild(m))+2) =gtable22a(nearcounti(i0,nchild(m)),nearcounti(i4,nchild(m))+2) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable23a(i5,1) =gtable23a(i5,1) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable23a(i5,nearcounti(i4,nchild(m))+2) =gtable23a(i5,nearcounti(i4,nchild(m))+2) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable24a(i7,1) =gtable24a(i7,1) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable24a(i7,nearcounti(i4,nchild(m))+2) =gtable24a(i7,nearcounti(i4,nchild(m))+2) +gtype1c(m,i7,i6,i5,i4,i0)
	    ENDDO
	    ENDDO
	    ENDDO
	    ENDDO
	    ENDDO

    CASE(3)

        DO i0=0, 7
        DO i4=0, 7
        DO i5=1, 3
        DO i6=0, 3
        DO i7=0, 6
            gtable14b(i6,1) =gtable14b(i6,1) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable15b(i7,1) =gtable15b(i7,1) +gtype1c(m,i7,i6,i5,i4,i0)
            IF (i0==0) THEN
            gtable14b(i6,2) =gtable14b(i6,2) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable15b(i7,2) =gtable15b(i7,2) +gtype1c(m,i7,i6,i5,i4,i0)
            ELSEIF (i0==1 .or. i0==2 .or. i0==4) THEN
            gtable14b(i6,3) =gtable14b(i6,3) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable15b(i7,3) =gtable15b(i7,3) +gtype1c(m,i7,i6,i5,i4,i0)
            ELSEIF (i0==3 .or. i0==5 .or. i0==6) THEN
            gtable14b(i6,4) =gtable14b(i6,4) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable15b(i7,4) =gtable15b(i7,4) +gtype1c(m,i7,i6,i5,i4,i0)
            ELSEIF (i0==7) THEN
            gtable14b(i6,5) =gtable14b(i6,5) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable15b(i7,5) =gtable15b(i7,5) +gtype1c(m,i7,i6,i5,i4,i0)
            ENDIF

            gtable16b(i6,1) =gtable16b(i6,1) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable17b(i7,1) =gtable17b(i7,1) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable16b(i6,i5+1) =gtable16b(i6,i5+1) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable17b(i7,i5+1) =gtable17b(i7,i5+1) +gtype1c(m,i7,i6,i5,i4,i0)

            gtable18b(i7,1) =gtable18b(i7,1) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable19b(i0,1) =gtable19b(i0,1) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable18b(i7,i6+2) =gtable18b(i7,i6+2) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable19b(i0,i6+2) =gtable19b(i0,i6+2) +gtype1c(m,i7,i6,i5,i4,i0)

            gtable20b(i5,1) =gtable20b(i5,1) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable21b(i4,1) =gtable21b(i4,1) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable20b(i5,i6+2) =gtable20b(i5,i6+2) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable21b(i4,i6+2) =gtable21b(i4,i6+2) +gtype1c(m,i7,i6,i5,i4,i0)

            gtable22b(nearcounti(i0,nchild(m)),1) =gtable22b(nearcounti(i0,nchild(m)),1) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable22b(nearcounti(i0,nchild(m)),nearcounti(i4,nchild(m))+2) =gtable22b(nearcounti(i0,nchild(m)),nearcounti(i4,nchild(m))+2) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable23b(i5,1) =gtable23b(i5,1) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable23b(i5,nearcounti(i4,nchild(m))+2) =gtable23b(i5,nearcounti(i4,nchild(m))+2) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable24b(i7,1) =gtable24b(i7,1) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable24b(i7,nearcounti(i4,nchild(m))+2) =gtable24b(i7,nearcounti(i4,nchild(m))+2) +gtype1c(m,i7,i6,i5,i4,i0)
	    ENDDO
	    ENDDO
	    ENDDO
	    ENDDO
	    ENDDO

    CASE(4)

        DO i0=0, 15
        DO i4=0, 15
        DO i5=1, 3
        DO i6=0, 4
        DO i7=0, 6
            gtable14c(i6,1) =gtable14c(i6,1) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable15c(i7,1) =gtable15c(i7,1) +gtype1c(m,i7,i6,i5,i4,i0)
            IF (i0==0) THEN
            gtable14c(i6,2) =gtable14c(i6,2) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable15c(i7,2) =gtable15c(i7,2) +gtype1c(m,i7,i6,i5,i4,i0)
            ELSEIF (i0==1 .or. i0==2 .or. i0==4 .or. i0==8) THEN
            gtable14c(i6,3) =gtable14c(i6,3) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable15c(i7,3) =gtable15c(i7,3) +gtype1c(m,i7,i6,i5,i4,i0)
            ELSEIF (i0==3 .or. i0==5 .or. i0==6 .or. i0==9 .or. i0==10 .or. i0==12) THEN
            gtable14c(i6,4) =gtable14c(i6,4) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable15c(i7,4) =gtable15c(i7,4) +gtype1c(m,i7,i6,i5,i4,i0)
            ELSEIF (i0==7 .or. i0==11 .or. i0==13 .or. i0==14) THEN
            gtable14c(i6,5) =gtable14c(i6,5) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable15c(i7,5) =gtable15c(i7,5) +gtype1c(m,i7,i6,i5,i4,i0)
            ELSEIF (i0==15) THEN
            gtable14c(i6,6) =gtable14c(i6,6) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable15c(i7,6) =gtable15c(i7,6) +gtype1c(m,i7,i6,i5,i4,i0)
            ENDIF

            gtable16c(i6,1) =gtable16c(i6,1) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable17c(i7,1) =gtable17c(i7,1) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable16c(i6,i5+1) =gtable16c(i6,i5+1) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable17c(i7,i5+1) =gtable17c(i7,i5+1) +gtype1c(m,i7,i6,i5,i4,i0)

            gtable18c(i7,1) =gtable18c(i7,1) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable19c(i0,1) =gtable19c(i0,1) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable18c(i7,i6+2) =gtable18c(i7,i6+2) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable19c(i0,i6+2) =gtable19c(i0,i6+2) +gtype1c(m,i7,i6,i5,i4,i0)

            gtable20c(i5,1) =gtable20c(i5,1) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable21c(i4,1) =gtable21c(i4,1) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable20c(i5,i6+2) =gtable20c(i5,i6+2) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable21c(i4,i6+2) =gtable21c(i4,i6+2) +gtype1c(m,i7,i6,i5,i4,i0)

            gtable22c(nearcounti(i0,nchild(m)),1) =gtable22c(nearcounti(i0,nchild(m)),1) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable22c(nearcounti(i0,nchild(m)),nearcounti(i4,nchild(m))+2) =gtable22c(nearcounti(i0,nchild(m)),nearcounti(i4,nchild(m))+2) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable23c(i5,1) =gtable23c(i5,1) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable23c(i5,nearcounti(i4,nchild(m))+2) =gtable23c(i5,nearcounti(i4,nchild(m))+2) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable24c(i7,1) =gtable24c(i7,1) +gtype1c(m,i7,i6,i5,i4,i0)
            gtable24c(i7,nearcounti(i4,nchild(m))+2) =gtable24c(i7,nearcounti(i4,nchild(m))+2) +gtype1c(m,i7,i6,i5,i4,i0)
	    ENDDO
	    ENDDO
	    ENDDO
	    ENDDO
	    ENDDO

    CASE DEFAULT
        PAUSE
    END SELECT

ENDDO

gtable14a(:,:) =gtable14a(:,:)/gtct_2c ;  gtable14b(:,:) =gtable14b(:,:)/gtct_3c ;  gtable14c(:,:) =gtable14c(:,:)/gtct_4c
gtable15a(:,:) =gtable15a(:,:)/gtct_2c ;  gtable15b(:,:) =gtable15b(:,:)/gtct_3c ;  gtable15c(:,:) =gtable15c(:,:)/gtct_4c

gtable16a(:,1) =gtable16a(:,1)/gtct_2c ; gtable16a(:,2)=gtable16a(:,2)/gtct_2ce1 ; gtable16a(:,3)=gtable16a(:,3)/gtct_2ce2 ; gtable16a(:,4)=gtable16a(:,4)/gtct_2ce3
gtable16b(:,1) =gtable16b(:,1)/gtct_3c ; gtable16b(:,2)=gtable16b(:,2)/gtct_3ce1 ; gtable16b(:,3)=gtable16b(:,3)/gtct_3ce2 ; gtable16b(:,4)=gtable16b(:,4)/gtct_3ce3
gtable16c(:,1) =gtable16c(:,1)/gtct_4c ; gtable16c(:,2)=gtable16c(:,2)/gtct_4ce1 ; gtable16c(:,3)=gtable16c(:,3)/gtct_4ce2 ; gtable16c(:,4)=gtable16c(:,4)/gtct_4ce3
gtable17a(:,1) =gtable17a(:,1)/gtct_2c ; gtable17a(:,2)=gtable17a(:,2)/gtct_2ce1 ; gtable17a(:,3)=gtable17a(:,3)/gtct_2ce2 ; gtable17a(:,4)=gtable17a(:,4)/gtct_2ce3
gtable17b(:,1) =gtable17b(:,1)/gtct_3c ; gtable17b(:,2)=gtable17b(:,2)/gtct_3ce1 ; gtable17b(:,3)=gtable17b(:,3)/gtct_3ce2 ; gtable17b(:,4)=gtable17b(:,4)/gtct_3ce3
gtable17c(:,1) =gtable17c(:,1)/gtct_4c ; gtable17c(:,2)=gtable17c(:,2)/gtct_4ce1 ; gtable17c(:,3)=gtable17c(:,3)/gtct_4ce2 ; gtable17c(:,4)=gtable17c(:,4)/gtct_4ce3

gtable18a(:,1) =gtable18a(:,1)/gtct_2c ; gtable18a(:,2) =gtable18a(:,2)/gtct_2cid0 ; gtable18a(:,3) =gtable18a(:,3)/gtct_2cid1 ; gtable18a(:,4) =gtable18a(:,4)/gtct_2cid2
gtable18b(:,1) =gtable18b(:,1)/gtct_3c ; gtable18b(:,2) =gtable18b(:,2)/gtct_3cid0 ; gtable18b(:,3) =gtable18b(:,3)/gtct_3cid1 ; gtable18b(:,4) =gtable18b(:,4)/gtct_3cid2 ; gtable18b(:,5) =gtable18b(:,5)/gtct_3cid3
gtable18c(:,1) =gtable18c(:,1)/gtct_4c ; gtable18c(:,2) =gtable18c(:,2)/gtct_4cid0 ; gtable18c(:,3) =gtable18c(:,3)/gtct_4cid1 ; gtable18c(:,4) =gtable18c(:,4)/gtct_4cid2 ; gtable18c(:,5) =gtable18c(:,5)/gtct_4cid3 ; gtable18c(:,6) =gtable18c(:,6)/gtct_4cid4

gtable19a(:,1) =gtable19a(:,1)/gtct_2c ; gtable19a(:,2) =gtable19a(:,2)/gtct_2cid0 ; gtable19a(:,3) =gtable19a(:,3)/gtct_2cid1 ; gtable19a(:,4) =gtable19a(:,4)/gtct_2cid2
gtable19b(:,1) =gtable19b(:,1)/gtct_3c ; gtable19b(:,2) =gtable19b(:,2)/gtct_3cid0 ; gtable19b(:,3) =gtable19b(:,3)/gtct_3cid1 ; gtable19b(:,4) =gtable19b(:,4)/gtct_3cid2 ; gtable19b(:,5) =gtable19b(:,5)/gtct_3cid3
gtable19c(:,1) =gtable19c(:,1)/gtct_4c ; gtable19c(:,2) =gtable19c(:,2)/gtct_4cid0 ; gtable19c(:,3) =gtable19c(:,3)/gtct_4cid1 ; gtable19c(:,4) =gtable19c(:,4)/gtct_4cid2 ; gtable19c(:,5) =gtable19c(:,5)/gtct_4cid3 ; gtable19c(:,6) =gtable19c(:,6)/gtct_4cid4

gtable20a(:,1) =gtable20a(:,1)/gtct_2c ; gtable20a(:,2) =gtable20a(:,2)/gtct_2cid0 ; gtable20a(:,3) =gtable20a(:,3)/gtct_2cid1 ; gtable20a(:,4) =gtable20a(:,4)/gtct_2cid2
gtable20b(:,1) =gtable20b(:,1)/gtct_3c ; gtable20b(:,2) =gtable20b(:,2)/gtct_3cid0 ; gtable20b(:,3) =gtable20b(:,3)/gtct_3cid1 ; gtable20b(:,4) =gtable20b(:,4)/gtct_3cid2 ; gtable20b(:,5) =gtable20b(:,5)/gtct_3cid3
gtable20c(:,1) =gtable20c(:,1)/gtct_4c ; gtable20c(:,2) =gtable20c(:,2)/gtct_4cid0 ; gtable20c(:,3) =gtable20c(:,3)/gtct_4cid1 ; gtable20c(:,4) =gtable20c(:,4)/gtct_4cid2 ; gtable20c(:,5) =gtable20c(:,5)/gtct_4cid3 ; gtable20c(:,6) =gtable20c(:,6)/gtct_4cid4

gtable21a(:,1) =gtable21a(:,1)/gtct_2c ; gtable21a(:,2) =gtable21a(:,2)/gtct_2cid0 ; gtable21a(:,3) =gtable21a(:,3)/gtct_2cid1 ; gtable21a(:,4) =gtable21a(:,4)/gtct_2cid2
gtable21b(:,1) =gtable21b(:,1)/gtct_3c ; gtable21b(:,2) =gtable21b(:,2)/gtct_3cid0 ; gtable21b(:,3) =gtable21b(:,3)/gtct_3cid1 ; gtable21b(:,4) =gtable21b(:,4)/gtct_3cid2 ; gtable21b(:,5) =gtable21b(:,5)/gtct_3cid3
gtable21c(:,1) =gtable21c(:,1)/gtct_4c ; gtable21c(:,2) =gtable21c(:,2)/gtct_4cid0 ; gtable21c(:,3) =gtable21c(:,3)/gtct_4cid1 ; gtable21c(:,4) =gtable21c(:,4)/gtct_4cid2 ; gtable21c(:,5) =gtable21c(:,5)/gtct_4cid3 ; gtable21c(:,6) =gtable21c(:,6)/gtct_4cid4

gtable22a(:,1) =gtable22a(:,1)/gtct_2c ; gtable22a(:,2) =gtable22a(:,2)/gtct_2cfb0 ; gtable22a(:,3) =gtable22a(:,3)/gtct_2cfb1 ; gtable22a(:,4) =gtable22a(:,4)/gtct_2cfb2
gtable22b(:,1) =gtable22b(:,1)/gtct_3c ; gtable22b(:,2) =gtable22b(:,2)/gtct_3cfb0 ; gtable22b(:,3) =gtable22b(:,3)/gtct_3cfb1 ; gtable22b(:,4) =gtable22b(:,4)/gtct_3cfb2 ; gtable22b(:,5) =gtable22b(:,5)/gtct_3cfb3
gtable22c(:,1) =gtable22c(:,1)/gtct_4c ; gtable22c(:,2) =gtable22c(:,2)/gtct_4cfb0 ; gtable22c(:,3) =gtable22c(:,3)/gtct_4cfb1 ; gtable22c(:,4) =gtable22c(:,4)/gtct_4cfb2 ; gtable22c(:,5) =gtable22c(:,5)/gtct_4cfb3 ; gtable22c(:,6) =gtable22c(:,6)/gtct_4cfb4

gtable23a(:,1) =gtable23a(:,1)/gtct_2c ; gtable23a(:,2) =gtable23a(:,2)/gtct_2cfb0 ; gtable23a(:,3) =gtable23a(:,3)/gtct_2cfb1 ; gtable23a(:,4) =gtable23a(:,4)/gtct_2cfb2
gtable23b(:,1) =gtable23b(:,1)/gtct_3c ; gtable23b(:,2) =gtable23b(:,2)/gtct_3cfb0 ; gtable23b(:,3) =gtable23b(:,3)/gtct_3cfb1 ; gtable23b(:,4) =gtable23b(:,4)/gtct_3cfb2 ; gtable23b(:,5) =gtable23b(:,5)/gtct_3cfb3
gtable23c(:,1) =gtable23c(:,1)/gtct_4c ; gtable23c(:,2) =gtable23c(:,2)/gtct_4cfb0 ; gtable23c(:,3) =gtable23c(:,3)/gtct_4cfb1 ; gtable23c(:,4) =gtable23c(:,4)/gtct_4cfb2 ; gtable23c(:,5) =gtable23c(:,5)/gtct_4cfb3 ; gtable23c(:,6) =gtable23c(:,6)/gtct_4cfb4

gtable24a(:,1) =gtable24a(:,1)/gtct_2c ; gtable24a(:,2) =gtable24a(:,2)/gtct_2cfb0 ; gtable24a(:,3) =gtable24a(:,3)/gtct_2cfb1 ; gtable24a(:,4) =gtable24a(:,4)/gtct_2cfb2
gtable24b(:,1) =gtable24b(:,1)/gtct_3c ; gtable24b(:,2) =gtable24b(:,2)/gtct_3cfb0 ; gtable24b(:,3) =gtable24b(:,3)/gtct_3cfb1 ; gtable24b(:,4) =gtable24b(:,4)/gtct_3cfb2 ; gtable24b(:,5) =gtable24b(:,5)/gtct_3cfb3
gtable24c(:,1) =gtable24c(:,1)/gtct_4c ; gtable24c(:,2) =gtable24c(:,2)/gtct_4cfb0 ; gtable24c(:,3) =gtable24c(:,3)/gtct_4cfb1 ; gtable24c(:,4) =gtable24c(:,4)/gtct_4cfb2 ; gtable24c(:,5) =gtable24c(:,5)/gtct_4cfb3 ; gtable24c(:,6) =gtable24c(:,6)/gtct_4cfb4



WRITE(4,fmt=*) 'Table 1-0: SPNE Living arrangement'
WRITE(4,fmt=*) '   0. Far , Far , Far , Far  (no one lives near)'
WRITE(4,fmt=*) '   1. Far , Far , Far , Near (only the youngest lives near)'
WRITE(4,fmt=*) '   2. Far , Far , Near, Far  (only the second youngest lives near)'
WRITE(4,fmt=*) '   3. Far , Far , Near, Near'
WRITE(4,fmt=*) '   4. Far , Near, Far , Far '
WRITE(4,fmt=*) '   5. Far , Near, Far , Near'
WRITE(4,fmt=*) '   6. Far , Near, Near, Far '
WRITE(4,fmt=*) '   7. Far , Near, Near, Near'
WRITE(4,fmt=*) '   8. Near, Far , Far , Far '
WRITE(4,fmt=*) '   9. Near, Far , Far , Near'
WRITE(4,fmt=*) '  10. Near, Far , Near, Far '
WRITE(4,fmt=*) '  11. Near, Far , Near, Near'
WRITE(4,fmt=*) '  12. Near, Near, Far , Far '
WRITE(4,fmt=*) '  13. Near, Near, Far , Near'
WRITE(4,fmt=*) '  14. Near, Near, Near, Far '
WRITE(4,fmt=*) '  15. Near, Near, Near, Near'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A10)') '     Type','     total','   1 child','   2 child','   3 child','   4 child'
DO k=0,15
    WRITE(4,fmt='(I10,5F10.4)') k,(gtable0(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-1: SPNE Living arrangement (Column shares)'
WRITE(4,fmt=*) '   0. Far , Far , Far , Far  (no one lives near)'
WRITE(4,fmt=*) '   1. Far , Far , Far , Near (only the youngest lives near)'
WRITE(4,fmt=*) '   2. Far , Far , Near, Far  (only the second youngest lives near)'
WRITE(4,fmt=*) '   3. Far , Far , Near, Near'
WRITE(4,fmt=*) '   4. Far , Near, Far , Far '
WRITE(4,fmt=*) '   5. Far , Near, Far , Near'
WRITE(4,fmt=*) '   6. Far , Near, Near, Far '
WRITE(4,fmt=*) '   7. Far , Near, Near, Near'
WRITE(4,fmt=*) '   8. Near, Far , Far , Far '
WRITE(4,fmt=*) '   9. Near, Far , Far , Near'
WRITE(4,fmt=*) '  10. Near, Far , Near, Far '
WRITE(4,fmt=*) '  11. Near, Far , Near, Near'
WRITE(4,fmt=*) '  12. Near, Near, Far , Far '
WRITE(4,fmt=*) '  13. Near, Near, Far , Near'
WRITE(4,fmt=*) '  14. Near, Near, Near, Far '
WRITE(4,fmt=*) '  15. Near, Near, Near, Near'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A10)') '     Type','     total','   1 child','   2 child','   3 child','   4 child'
DO k=0,15
    WRITE(4,fmt='(I10,5F10.4)') k,(gtable1(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-2: Efficiency type'
WRITE(4,fmt=*) '   1. Joint utility best (KH efficient): Normal form N.E.'
WRITE(4,fmt=*) '   2. Joint utility best (KH efficient): Not normal form N.E.'
WRITE(4,fmt=*) '   3. KH inefficient, not Pareto dominated: Normal form N.E.'
WRITE(4,fmt=*) '   4. KH inefficient, not Pareto dominated: Not normal form N.E.'
WRITE(4,fmt=*) '   5. Pareto Dominated (P.D.): Normal form N.E.'
WRITE(4,fmt=*) '   6. Pareto Dominated (P.D.): Not normal form N.E.'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A10)') '     Type','     total','   1 child','   2 child','   3 child','   4 child'
DO k=1,6
    WRITE(4,fmt='(I10,5F10.4)') k,(gtable2(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-3: Children who have dominant strategy in normal form'
WRITE(4,fmt=*) '   0. 0, 0, 0, 0 (no one has d.s.)'
WRITE(4,fmt=*) '   1. 0, 0, 0, 1 (only the youngest has d.s.)'
WRITE(4,fmt=*) '   2. 0, 0, 1, 0 (only the second youngest has d.s.)'
WRITE(4,fmt=*) '   3. 0, 0, 1, 1'
WRITE(4,fmt=*) '   4. 0, 1, 0, 0'
WRITE(4,fmt=*) '   5. 0, 1, 0, 1'
WRITE(4,fmt=*) '   6. 0, 1, 1, 0'
WRITE(4,fmt=*) '   7. 0, 1, 1, 1'
WRITE(4,fmt=*) '   8. 1, 0, 0, 0'
WRITE(4,fmt=*) '   9. 1, 0, 0, 1'
WRITE(4,fmt=*) '  10. 1, 0, 1, 0'
WRITE(4,fmt=*) '  11. 1, 0, 1, 1'
WRITE(4,fmt=*) '  12. 1, 1, 0, 0'
WRITE(4,fmt=*) '  13. 1, 1, 0, 1'
WRITE(4,fmt=*) '  14. 1, 1, 1, 0'
WRITE(4,fmt=*) '  15. 1, 1, 1, 1'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A10)') '     Type','     total','   1 child','   2 child','   3 child','   4 child'
DO k=0,15
    WRITE(4,fmt='(I10,5F10.4)') k,(gtable3(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-4: Children who play a normal-form dominant strategy in SPNE'
WRITE(4,fmt=*) '   0. 0, 0, 0, 0 (no one does)'
WRITE(4,fmt=*) '   1. 0, 0, 0, 1 (only the youngest does)'
WRITE(4,fmt=*) '   2. 0, 0, 1, 0'
WRITE(4,fmt=*) '   3. 0, 0, 1, 1'
WRITE(4,fmt=*) '   4. 0, 1, 0, 0'
WRITE(4,fmt=*) '   5. 0, 1, 0, 1'
WRITE(4,fmt=*) '   6. 0, 1, 1, 0'
WRITE(4,fmt=*) '   7. 0, 1, 1, 1'
WRITE(4,fmt=*) '   8. 1, 0, 0, 0'
WRITE(4,fmt=*) '   9. 1, 0, 0, 1'
WRITE(4,fmt=*) '  10. 1, 0, 1, 0'
WRITE(4,fmt=*) '  11. 1, 0, 1, 1'
WRITE(4,fmt=*) '  12. 1, 1, 0, 0'
WRITE(4,fmt=*) '  13. 1, 1, 0, 1'
WRITE(4,fmt=*) '  14. 1, 1, 1, 0'
WRITE(4,fmt=*) '  15. 1, 1, 1, 1'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A10)') '     Type','     total','   1 child','   2 child','   3 child','   4 child'
DO k=0,15
    WRITE(4,fmt='(I10,5F10.4)') k,(gtable4(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-5: Location pattern of the joint optimal configuration'
WRITE(4,fmt=*) '   0. 0, 0, 0, 0 (no one near)'
WRITE(4,fmt=*) '   1. 0, 0, 0, 1 (only the youngest near)'
WRITE(4,fmt=*) '   2. 0, 0, 1, 0 (only the second youngest near)'
WRITE(4,fmt=*) '   3. 0, 0, 1, 1'
WRITE(4,fmt=*) '   4. 0, 1, 0, 0'
WRITE(4,fmt=*) '   5. 0, 1, 0, 1'
WRITE(4,fmt=*) '   6. 0, 1, 1, 0'
WRITE(4,fmt=*) '   7. 0, 1, 1, 1'
WRITE(4,fmt=*) '   8. 1, 0, 0, 0'
WRITE(4,fmt=*) '   9. 1, 0, 0, 1'
WRITE(4,fmt=*) '  10. 1, 0, 1, 0'
WRITE(4,fmt=*) '  11. 1, 0, 1, 1'
WRITE(4,fmt=*) '  12. 1, 1, 0, 0'
WRITE(4,fmt=*) '  13. 1, 1, 0, 1'
WRITE(4,fmt=*) '  14. 1, 1, 1, 0'
WRITE(4,fmt=*) '  15. 1, 1, 1, 1'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A10)') '     Type','     total','   1 child','   2 child','   3 child','   4 child'
DO k=0,15
    WRITE(4,fmt='(I10,5F10.4)') k,(gtable5(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '


WRITE(4,fmt=*) 'Table 1-6a Children who have dominant strategy in normal form'
WRITE(4,fmt=*) '  By SPNE Efficiency type: 2 child families (Column share)'
!WRITE(4,fmt=*) '   0. 0, 0 (no one has d.s.)'
!WRITE(4,fmt=*) '   1. 0, 1 (only the youngest has d.s.)'
!WRITE(4,fmt=*) '   2. 1, 0 (only the second youngest has d.s.)'
!WRITE(4,fmt=*) '   3. 1, 1'
!WRITE(4,fmt=*) ' '
!WRITE(4,fmt='(5A10)') '     Type','     Total','   KH best','      P.O.','      P.D.'
!WRITE(4,fmt=*) 'Not available for the time being'
!DO k=0,3
!    WRITE(4,fmt='(I10,4F10.4)') k,(gtable6a(k,m),m=1,4)
!ENDDO
!WRITE(4,fmt=*) ' '
!
WRITE(4,fmt=*) 'Table 1-6b Children who have dominant strategy in normal form'
WRITE(4,fmt=*) '  By SPNE Efficiency type: 3 child families (Column share)'
!WRITE(4,fmt=*) '   0. 0, 0, 0 (no one has d.s.)'
!WRITE(4,fmt=*) '   1. 0, 0, 1 (only the youngest has d.s.)'
!WRITE(4,fmt=*) '   2. 0, 1, 0 (only the second youngest has d.s.)'
!WRITE(4,fmt=*) '   3. 0, 1, 1'
!WRITE(4,fmt=*) '   4. 1, 0, 0'
!WRITE(4,fmt=*) '   5. 1, 0, 1'
!WRITE(4,fmt=*) '   6. 1, 1, 0'
!WRITE(4,fmt=*) '   7. 1, 1, 1'
!WRITE(4,fmt=*) ' '
!WRITE(4,fmt='(5A10)') '     Type','     Total','   KH best','      P.O.','      P.D.'
!WRITE(4,fmt=*) 'Not available for the time being'
!DO k=0,7
!    WRITE(4,fmt='(I10,4F10.4)') k,(gtable6b(k,m),m=1,4)
!ENDDO
!WRITE(4,fmt=*) ' '
!
WRITE(4,fmt=*) 'Table 1-6c Children who have dominant strategy in normal form'
WRITE(4,fmt=*) '  By SPNE Efficiency type: 4 child families (Column share)'
!WRITE(4,fmt=*) '   0. 0, 0, 0, 0 (no one has d.s.)'
!WRITE(4,fmt=*) '   1. 0, 0, 0, 1 (only the youngest has d.s.)'
!WRITE(4,fmt=*) '   2. 0, 0, 1, 0 (only the second youngest has d.s.)'
!WRITE(4,fmt=*) '   3. 0, 0, 1, 1'
!WRITE(4,fmt=*) '   4. 0, 1, 0, 0'
!WRITE(4,fmt=*) '   5. 0, 1, 0, 1'
!WRITE(4,fmt=*) '   6. 0, 1, 1, 0'
!WRITE(4,fmt=*) '   7. 0, 1, 1, 1'
!WRITE(4,fmt=*) '   8. 1, 0, 0, 0'
!WRITE(4,fmt=*) '   9. 1, 0, 0, 1'
!WRITE(4,fmt=*) '  10. 1, 0, 1, 0'
!WRITE(4,fmt=*) '  11. 1, 0, 1, 1'
!WRITE(4,fmt=*) '  12. 1, 1, 0, 0'
!WRITE(4,fmt=*) '  13. 1, 1, 0, 1'
!WRITE(4,fmt=*) '  14. 1, 1, 1, 0'
!WRITE(4,fmt=*) '  15. 1, 1, 1, 1'
WRITE(4,fmt=*) ' '
!WRITE(4,fmt='(5A10)') '     Type','     Total','   KH best','      P.O.','      P.D.'
WRITE(4,fmt=*) 'Not available for the time being'
!DO k=0,15
!    WRITE(4,fmt='(I10,4F10.4)') k,(gtable6c(k,m),m=1,4)
!ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-7a: SPNE (observed) location pattern'
WRITE(4,fmt=*) '  By SPNE Efficiency type: 2 child families (Column share)'
WRITE(4,fmt=*) '   0. 0, 0 (no one near)'
WRITE(4,fmt=*) '   1. 0, 1 (only the youngest near)'
WRITE(4,fmt=*) '   2. 1, 0 (only the second youngest near)'
WRITE(4,fmt=*) '   3. 1, 1'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(5A10)') '     Type','     Total','   KH best','      P.O.','      P.D.'
DO k=0,3
    WRITE(4,fmt='(I10,4F10.4)') k,(gtable7a(k,m),m=1,4)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-7b: SPNE (observed) location pattern'
WRITE(4,fmt=*) '  By SPNE Efficiency type: 3 child families (Column share)'
WRITE(4,fmt=*) '   0. 0, 0, 0 (no one near)'
WRITE(4,fmt=*) '   1. 0, 0, 1 (only the youngest near)'
WRITE(4,fmt=*) '   2. 0, 1, 0 (only the second youngest near)'
WRITE(4,fmt=*) '   3. 0, 1, 1'
WRITE(4,fmt=*) '   4. 1, 0, 0'
WRITE(4,fmt=*) '   5. 1, 0, 1'
WRITE(4,fmt=*) '   6. 1, 1, 0'
WRITE(4,fmt=*) '   7. 1, 1, 1'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(5A10)') '     Type','     Total','   KH best','      P.O.','      P.D.'
DO k=0,7
    WRITE(4,fmt='(I10,4F10.4)') k,(gtable7b(k,m),m=1,4)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-7c: SPNE (observed) location pattern'
WRITE(4,fmt=*) '  By SPNE Efficiency type: 4 child families (Column share)'
WRITE(4,fmt=*) '   0. 0, 0, 0, 0 (no one near)'
WRITE(4,fmt=*) '   1. 0, 0, 0, 1 (only the youngest near)'
WRITE(4,fmt=*) '   2. 0, 0, 1, 0 (only the second youngest near)'
WRITE(4,fmt=*) '   3. 0, 0, 1, 1'
WRITE(4,fmt=*) '   4. 0, 1, 0, 0'
WRITE(4,fmt=*) '   5. 0, 1, 0, 1'
WRITE(4,fmt=*) '   6. 0, 1, 1, 0'
WRITE(4,fmt=*) '   7. 0, 1, 1, 1'
WRITE(4,fmt=*) '   8. 1, 0, 0, 0'
WRITE(4,fmt=*) '   9. 1, 0, 0, 1'
WRITE(4,fmt=*) '  10. 1, 0, 1, 0'
WRITE(4,fmt=*) '  11. 1, 0, 1, 1'
WRITE(4,fmt=*) '  12. 1, 1, 0, 0'
WRITE(4,fmt=*) '  13. 1, 1, 0, 1'
WRITE(4,fmt=*) '  14. 1, 1, 1, 0'
WRITE(4,fmt=*) '  15. 1, 1, 1, 1'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(5A10)') '     Type','     Total','   KH best','      P.O.','      P.D.'
DO k=0,15
    WRITE(4,fmt='(I10,4F10.4)') k,(gtable7c(k,m),m=1,4)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-8a: Location pattern of the joint optimal configuration'
WRITE(4,fmt=*) '  By SPNE Efficiency type: 2 child families (Column share)'
WRITE(4,fmt=*) '   0. 0, 0 (no one near)'
WRITE(4,fmt=*) '   1. 0, 1 (only the youngest near)'
WRITE(4,fmt=*) '   2. 1, 0 (only the second youngest near)'
WRITE(4,fmt=*) '   3. 1, 1'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(5A10)') '     Type','     Total','   KH best','      P.O.','      P.D.'
DO k=0,3
    WRITE(4,fmt='(I10,4F10.4)') k,(gtable8a(k,m),m=1,4)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-8b: Location pattern of the joint optimal configuration'
WRITE(4,fmt=*) '  By SPNE Efficiency type: 3 child families (Column share)'
WRITE(4,fmt=*) '   0. 0, 0, 0 (no one near)'
WRITE(4,fmt=*) '   1. 0, 0, 1 (only the youngest near)'
WRITE(4,fmt=*) '   2. 0, 1, 0 (only the second youngest near)'
WRITE(4,fmt=*) '   3. 0, 1, 1'
WRITE(4,fmt=*) '   4. 1, 0, 0'
WRITE(4,fmt=*) '   5. 1, 0, 1'
WRITE(4,fmt=*) '   6. 1, 1, 0'
WRITE(4,fmt=*) '   7. 1, 1, 1'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(5A10)') '     Type','     Total','   KH best','      P.O.','      P.D.'
DO k=0,7
    WRITE(4,fmt='(I10,4F10.4)') k,(gtable8b(k,m),m=1,4)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-8c: Location pattern of the joint optimal configuration'
WRITE(4,fmt=*) '  By SPNE Efficiency type: 4 child families (Column share)'
WRITE(4,fmt=*) '   0. 0, 0, 0, 0 (no one near)'
WRITE(4,fmt=*) '   1. 0, 0, 0, 1 (only the youngest near)'
WRITE(4,fmt=*) '   2. 0, 0, 1, 0 (only the second youngest near)'
WRITE(4,fmt=*) '   3. 0, 0, 1, 1'
WRITE(4,fmt=*) '   4. 0, 1, 0, 0'
WRITE(4,fmt=*) '   5. 0, 1, 0, 1'
WRITE(4,fmt=*) '   6. 0, 1, 1, 0'
WRITE(4,fmt=*) '   7. 0, 1, 1, 1'
WRITE(4,fmt=*) '   8. 1, 0, 0, 0'
WRITE(4,fmt=*) '   9. 1, 0, 0, 1'
WRITE(4,fmt=*) '  10. 1, 0, 1, 0'
WRITE(4,fmt=*) '  11. 1, 0, 1, 1'
WRITE(4,fmt=*) '  12. 1, 1, 0, 0'
WRITE(4,fmt=*) '  13. 1, 1, 0, 1'
WRITE(4,fmt=*) '  14. 1, 1, 1, 0'
WRITE(4,fmt=*) '  15. 1, 1, 1, 1'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(5A10)') '      Type','     Total','   KH best','      P.O.','      P.D.'
DO k=0,15
    WRITE(4,fmt='(I10,4F10.4)') k,(gtable8c(k,m),m=1,4)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-9a # of Children who have dominant strategy in normal form'
WRITE(4,fmt=*) '  By # of children living near: 2 child families'
WRITE(4,fmt=*) '   0. 0 child'
WRITE(4,fmt=*) '   1. 1 child'
WRITE(4,fmt=*) '   2. 2 children'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(5A8)') '    Type','   Total',' 0 child',' 1 child',' 2 child'
DO k=0,2
    WRITE(4,fmt='(I8,4F8.4)') k,(gtable9a(k,m),m=1,4)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-9b # of Children who have dominant strategy in normal form'
WRITE(4,fmt=*) '  By # of children living near: 3 child families'
WRITE(4,fmt=*) '   0. 0 child'
WRITE(4,fmt=*) '   1. 1 child'
WRITE(4,fmt=*) '   2. 2 children'
WRITE(4,fmt=*) '   3. 3 children'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A8)') '    Type','   Total',' 0 child',' 1 child',' 2 child',' 3 child'
DO k=0,3
    WRITE(4,fmt='(I8,5F8.4)') k,(gtable9b(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-9c # of Children who have dominant strategy in normal form'
WRITE(4,fmt=*) '  By # of children living near: 4 child families'
WRITE(4,fmt=*) '   0. 0 child'
WRITE(4,fmt=*) '   1. 1 child'
WRITE(4,fmt=*) '   2. 2 children'
WRITE(4,fmt=*) '   3. 3 children'
WRITE(4,fmt=*) '   4. 4 children'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(7A8)') '    Type','   Total',' 0 child',' 1 child',' 2 child',' 3 child',' 4 child'
DO k=0,4
    WRITE(4,fmt='(I8,6F8.4)') k,(gtable9c(k,m),m=1,6)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-10a: # of children who play a normal-form dominant strategy in SPNE'
WRITE(4,fmt=*) '  By # of children living near: 2 child families'
WRITE(4,fmt=*) '   0. 0 child'
WRITE(4,fmt=*) '   1. 1 child'
WRITE(4,fmt=*) '   2. 2 children'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(5A8)') '    Type','   Total',' 0 child',' 1 child',' 2 child'
DO k=0,2
    WRITE(4,fmt='(I8,4F8.4)') k,(gtable10a(k,m),m=1,4)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-10b: # of children who play a normal-form dominant strategy in SPNE'
WRITE(4,fmt=*) '  By # of children living near: 3 child families'
WRITE(4,fmt=*) '   0. 0 child'
WRITE(4,fmt=*) '   1. 1 child'
WRITE(4,fmt=*) '   2. 2 children'
WRITE(4,fmt=*) '   3. 3 children'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A8)') '    Type','   Total',' 0 child',' 1 child',' 2 child',' 3 child'
DO k=0,3
    WRITE(4,fmt='(I8,5F8.4)') k,(gtable10b(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-10c: # of children who play a normal-form dominant strategy in SPNE'
WRITE(4,fmt=*) '  By # of children living near: 4 child families'
WRITE(4,fmt=*) '   0. 0 child'
WRITE(4,fmt=*) '   1. 1 child'
WRITE(4,fmt=*) '   2. 2 children'
WRITE(4,fmt=*) '   3. 3 children'
WRITE(4,fmt=*) '   4. 4 children'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(7A8)') '    Type','   Total',' 0 child',' 1 child',' 2 child',' 3 child',' 4 child'
DO k=0,4
    WRITE(4,fmt='(I8,6F8.4)') k,(gtable10c(k,m),m=1,6)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-11a: # of children who live near in the joint optimal configuration'
WRITE(4,fmt=*) '  By # of children living near: 2 child families'
WRITE(4,fmt=*) '   0. 0 child'
WRITE(4,fmt=*) '   1. 1 child'
WRITE(4,fmt=*) '   2. 2 children'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(5A8)') '    Type','   Total',' 0 child',' 1 child',' 2 child'
DO k=0,2
    WRITE(4,fmt='(I8,4F8.4)') k,(gtable11a(k,m),m=1,4)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-11b: # of children who live near in the joint optimal configuration'
WRITE(4,fmt=*) '  By # of children living near: 3 child families'
WRITE(4,fmt=*) '   0. 0 child'
WRITE(4,fmt=*) '   1. 1 child'
WRITE(4,fmt=*) '   2. 2 children'
WRITE(4,fmt=*) '   3. 3 children'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A8)') '    Type','   Total',' 0 child',' 1 child',' 2 child',' 3 child'
DO k=0,3
    WRITE(4,fmt='(I8,5F8.5)') k,(gtable11b(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-11c: # of children who live near in the joint optimal configuration'
WRITE(4,fmt=*) '  By # of children living near: 4 child families'
WRITE(4,fmt=*) '   0. 0 child'
WRITE(4,fmt=*) '   1. 1 child'
WRITE(4,fmt=*) '   2. 2 children'
WRITE(4,fmt=*) '   3. 3 children'
WRITE(4,fmt=*) '   4. 4 children'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(7A8)') '    Type','   Total',' 0 child',' 1 child',' 2 child',' 3 child',' 4 child'
DO k=0,4
    WRITE(4,fmt='(I8,6F8.4)') k,(gtable11c(k,m),m=1,6)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-12: The number of children with increasing difference'
WRITE(4,fmt=*) '   N.B. Increasing difference is defined as the difference between'
WRITE(4,fmt=*) '   when there is no other sibling living nearby and when there is'
WRITE(4,fmt=*) '   one such sibling.'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A10)') '     Type','     total','   1 child','   2 child','   3 child','   4 child'
DO k=0,4
    WRITE(4,fmt='(I10,5F10.4)') k,(gtable12(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '


WRITE(4,fmt=*) 'Table 1-12a: The number of children with increasing difference (Column share)'
WRITE(4,fmt=*) '   N.B. Increasing difference is defined as the difference between'
WRITE(4,fmt=*) '   when there is no other sibling living nearby and when there is'
WRITE(4,fmt=*) '   one such sibling.'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A10)') '     Type','     total','   1 child','   2 child','   3 child','   4 child'
DO k=0,4
    WRITE(4,fmt='(I10,5F10.4)') k,(gtable12a(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '


WRITE(4,fmt=*) 'Table 1-13: The number of normal form Nash Equilibria'
WRITE(4,fmt=*) '   0. 0'
WRITE(4,fmt=*) '   1. 1'
WRITE(4,fmt=*) '   2. 2'
WRITE(4,fmt=*) '   3. 3'
WRITE(4,fmt=*) '   4. 4'
WRITE(4,fmt=*) '   5. 5'
WRITE(4,fmt=*) '   6. 6 or more'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A10)') '     Type','     total','   1 child','   2 child','   3 child','   4 child'
DO k=0,6
    WRITE(4,fmt='(I10,5F10.4)') k,(gtable13(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '


WRITE(4,fmt=*) 'Table 1-13a: The number of normal form Nash Equilibria (Column share)'
WRITE(4,fmt=*) '   0. 0'
WRITE(4,fmt=*) '   1. 1'
WRITE(4,fmt=*) '   2. 2'
WRITE(4,fmt=*) '   3. 3'
WRITE(4,fmt=*) '   4. 4'
WRITE(4,fmt=*) '   5. 5'
WRITE(4,fmt=*) '   6. 6 or more'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A10)') '     Type','     total','   1 child','   2 child','   3 child','   4 child'
DO k=0,6
    WRITE(4,fmt='(I10,5F10.4)') k,(gtable13a(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '


WRITE(4,fmt=*) 'Table 1-14a: The number of children with increasing difference'
WRITE(4,fmt=*) '  By # of children living near: 2 child families'
WRITE(4,fmt=*) '   0. 0 child'
WRITE(4,fmt=*) '   1. 1 child'
WRITE(4,fmt=*) '   2. 2 children'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(5A8)') '    Type','   Total',' 0 child',' 1 child',' 2 child'
DO k=0,2
    WRITE(4,fmt='(I8,4F8.4)') k,(gtable14a(k,m),m=1,4)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-14b: The number of children with increasing difference'
WRITE(4,fmt=*) '  By # of children living near: 3 child families'
WRITE(4,fmt=*) '   0. 0 child'
WRITE(4,fmt=*) '   1. 1 child'
WRITE(4,fmt=*) '   2. 2 children'
WRITE(4,fmt=*) '   3. 3 children'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A8)') '    Type','   Total',' 0 child',' 1 child',' 2 child',' 3 child'
DO k=0,3
    WRITE(4,fmt='(I8,5F8.4)') k,(gtable14b(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-14c: The number of children with increasing difference'
WRITE(4,fmt=*) '  By # of children living near: 4 child families'
WRITE(4,fmt=*) '   0. 0 child'
WRITE(4,fmt=*) '   1. 1 child'
WRITE(4,fmt=*) '   2. 2 children'
WRITE(4,fmt=*) '   3. 3 children'
WRITE(4,fmt=*) '   4. 4 children'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(7A8)') '    Type','   Total',' 0 child',' 1 child',' 2 child',' 3 child',' 4 child'
DO k=0,4
    WRITE(4,fmt='(I8,6F8.4)') k,(gtable14c(k,m),m=1,6)
ENDDO
WRITE(4,fmt=*) ' '


WRITE(4,fmt=*) 'Table 1-15a: The number of normal form Nash Equilibria'
WRITE(4,fmt=*) '  By # of children living near: 2 child families'
WRITE(4,fmt=*) '   0. 0'
WRITE(4,fmt=*) '   1. 1'
WRITE(4,fmt=*) '   2. 2'
WRITE(4,fmt=*) '   3. 3'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(5A8)') '    Type','   Total',' 0 child',' 1 child',' 2 child'
DO k=0,3
    WRITE(4,fmt='(I8,4F8.4)') k,(gtable15a(k,m),m=1,4)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-15b: The number of normal form Nash Equilibria'
WRITE(4,fmt=*) '  By # of children living near: 3 child families'
WRITE(4,fmt=*) '   0. 0'
WRITE(4,fmt=*) '   1. 1'
WRITE(4,fmt=*) '   2. 2'
WRITE(4,fmt=*) '   3. 3'
WRITE(4,fmt=*) '   4. 4'
WRITE(4,fmt=*) '   5. 5'
WRITE(4,fmt=*) '   6. 6 or more'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A8)') '    Type','   Total',' 0 child',' 1 child',' 2 child',' 3 child'
DO k=0,6
    WRITE(4,fmt='(I8,5F8.4)') k,(gtable15b(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-15c: The number of normal form Nash Equilibria'
WRITE(4,fmt=*) '  By # of children living near: 4 child families'
WRITE(4,fmt=*) '   0. 0'
WRITE(4,fmt=*) '   1. 1'
WRITE(4,fmt=*) '   2. 2'
WRITE(4,fmt=*) '   3. 3'
WRITE(4,fmt=*) '   4. 4'
WRITE(4,fmt=*) '   5. 5'
WRITE(4,fmt=*) '   6. 6 or more'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(7A8)') '    Type','   Total',' 0 child',' 1 child',' 2 child',' 3 child',' 4 child'
DO k=0,6
    WRITE(4,fmt='(I8,6F8.4)') k,(gtable15c(k,m),m=1,6)
ENDDO
WRITE(4,fmt=*) ' '


WRITE(4,fmt=*) 'Table 1-16a: The number of children with increasing difference'
WRITE(4,fmt=*) '  By SPNE Efficiency type: 2 child families (Column share)'
WRITE(4,fmt=*) '   0. 0 child'
WRITE(4,fmt=*) '   1. 1 child'
WRITE(4,fmt=*) '   2. 2 children'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(5A10)') '     Type','     Total','   KH best','      P.O.','      P.D.'
DO k=0,2
    WRITE(4,fmt='(I10,4F10.4)') k,(gtable16a(k,m),m=1,4)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-16b: The number of children with increasing difference'
WRITE(4,fmt=*) '  By SPNE Efficiency type: 3 child families (Column share)'
WRITE(4,fmt=*) '   0. 0 child'
WRITE(4,fmt=*) '   1. 1 child'
WRITE(4,fmt=*) '   2. 2 children'
WRITE(4,fmt=*) '   3. 3 children'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(5A10)') '     Type','     Total','   KH best','      P.O.','      P.D.'
DO k=0,3
    WRITE(4,fmt='(I10,4F10.4)') k,(gtable16b(k,m),m=1,4)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-16c: The number of children with increasing difference'
WRITE(4,fmt=*) '  By SPNE Efficiency type: 4 child families (Column share)'
WRITE(4,fmt=*) '   0. 0 child'
WRITE(4,fmt=*) '   1. 1 child'
WRITE(4,fmt=*) '   2. 2 children'
WRITE(4,fmt=*) '   3. 3 children'
WRITE(4,fmt=*) '   4. 4 children'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(5A10)') '     Type','     Total','   KH best','      P.O.','      P.D.'
DO k=0,4
    WRITE(4,fmt='(I10,4F10.4)') k,(gtable16c(k,m),m=1,4)
ENDDO
WRITE(4,fmt=*) ' '


WRITE(4,fmt=*) 'Table 1-17a: The number of normal form Nash Equilibria'
WRITE(4,fmt=*) '  By SPNE Efficiency type: 2 child families (Column share)'
WRITE(4,fmt=*) '   0. 0'
WRITE(4,fmt=*) '   1. 1'
WRITE(4,fmt=*) '   2. 2'
WRITE(4,fmt=*) '   3. 3'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(5A10)') '     Type','     Total','   KH best','      P.O.','      P.D.'
DO k=0,3
    WRITE(4,fmt='(I10,4F10.4)') k,(gtable17a(k,m),m=1,4)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-17b: The number of normal form Nash Equilibria'
WRITE(4,fmt=*) '  By SPNE Efficiency type: 3 child families (Column share)'
WRITE(4,fmt=*) '   0. 0'
WRITE(4,fmt=*) '   1. 1'
WRITE(4,fmt=*) '   2. 2'
WRITE(4,fmt=*) '   3. 3'
WRITE(4,fmt=*) '   4. 4'
WRITE(4,fmt=*) '   5. 5'
WRITE(4,fmt=*) '   6. 6 or more'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(5A10)') '     Type','     Total','   KH best','      P.O.','      P.D.'
DO k=0,6
    WRITE(4,fmt='(I10,4F10.4)') k,(gtable17b(k,m),m=1,4)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-17c: The number of normal form Nash Equilibria'
WRITE(4,fmt=*) '  By SPNE Efficiency type: 4 child families (Column share)'
WRITE(4,fmt=*) '   0. 0'
WRITE(4,fmt=*) '   1. 1'
WRITE(4,fmt=*) '   2. 2'
WRITE(4,fmt=*) '   3. 3'
WRITE(4,fmt=*) '   4. 4'
WRITE(4,fmt=*) '   5. 5'
WRITE(4,fmt=*) '   6. 6 or more'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(5A10)') '     Type','     Total','   KH best','      P.O.','      P.D.'
DO k=0,6
    WRITE(4,fmt='(I10,4F10.4)') k,(gtable17c(k,m),m=1,4)
ENDDO
WRITE(4,fmt=*) ' '


WRITE(4,fmt=*) 'Table 1-18a: The number of normal form Nash Equilibria'
WRITE(4,fmt=*) '  By # of children with increasing difference: 2 child families'
WRITE(4,fmt=*) '   0. 0'
WRITE(4,fmt=*) '   1. 1'
WRITE(4,fmt=*) '   2. 2'
WRITE(4,fmt=*) '   3. 3'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(5A8)') '    Type','   Total',' 0 child',' 1 child',' 2 child'
DO k=0,3
    WRITE(4,fmt='(I8,4F8.4)') k,(gtable18a(k,m),m=1,4)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-18b: The number of normal form Nash Equilibria'
WRITE(4,fmt=*) '  By # of children with increasing difference: 3 child families'
WRITE(4,fmt=*) '   0. 0'
WRITE(4,fmt=*) '   1. 1'
WRITE(4,fmt=*) '   2. 2'
WRITE(4,fmt=*) '   3. 3'
WRITE(4,fmt=*) '   4. 4'
WRITE(4,fmt=*) '   5. 5'
WRITE(4,fmt=*) '   6. 6 or more'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A8)') '    Type','   Total',' 0 child',' 1 child',' 2 child',' 3 child'
DO k=0,6
    WRITE(4,fmt='(I8,5F8.4)') k,(gtable18b(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-18c: The number of normal form Nash Equilibria'
WRITE(4,fmt=*) '  By # of children with increasing difference: 4 child families'
WRITE(4,fmt=*) '   0. 0'
WRITE(4,fmt=*) '   1. 1'
WRITE(4,fmt=*) '   2. 2'
WRITE(4,fmt=*) '   3. 3'
WRITE(4,fmt=*) '   4. 4'
WRITE(4,fmt=*) '   5. 5'
WRITE(4,fmt=*) '   6. 6 or more'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(7A8)') '    Type','   Total',' 0 child',' 1 child',' 2 child',' 3 child',' 4 child'
DO k=0,6
    WRITE(4,fmt='(I8,6F8.4)') k,(gtable18c(k,m),m=1,6)
ENDDO
WRITE(4,fmt=*) ' '



WRITE(4,fmt=*) 'Table 1-19a: SPNE (observed) location pattern'
WRITE(4,fmt=*) '  By # of children with increasing difference: 2 child families'
WRITE(4,fmt=*) '   0. 0, 0 (no one near)'
WRITE(4,fmt=*) '   1. 0, 1 (only the youngest near)'
WRITE(4,fmt=*) '   2. 1, 0 (only the second youngest near)'
WRITE(4,fmt=*) '   3. 1, 1'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(5A8)') '    Type','   Total',' 0 child',' 1 child',' 2 child'
DO k=0,3
    WRITE(4,fmt='(I8,4F8.4)') k,(gtable19a(k,m),m=1,4)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-19b: SPNE (observed) location pattern'
WRITE(4,fmt=*) '  By # of children with increasing difference: 3 child families'
WRITE(4,fmt=*) '   0. 0, 0, 0 (no one near)'
WRITE(4,fmt=*) '   1. 0, 0, 1 (only the youngest near)'
WRITE(4,fmt=*) '   2. 0, 1, 0 (only the second youngest near)'
WRITE(4,fmt=*) '   3. 0, 1, 1'
WRITE(4,fmt=*) '   4. 1, 0, 0'
WRITE(4,fmt=*) '   5. 1, 0, 1'
WRITE(4,fmt=*) '   6. 1, 1, 0'
WRITE(4,fmt=*) '   7. 1, 1, 1'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A8)') '    Type','   Total',' 0 child',' 1 child',' 2 child',' 3 child'
DO k=0,7
    WRITE(4,fmt='(I8,5F8.4)') k,(gtable19b(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-19c: SPNE (observed) location pattern'
WRITE(4,fmt=*) '  By # of children with increasing difference: 4 child families'
WRITE(4,fmt=*) '   0. 0, 0, 0, 0 (no one near)'
WRITE(4,fmt=*) '   1. 0, 0, 0, 1 (only the youngest near)'
WRITE(4,fmt=*) '   2. 0, 0, 1, 0 (only the second youngest near)'
WRITE(4,fmt=*) '   3. 0, 0, 1, 1'
WRITE(4,fmt=*) '   4. 0, 1, 0, 0'
WRITE(4,fmt=*) '   5. 0, 1, 0, 1'
WRITE(4,fmt=*) '   6. 0, 1, 1, 0'
WRITE(4,fmt=*) '   7. 0, 1, 1, 1'
WRITE(4,fmt=*) '   8. 1, 0, 0, 0'
WRITE(4,fmt=*) '   9. 1, 0, 0, 1'
WRITE(4,fmt=*) '  10. 1, 0, 1, 0'
WRITE(4,fmt=*) '  11. 1, 0, 1, 1'
WRITE(4,fmt=*) '  12. 1, 1, 0, 0'
WRITE(4,fmt=*) '  13. 1, 1, 0, 1'
WRITE(4,fmt=*) '  14. 1, 1, 1, 0'
WRITE(4,fmt=*) '  15. 1, 1, 1, 1'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(7A8)') '    Type','   Total',' 0 child',' 1 child',' 2 child',' 3 child',' 4 child'
DO k=0,15
    WRITE(4,fmt='(I8,6F8.4)') k,(gtable19c(k,m),m=1,6)
ENDDO
WRITE(4,fmt=*) ' '





WRITE(4,fmt=*) 'Table 1-20a: Efficiency type'
WRITE(4,fmt=*) '  By # of children with increasing difference: 2 child families'
WRITE(4,fmt=*) '   1. Joint utility best (KH efficient)'
WRITE(4,fmt=*) '   2. KH inefficient, not Pareto dominated'
WRITE(4,fmt=*) '   3. Pareto Dominated (P.D.)'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(5A8)') '    Type','   Total',' 0 child',' 1 child',' 2 child'
DO k=1,3
    WRITE(4,fmt='(I8,4F8.4)') k,(gtable20a(k,m),m=1,4)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-20b: Efficiency type'
WRITE(4,fmt=*) '  By # of children with increasing difference: 3 child families'
WRITE(4,fmt=*) '   1. Joint utility best (KH efficient)'
WRITE(4,fmt=*) '   2. KH inefficient, not Pareto dominated'
WRITE(4,fmt=*) '   3. Pareto Dominated (P.D.)'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A8)') '    Type','   Total',' 0 child',' 1 child',' 2 child',' 3 child'
DO k=1,3
    WRITE(4,fmt='(I8,5F8.4)') k,(gtable20b(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-20c: Efficiency type'
WRITE(4,fmt=*) '  By # of children with increasing difference: 4 child families'
WRITE(4,fmt=*) '   1. Joint utility best (KH efficient)'
WRITE(4,fmt=*) '   2. KH inefficient, not Pareto dominated'
WRITE(4,fmt=*) '   3. Pareto Dominated (P.D.)'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(7A8)') '    Type','   Total',' 0 child',' 1 child',' 2 child',' 3 child',' 4 child'
DO k=1,3
    WRITE(4,fmt='(I8,6F8.4)') k,(gtable20c(k,m),m=1,6)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-21a: Location pattern of the joint optimal configuration'
WRITE(4,fmt=*) '  By # of children with increasing difference: 2 child families'
WRITE(4,fmt=*) '   0. 0, 0 (no one near)'
WRITE(4,fmt=*) '   1. 0, 1 (only the youngest near)'
WRITE(4,fmt=*) '   2. 1, 0 (only the second youngest near)'
WRITE(4,fmt=*) '   3. 1, 1'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(5A8)') '    Type','   Total',' 0 child',' 1 child',' 2 child'
DO k=0,3
    WRITE(4,fmt='(I8,4F8.4)') k,(gtable21a(k,m),m=1,4)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-21b: Location pattern of the joint optimal configuration'
WRITE(4,fmt=*) '  By # of children with increasing difference: 3 child families'
WRITE(4,fmt=*) '   0. 0, 0, 0 (no one near)'
WRITE(4,fmt=*) '   1. 0, 0, 1 (only the youngest near)'
WRITE(4,fmt=*) '   2. 0, 1, 0 (only the second youngest near)'
WRITE(4,fmt=*) '   3. 0, 1, 1'
WRITE(4,fmt=*) '   4. 1, 0, 0'
WRITE(4,fmt=*) '   5. 1, 0, 1'
WRITE(4,fmt=*) '   6. 1, 1, 0'
WRITE(4,fmt=*) '   7. 1, 1, 1'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A8)') '    Type','   Total',' 0 child',' 1 child',' 2 child',' 3 child'
DO k=0,7
    WRITE(4,fmt='(I8,5F8.4)') k,(gtable21b(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-21c: Location pattern of the joint optimal configuration'
WRITE(4,fmt=*) '  By # of children with increasing difference: 4 child families'
WRITE(4,fmt=*) '   0. 0, 0, 0, 0 (no one near)'
WRITE(4,fmt=*) '   1. 0, 0, 0, 1 (only the youngest near)'
WRITE(4,fmt=*) '   2. 0, 0, 1, 0 (only the second youngest near)'
WRITE(4,fmt=*) '   3. 0, 0, 1, 1'
WRITE(4,fmt=*) '   4. 0, 1, 0, 0'
WRITE(4,fmt=*) '   5. 0, 1, 0, 1'
WRITE(4,fmt=*) '   6. 0, 1, 1, 0'
WRITE(4,fmt=*) '   7. 0, 1, 1, 1'
WRITE(4,fmt=*) '   8. 1, 0, 0, 0'
WRITE(4,fmt=*) '   9. 1, 0, 0, 1'
WRITE(4,fmt=*) '  10. 1, 0, 1, 0'
WRITE(4,fmt=*) '  11. 1, 0, 1, 1'
WRITE(4,fmt=*) '  12. 1, 1, 0, 0'
WRITE(4,fmt=*) '  13. 1, 1, 0, 1'
WRITE(4,fmt=*) '  14. 1, 1, 1, 0'
WRITE(4,fmt=*) '  15. 1, 1, 1, 1'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(7A8)') '    Type','   Total',' 0 child',' 1 child',' 2 child',' 3 child',' 4 child'
DO k=0,15
    WRITE(4,fmt='(I8,6F8.4)') k,(gtable21c(k,m),m=1,6)
ENDDO
WRITE(4,fmt=*) ' '


WRITE(4,fmt=*) 'Table 1-22a: The number of children living near'
WRITE(4,fmt=*) '  By # of child near in first best: 2 child families'
WRITE(4,fmt=*) '   0. 0'
WRITE(4,fmt=*) '   1. 1'
WRITE(4,fmt=*) '   2. 2'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(5A8)') '    Type','   Total',' 0 child',' 1 child',' 2 child'
DO k=0,2
    WRITE(4,fmt='(I8,4F8.4)') k,(gtable22a(k,m),m=1,4)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-22b: The number of children living near'
WRITE(4,fmt=*) '  By # of child near in first best: 3 child families'
WRITE(4,fmt=*) '   0. 0'
WRITE(4,fmt=*) '   1. 1'
WRITE(4,fmt=*) '   2. 2'
WRITE(4,fmt=*) '   3. 3'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A8)') '    Type','   Total',' 0 child',' 1 child',' 2 child',' 3 child'
DO k=0,3
    WRITE(4,fmt='(I8,5F8.4)') k,(gtable22b(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-22c: The number of children living near'
WRITE(4,fmt=*) '  By # of child near in first best: 4 child families'
WRITE(4,fmt=*) '   0. 0'
WRITE(4,fmt=*) '   1. 1'
WRITE(4,fmt=*) '   2. 2'
WRITE(4,fmt=*) '   3. 3'
WRITE(4,fmt=*) '   4. 4'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(7A8)') '    Type','   Total',' 0 child',' 1 child',' 2 child',' 3 child',' 4 child'
DO k=0,4
    WRITE(4,fmt='(I8,6F8.4)') k,(gtable22c(k,m),m=1,6)
ENDDO
WRITE(4,fmt=*) ' '


WRITE(4,fmt=*) 'Table 1-23a: Efficiency type'
WRITE(4,fmt=*) '  By # of child near in first best: 2 child families'
WRITE(4,fmt=*) '   1. Joint utility best (KH efficient)'
WRITE(4,fmt=*) '   2. KH inefficient, not Pareto dominated'
WRITE(4,fmt=*) '   3. Pareto Dominated (P.D.)'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(5A8)') '    Type','   Total',' 0 child',' 1 child',' 2 child'
DO k=1,3
    WRITE(4,fmt='(I8,4F8.4)') k,(gtable23a(k,m),m=1,4)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-23b: Efficiency type'
WRITE(4,fmt=*) '  By # of child near in first best: 2 child families'
WRITE(4,fmt=*) '   1. Joint utility best (KH efficient)'
WRITE(4,fmt=*) '   2. KH inefficient, not Pareto dominated'
WRITE(4,fmt=*) '   3. Pareto Dominated (P.D.)'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A8)') '    Type','   Total',' 0 child',' 1 child',' 2 child',' 3 child'
DO k=1,3
    WRITE(4,fmt='(I8,5F8.4)') k,(gtable23b(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-23c: Efficiency type'
WRITE(4,fmt=*) '  By # of child near in first best: 2 child families'
WRITE(4,fmt=*) '   1. Joint utility best (KH efficient)'
WRITE(4,fmt=*) '   2. KH inefficient, not Pareto dominated'
WRITE(4,fmt=*) '   3. Pareto Dominated (P.D.)'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(7A8)') '    Type','   Total',' 0 child',' 1 child',' 2 child',' 3 child',' 4 child'
DO k=1,3
    WRITE(4,fmt='(I8,6F8.4)') k,(gtable23c(k,m),m=1,6)
ENDDO
WRITE(4,fmt=*) ' '


WRITE(4,fmt=*) 'Table 1-24a: The number of normal form Nash Equilibria'
WRITE(4,fmt=*) '  By # of child near in first best: 2 child families'
WRITE(4,fmt=*) '   0. 0'
WRITE(4,fmt=*) '   1. 1'
WRITE(4,fmt=*) '   2. 2'
WRITE(4,fmt=*) '   3. 3'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(5A8)') '    Type','   Total',' 0 child',' 1 child',' 2 child'
DO k=0,3
    WRITE(4,fmt='(I8,4F8.4)') k,(gtable24a(k,m),m=1,4)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-24b: The number of normal form Nash Equilibria'
WRITE(4,fmt=*) '  By # of child near in first best: 2 child families'
WRITE(4,fmt=*) '   0. 0'
WRITE(4,fmt=*) '   1. 1'
WRITE(4,fmt=*) '   2. 2'
WRITE(4,fmt=*) '   3. 3'
WRITE(4,fmt=*) '   4. 4'
WRITE(4,fmt=*) '   5. 5'
WRITE(4,fmt=*) '   6. 6 or more'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A8)') '    Type','   Total',' 0 child',' 1 child',' 2 child',' 3 child'
DO k=0,6
    WRITE(4,fmt='(I8,5F8.4)') k,(gtable24b(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 1-24c: The number of normal form Nash Equilibria'
WRITE(4,fmt=*) '  By # of child near in first best: 2 child families'
WRITE(4,fmt=*) '   0. 0'
WRITE(4,fmt=*) '   1. 1'
WRITE(4,fmt=*) '   2. 2'
WRITE(4,fmt=*) '   3. 3'
WRITE(4,fmt=*) '   4. 4'
WRITE(4,fmt=*) '   5. 5'
WRITE(4,fmt=*) '   6. 6 or more'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(7A8)') '    Type','   Total',' 0 child',' 1 child',' 2 child',' 3 child',' 4 child'
DO k=0,6
    WRITE(4,fmt='(I8,6F8.4)') k,(gtable24c(k,m),m=1,6)
ENDDO
WRITE(4,fmt=*) ' '




CLOSE(4)



END PROGRAM MAIN
