MODULE Globvar
IMPLICIT NONE
! This module declares the variables that are going to be used throughout the program.
! To avoid stack overflow error --> Stack Reserve: 70,000,000 (121222)

INTEGER, PARAMETER :: nchildmax=4  ! max number of children
INTEGER, PARAMETER :: sm2=2**(1+1)  ! necessary amount for strategy patterns
INTEGER, PARAMETER :: sm3=2**(1+1)*2**(2+1)  ! necessary amount for strategy patterns
INTEGER, PARAMETER :: sm4=2**(1+1)*2**(2+1)*2**(3+1)  ! necessary amount for strategy patterns
	! The 4th child in a 4-child family has 2**(3+1) strategies, because 
	! the number of possible elder siblings in the upstream from 0 to 3, i.e. 4.
INTEGER :: nobs
INTEGER :: nfam
INTEGER :: ns              ! number of simulation draws
INTEGER :: ntheta
INTEGER :: nthetab     ! # of parameters in beta term
INTEGER :: nthetaa     ! # of parameters in alpha term
INTEGER :: nthetag     ! # of parameters in gamma term
INTEGER :: nthetar     ! # of parameters in rho term
REAL(4) :: amp1         ! alpha multiplier for counterfactual simulation
REAL(4) :: amp2         ! gamma multiplier for counterfactual simulation
INTEGER, ALLOCATABLE :: famid(:)   ! famid(nobs)
INTEGER, ALLOCATABLE :: nchild(:), famindex(:)   ! nchild(nfam), famindex(nfam)
REAL(8), ALLOCATABLE :: theta(:) ! theta(ntheta)
REAL(8), ALLOCATABLE :: xb(:,:)   ! xb(nthetab,nobs)
REAL(8), ALLOCATABLE :: xa(:,:)   ! xa(nthetaa,nobs)
REAL(8), ALLOCATABLE :: xg(:,:)   ! xg(nthetag,nobs)
INTEGER :: sa, sg		! actual size of xa and xg
REAL(8), ALLOCATABLE :: xr(:,:,:)   ! xr(nthetar,nchildmax,nobs)
INTEGER, ALLOCATABLE :: y(:)     ! y(nobs)
INTEGER, ALLOCATABLE :: near(:)   ! near(nfam)
REAL(8), ALLOCATABLE :: near_cf(:,:)   ! near_cf(nfam,0:15) ... proportions of living arrangement types for counterfactual simulation
REAL(8), ALLOCATABLE :: utilxb(:) ! utilxb(nobs)

REAL(8), ALLOCATABLE :: draws2(:,:,:)		! nchild=2. draws(ns,necessary amount,2)
REAL(8), ALLOCATABLE :: draws3(:,:,:)		! nchild=3. draws(ns,necessary amount,3)
REAL(8), ALLOCATABLE :: draws4(:,:,:)		! nchild=4. draws(ns,necessary amount,4)
REAL(8), ALLOCATABLE :: draws2seed(:,:,:)	! nchild=2. draws(ns,necessary amount,2)
REAL(8), ALLOCATABLE :: draws3seed(:,:,:)	! nchild=3. draws(ns,necessary amount,3)
REAL(8), ALLOCATABLE :: draws4seed(:,:,:)	! nchild=4. draws(ns,necessary amount,4)

REAL(8), ALLOCATABLE :: exputil(:)   ! exputil(nobs) ... expected utility
REAL(8), ALLOCATABLE :: dst(:)   ! dst(nobs) ... whether a chidl is playing a daminant strategy (in a simultaneous move game sense)
                                    !!! This is not whether the child has a dominant strategy. A child who has a dominant strategy may not be playing it in SPNE.
REAL(8), ALLOCATABLE :: famutil(:) ! famutil(nfam)
REAL(8), ALLOCATABLE :: fbfamutil(:) !  fbfamutil(nfam)
REAL(8), ALLOCATABLE :: pbfamutil(:) !  pbfamutil(nfam)
REAL(8), ALLOCATABLE :: y_any_fb(:) !  y_any_fb(nfam) ... probability that a parent has anyone near in the fb outcome
REAL(8), ALLOCATABLE :: gtype1a(:,:,:,:) !  gtype1a(nfam,0:15,1:6,0:15) ... proportions of game type patterns (1-4 child families)
REAL(8), ALLOCATABLE :: gtype2a(:,:,:,:,:) !  gtype2a(nfam,0:16,0:11,1:15,0:3) ... proportions of game type patterns (2 child families)
REAL(8), ALLOCATABLE :: gtype2b(:,:,:,:)   !  gtype2b(nfam,0:15,0:11,0:3) ... proportions of game type patterns (2 child families)
REAL(8), ALLOCATABLE :: gtype2c(:,:,:,:,:) !  gtype2c(nfam,0:5,0:6,0:11,0:3) ... proportions of game type patterns (2 child families)


CHARACTER(10) :: path                          ! Specify the data directory. To be read from setting.prn

END MODULE Globvar






MODULE INIT
	USE Globvar
	IMPLICIT NONE

CONTAINS

	SUBROUTINE READDATA(nobs)
		USE Globvar
		IMPLICIT NONE
		INTEGER, INTENT(IN) :: nobs
		INTEGER :: i,j,k,m,io

        CHARACTER(60) :: readfile_x      ! File names to be read

readfile_x='data.raw'

xb=0.0d0; y=0.0d0
xb(1,:)=1.00000000d0
xa(1,:)=1.0000000000d0
xr(1,:,:)=1.00000000d0
IF (nthetag>=1) xg(1,:)=1.00000000d0

print *,'Reading Data ... '

	OPEN(1,file=(readfile_x),status='old')

    IF (nthetaa>3 .and. nthetar>1 .and. nthetag>2) THEN
        DO i = 1, nobs
	    READ(1,fmt=*,iostat=io) y(i), famid(i), xb(2:,i), xa(2:,i), xg(2:,i), xr(2:nthetar,1,i), xr(2:nthetar,2,i), xr(2:nthetar,3,i), xr(2:nthetar,4,i)
        IF (io<0) EXIT
        END DO
    ELSE IF (nthetaa>3 .and. nthetar>1) THEN
        DO i = 1, nobs
	    READ(1,fmt=*,iostat=io) y(i), famid(i), xb(2:,i), xa(2:,i), xr(2:nthetar,1,i), xr(2:nthetar,2,i), xr(2:nthetar,3,i), xr(2:nthetar,4,i)
        IF (io<0) EXIT
        END DO
    ELSE IF (nthetaa>1) THEN
        DO i = 1, nobs
	    READ(1,fmt=*,iostat=io) y(i), famid(i), xb(2:,i), xa(2:,i)
        IF (io<0) EXIT
        END DO
    ELSE
        DO i = 1, nobs
	    READ(1,fmt=*,iostat=io) y(i), famid(i), xb(2:,i)
        IF (io<0) EXIT
        END DO
    ENDIF
    CLOSE(1)

nchild=0
j=1
nchild(1)=1
DO i=2,nobs
	IF (famid(i)/=famid(i-1)) THEN
		j=j+1
	ENDIF
		nchild(j)=nchild(j)+1
ENDDO

famindex(1)=nchild(1)
DO i=2,nfam
	famindex(i)=famindex(i-1)+nchild(i)
ENDDO


DO m=1,nfam
near(m)=0
DO i=1,nchild(m)
near(m)=near(m)*2
near(m)=near(m)+y(famindex(m)-nchild(m)+i)
END DO
END DO



print *,'Reading Data ... Finished '

	END SUBROUTINE READDATA


END MODULE INIT

MODULE Object
	USE Globvar
	USE GHK
	USE MATRIX
    USE PROBABILITY
	IMPLICIT NONE

CONTAINS


FUNCTION SIMULATION(param,n) RESULT(lnmlobj)	! This subroutine conducts simulations
	USE Globvar
	IMPLICIT NONE
	INTEGER, INTENT(IN) :: n 
	REAL(8), INTENT(IN) :: param(n)
	REAL(8) :: mlobj
	REAL(8) :: lnmlobj
	REAL(8) :: famprob(nfam)			! The family-level likelihood
	INTEGER :: i, j, k, m
	REAL(4) :: tms

	print *,'SUBROUTINE SIMULATION ... In '

	tms = SECNDS(0.0)

		exputil=0.0d0
		dst=0.0d0
		famutil=0.0d0
		fbfamutil=0.0d0
		pbfamutil=0.0d0
		y_any_fb=0.0d0
		gtype1a=0.0d0
		gtype2a=0.0d0
		gtype2b=0.0d0
		gtype2c=0.0d0
		near_cf=0.0d0



draws2=draws2seed
draws3=draws3seed
draws4=draws4seed

DO m = 1, nfam

IF (MOD(m,500)==0) THEN
    PRINT *,' ... current family # / nfam: ', m, '/',nfam
ENDIF

SELECT CASE(nchild(m))
CASE(2)
    draws2=CSHIFT(draws2,1,2)
CASE(3)
    draws3=CSHIFT(draws3,7,2)
CASE(4)
    draws4=CSHIFT(draws4,9,2)
END SELECT


	SELECT CASE(nchild(m))
	CASE(1)
		famprob(m)=DLOG(SIM1(m,param,n))
	CASE(2)
		famprob(m)=DLOG(SIM2(m,param,n))
	CASE(3)
		famprob(m)=DLOG(SIM3(m,param,n))
	CASE(4)
		famprob(m)=DLOG(SIM4(m,param,n))
	CASE DEFAULT
		PRINT *,'nchild(m): ', nchild(m)
		PRINT *,'error'
		PAUSE
	END SELECT

END DO


	lnmlobj=(0-1)*SUM(famprob)
	print *,'ln ML objective: ',-lnmlobj
	print *,'Time spent for Simulation: ',SECNDS(tms)

END FUNCTION SIMULATION




FUNCTION SIM1(m,param,n) RESULT(ghkval)
	! This function calculates the GHK simulator for each market*year
USE Globvar
USE PROBABILITY
IMPLICIT NONE
INTEGER, INTENT(IN) :: m			! market id
INTEGER, INTENT(IN) :: n 
REAL(8), INTENT(IN) :: param(n)
REAL(8) :: v(ns)   ! GHK simulation draws constructed from draws()
INTEGER :: i, j, k, l
REAL(8) :: ghkval		! The calculated GHK value
REAL(8) :: utilsub, alpha

ghkval=0.000d0

k=famindex(m)

	utilsub=DOT_PRODUCT(param(1:nthetab),xb(:,k))
    IF(nthetaa>=3) THEN
    alpha  = EXP( DOT_PRODUCT(param(nthetab+1:nthetab+nthetaa-2),xa(:,k)) )
    ELSEIF(nthetaa==1) THEN
    alpha  =param(nthetab+nthetaa)
    ELSE
    alpha = 0.0d0
    ENDIF

	IF (IBITS(near(m),0,1)==1) THEN
    ghkval=1.0d0-NORMCDF(-utilsub-alpha,0.0d0,1.0d0)

    DO j=1, ns
        v(j)=INVNORMCDF((NORMCDF(-utilsub-alpha,0.0d0,1.0d0)*(1-draws2(j,1,1))+draws2(j,1,1)),0.0d0,1.0d0)
        IF (v(j)+utilsub+alpha*amp1 > 0.0d0) THEN
            near_cf(m,1)=near_cf(m,1)+1.0d0
            gtype1a(m,1,1,1)=gtype1a(m,1,1,1)+1.0d0
            exputil(k)=exputil(k)+(v(j)+utilsub+alpha*amp1)
            y_any_fb(m)=y_any_fb(m)+1.0d0
        ELSE
            near_cf(m,0)=near_cf(m,0)+1.0d0
            gtype1a(m,0,1,0)=gtype1a(m,0,1,0)+1.0d0
        ENDIF
    END DO

	ELSE
    ghkval=NORMCDF(-utilsub-alpha,0.0d0,1.0d0)
    DO j=1, ns
        v(j)=INVNORMCDF((NORMCDF(-utilsub-alpha,0.0d0,1.0d0)*draws2(j,1,1)),0.0d0,1.0d0)
        IF (v(j)+utilsub+alpha*amp1 > 0.0d0) THEN
            near_cf(m,1)=near_cf(m,1)+1.0d0
            gtype1a(m,1,1,1)=gtype1a(m,1,1,1)+1.0d0
            exputil(k)=exputil(k)+(v(j)+utilsub+alpha*amp1)
            y_any_fb(m)=y_any_fb(m)+1.0d0
        ELSE
            near_cf(m,0)=near_cf(m,0)+1.0d0
            gtype1a(m,0,1,0)=gtype1a(m,0,1,0)+1.0d0
        ENDIF
    END DO

    ENDIF

    exputil(k)=exputil(k)/ns
    famutil(m)=exputil(k)
    fbfamutil(m)=exputil(k)
    pbfamutil(m)=exputil(k)
    gtype1a(m,:,:,:)=gtype1a(m,:,:,:)/ns
    near_cf(m,:)=near_cf(m,:)/ns
    y_any_fb(m)=y_any_fb(m)/ns


dst(k)=1.0d0

END FUNCTION SIM1





FUNCTION SIM2(m,param,n) RESULT(ghkval)
	! This function calculates the GHK simulator for each market*year
USE Globvar
USE PROBABILITY
IMPLICIT NONE
INTEGER, INTENT(IN) :: m			! market id
INTEGER, INTENT(IN) :: n 
REAL(8), INTENT(IN) :: param(n)
INTEGER :: i, j, k, l
INTEGER :: s1, s2, s3, s4
REAL(8) :: wgv(ns,0:sm2-1,0:2)		! wgv(ns, # strategy combitnations, 1 + # child) ... wgv(ns,s,0) for wg.
REAL(8) :: wg(ns,0:sm2-1)
REAL(8) :: ta(ns), tb(ns)
REAL(8) :: v(ns,0:sm2-1,nchild(m))   ! GHK simulation draws constructed from draws()
REAL(8) :: gtemp1a(0:15,1:6,0:15)
REAL(8) :: gtemp2a(0:16,0:11,1:15,0:3)
REAL(8) :: gtemp2b(0:15,0:11,0:3)
REAL(8) :: gtemp2c(0:5,0:6,0:11,0:3)
REAL(8) :: nearcftemp(0:15)
REAL(8) :: utiltemp
REAL(8) :: utemp(nchild(m)),dstemp(nchild(m))
REAL(8) :: fbutemp, pbutemp, fbytemp
REAL(8) :: ghkval, ghkvals		! The calculated GHK value
REAL(8) :: utilsub(2), alpha(2,3), gamma(2,3)		! 1 ... oldest child
REAL(8) :: lchol(2,2)

wgv=0.000d0
wg=0.000d0
v=0.000d0
ghkval=0.000d0

k=famindex(m)

utemp=0.000d0
fbutemp=0.000d0 ; pbutemp=0.000d0 ; fbytemp=0.000d0

lchol=lcholbase(m,param,n)
wgv=GHK2(m,param,n,ns,draws2,lchol,nthetab, nthetaa, nthetag, xb(:,famindex(m)-1:famindex(m)), xa(:,famindex(m)-1:famindex(m)), xg(:,famindex(m)-1:famindex(m)), sa, sg, near(m),2)

	utilsub(1:2)=MATMUL(param(1:nthetab),xb(:,k-1:k))
	IF (nthetaa>=3) THEN
	alpha(1:2,1) = EXP( MATMUL(param(nthetab+1:nthetab+nthetaa-2),xa(:,k-1:k)) )
	alpha(1:2,2) = EXP( MATMUL(param(nthetab+1:nthetab+nthetaa-2),xa(:,k-1:k)) + param(nthetab+nthetaa-1) )
	alpha(1:2,3) = EXP( MATMUL(param(nthetab+1:nthetab+nthetaa-2),xa(:,k-1:k)) + param(nthetab+nthetaa-1) + param(nthetab+nthetaa) )
	ELSEIF (nthetaa==1) THEN
	alpha(1:2,1:3)=param(nthetab+nthetaa)
	ELSE
	alpha=0.0d0
	ENDIF
	IF (nthetag>=2) THEN
	gamma(1:2,1) =MATMUL(param(nthetab+nthetaa+1:nthetab+nthetaa+nthetag-1),xg(:,k-1:k))
	gamma(1:2,2) =MATMUL(param(nthetab+nthetaa+1:nthetab+nthetaa+nthetag-1),xg(:,k-1:k))*(1.0d0 + param(nthetab+nthetaa+nthetag))
	gamma(1:2,3) =MATMUL(param(nthetab+nthetaa+1:nthetab+nthetaa+nthetag-1),xg(:,k-1:k))*(1.0d0 + param(nthetab+nthetaa+nthetag)*2.0d0)
	ELSEIF (nthetag==1) THEN
	gamma(1:2,1:3)=param(nthetab+nthetaa+nthetag)
	ELSE
	gamma=0.0d0
	ENDIF


alpha(:,:)=alpha(:,:)*amp1
gamma(:,:)=gamma(:,:)*amp2

DO s1=0,2**2-1       ! Strategy loop

IF (ghkcheck(s1,2,m)==1 .and. wgv(1,s1,0)>0.0d0) THEN
    CALL geval(gtemp1a(:,:,:),gtemp2a(:,:,:,:),gtemp2b(:,:,:),gtemp2c(:,:,:,:),nearcftemp(:),utemp,pbutemp,fbutemp,fbytemp,dstemp,m,wgv(:,s1,1:2),alpha,gamma,lchol)
ENDIF

ghkvals=SUM(wgv(:,s1,0))/ns
ghkval=ghkval+ghkvals
exputil(k-1:k)=exputil(k-1:k)+utemp(1:2)*ghkvals
dst(k-1:k)=dst(k-1:k)+dstemp(1:2)*ghkvals
fbfamutil(m)=fbfamutil(m)+fbutemp*ghkvals
pbfamutil(m)=pbfamutil(m)+pbutemp*ghkvals
y_any_fb(m)=y_any_fb(m)+fbytemp*ghkvals
gtype1a(m,:,:,:)=gtype1a(m,:,:,:)+gtemp1a(:,:,:)*ghkvals
gtype2a(m,:,:,:,:)=gtype2a(m,:,:,:,:)+gtemp2a(:,:,:,:)*ghkvals
gtype2b(m,:,:,:)=gtype2b(m,:,:,:)+gtemp2b(:,:,:)*ghkvals
gtype2c(m,:,:,:,:)=gtype2c(m,:,:,:,:)+gtemp2c(:,:,:,:)*ghkvals
near_cf(m,:)=near_cf(m,:)+nearcftemp(:)*ghkvals

END DO         ! Strategy loop

exputil(k-1:k)=exputil(k-1:k)/ghkval
dst(k-1:k)=dst(k-1:k)/ghkval
famutil(m)=SUM(exputil(k-1:k))
fbfamutil(m)=fbfamutil(m)/ghkval
pbfamutil(m)=pbfamutil(m)/ghkval
y_any_fb(m)=y_any_fb(m)/ghkval
gtype1a(m,:,:,:)  =gtype1a(m,:,:,:)/ghkval
gtype2a(m,:,:,:,:)=gtype2a(m,:,:,:,:)/ghkval
gtype2b(m,:,:,:)  =gtype2b(m,:,:,:)/ghkval
gtype2c(m,:,:,:,:)=gtype2c(m,:,:,:,:)/ghkval
near_cf(m,:)=near_cf(m,:)/ghkval


END FUNCTION SIM2

FUNCTION SIM3(m,param,n) RESULT(ghkval)
	! This function calculates the GHK simulator for each market*year
USE Globvar
USE PROBABILITY
IMPLICIT NONE
INTEGER, INTENT(IN) :: m			! market id
INTEGER, INTENT(IN) :: n 
REAL(8), INTENT(IN) :: param(n)
INTEGER :: i, j, k, l
INTEGER :: s1, s2, s3, s4
REAL(8) :: wgv(ns,0:sm3-1,0:3)		! wgv(ns, # strategy combitnations, 1 + # child) ... wgv(ns,s,0) for wg.
REAL(8) :: wg(ns,0:sm3-1)
REAL(8) :: ta(ns), tb(ns)
REAL(8) :: v(ns,0:sm3-1,3)   ! GHK simulation draws constructed from draws()
REAL(8) :: gtemp1a(0:15,1:6,0:15)
REAL(8) :: gtemp2a(0:16,0:11,1:15,0:3)
REAL(8) :: gtemp2b(0:15,0:11,0:3)
REAL(8) :: gtemp2c(0:5,0:6,0:11,0:3)
REAL(8) :: nearcftemp(0:15)
REAL(8) :: utiltemp
REAL(8) :: utemp(nchild(m)), dstemp(nchild(m))
REAL(8) :: fbutemp, pbutemp, fbytemp
REAL(8) :: ghkval, ghkvals		! The calculated GHK value
REAL(8) :: utilsub(3), alpha(3,3), gamma(3,3)		! 1 ... oldest child
REAL(8) :: lchol(3,3)

wgv=0.000d0
wg=0.000d0
v=0.000d0
ghkval=0.000d0

k=famindex(m)

utemp=0.000d0
fbutemp=0.000d0 ; pbutemp=0.000d0 ; fbytemp=0.000d0

lchol=lcholbase(m,param,n)
wgv=GHK3(m,param,n,ns,draws3,lchol,nthetab, nthetaa, nthetag, xb(:,famindex(m)-2:famindex(m)), xa(:,famindex(m)-2:famindex(m)), xg(:,famindex(m)-2:famindex(m)), sa, sg, near(m),2)

utilsub(1:3)=MATMUL(param(1:nthetab),xb(:,k-2:k))
	IF (nthetaa>=3) THEN
	alpha(1:3,1) = EXP( MATMUL(param(nthetab+1:nthetab+nthetaa-2),xa(:,k-2:k)) )
	alpha(1:3,2) = EXP( MATMUL(param(nthetab+1:nthetab+nthetaa-2),xa(:,k-2:k)) + param(nthetab+nthetaa-1) )
	alpha(1:3,3) = EXP( MATMUL(param(nthetab+1:nthetab+nthetaa-2),xa(:,k-2:k)) + param(nthetab+nthetaa-1) + param(nthetab+nthetaa) )
	ELSEIF (nthetaa==1) THEN
	alpha(1:3,1:3)=param(nthetab+nthetaa)
	ELSE
	alpha=0.0d0
	ENDIF
	IF (nthetag>=2) THEN
	gamma(1:3,1) =MATMUL(param(nthetab+nthetaa+1:nthetab+nthetaa+nthetag-1),xg(:,k-2:k))
	gamma(1:3,2) =MATMUL(param(nthetab+nthetaa+1:nthetab+nthetaa+nthetag-1),xg(:,k-2:k))*(1.0d0 + param(nthetab+nthetaa+nthetag))
	gamma(1:3,3) =MATMUL(param(nthetab+nthetaa+1:nthetab+nthetaa+nthetag-1),xg(:,k-2:k))*(1.0d0 + param(nthetab+nthetaa+nthetag)*2.0d0)
	ELSEIF (nthetag==1) THEN
	gamma(1:3,1:3)=param(nthetab+nthetaa+nthetag)
	ELSE
	gamma=0.0d0
	ENDIF


alpha(:,:)=alpha(:,:)*amp1
gamma(:,:)=gamma(:,:)*amp2

DO s1=0,2**2-1       ! Strategy loop
DO s2=0,2**3-1

IF (ghkcheck(s1,2,m)==1 .and. ghkcheck(s2,3,m)==1 .and. wgv(1,s1*8+s2,0)>0.0d0) THEN
    CALL geval(gtemp1a(:,:,:),gtemp2a(:,:,:,:),gtemp2b(:,:,:),gtemp2c(:,:,:,:),nearcftemp(:),utemp,pbutemp,fbutemp,fbytemp,dstemp,m,wgv(:,s1*8+s2,1:3),alpha,gamma,lchol)
ENDIF


ghkvals=SUM(wgv(:,s1*8+s2,0))/ns
ghkval=ghkval+ghkvals
exputil(k-2:k)=exputil(k-2:k)+utemp(1:3)*ghkvals
dst(k-2:k)=dst(k-2:k)+dstemp(1:3)*ghkvals
fbfamutil(m)=fbfamutil(m)+fbutemp*ghkvals
pbfamutil(m)=pbfamutil(m)+pbutemp*ghkvals
y_any_fb(m) = y_any_fb(m)+fbytemp*ghkvals
gtype1a(m,:,:,:)=gtype1a(m,:,:,:)+gtemp1a(:,:,:)*ghkvals
near_cf(m,:)=near_cf(m,:)+nearcftemp(:)*ghkvals
END DO
END DO         ! Strategy loop
exputil(k-2:k)=exputil(k-2:k)/ghkval
dst(k-2:k)=dst(k-2:k)/ghkval
famutil(m)=SUM(exputil(k-2:k))
fbfamutil(m)=fbfamutil(m)/ghkval
pbfamutil(m)=pbfamutil(m)/ghkval
y_any_fb(m)=y_any_fb(m)/ghkval
gtype1a(m,:,:,:)=gtype1a(m,:,:,:)/ghkval
near_cf(m,:)=near_cf(m,:)/ghkvals

END FUNCTION SIM3



FUNCTION SIM4(m,param,n) RESULT(ghkval)
	! This function calculates the GHK simulator for each market*year
USE Globvar
USE PROBABILITY
IMPLICIT NONE
INTEGER, INTENT(IN) :: m			! market id
INTEGER, INTENT(IN) :: n 
REAL(8), INTENT(IN) :: param(n)
INTEGER :: i, j, k, l
INTEGER :: s1, s2, s3, s4, stmp
REAL(8) :: wgv(ns,0:sm4-1,0:4)		! wgv(ns, # strategy combitnations, 1 + # child) ... wgv(ns,s,0) for wg.
REAL(8) :: wg(ns,0:sm4-1)
REAL(8) :: ta(ns), tb(ns)
REAL(8) :: v(ns,0:sm4-1,nchild(m))   ! GHK simulation draws constructed from draws()
REAL(8) :: gtemp1a(0:15,1:6,0:15)
REAL(8) :: gtemp2a(0:16,0:11,1:15,0:3)
REAL(8) :: gtemp2b(0:15,0:11,0:3)
REAL(8) :: gtemp2c(0:5,0:6,0:11,0:3)
REAL(8) :: nearcftemp(0:15)
REAL(8) :: utiltemp
REAL(8) :: utemp(nchild(m)), dstemp(nchild(m))
REAL(8) :: fbutemp, pbutemp, fbytemp
REAL(8) :: ghkval, ghkvals		! The calculated GHK value
REAL(8) :: utilsub(4), alpha(4,3), gamma(4,3)		! 1 ... oldest child
REAL(8) :: lchol(4,4)

wgv=0.000d0
wg=0.000d0
v=0.000d0
ghkval=0.000d0

k=famindex(m)

utemp=0.000d0
fbutemp=0.000d0 ; pbutemp=0.000d0 ; fbytemp=0.000d0

lchol=lcholbase(m,param,n)

wgv=GHK4(m,param,n,ns,draws4,lchol,nthetab, nthetaa, nthetag, xb(:,famindex(m)-3:famindex(m)), xa(:,famindex(m)-3:famindex(m)), xg(:,famindex(m)-3:famindex(m)), sa, sg, near(m),2)

utilsub(1:4)=MATMUL(param(1:nthetab),xb(:,k-3:k))
	IF (nthetaa>=3) THEN
	alpha(1:4,1) = EXP( MATMUL(param(nthetab+1:nthetab+nthetaa-2),xa(:,k-3:k)) )
	alpha(1:4,2) = EXP( MATMUL(param(nthetab+1:nthetab+nthetaa-2),xa(:,k-3:k)) + param(nthetab+nthetaa-1) )
	alpha(1:4,3) = EXP( MATMUL(param(nthetab+1:nthetab+nthetaa-2),xa(:,k-3:k)) + param(nthetab+nthetaa-1) + param(nthetab+nthetaa) )
	ELSEIF (nthetaa==1) THEN
	alpha(1:4,1:3)=param(nthetab+nthetaa)
	ELSE
	alpha=0.0d0
	ENDIF
	IF (nthetag>=2) THEN
	gamma(1:4,1) =MATMUL(param(nthetab+nthetaa+1:nthetab+nthetaa+nthetag-1),xg(:,k-3:k))
	gamma(1:4,2) =MATMUL(param(nthetab+nthetaa+1:nthetab+nthetaa+nthetag-1),xg(:,k-3:k))*(1.0d0 + param(nthetab+nthetaa+nthetag))
	gamma(1:4,3) =MATMUL(param(nthetab+nthetaa+1:nthetab+nthetaa+nthetag-1),xg(:,k-3:k))*(1.0d0 + param(nthetab+nthetaa+nthetag)*2.0d0)
	ELSEIF (nthetag==1) THEN
	gamma(1:4,1:3)=param(nthetab+nthetaa+nthetag)
	ELSE
	gamma=0.0d0
	ENDIF

alpha(:,:)=alpha(:,:)*amp1
gamma(:,:)=gamma(:,:)*amp2

DO s1=0,2**2-1       ! Strategy loop
DO s2=0,2**3-1
DO s3=0,2**4-1

IF (ghkcheck(s1,2,m)==1 .and. ghkcheck(s2,3,m)==1 .and. ghkcheck(s3,4,m)==1 .and. SUM(wgv(:,s1*128+s2*16+s3,0))>0.0d0) THEN
    CALL geval(gtemp1a(:,:,:),gtemp2a(:,:,:,:),gtemp2b(:,:,:),gtemp2c(:,:,:,:),nearcftemp(:),utemp,pbutemp,fbutemp,fbytemp,dstemp,m,wgv(:,s1*128+s2*16+s3,1:4),alpha,gamma,lchol)
ENDIF

ghkvals=SUM(wgv(:,s1*128+s2*16+s3,0))/ns
ghkval=ghkval+ghkvals
exputil(k-3:k)=exputil(k-3:k)+utemp(1:4)*ghkvals
dst(k-3:k)=dst(k-3:k)+dstemp(1:4)*ghkvals
fbfamutil(m)=fbfamutil(m)+fbutemp*ghkvals
pbfamutil(m)=pbfamutil(m)+pbutemp*ghkvals
y_any_fb(m) =y_any_fb(m) +fbytemp*ghkvals
gtype1a(m,:,:,:)=gtype1a(m,:,:,:)+gtemp1a(:,:,:)*ghkvals
near_cf(m,:)=near_cf(m,:)+nearcftemp(:)*ghkvals
END DO
END DO
END DO         ! Strategy loop

exputil(k-3:k)=exputil(k-3:k)/ghkval
dst(k-3:k)=dst(k-3:k)/ghkval
famutil(m)=SUM(exputil(k-3:k))
fbfamutil(m)=fbfamutil(m)/ghkval
pbfamutil(m)=pbfamutil(m)/ghkval
y_any_fb(m)=y_any_fb(m)/ghkval
gtype1a(m,:,:,:)=gtype1a(m,:,:,:)/ghkval
near_cf(m,:)=near_cf(m,:)/ghkvals

END FUNCTION SIM4



FUNCTION ghkcheck(s,n,m)
USE Globvar
IMPLICIT NONE
INTEGER, INTENT(IN) :: m     ! mkt id
INTEGER, INTENT(IN) :: n      ! child id
INTEGER, INTENT(IN) :: s      ! player n's strategy
INTEGER :: i, nup
INTEGER :: ghkcheck

ghkcheck=0
IF (n>=2) THEN
nup=nearcounti(IBITS(near(m),nchild(m)-n+1,n-1),n-1)
    IF (IBITS(s,nup,1)==1 .and. IBITS(near(m),nchild(m)-n,1)==1) THEN
        ghkcheck=1
    ELSEIF (IBITS(s,nup,1)==0 .and. IBITS(near(m),nchild(m)-n,1)==0) THEN
        ghkcheck=1
    END IF
ELSE
	PRINT *,'Error found!'
	PRINT *,'player #: ', n
	STOP
ENDIF

END FUNCTION ghkcheck





SUBROUTINE geval(gtp1a,gtp2a,gtp2b,gtp2c,nearcfval,uval,pbuval,fbuval,fbyval,dsval,m,v,tmpalpha,tmpgamma,lchol)        !compute first best
USE Globvar
IMPLICIT NONE
INTEGER, INTENT(IN) :: m    	! family id
REAL(8), INTENT(IN) :: v(ns,nchild(m))	! error term. 1 ... youngest
REAL(8), INTENT(IN) :: tmpalpha(nchild(m),3)
REAL(8), INTENT(IN) :: tmpgamma(nchild(m),3)
REAL(8), INTENT(IN) :: lchol(nchild(m),nchild(m))
REAL(8), INTENT(OUT) :: gtp1a(0:15,1:6,0:15)
REAL(8), INTENT(OUT) :: gtp2a(0:16,0:11,1:15,0:3)
REAL(8), INTENT(OUT) :: gtp2b(0:15,0:11,0:3)
REAL(8), INTENT(OUT) :: gtp2c(0:5,0:6,0:11,0:3)
REAL(8), INTENT(OUT) :: nearcfval(0:15)
REAL(8), INTENT(OUT) :: uval(nchild(m))
REAL(8), INTENT(OUT) :: dsval(nchild(m))
REAL(8), INTENT(OUT) :: pbuval, fbuval, fbyval
INTEGER :: i, j, k, st, i_pareto, i_ne
INTEGER :: fbnear, pbnear
INTEGER :: g1check3, g1check4
INTEGER :: g2check1, g2check2, g2check3, g2check4, g2check5, g2check6
REAL(8) :: error(ns,nchild(m)) ! error(:,1) ... oldest
INTEGER :: near_cf_s(ns)
REAL(8) :: uval_s(ns,nchild(m)), pbu_s(ns), fbu_s(ns)
REAL(8) :: uval_s_dev(ns,nchild(m))
INTEGER :: dsval_s(ns,nchild(m))
REAL(8) :: temp(nchild(m)), tempu1, tempu2
REAL(8) :: u1(0:1,0:1), u2(0:1,0:1)
INTEGER :: nearr, ne(0:1,0:1), neunique

uval=0.0d0
uval_s=0.0d0
uval_s_dev=0.0d0
dsval=0.0d0
pbuval=0.0d0
pbu_s=0.0d0
fbuval=0.0d0
fbyval=0.0d0
fbu_s=0.0d0
gtp1a=0.0d0
gtp2a=0.0d0
gtp2b=0.0d0
gtp2c=0.0d0
nearcfval=0.0d0

!!! Construct the error term !!!
DO j=1,nchild(m)    ! j=1 means oldest
    k=nchild(m)-j+1
    error(:,j)=MATMUL(v(:,nchild(m):j:-1),(lchol(k,1:k)))
END DO



DO i=1,ns
    g1check3=0 ; g1check4=0

    !!! Solve for SPNE if amp/=1.0
    IF (amp1==1.0d0 .and. amp2==1.0d0) THEN
        near_cf_s(i)=near(m)
    ELSE
        CALL NESOLVE(near_cf_s(i),m,error(i,:),utilxb(famindex(m)-nchild(m)+1:famindex(m)),tmpalpha,tmpgamma)
    ENDIF

    !!! Compute current (SPNE) utility values !!!
    DO j=1,nchild(m)    ! j=1 means oldest
        IF (IBITS(near_cf_s(i),nchild(m)-j,1)==1) THEN		! when near
            uval_s(i,j)=error(i,j)+utilxb(famindex(m)-nchild(m)+j)
            IF     (nearcounti(near_cf_s(i),nchild(m))==4) THEN
                uval_s(i,j)    =uval_s(i,j)+tmpalpha(j,3)
                uval_s(i,j)    =uval_s(i,j)+tmpgamma(j,3)
                uval_s_dev(i,j)=tmpalpha(j,3)
            ELSEIF (nearcounti(near_cf_s(i),nchild(m))==3) THEN
                uval_s(i,j)    =uval_s(i,j)+tmpalpha(j,3)
                uval_s(i,j)    =uval_s(i,j)+tmpgamma(j,2)
                uval_s_dev(i,j)=tmpalpha(j,2)
            ELSEIF (nearcounti(near_cf_s(i),nchild(m))==2) THEN
                uval_s(i,j)    =uval_s(i,j)+tmpalpha(j,2)
                uval_s(i,j)    =uval_s(i,j)+tmpgamma(j,1)
                uval_s_dev(i,j)=tmpalpha(j,1)
            ELSEIF (nearcounti(near_cf_s(i),nchild(m))==1) THEN
                uval_s(i,j)    =uval_s(i,j)+tmpalpha(j,1)
                uval_s_dev(i,j)=0.0d0
            ENDIF
        ELSE
            uval_s_dev(i,j)=error(i,j)+utilxb(famindex(m)-nchild(m)+j)
            IF     (nearcounti(near_cf_s(i),nchild(m))==3) THEN
                uval_s(i,j)    =tmpalpha(j,3)
                uval_s_dev(i,j)=uval_s_dev(i,j)+tmpalpha(j,3)
                uval_s_dev(i,j)=uval_s_dev(i,j)+tmpgamma(j,3)
            ELSEIF (nearcounti(near_cf_s(i),nchild(m))==2) THEN
                uval_s(i,j)    =tmpalpha(j,2)
                uval_s_dev(i,j)=uval_s_dev(i,j)+tmpalpha(j,3)
                uval_s_dev(i,j)=uval_s_dev(i,j)+tmpgamma(j,2)
            ELSEIF (nearcounti(near_cf_s(i),nchild(m))==1) THEN
                uval_s(i,j)    =tmpalpha(j,1)
                uval_s_dev(i,j)=uval_s_dev(i,j)+tmpalpha(j,2)
                uval_s_dev(i,j)=uval_s_dev(i,j)+tmpgamma(j,1)
            ELSEIF (nearcounti(near_cf_s(i),nchild(m))==0) THEN
                uval_s(i,j)    =0.0d0
                uval_s_dev(i,j)=uval_s_dev(i,j)+tmpalpha(j,1)
            ENDIF
        ENDIF
    END DO

    !!! Check whether SPNE (near) is a normal form NE.
    i_ne=1
    DO j=1,nchild(m)    ! j=1 means oldest
        IF (uval_s_dev(i,j)>uval_s(i,j)) THEN
            i_ne=0
        ENDIF
    ENDDO

    !!! Compute Pareto best !!!
    pbnear=near_cf_s(i)
    pbu_s(i)=SUM(uval_s(i,:))
    DO st=0,2**nchild(m)-1       ! Strategy loop
	    temp=0.0d0
	    i_pareto=1
	    DO j=1,nchild(m)
		    IF (IBITS(st,nchild(m)-j,1)==1) THEN	! when near
		        IF (nearcounti(st,nchild(m))==1) THEN
    		        temp(j)=error(i,j)+utilxb(famindex(m)-nchild(m)+j)+tmpalpha(j,1)
		        ELSEIF (nearcounti(st,nchild(m))==2) THEN
    		        temp(j)=error(i,j)+utilxb(famindex(m)-nchild(m)+j)+tmpalpha(j,2)+tmpgamma(j,1)
		        ELSEIF (nearcounti(st,nchild(m))==3) THEN
    		        temp(j)=error(i,j)+utilxb(famindex(m)-nchild(m)+j)+tmpalpha(j,3)+tmpgamma(j,2)
		        ELSEIF (nearcounti(st,nchild(m))==4) THEN
    		        temp(j)=error(i,j)+utilxb(famindex(m)-nchild(m)+j)+tmpalpha(j,3)+tmpgamma(j,3)
		        ENDIF
		    ELSE
		        IF (nearcounti(st,nchild(m))==0) THEN
    		        temp(j)=0.0d0
		        ELSEIF (nearcounti(st,nchild(m))==1) THEN
    		        temp(j)=tmpalpha(j,1)
		        ELSEIF (nearcounti(st,nchild(m))==2) THEN
    		        temp(j)=tmpalpha(j,2)
		        ELSEIF (nearcounti(st,nchild(m))>2) THEN
    		        temp(j)=tmpalpha(j,3)
		        ENDIF
		    ENDIF
		    IF (temp(j)<uval_s(i,j)) THEN
		        i_pareto=0
		    ENDIF
	    ENDDO
	    IF (SUM(temp(:))>pbu_s(i) .and. i_pareto==1) THEN
		    pbu_s(i)=SUM(temp(:))
		    pbnear=st
	    ENDIF
    ENDDO         ! Strategy loop
    IF (pbnear==near_cf_s(i)) THEN   !!! No Pareto improvement
        IF (i_ne==1) THEN
            g1check3=3
        ELSE
            g1check3=4
        ENDIF
    ELSE   !!! Pareto dominated
        IF (i_ne==1) THEN
            g1check3=5
        ELSE
            g1check3=6
        ENDIF
    ENDIF

    !!! Compute first best !!!
    fbnear=0
    DO st=0,2**nchild(m)-1       ! Strategy loop
	    temp=0.0d0
	    DO j=1,nchild(m)
		    IF (IBITS(st,nchild(m)-j,1)==1) THEN
		        IF (nearcounti(st,nchild(m))==1) THEN
    		        temp(j)=error(i,j)+utilxb(famindex(m)-nchild(m)+j)+tmpalpha(j,1)
		        ELSEIF (nearcounti(st,nchild(m))==2) THEN
    		        temp(j)=error(i,j)+utilxb(famindex(m)-nchild(m)+j)+tmpalpha(j,2)+tmpgamma(j,1)
		        ELSEIF (nearcounti(st,nchild(m))==3) THEN
    		        temp(j)=error(i,j)+utilxb(famindex(m)-nchild(m)+j)+tmpalpha(j,3)+tmpgamma(j,2)
		        ELSEIF (nearcounti(st,nchild(m))==4) THEN
    		        temp(j)=error(i,j)+utilxb(famindex(m)-nchild(m)+j)+tmpalpha(j,3)+tmpgamma(j,3)
		        ENDIF
		    ELSE
		        IF (nearcounti(st,nchild(m))==0) THEN
    		        temp(j)=0.0d0
		        ELSEIF (nearcounti(st,nchild(m))==1) THEN
    		        temp(j)=tmpalpha(j,1)
		        ELSEIF (nearcounti(st,nchild(m))==2) THEN
    		        temp(j)=tmpalpha(j,2)
		        ELSEIF (nearcounti(st,nchild(m))>2) THEN
    		        temp(j)=tmpalpha(j,3)
		        ENDIF
		    ENDIF
	    ENDDO
	    IF (SUM(temp(:))>fbu_s(i)) THEN
		    fbu_s(i)=SUM(temp(:))
		    fbnear=st
	    ENDIF
    ENDDO         ! Strategy loop
    g1check4=fbnear
    IF (fbnear>0) THEN
        fbyval=fbyval+1.0d0
    ENDIF
    IF (fbnear==near_cf_s(i) .and. pbnear==near_cf_s(i)) THEN
        IF (i_ne==1) THEN
            g1check3=1
        ELSE
            g1check3=2
        ENDIF
    ELSEIF (fbnear==near_cf_s(i) .and. pbnear/=near_cf_s(i)) THEN
        PRINT *,'Error: First best but Pareto dominated'
        PAUSE
    ENDIF


    !!! Check dominant strategies !!!
    DO j=1,nchild(m)
        dsval_s(i,j)=1                      !!! check if near is a dominant strategy
        DO st=0,2**nchild(m)-1       ! Strategy loop
            IF (IBITS(st,nchild(m)-j,1)==1) THEN
		        IF (nearcounti(st,nchild(m))==1) THEN
    		        tempu1=error(i,j)+utilxb(famindex(m)-nchild(m)+j)+tmpalpha(j,1)
    		        tempu2=0.0d0
		        ELSEIF (nearcounti(st,nchild(m))==2) THEN
    		        tempu1=error(i,j)+utilxb(famindex(m)-nchild(m)+j)+tmpalpha(j,2)+tmpgamma(j,1)
    		        tempu2=tmpalpha(j,1)
		        ELSEIF (nearcounti(st,nchild(m))==3) THEN
    		        tempu1=error(i,j)+utilxb(famindex(m)-nchild(m)+j)+tmpalpha(j,3)+tmpgamma(j,2)
    		        tempu2=tmpalpha(j,2)
		        ELSEIF (nearcounti(st,nchild(m))==4) THEN
    		        tempu1=error(i,j)+utilxb(famindex(m)-nchild(m)+j)+tmpalpha(j,3)+tmpgamma(j,3)
    		        tempu2=tmpalpha(j,3)
		        ENDIF
	            IF (tempu1<tempu2) THEN
	                dsval_s(i,j)=2
	            ENDIF
            ENDIF
        ENDDO
        IF (dsval_s(i,j)==2) THEN           !!! When near is not a dominant strategy, check if far is dominant
            dsval_s(i,j)=0
            DO st=0,2**nchild(m)-1       ! Strategy loop
                IF (IBITS(st,nchild(m)-j,1)==0) THEN
		            IF (nearcounti(st,nchild(m))==0) THEN
    		            tempu1=0.0d0
    		            tempu2=error(i,j)+utilxb(famindex(m)-nchild(m)+j)+tmpalpha(j,1)
		            ELSEIF (nearcounti(st,nchild(m))==1) THEN
    		            tempu1=tmpalpha(j,1)
    		            tempu2=error(i,j)+utilxb(famindex(m)-nchild(m)+j)+tmpalpha(j,2)+tmpgamma(j,1)
		            ELSEIF (nearcounti(st,nchild(m))==2) THEN
    		            tempu1=tmpalpha(j,2)
    		            tempu2=error(i,j)+utilxb(famindex(m)-nchild(m)+j)+tmpalpha(j,3)+tmpgamma(j,2)
		            ELSEIF (nearcounti(st,nchild(m))==3) THEN
    		            tempu1=tmpalpha(j,3)
    		            tempu2=error(i,j)+utilxb(famindex(m)-nchild(m)+j)+tmpalpha(j,3)+tmpgamma(j,3)
		            ENDIF
	                IF (tempu1<tempu2) THEN
	                    dsval_s(i,j)=2
	                ENDIF
                ENDIF
            ENDDO
        ENDIF

    ENDDO

    !!! Analyze Normal Form Game for 2 child families !!!
    IF (nchild(m)==2) THEN

        !!! Determine the utility values for each living arrangement
        u1(0,0)=0.0d0
        u1(0,1)=tmpalpha(1,1)
        u1(1,0)=error(i,1)+utilxb(famindex(m)-1)+tmpalpha(1,1)
        u1(1,1)=error(i,1)+utilxb(famindex(m)-1)+tmpalpha(1,2)+tmpgamma(1,1)
        u2(0,0)=0.0d0
        u2(1,0)=tmpalpha(2,1)
        u2(0,1)=error(i,2)+utilxb(famindex(m)  )+tmpalpha(2,1)
        u2(1,1)=error(i,2)+utilxb(famindex(m)  )+tmpalpha(2,2)+tmpgamma(2,1)

        !!! Increasing difference?
        IF (u1(1,1)-u1(0,1) > u1(1,0)-u1(0,0) .and. u2(1,1)-u2(1,0) > u2(0,1)-u2(0,0)) g2check2=3
        IF (u1(1,1)-u1(0,1) > u1(1,0)-u1(0,0) .and. u2(1,1)-u2(1,0) < u2(0,1)-u2(0,0)) g2check2=2
        IF (u1(1,1)-u1(0,1) < u1(1,0)-u1(0,0) .and. u2(1,1)-u2(1,0) > u2(0,1)-u2(0,0)) g2check2=1
        IF (u1(1,1)-u1(0,1) < u1(1,0)-u1(0,0) .and. u2(1,1)-u2(1,0) < u2(0,1)-u2(0,0)) g2check2=0
        IF (near_cf_s(i)/=fbnear) THEN
            IF (near_cf_s(i)==pbnear) THEN
                g2check2=g2check2+4
            ELSE
                g2check2=g2check2+8
            ENDIF
        ENDIF
        

        !!! Solve for the SPNE of the reverse order
        IF (u1(0,0)>u1(1,0) .and. u1(0,1)>u1(1,1) .and. u2(0,0)>u2(0,1)) THEN
            nearr=0
        ELSEIF (u1(0,0)>u1(1,0) .and. u1(0,1)>u1(1,1) .and. u2(0,0)<u2(0,1)) THEN
            nearr=1
        ELSEIF (u1(0,0)<u1(1,0) .and. u1(0,1)>u1(1,1) .and. u2(1,0)>u2(0,1)) THEN
            nearr=2
        ELSEIF (u1(0,0)<u1(1,0) .and. u1(0,1)>u1(1,1) .and. u2(1,0)<u2(0,1)) THEN
            nearr=1
        ELSEIF (u1(0,0)>u1(1,0) .and. u1(0,1)<u1(1,1) .and. u2(0,0)>u2(1,1)) THEN
            nearr=0
        ELSEIF (u1(0,0)>u1(1,0) .and. u1(0,1)<u1(1,1) .and. u2(0,0)<u2(1,1)) THEN
            nearr=3
        ELSEIF (u1(0,0)<u1(1,0) .and. u1(0,1)<u1(1,1) .and. u2(1,0)>u2(1,1)) THEN
            nearr=2
        ELSEIF (u1(0,0)<u1(1,0) .and. u1(0,1)<u1(1,1) .and. u2(1,0)<u2(1,1)) THEN
            nearr=3
        ELSE
            print *,'Error! 00'
            pause
        ENDIF
        
        g2check4=g1check4*4+nearr
        

        IF (near_cf_s(i)==nearr) THEN
            g2check5=0
        ELSEIF (uval_s(i,1)+uval_s(i,2) > u1(IBITS(nearr,1,1),IBITS(nearr,0,1))+u2(IBITS(nearr,1,1),IBITS(nearr,0,1)) .and. uval_s(i,1) > u1(IBITS(nearr,1,1),IBITS(nearr,0,1)) .and. uval_s(i,2) > u2(IBITS(nearr,1,1),IBITS(nearr,0,1))) THEN
            g2check5=1
        ELSEIF (uval_s(i,1)+uval_s(i,2) > u1(IBITS(nearr,1,1),IBITS(nearr,0,1))+u2(IBITS(nearr,1,1),IBITS(nearr,0,1)) .and. uval_s(i,1) > u1(IBITS(nearr,1,1),IBITS(nearr,0,1)) .and. uval_s(i,2) < u2(IBITS(nearr,1,1),IBITS(nearr,0,1))) THEN
            g2check5=2
        ELSEIF (uval_s(i,1)+uval_s(i,2) > u1(IBITS(nearr,1,1),IBITS(nearr,0,1))+u2(IBITS(nearr,1,1),IBITS(nearr,0,1)) .and. uval_s(i,1) < u1(IBITS(nearr,1,1),IBITS(nearr,0,1)) .and. uval_s(i,2) > u2(IBITS(nearr,1,1),IBITS(nearr,0,1))) THEN
            g2check5=3
        ELSEIF (uval_s(i,1)+uval_s(i,2) < u1(IBITS(nearr,1,1),IBITS(nearr,0,1))+u2(IBITS(nearr,1,1),IBITS(nearr,0,1)) .and. uval_s(i,1) > u1(IBITS(nearr,1,1),IBITS(nearr,0,1)) .and. uval_s(i,2) < u2(IBITS(nearr,1,1),IBITS(nearr,0,1))) THEN
            g2check5=4
        ELSEIF (uval_s(i,1)+uval_s(i,2) < u1(IBITS(nearr,1,1),IBITS(nearr,0,1))+u2(IBITS(nearr,1,1),IBITS(nearr,0,1)) .and. uval_s(i,1) < u1(IBITS(nearr,1,1),IBITS(nearr,0,1)) .and. uval_s(i,2) > u2(IBITS(nearr,1,1),IBITS(nearr,0,1))) THEN
            g2check5=5
        ELSEIF (uval_s(i,1)+uval_s(i,2) < u1(IBITS(nearr,1,1),IBITS(nearr,0,1))+u2(IBITS(nearr,1,1),IBITS(nearr,0,1)) .and. uval_s(i,1) < u1(IBITS(nearr,1,1),IBITS(nearr,0,1)) .and. uval_s(i,2) < u2(IBITS(nearr,1,1),IBITS(nearr,0,1))) THEN
            g2check5=6
        ELSE
            print *,'Error! 01'
            pause
        ENDIF

        !!! Check Nash Equilibrium
        ne(0:1,0:1)=0
        IF (u1(1,1)>u1(0,1) .and. u2(1,1)>u2(1,0))      ne(1,1)=1
        IF (u1(0,1)>u1(1,1) .and. u2(0,1)>u2(0,0))      ne(0,1)=1
        IF (u1(1,0)>u1(0,0) .and. u2(1,0)>u2(1,1))      ne(1,0)=1
        IF (u1(0,0)>u1(1,0) .and. u2(0,0)>u2(0,1))      ne(0,0)=1
        neunique=0
        IF (ne(1,1)+ne(1,0)+ne(0,1)+ne(0,0)==1) THEN
            IF (ne(0,0)==1) THEN
                neunique=0
            ELSEIF (ne(0,1)==1) THEN
                neunique=1
            ELSEIF (ne(1,0)==1) THEN
                neunique=2
            ELSEIF (ne(1,1)==1) THEN
                neunique=3
            ENDIF
        ENDIF

        !!! Compare SPNE and normal form NE
        IF (ne(1,1)+ne(1,0)+ne(0,1)+ne(0,0)==2) THEN
            g2check6=4
        ELSEIF (ne(1,1)+ne(1,0)+ne(0,1)+ne(0,0)==0) THEN
            g2check6=5
        ELSEIF (near_cf_s(i)==neunique) THEN
            g2check6=0
        ELSEIF (uval_s(i,1)+uval_s(i,2) < u1(IBITS(neunique,1,1),IBITS(neunique,0,1))+u2(IBITS(neunique,1,1),IBITS(neunique,0,1)) .and. uval_s(i,2) < u2(IBITS(neunique,1,1),IBITS(neunique,0,1))) THEN
            g2check6=1
        ELSEIF (uval_s(i,1)+uval_s(i,2) > u1(IBITS(neunique,1,1),IBITS(neunique,0,1))+u2(IBITS(neunique,1,1),IBITS(neunique,0,1)) .and. uval_s(i,2) < u2(IBITS(neunique,1,1),IBITS(neunique,0,1))) THEN
            g2check6=2
        ELSEIF (uval_s(i,1)+uval_s(i,2) > u1(IBITS(neunique,1,1),IBITS(neunique,0,1))+u2(IBITS(neunique,1,1),IBITS(neunique,0,1)) .and. uval_s(i,2) > u2(IBITS(neunique,1,1),IBITS(neunique,0,1))) THEN
            g2check6=3
        ELSE
            print *,'Error! 01'
            pause
        ENDIF


        !!! Detremine the game type based on normal form matrix
        g2check3=0
        IF (ne(1,1)+ne(1,0)+ne(0,1)+ne(0,0)==1) THEN    !!! Unique equilibrium cases
            SELECT CASE(neunique)
            CASE(0)
                IF     (fbnear==0) THEN
                    g2check3=1
                ELSEIF (u1(1,1)>u1(0,0) .and. u2(1,1)>u2(0,0)) THEN
                    g2check3=3
                ELSE
                    g2check3=2
                ENDIF
            CASE(1)
                IF     (fbnear==1) THEN
                    g2check3=4
                ELSEIF (u1(1,0)>u1(0,1) .and. u2(1,0)>u2(0,1)) THEN
                    g2check3=6
                ELSE
                    g2check3=5
                ENDIF
            CASE(2)
                IF     (fbnear==2) THEN
                    g2check3=7
                ELSEIF (u1(0,1)>u1(1,0) .and. u2(0,1)>u2(1,0)) THEN
                    g2check3=9
                ELSE
                    g2check3=8
                ENDIF
            CASE(3)
                IF     (fbnear==3) THEN
                    g2check3=10
                ELSEIF (u1(0,0)>u1(1,1) .and. u2(0,0)>u2(1,1)) THEN
                    g2check3=12
                ELSE
                    g2check3=11
                ENDIF
            CASE DEFAULT
                PAUSE
            END SELECT

        ELSEIF (ne(1,1)+ne(1,0)+ne(0,1)+ne(0,0)==2) THEN
            IF (ne(1,1)+ne(0,0)==2) THEN       !!! Coordination game
                IF (fbnear==0 .or. fbnear==3) THEN
                    g2check3=13
                ELSE
                    g2check3=14
                ENDIF
            ELSEIF (ne(1,0)+ne(0,1)==2) THEN   !!! Anti-coordination game
                IF (fbnear==1 .or. fbnear==2) THEN
                    g2check3=15
                ELSE
                    g2check3=16
                ENDIF
            ELSE
                print *,'Error! 06'
                pause
            ENDIF
        ELSEIF (ne(1,1)+ne(1,0)+ne(0,1)+ne(0,0)==0) THEN       !!! No-equilibrium
            g2check3=0
        ELSE
            print *,'Error! 07'
            pause
        ENDIF

        !!! Check normal form dominant strategy
        IF (dsval_s(i,1)==1 .and. IBITS(near_cf_s(i),1,1)==1 .and. dsval_s(i,2)==1) g2check1= 1
        IF (dsval_s(i,1)==1 .and. IBITS(near_cf_s(i),1,1)==1 .and. dsval_s(i,2)==0) g2check1= 2
        IF (dsval_s(i,1)==1 .and. IBITS(near_cf_s(i),1,1)==1 .and. dsval_s(i,2)==2) g2check1= 3
        IF (dsval_s(i,1)==1 .and. IBITS(near_cf_s(i),1,1)==0 .and. dsval_s(i,2)==1) g2check1= 4
        IF (dsval_s(i,1)==1 .and. IBITS(near_cf_s(i),1,1)==0 .and. dsval_s(i,2)==0) g2check1= 5
        IF (dsval_s(i,1)==1 .and. IBITS(near_cf_s(i),1,1)==0 .and. dsval_s(i,2)==2) g2check1= 6
        IF (dsval_s(i,1)==0 .and. IBITS(near_cf_s(i),1,1)==0 .and. dsval_s(i,2)==1) g2check1= 7
        IF (dsval_s(i,1)==0 .and. IBITS(near_cf_s(i),1,1)==0 .and. dsval_s(i,2)==0) g2check1= 8
        IF (dsval_s(i,1)==0 .and. IBITS(near_cf_s(i),1,1)==0 .and. dsval_s(i,2)==2) g2check1= 9
        IF (dsval_s(i,1)==0 .and. IBITS(near_cf_s(i),1,1)==1 .and. dsval_s(i,2)==1) g2check1=10
        IF (dsval_s(i,1)==0 .and. IBITS(near_cf_s(i),1,1)==1 .and. dsval_s(i,2)==0) g2check1=11
        IF (dsval_s(i,1)==0 .and. IBITS(near_cf_s(i),1,1)==1 .and. dsval_s(i,2)==2) g2check1=12
        IF (dsval_s(i,1)==2 .and.                             dsval_s(i,2)==1) g2check1=13
        IF (dsval_s(i,1)==2 .and.                             dsval_s(i,2)==0) g2check1=14
        IF (dsval_s(i,1)==2 .and.                             dsval_s(i,2)==2) g2check1=15

    ENDIF



    gtp1a(g1check4,g1check3,near_cf_s(i))=gtp1a(g1check4,g1check3,near_cf_s(i))+1.0d0
    IF (nchild(m)==2) THEN
    gtp2a(g2check3,g2check2,g2check1,near_cf_s(i))=gtp2a(g2check3,g2check2,g2check1,near_cf_s(i))+1.0d0
    gtp2b(g2check4,g2check2,near_cf_s(i))=gtp2b(g2check4,g2check2,near_cf_s(i))+1.0d0
    gtp2c(g2check6,g2check5,g2check2,near_cf_s(i))=gtp2c(g2check6,g2check5,g2check2,near_cf_s(i))+1.0d0
    ELSE
    gtp2a=0
    gtp2b=0
    gtp2c=0
    ENDIF
    nearcfval(near_cf_s(i))=nearcfval(near_cf_s(i))+1.0d0

    !!! re-code dsval_s
    DO j=1,nchild(m)
        IF (dsval_s(i,j)==IBITS(near_cf_s(i),nchild(m)-j,1)) THEN
            dsval_s(i,j)=1
        ELSE
            dsval_s(i,j)=0
        ENDIF
    ENDDO


ENDDO          ! Sim draw loop
uval=SUM(uval_s(:,:),1)/ns

dsval=SUM(dsval_s(:,:),1)*1.0d0/ns
pbuval=SUM(pbu_s(:))/ns
fbuval=SUM(fbu_s(:))/ns
fbyval=fbyval/ns
gtp1a=gtp1a/ns
gtp2a=gtp2a/ns
gtp2b=gtp2b/ns
gtp2c=gtp2c/ns
nearcfval=nearcfval/ns

END SUBROUTINE geval




END MODULE Object











PROGRAM MAIN
USE Globvar
USE INIT
USE Object
IMPLICIT NONE

REAL(8) :: resultmlv, resultmlse
REAL(8) :: objsaved
REAL(4) :: time0, iniadj
REAL(8) :: g2table1(0:3,1:4), g2table2a(0:6,1:5), g2table2b(0:16,1:5)      !! temporary variables to make tables
REAL(8) :: g2table3a(0:6,1:5), g2table3b(0:16,1:5)       !! temporary variables to make tables
REAL(8) :: g2table4(1:4,1:5), g2table4a(1:9,1:5), g2table4b(1:15,1:5), g2table5(0:3,1:5)      !! temporary variables to make tables
REAL(8) :: g2table6a(0:3,1:5), g2table6b(0:3,1:5), g2table6c(0:15,1:5)      !! temporary variables to make tables
REAL(8) :: g2table6a2(0:3,1:5), g2table6b2(0:3,1:5), g2table6c2(0:15,1:5)      !! temporary variables to make tables
REAL(8) :: g2table7(0:5,1:5), g2table8(0:6,1:5)      !! temporary variables to make tables
REAL(8) :: g2table9a(0:6,1:4), g2table9b(0:16,1:4)      !! temporary variables to make tables
REAL(8) :: g2table10a(0:6,1:4), g2table10b(0:16,1:4)      !! temporary variables to make tables
REAL(8) :: g2table11a(1:9,1:4), g2table11b(1:15,1:4), g2table12(0:3,1:4)      !! temporary variables to make tables
REAL(8) :: g2table13a(0:3,1:4), g2table13b(0:3,1:4), g2table13c(0:15,1:4)      !! temporary variables to make tables
REAL(8) :: g2table14(0:5,1:4), g2table15(0:6,1:4)      !! temporary variables to make tables
REAL(8) :: g2table16a(0:6,1:5), g2table16b(0:16,1:5)      !! temporary variables to make tables
REAL(8) :: g2table17a(1:9,1:5), g2table17b(1:15,1:5)      !! temporary variables to make tables
REAL(8) :: g2table18a(0:3,1:5), g2table18b(0:3,1:5), g2table18c(0:15,1:5)      !! temporary variables to make tables
REAL(8) :: g2table19(0:5,1:5), g2table20(0:6,1:5)      !! temporary variables to make tables
REAL(8) :: g2table19a(0:5,1:5), g2table20a(0:6,1:5)      !! temporary variables to make tables
REAL(8) :: g2table21(0:6,1:6)      !! temporary variables to make tables
REAL(8) :: g2table22(0:6,1:5), g2table23(0:3,1:5)      !! temporary variables to make tables
REAL(8) :: g2table24(0:2,1:5), g2table25(0:3,1:5)      !! temporary variables to make tables
REAL(8) :: g2table26(0:3,1:5)      !! temporary variables to make tables

INTEGER :: i, j, k, l, m, n, io, nnear
INTEGER :: i0, i1, i2, i3, i4, i5, i6, i7
INTEGER :: gtct_1c, gtct_2c, gtct_3c, gtct_4c
REAL(8) :: gtct_2ce1, gtct_3ce1, gtct_4ce1
REAL(8) :: gtct_2ce2, gtct_3ce2, gtct_4ce2
REAL(8) :: gtct_2ce3, gtct_3ce3, gtct_4ce3
REAL(8) :: gtct2_eff1, gtct2_eff2, gtct2_eff3
REAL(8) :: gtct2_la0, gtct2_la1, gtct2_la2, gtct2_la3
REAL(8) :: gtct2_fbla0, gtct2_fbla1, gtct2_fbla2, gtct2_fbla3
REAL(8) :: gtct2_idf0, gtct2_idf1, gtct2_idf2, gtct2_idf3
REAL(8) :: gtct2_spt1, gtct2_spt2, gtct2_spt3, gtct2_spt4, gtct2_spt5, gtct2_spt6
CHARACTER(12) :: horz(3)
CHARACTER(20) :: flname
CHARACTER(80) :: flname2
CHARACTER(20) :: fmtname           ! To accomodate different number of parameters
CHARACTER(50) :: inivalnote
CHARACTER(60) :: setupnote

CHARACTER(16), ALLOCATABLE :: vert(:)

horz(1)='        coef' ; horz(2)='        s.e.' ; horz(3)=' asy t-val'

! Read the setting file
    flname2='setting.prn'
    OPEN(1,file=(flname2))
    READ(1,fmt='(A60)') setupnote
    READ(1,fmt=*) nthetab
    READ(1,fmt=*) nthetaa
    READ(1,fmt=*) nthetag
    READ(1,fmt=*) nthetar
    READ(1,fmt=*) ntheta
    READ(1,fmt=*) nobs
    READ(1,fmt=*) nfam
    READ(1,fmt=*) ns
ALLOCATE (vert(ntheta))
    vert(1)='cons      '
    READ(1,fmt=*) vert(2:ntheta)
    READ(1,fmt=*) amp1
    READ(1,fmt=*) amp2
    CLOSE(1)

ALLOCATE (theta(ntheta))
ALLOCATE (famid(nobs))
ALLOCATE (famindex(nfam))
ALLOCATE (nchild(nfam))
ALLOCATE (y(nobs))
ALLOCATE (xb(nthetab,nobs))
IF (nthetaa>=3) THEN
	sa=nthetaa-2
	ALLOCATE (xa(nthetaa-2,nobs))
ELSEIF (nthetaa==1) THEN
	sa=1
	ALLOCATE (xa(1,nobs))
ENDIF
IF (nthetag>=2) THEN
	sg=nthetag-1
	ALLOCATE (xg(nthetag-1,nobs))
ELSEIF (nthetag==1) THEN
	sg=1
	ALLOCATE (xg(1,nobs))
ENDIF
ALLOCATE (xr(nthetar,nchildmax,nobs))
ALLOCATE (near(nfam))
ALLOCATE (near_cf(nfam,0:15))
ALLOCATE (utilxb(nobs))
ALLOCATE (exputil(nobs))
ALLOCATE (dst(nobs))
ALLOCATE (famutil(nfam))
ALLOCATE (fbfamutil(nfam))
ALLOCATE (pbfamutil(nfam))
ALLOCATE (y_any_fb(nfam))
ALLOCATE (gtype1a(nfam,0:15,1:6,0:15))
ALLOCATE (gtype2a(nfam,0:16,0:11,1:15,0:3))
ALLOCATE (gtype2b(nfam,0:15,0:11,0:3))
ALLOCATE (gtype2c(nfam,0:5,0:6,0:11,0:3))

OPEN(1,file='param1.out')
    READ(1,fmt=*) theta(1:nthetab)
CLOSE(1)
IF (ntheta>nthetab) THEN
OPEN(1,file='param2.out')
    READ(1,fmt=*) theta(nthetab+1:ntheta)
CLOSE(1)
ENDIF


time0 = SECNDS(0.0)

CALL GETTIM(i, j, k, l)
CALL GETDAT(l, m, n)
flname2=CHAR(ICHAR('0')+(m/10))//CHAR(ICHAR('0')+(m-m/10*10))//CHAR(ICHAR('0')+(n/10))//CHAR(ICHAR('0')+(n-n/10*10))//CHAR(ICHAR('0')+(i/10))//CHAR(ICHAR('0')+(i-i/10*10))//CHAR(ICHAR('0')+(j/10))//CHAR(ICHAR('0')+(j-j/10*10))//CHAR(ICHAR('0')+(k/10))//CHAR(ICHAR('0')+(k-k/10*10))//'.prn'

OPEN(4,file=(flname2),STATUS='REPLACE')
WRITE(4,fmt='(A20,4x,A60)') 'Estimation Setting: ', setupnote
WRITE(4,fmt=*) '# of families     :  ', nfam
WRITE(4,fmt=*) '# of observations:  ', nobs
WRITE(4,fmt=*) '# of simuln draws  :  ', ns
WRITE(4,fmt=*) ' '
WRITE(4,fmt=*) 'Counterfactual simulation. Alpha multiplier  :  ', amp1
WRITE(4,fmt=*) 'Counterfactual simulation. Gamma multiplier  :  ', amp2
WRITE(4,fmt=*) ' '
WRITE(4,fmt=*) 'Output: Tables for 2 child families'
WRITE(4,fmt=*) ' '



! Load the data set
CALL READDATA(nobs)

ALLOCATE (draws2seed(ns,0:sm2-1,2))
ALLOCATE (draws3seed(ns,0:sm3-1,3))
ALLOCATE (draws4seed(ns,0:sm4-1,4))
ALLOCATE (draws2(ns,0:sm2-1,2))
ALLOCATE (draws3(ns,0:sm3-1,3))
ALLOCATE (draws4(ns,0:sm4-1,4))
CALL RANDOM_SEED(PUT=(/2000/))
CALL RANDOM_NUMBER (draws2seed)
CALL RANDOM_NUMBER (draws3seed)
CALL RANDOM_NUMBER (draws4seed)
draws2seed(ns/2+1:ns,:,:)=1-draws2seed(1:ns/2,:,:)        ! This line makes antithetics
draws3seed(ns/2+1:ns,:,:)=1-draws3seed(1:ns/2,:,:)        ! This line makes antithetics
draws4seed(ns/2+1:ns,:,:)=1-draws4seed(1:ns/2,:,:)        ! This line makes antithetics


utilxb(1:nobs)=MATMUL(theta(1:nthetab),xb(:,1:nobs))

objsaved=SIMULATION(theta,ntheta)

WRITE(4,fmt=*) ' '
WRITE(4,fmt=*) 'ML objective:  ', objsaved
WRITE(4,fmt=*) ' '
WRITE(4,fmt=*) 'run time(seconds):  ', SECNDS(time0)
WRITE(4,fmt=*) ' '


print *,'Making Tables ... '

!!! Make counts for game type analysis

gtct_1c=0 ; gtct_2c=0 ; gtct_3c=0 ; gtct_4c=0
gtct_2ce1=0.0d0 ; gtct_3ce1=0.0d0 ; gtct_4ce1=0.0d0
gtct_2ce2=0.0d0 ; gtct_3ce2=0.0d0 ; gtct_4ce2=0.0d0
gtct_2ce3=0.0d0 ; gtct_3ce3=0.0d0 ; gtct_4ce3=0.0d0


DO m=1,nfam
    SELECT CASE(nchild(m))
    CASE(1)
        gtct_1c=gtct_1c+1
    CASE(2)
        gtct_2c=gtct_2c+1
        DO i0=0, 3
        DO i3=1, 6
        DO i4=0, 3
            IF     (i3==1 .or. i3==2) THEN
                gtct_2ce1=gtct_2ce1+gtype1a(m,i4,i3,i0)
            ELSEIF (i3==3 .or. i3==4) THEN
                gtct_2ce2=gtct_2ce2+gtype1a(m,i4,i3,i0)
            ELSEIF (i3==5 .or. i3==6) THEN
                gtct_2ce3=gtct_2ce3+gtype1a(m,i4,i3,i0)
            ENDIF
        ENDDO
        ENDDO
        ENDDO
    CASE(3)
        gtct_3c=gtct_3c+1
        DO i0=0, 7
        DO i3=1, 6
        DO i4=0, 7
            IF     (i3==1 .or. i3==2) THEN
                gtct_3ce1=gtct_3ce1+gtype1a(m,i4,i3,i0)
            ELSEIF (i3==3 .or. i3==4) THEN
                gtct_3ce2=gtct_3ce2+gtype1a(m,i4,i3,i0)
            ELSEIF (i3==5 .or. i3==6) THEN
                gtct_3ce3=gtct_3ce3+gtype1a(m,i4,i3,i0)
            ENDIF
        ENDDO
        ENDDO
        ENDDO
    CASE(4)
        gtct_4c=gtct_4c+1
        DO i0=0, 15
        DO i3=1, 6
        DO i4=0, 15
            IF     (i3==1 .or. i3==2) THEN
                gtct_4ce1=gtct_4ce1+gtype1a(m,i4,i3,i0)
            ELSEIF (i3==3 .or. i3==4) THEN
                gtct_4ce2=gtct_4ce2+gtype1a(m,i4,i3,i0)
            ELSEIF (i3==5 .or. i3==6) THEN
                gtct_4ce3=gtct_4ce3+gtype1a(m,i4,i3,i0)
            ENDIF
        ENDDO
        ENDDO
        ENDDO
    CASE DEFAULT
        PAUSE
    END SELECT
ENDDO

gtct2_eff1=0.0d0 ; gtct2_eff2=0.0d0 ; gtct2_eff3=0.0d0          !!! Efficiency type in 2 child families
gtct2_la0 =0.0d0 ; gtct2_la1 =0.0d0 ; gtct2_la2 =0.0d0 ; gtct2_la3 =0.0d0           !!! Observed Living Arrangement type in 2 child families
gtct2_fbla0 =0.0d0 ; gtct2_fbla1 =0.0d0 ; gtct2_fbla2 =0.0d0 ; gtct2_fbla3 =0.0d0           !!! First best Living Arrangement type in 2 child families
gtct2_idf0=0.0d0 ; gtct2_idf1=0.0d0 ; gtct2_idf2=0.0d0 ; gtct2_idf3=0.0d0          !!! Increasing difference type in 2 child families

gtct2_spt1=0.0d0 ; gtct2_spt2=0.0d0 ; gtct2_spt3=0.0d0 ; gtct2_spt4=0.0d0 ; gtct2_spt5=0.0d0 ; gtct2_spt6=0.0d0

DO m=1,nfam
    IF (nchild(m)==2) THEN
        gtct2_la0=gtct2_la0+near_cf(m,0)
        gtct2_la1=gtct2_la1+near_cf(m,1)
        gtct2_la2=gtct2_la2+near_cf(m,2)
        gtct2_la3=gtct2_la3+near_cf(m,3)

        DO i0=0, 3
        DO i3=1, 6
        DO i4=0, 3
            IF     (i3==1 .or. i3==2) THEN
                gtct2_eff1=gtct2_eff1+gtype1a(m,i4,i3,i0)
            ELSEIF (i3==3 .or. i3==4) THEN
                gtct2_eff2=gtct2_eff2+gtype1a(m,i4,i3,i0)
            ELSEIF (i3==5 .or. i3==6) THEN
                gtct2_eff3=gtct2_eff3+gtype1a(m,i4,i3,i0)
            ENDIF
        ENDDO
        ENDDO
        ENDDO

        DO i0=0,  3
        DO i1=1, 15
        DO i2=0, 11
        DO i3=0, 16
            IF     (MOD(i2,4)==0) THEN
                gtct2_idf0=gtct2_idf0+gtype2a(m,i3,i2,i1,i0)
            ELSEIF (MOD(i2,4)==1) THEN
                gtct2_idf1=gtct2_idf1+gtype2a(m,i3,i2,i1,i0)
            ELSEIF (MOD(i2,4)==2) THEN
                gtct2_idf2=gtct2_idf2+gtype2a(m,i3,i2,i1,i0)
            ELSEIF (MOD(i2,4)==3) THEN
                gtct2_idf3=gtct2_idf3+gtype2a(m,i3,i2,i1,i0)
            ENDIF
        ENDDO
        ENDDO
        ENDDO
        ENDDO

        DO i0=0,  3
        DO i2=0, 11
        DO i4=0, 15
            IF     (i4/4==0) THEN
                gtct2_fbla0=gtct2_fbla0+gtype2b(m,i4,i2,i0)
            ELSEIF (i4/4==1) THEN
                gtct2_fbla1=gtct2_fbla1+gtype2b(m,i4,i2,i0)
            ELSEIF (i4/4==2) THEN
                gtct2_fbla2=gtct2_fbla2+gtype2b(m,i4,i2,i0)
            ELSEIF (i4/4==3) THEN
                gtct2_fbla3=gtct2_fbla3+gtype2b(m,i4,i2,i0)
            ENDIF
        ENDDO
        ENDDO
        ENDDO

        DO i0=0,  3
        DO i2=0, 11
        DO i5=0, 6
        DO i6=0, 5
            IF     (i6==0) THEN
                gtct2_spt1=gtct2_spt1+gtype2c(m,i6,i5,i2,i0)
            ELSEIF (i6==1) THEN
                gtct2_spt2=gtct2_spt2+gtype2c(m,i6,i5,i2,i0)
            ELSEIF (i6==2) THEN
                gtct2_spt3=gtct2_spt3+gtype2c(m,i6,i5,i2,i0)
            ELSEIF (i6==3) THEN
                gtct2_spt4=gtct2_spt4+gtype2c(m,i6,i5,i2,i0)
            ELSEIF (i6==4) THEN
                gtct2_spt5=gtct2_spt5+gtype2c(m,i6,i5,i2,i0)
            ELSEIF (i6==5) THEN
                gtct2_spt6=gtct2_spt6+gtype2c(m,i6,i5,i2,i0)
            ENDIF
        ENDDO
        ENDDO
        ENDDO
        ENDDO

    ENDIF
ENDDO




!!! Game type analysis (nchild 0-4)




!!! Game type analysis (nchild==2)

g2table2a=0.0d0 ; g2table2b=0.0d0
g2table3a=0.0d0 ; g2table3b=0.0d0
g2table4 =0.0d0 ; g2table4a=0.0d0 ; g2table4b=0.0d0
g2table5=0.0d0
DO m=1,nfam
    IF (nchild(m)==2) THEN
        DO i0=0, 3
        DO i1=1, 15
        DO i2=0, 11
        DO i3=0, 16
                IF (i3==0) THEN
                    g2table2a(i3,1)=g2table2a(i3,1)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==1 .or. i3==2 .or. i3==3) THEN
                    g2table2a(1,1)=g2table2a(1,1)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==4 .or. i3==5 .or. i3==6) THEN
                    g2table2a(2,1)=g2table2a(2,1)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==7 .or. i3==8 .or. i3==9) THEN
                    g2table2a(3,1)=g2table2a(3,1)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==10 .or. i3==11 .or. i3==12) THEN
                    g2table2a(4,1)=g2table2a(4,1)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==13 .or. i3==14) THEN
                    g2table2a(5,1)=g2table2a(5,1)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==15 .or. i3==16) THEN
                    g2table2a(6,1)=g2table2a(6,1)+gtype2a(m,i3,i2,i1,i0)
                ENDIF
                g2table2b(i3,1)=g2table2b(i3,1)+gtype2a(m,i3,i2,i1,i0)
                IF (i1==1 .or. i1==4) THEN
                    g2table4a(1,1)=g2table4a(1,1)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(1,1) =g2table4(1,1) +gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==2 .or. i1==5) THEN
                    g2table4a(2,1)=g2table4a(2,1)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(1,1) =g2table4(1,1) +gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==3 .or. i1==6) THEN
                    g2table4a(3,1)=g2table4a(3,1)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(2,1) =g2table4(2,1) +gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==7 .or. i1==10) THEN
                    g2table4a(4,1)=g2table4a(4,1)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(1,1) =g2table4(1,1) +gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==8 .or. i1==11) THEN
                    g2table4a(5,1)=g2table4a(5,1)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(1,1) =g2table4(1,1) +gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==9 .or. i1==12) THEN
                    g2table4a(6,1)=g2table4a(6,1)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(2,1) =g2table4(2,1) +gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==13.or. i1==14) THEN
                    g2table4a(i1-6,1)=g2table4a(i1-6,1)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(3,1) =g2table4(3,1) +gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==15) THEN
                    g2table4a(9,1) =g2table4a(9,1)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(4,1) =g2table4(4,1) +gtype2a(m,i3,i2,i1,i0)
                ENDIF
                g2table4b(i1,1)=g2table4b(i1,1)+gtype2a(m,i3,i2,i1,i0)
                g2table5(MOD(i2,4),1)=g2table5(MOD(i2,4),1)+gtype2a(m,i3,i2,i1,i0)
            IF     (i0==0) THEN
                IF (i3==0) THEN
                    g2table2a(i3,2)=g2table2a(i3,2)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==1 .or. i3==2 .or. i3==3) THEN
                    g2table2a(1,2)=g2table2a(1,2)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==4 .or. i3==5 .or. i3==6) THEN
                    g2table2a(2,2)=g2table2a(2,2)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==7 .or. i3==8 .or. i3==9) THEN
                    g2table2a(3,2)=g2table2a(3,2)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==10 .or. i3==11 .or. i3==12) THEN
                    g2table2a(4,2)=g2table2a(4,2)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==13 .or. i3==14) THEN
                    g2table2a(5,2)=g2table2a(5,2)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==15 .or. i3==16) THEN
                    g2table2a(6,2)=g2table2a(6,2)+gtype2a(m,i3,i2,i1,i0)
                ENDIF
                g2table2b(i3,2)=g2table2b(i3,2)+gtype2a(m,i3,i2,i1,i0)
                IF (i1==1 .or. i1==4) THEN
                    g2table4a(1,2)=g2table4a(1,2)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(1,2) =g2table4(1,2) +gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==2 .or. i1==5) THEN
                    g2table4a(2,2)=g2table4a(2,2)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(1,2) =g2table4(1,2) +gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==3 .or. i1==6) THEN
                    g2table4a(3,2)=g2table4a(3,2)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(2,2) =g2table4(2,2) +gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==7 .or. i1==10) THEN
                    g2table4a(4,2)=g2table4a(4,2)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(1,2) =g2table4(1,2) +gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==8 .or. i1==11) THEN
                    g2table4a(5,2)=g2table4a(5,2)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(1,2) =g2table4(1,2) +gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==9 .or. i1==12) THEN
                    g2table4a(6,2)=g2table4a(6,2)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(2,2) =g2table4(2,2) +gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==13.or. i1==14) THEN
                    g2table4a(i1-6,2)=g2table4a(i1-6,2)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(3,2) =g2table4(3,2) +gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==15) THEN
                    g2table4a(9,2) =g2table4a(9,2)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(4,2) =g2table4(4,2) +gtype2a(m,i3,i2,i1,i0)
                ENDIF
                g2table4b(i1,2)=g2table4b(i1,2)+gtype2a(m,i3,i2,i1,i0)
                g2table5(MOD(i2,4),2)=g2table5(MOD(i2,4),2)+gtype2a(m,i3,i2,i1,i0)
            ELSEIF (i0==1) THEN
                IF (i3==0) THEN
                    g2table2a(i3,3)=g2table2a(i3,3)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==1 .or. i3==2 .or. i3==3) THEN
                    g2table2a(1,3)=g2table2a(1,3)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==4 .or. i3==5 .or. i3==6) THEN
                    g2table2a(2,3)=g2table2a(2,3)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==7 .or. i3==8 .or. i3==9) THEN
                    g2table2a(3,3)=g2table2a(3,3)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==10 .or. i3==11 .or. i3==12) THEN
                    g2table2a(4,3)=g2table2a(4,3)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==13 .or. i3==14) THEN
                    g2table2a(5,3)=g2table2a(5,3)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==15 .or. i3==16) THEN
                    g2table2a(6,3)=g2table2a(6,3)+gtype2a(m,i3,i2,i1,i0)
                ENDIF
                g2table2b(i3,3)=g2table2b(i3,3)+gtype2a(m,i3,i2,i1,i0)
                IF (i1==1 .or. i1==4) THEN
                    g2table4a(1,3)=g2table4a(1,3)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(1,3) =g2table4(1,3) +gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==2 .or. i1==5) THEN
                    g2table4a(2,3)=g2table4a(2,3)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(1,3) =g2table4(1,3) +gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==3 .or. i1==6) THEN
                    g2table4a(3,3)=g2table4a(3,3)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(2,3) =g2table4(2,3) +gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==7 .or. i1==10) THEN
                    g2table4a(4,3)=g2table4a(4,3)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(1,3) =g2table4(1,3) +gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==8 .or. i1==11) THEN
                    g2table4a(5,3)=g2table4a(5,3)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(1,3) =g2table4(1,3) +gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==9 .or. i1==12) THEN
                    g2table4a(6,3)=g2table4a(6,3)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(2,3) =g2table4(2,3) +gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==13.or. i1==14) THEN
                    g2table4a(i1-6,3)=g2table4a(i1-6,3)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(3,3) =g2table4(3,3) +gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==15) THEN
                    g2table4a(9,3) =g2table4a(9,3)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(4,3) =g2table4(4,3) +gtype2a(m,i3,i2,i1,i0)
                ENDIF
                g2table4b(i1,3)=g2table4b(i1,3)+gtype2a(m,i3,i2,i1,i0)
                g2table5(MOD(i2,4),3)=g2table5(MOD(i2,4),3)+gtype2a(m,i3,i2,i1,i0)
            ELSEIF (i0==2) THEN
                IF (i3==0) THEN
                    g2table2a(i3,4)=g2table2a(i3,4)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==1 .or. i3==2 .or. i3==3) THEN
                    g2table2a(1,4)=g2table2a(1,4)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==4 .or. i3==5 .or. i3==6) THEN
                    g2table2a(2,4)=g2table2a(2,4)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==7 .or. i3==8 .or. i3==9) THEN
                    g2table2a(3,4)=g2table2a(3,4)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==10 .or. i3==11 .or. i3==12) THEN
                    g2table2a(4,4)=g2table2a(4,4)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==13 .or. i3==14) THEN
                    g2table2a(5,4)=g2table2a(5,4)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==15 .or. i3==16) THEN
                    g2table2a(6,4)=g2table2a(6,4)+gtype2a(m,i3,i2,i1,i0)
                ENDIF
                g2table2b(i3,4)=g2table2b(i3,4)+gtype2a(m,i3,i2,i1,i0)
                IF (i1==1 .or. i1==4) THEN
                    g2table4a(1,4)=g2table4a(1,4)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(1,4) =g2table4(1,4) +gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==2 .or. i1==5) THEN
                    g2table4a(2,4)=g2table4a(2,4)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(1,4) =g2table4(1,4) +gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==3 .or. i1==6) THEN
                    g2table4a(3,4)=g2table4a(3,4)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(2,4) =g2table4(2,4) +gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==7 .or. i1==10) THEN
                    g2table4a(4,4)=g2table4a(4,4)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(1,4) =g2table4(1,4) +gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==8 .or. i1==11) THEN
                    g2table4a(5,4)=g2table4a(5,4)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(1,4) =g2table4(1,4) +gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==9 .or. i1==12) THEN
                    g2table4a(6,4)=g2table4a(6,4)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(2,4) =g2table4(2,4) +gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==13.or. i1==14) THEN
                    g2table4a(i1-6,4)=g2table4a(i1-6,4)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(3,4) =g2table4(3,4) +gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==15) THEN
                    g2table4a(9,4) =g2table4a(9,4)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(4,4) =g2table4(4,4) +gtype2a(m,i3,i2,i1,i0)
                ENDIF
                g2table4b(i1,4)=g2table4b(i1,4)+gtype2a(m,i3,i2,i1,i0)
                g2table5(MOD(i2,4),4)=g2table5(MOD(i2,4),4)+gtype2a(m,i3,i2,i1,i0)
            ELSEIF (i0==3) THEN
                IF (i3==0) THEN
                    g2table2a(i3,5)=g2table2a(i3,5)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==1 .or. i3==2 .or. i3==3) THEN
                    g2table2a(1,5)=g2table2a(1,5)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==4 .or. i3==5 .or. i3==6) THEN
                    g2table2a(2,5)=g2table2a(2,5)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==7 .or. i3==8 .or. i3==9) THEN
                    g2table2a(3,5)=g2table2a(3,5)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==10 .or. i3==11 .or. i3==12) THEN
                    g2table2a(4,5)=g2table2a(4,5)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==13 .or. i3==14) THEN
                    g2table2a(5,5)=g2table2a(5,5)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==15 .or. i3==16) THEN
                    g2table2a(6,5)=g2table2a(6,5)+gtype2a(m,i3,i2,i1,i0)
                ENDIF
                g2table2b(i3,5)=g2table2b(i3,5)+gtype2a(m,i3,i2,i1,i0)
                IF (i1==1 .or. i1==4) THEN
                    g2table4a(1,5)=g2table4a(1,5)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(1,5) =g2table4(1,5) +gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==2 .or. i1==5) THEN
                    g2table4a(2,5)=g2table4a(2,5)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(1,5) =g2table4(1,5) +gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==3 .or. i1==6) THEN
                    g2table4a(3,5)=g2table4a(3,5)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(2,5) =g2table4(2,5) +gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==7 .or. i1==10) THEN
                    g2table4a(4,5)=g2table4a(4,5)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(1,5) =g2table4(1,5) +gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==8 .or. i1==11) THEN
                    g2table4a(5,5)=g2table4a(5,5)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(1,5) =g2table4(1,5) +gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==9 .or. i1==12) THEN
                    g2table4a(6,5)=g2table4a(6,5)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(2,5) =g2table4(2,5) +gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==13.or. i1==14) THEN
                    g2table4a(i1-6,5)=g2table4a(i1-6,5)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(3,5) =g2table4(3,5) +gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==15) THEN
                    g2table4a(9,5) =g2table4a(9,5)+gtype2a(m,i3,i2,i1,i0)
                    g2table4(4,5) =g2table4(4,5) +gtype2a(m,i3,i2,i1,i0)
                ENDIF
                g2table4b(i1,5)=g2table4b(i1,5)+gtype2a(m,i3,i2,i1,i0)
                g2table5(MOD(i2,4),5)=g2table5(MOD(i2,4),5)+gtype2a(m,i3,i2,i1,i0)
            ENDIF

        ENDDO
        ENDDO
        ENDDO
        ENDDO
    ENDIF
ENDDO
    g2table3a(:,:) =g2table2a(:,:)
    g2table2a(:,:) =g2table2a(:,:)/gtct_2c
        g2table3a(:,1) =g2table3a(:,1)/gtct_2c
        g2table3a(:,2) =g2table3a(:,2)/gtct2_la0
        g2table3a(:,3) =g2table3a(:,3)/gtct2_la1
        g2table3a(:,4) =g2table3a(:,4)/gtct2_la2
        g2table3a(:,5) =g2table3a(:,5)/gtct2_la3

    g2table3b(:,:) =g2table2b(:,:)
    g2table2b(:,:) =g2table2b(:,:)/gtct_2c
        g2table3b(:,1) =g2table3b(:,1)/gtct_2c
        g2table3b(:,2) =g2table3b(:,2)/gtct2_la0
        g2table3b(:,3) =g2table3b(:,3)/gtct2_la1
        g2table3b(:,4) =g2table3b(:,4)/gtct2_la2
        g2table3b(:,5) =g2table3b(:,5)/gtct2_la3

    g2table4(:,1) =g2table4(:,1)/gtct_2c
    g2table4(:,2) =g2table4(:,2)/gtct2_la0
    g2table4(:,3) =g2table4(:,3)/gtct2_la1
    g2table4(:,4) =g2table4(:,4)/gtct2_la2
    g2table4(:,5) =g2table4(:,5)/gtct2_la3

    g2table4a(:,1) =g2table4a(:,1)/gtct_2c
    g2table4a(:,2) =g2table4a(:,2)/gtct2_la0
    g2table4a(:,3) =g2table4a(:,3)/gtct2_la1
    g2table4a(:,4) =g2table4a(:,4)/gtct2_la2
    g2table4a(:,5) =g2table4a(:,5)/gtct2_la3

    g2table4b(:,1) =g2table4b(:,1)/gtct_2c
    g2table4b(:,2) =g2table4b(:,2)/gtct2_la0
    g2table4b(:,3) =g2table4b(:,3)/gtct2_la1
    g2table4b(:,4) =g2table4b(:,4)/gtct2_la2
    g2table4b(:,5) =g2table4b(:,5)/gtct2_la3

    g2table5(:,1) =g2table5(:,1)/gtct_2c
    g2table5(:,2) =g2table5(:,2)/gtct2_la0
    g2table5(:,3) =g2table5(:,3)/gtct2_la1
    g2table5(:,4) =g2table5(:,4)/gtct2_la2
    g2table5(:,5) =g2table5(:,5)/gtct2_la3

g2table6a=0.0d0 ; g2table6a2=0.0d0
g2table6b=0.0d0 ; g2table6b2=0.0d0
g2table6c=0.0d0 ; g2table6c2=0.0d0

DO m=1,nfam
    IF (nchild(m)==2) THEN
        DO i0=0, 3
        DO i2=0, 11
        DO i4=0, 15
                g2table6c(i4,1)=g2table6c(i4,1)+gtype2b(m,i4,i2,i0)
                g2table6a(i4/4,1)=g2table6a(i4/4,1)+gtype2b(m,i4,i2,i0)
                g2table6b(MOD(i4,4),1)=g2table6b(MOD(i4,4),1)+gtype2b(m,i4,i2,i0)
            IF     (i0==0) THEN
                g2table6c(i4,2)=g2table6c(i4,2)+gtype2b(m,i4,i2,i0)
                g2table6a(i4/4,2)=g2table6a(i4/4,2)+gtype2b(m,i4,i2,i0)
                g2table6b(MOD(i4,4),2)=g2table6b(MOD(i4,4),2)+gtype2b(m,i4,i2,i0)
            ELSEIF (i0==1) THEN
                g2table6c(i4,3)=g2table6c(i4,3)+gtype2b(m,i4,i2,i0)
                g2table6a(i4/4,3)=g2table6a(i4/4,3)+gtype2b(m,i4,i2,i0)
                g2table6b(MOD(i4,4),3)=g2table6b(MOD(i4,4),3)+gtype2b(m,i4,i2,i0)
            ELSEIF (i0==2) THEN
                g2table6c(i4,4)=g2table6c(i4,4)+gtype2b(m,i4,i2,i0)
                g2table6a(i4/4,4)=g2table6a(i4/4,4)+gtype2b(m,i4,i2,i0)
                g2table6b(MOD(i4,4),4)=g2table6b(MOD(i4,4),4)+gtype2b(m,i4,i2,i0)
            ELSEIF (i0==3) THEN
                g2table6c(i4,5)=g2table6c(i4,5)+gtype2b(m,i4,i2,i0)
                g2table6a(i4/4,5)=g2table6a(i4/4,5)+gtype2b(m,i4,i2,i0)
                g2table6b(MOD(i4,4),5)=g2table6b(MOD(i4,4),5)+gtype2b(m,i4,i2,i0)
            ENDIF
        ENDDO
        ENDDO
        ENDDO
    ENDIF
ENDDO

    g2table6a2(:,:) =g2table6a(:,:)
    g2table6b2(:,:) =g2table6b(:,:)
    g2table6c2(:,:) =g2table6c(:,:)

    g2table6a(:,:) =g2table6a(:,:)/gtct_2c
    g2table6b(:,:) =g2table6b(:,:)/gtct_2c
    g2table6c(:,:) =g2table6c(:,:)/gtct_2c

    g2table6a2(:,1) =g2table6a2(:,1)/gtct_2c ;  g2table6a2(:,2) =g2table6a2(:,2)/gtct2_la0 ;  g2table6a2(:,3) =g2table6a2(:,3)/gtct2_la1 ;  g2table6a2(:,4) =g2table6a2(:,4)/gtct2_la2 ;  g2table6a2(:,5) =g2table6a2(:,5)/gtct2_la3
    g2table6b2(:,1) =g2table6b2(:,1)/gtct_2c ;  g2table6b2(:,2) =g2table6b2(:,2)/gtct2_la0 ;  g2table6b2(:,3) =g2table6b2(:,3)/gtct2_la1 ;  g2table6b2(:,4) =g2table6b2(:,4)/gtct2_la2 ;  g2table6b2(:,5) =g2table6b2(:,5)/gtct2_la3
    g2table6c2(:,1) =g2table6c2(:,1)/gtct_2c ;  g2table6c2(:,2) =g2table6c2(:,2)/gtct2_la0 ;  g2table6c2(:,3) =g2table6c2(:,3)/gtct2_la1 ;  g2table6c2(:,4) =g2table6c2(:,4)/gtct2_la2 ;  g2table6c2(:,5) =g2table6c2(:,5)/gtct2_la3




g2table7=0.0d0
g2table8=0.0d0
DO m=1,nfam
    IF (nchild(m)==2) THEN
        DO i0=0, 3
        DO i2=0, 11
        DO i5=0, 6
        DO i6=0, 5
                g2table7(i6,1)=g2table7(i6,1)+gtype2c(m,i6,i5,i2,i0)
                g2table8(i5,1)=g2table8(i5,1)+gtype2c(m,i6,i5,i2,i0)
            IF     (i0==0) THEN
                g2table7(i6,2)=g2table7(i6,2)+gtype2c(m,i6,i5,i2,i0)
                g2table8(i5,2)=g2table8(i5,2)+gtype2c(m,i6,i5,i2,i0)
            ELSEIF (i0==1) THEN
                g2table7(i6,3)=g2table7(i6,3)+gtype2c(m,i6,i5,i2,i0)
                g2table8(i5,3)=g2table8(i5,3)+gtype2c(m,i6,i5,i2,i0)
            ELSEIF (i0==2) THEN
                g2table7(i6,4)=g2table7(i6,4)+gtype2c(m,i6,i5,i2,i0)
                g2table8(i5,4)=g2table8(i5,4)+gtype2c(m,i6,i5,i2,i0)
            ELSEIF (i0==3) THEN
                g2table7(i6,5)=g2table7(i6,5)+gtype2c(m,i6,i5,i2,i0)
                g2table8(i5,5)=g2table8(i5,5)+gtype2c(m,i6,i5,i2,i0)
            ENDIF
        ENDDO
        ENDDO
        ENDDO
        ENDDO
    ENDIF
ENDDO
    g2table7(:,1) =g2table7(:,1)/gtct_2c
    g2table7(:,2) =g2table7(:,2)/gtct2_la0
    g2table7(:,3) =g2table7(:,3)/gtct2_la1
    g2table7(:,4) =g2table7(:,4)/gtct2_la2
    g2table7(:,5) =g2table7(:,5)/gtct2_la3

    g2table8(:,1) =g2table8(:,1)/gtct_2c
    g2table8(:,2) =g2table8(:,2)/gtct2_la0
    g2table8(:,3) =g2table8(:,3)/gtct2_la1
    g2table8(:,4) =g2table8(:,4)/gtct2_la2
    g2table8(:,5) =g2table8(:,5)/gtct2_la3


g2table1=0.0d0 
g2table9a=0.0d0 ; g2table10a=0.0d0
g2table9b=0.0d0 ; g2table10b=0.0d0
g2table11a=0.0d0 ; g2table11b=0.0d0
g2table12=0.0d0
DO m=1,nfam
    IF (nchild(m)==2) THEN
        DO i0=0, 3
        DO i1=1, 15
        DO i2=0, 11
        DO i3=0, 16
                g2table1(i0,1)=g2table1(i0,1)+gtype2a(m,i3,i2,i1,i0)
                IF (i3==0) THEN
                    g2table9a(i3,1)=g2table9a(i3,1)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==1 .or. i3==2 .or. i3==3) THEN
                    g2table9a(1,1)=g2table9a(1,1)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==4 .or. i3==5 .or. i3==6) THEN
                    g2table9a(2,1)=g2table9a(2,1)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==7 .or. i3==8 .or. i3==9) THEN
                    g2table9a(3,1)=g2table9a(3,1)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==10 .or. i3==11 .or. i3==12) THEN
                    g2table9a(4,1)=g2table9a(4,1)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==13 .or. i3==14) THEN
                    g2table9a(5,1)=g2table9a(5,1)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==15 .or. i3==16) THEN
                    g2table9a(6,1)=g2table9a(6,1)+gtype2a(m,i3,i2,i1,i0)
                ENDIF
                g2table9b(i3,1)=g2table9b(i3,1)+gtype2a(m,i3,i2,i1,i0)
                IF (i1==1 .or. i1==4) THEN
                    g2table11a(1,1)=g2table11a(1,1)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==2 .or. i1==5) THEN
                    g2table11a(2,1)=g2table11a(2,1)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==3 .or. i1==6) THEN
                    g2table11a(3,1)=g2table11a(3,1)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==7 .or. i1==10) THEN
                    g2table11a(4,1)=g2table11a(4,1)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==8 .or. i1==11) THEN
                    g2table11a(5,1)=g2table11a(5,1)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==9 .or. i1==12) THEN
                    g2table11a(6,1)=g2table11a(6,1)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1>=13) THEN
                    g2table11a(i1-6,1)=g2table11a(i1-6,1)+gtype2a(m,i3,i2,i1,i0)
                ENDIF
                g2table11b(i1,1)=g2table11b(i1,1)+gtype2a(m,i3,i2,i1,i0)
                g2table12(MOD(i2,4),1)=g2table12(MOD(i2,4),1)+gtype2a(m,i3,i2,i1,i0)
            IF     (i2==0 .or. i2==1 .or. i2==2 .or. i2==3) THEN
                g2table1(i0,2)=g2table1(i0,2)+gtype2a(m,i3,i2,i1,i0)
                IF (i3==0) THEN
                    g2table9a(i3,2)=g2table9a(i3,2)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==1 .or. i3==2 .or. i3==3) THEN
                    g2table9a(1,2)=g2table9a(1,2)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==4 .or. i3==5 .or. i3==6) THEN
                    g2table9a(2,2)=g2table9a(2,2)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==7 .or. i3==8 .or. i3==9) THEN
                    g2table9a(3,2)=g2table9a(3,2)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==10 .or. i3==11 .or. i3==12) THEN
                    g2table9a(4,2)=g2table9a(4,2)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==13 .or. i3==14) THEN
                    g2table9a(5,2)=g2table9a(5,2)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==15 .or. i3==16) THEN
                    g2table9a(6,2)=g2table9a(6,2)+gtype2a(m,i3,i2,i1,i0)
                ENDIF
                g2table9b(i3,2)=g2table9b(i3,2)+gtype2a(m,i3,i2,i1,i0)
                IF (i1==1 .or. i1==4) THEN
                    g2table11a(1,2)=g2table11a(1,2)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==2 .or. i1==5) THEN
                    g2table11a(2,2)=g2table11a(2,2)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==3 .or. i1==6) THEN
                    g2table11a(3,2)=g2table11a(3,2)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==7 .or. i1==10) THEN
                    g2table11a(4,2)=g2table11a(4,2)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==8 .or. i1==11) THEN
                    g2table11a(5,2)=g2table11a(5,2)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==9 .or. i1==12) THEN
                    g2table11a(6,2)=g2table11a(6,2)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1>=13) THEN
                    g2table11a(i1-6,2)=g2table11a(i1-6,2)+gtype2a(m,i3,i2,i1,i0)
                ENDIF
                g2table11b(i1,2)=g2table11b(i1,2)+gtype2a(m,i3,i2,i1,i0)
                g2table12(MOD(i2,4),2)=g2table12(MOD(i2,4),2)+gtype2a(m,i3,i2,i1,i0)
            ELSEIF (i2==4 .or. i2==5 .or. i2==6 .or. i2==7) THEN
                g2table1(i0,3)=g2table1(i0,3)+gtype2a(m,i3,i2,i1,i0)
                IF (i3==0) THEN
                    g2table9a(i3,3)=g2table9a(i3,3)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==1 .or. i3==2 .or. i3==3) THEN
                    g2table9a(1,3)=g2table9a(1,3)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==4 .or. i3==5 .or. i3==6) THEN
                    g2table9a(2,3)=g2table9a(2,3)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==7 .or. i3==8 .or. i3==9) THEN
                    g2table9a(3,3)=g2table9a(3,3)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==10 .or. i3==11 .or. i3==12) THEN
                    g2table9a(4,3)=g2table9a(4,3)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==13 .or. i3==14) THEN
                    g2table9a(5,3)=g2table9a(5,3)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==15 .or. i3==16) THEN
                    g2table9a(6,3)=g2table9a(6,3)+gtype2a(m,i3,i2,i1,i0)
                ENDIF
                g2table9b(i3,3)=g2table9b(i3,3)+gtype2a(m,i3,i2,i1,i0)
                IF (i1==1 .or. i1==4) THEN
                    g2table11a(1,3)=g2table11a(1,3)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==2 .or. i1==5) THEN
                    g2table11a(2,3)=g2table11a(2,3)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==3 .or. i1==6) THEN
                    g2table11a(3,3)=g2table11a(3,3)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==7 .or. i1==10) THEN
                    g2table11a(4,3)=g2table11a(4,3)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==8 .or. i1==11) THEN
                    g2table11a(5,3)=g2table11a(5,3)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==9 .or. i1==12) THEN
                    g2table11a(6,3)=g2table11a(6,3)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1>=13) THEN
                    g2table11a(i1-6,3)=g2table11a(i1-6,3)+gtype2a(m,i3,i2,i1,i0)
                ENDIF
                g2table11b(i1,3)=g2table11b(i1,3)+gtype2a(m,i3,i2,i1,i0)
                g2table12(MOD(i2,4),3)=g2table12(MOD(i2,4),3)+gtype2a(m,i3,i2,i1,i0)
            ELSEIF (i2==8 .or. i2==9 .or. i2==10 .or. i2==11) THEN
                g2table1(i0,4)=g2table1(i0,4)+gtype2a(m,i3,i2,i1,i0)
                IF (i3==0) THEN
                    g2table9a(i3,4)=g2table9a(i3,4)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==1 .or. i3==2 .or. i3==3) THEN
                    g2table9a(1,4)=g2table9a(1,4)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==4 .or. i3==5 .or. i3==6) THEN
                    g2table9a(2,4)=g2table9a(2,4)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==7 .or. i3==8 .or. i3==9) THEN
                    g2table9a(3,4)=g2table9a(3,4)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==10 .or. i3==11 .or. i3==12) THEN
                    g2table9a(4,4)=g2table9a(4,4)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==13 .or. i3==14) THEN
                    g2table9a(5,4)=g2table9a(5,4)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==15 .or. i3==16) THEN
                    g2table9a(6,4)=g2table9a(6,4)+gtype2a(m,i3,i2,i1,i0)
                ENDIF
                g2table9b(i3,4)=g2table9b(i3,4)+gtype2a(m,i3,i2,i1,i0)
                IF (i1==1 .or. i1==4) THEN
                    g2table11a(1,4)=g2table11a(1,4)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==2 .or. i1==5) THEN
                    g2table11a(2,4)=g2table11a(2,4)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==3 .or. i1==6) THEN
                    g2table11a(3,4)=g2table11a(3,4)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==7 .or. i1==10) THEN
                    g2table11a(4,4)=g2table11a(4,4)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==8 .or. i1==11) THEN
                    g2table11a(5,4)=g2table11a(5,4)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==9 .or. i1==12) THEN
                    g2table11a(6,4)=g2table11a(6,4)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1>=13) THEN
                    g2table11a(i1-6,4)=g2table11a(i1-6,4)+gtype2a(m,i3,i2,i1,i0)
                ENDIF
                g2table11b(i1,4)=g2table11b(i1,4)+gtype2a(m,i3,i2,i1,i0)
                g2table12(MOD(i2,4),4)=g2table12(MOD(i2,4),4)+gtype2a(m,i3,i2,i1,i0)
            ENDIF

        ENDDO
        ENDDO
        ENDDO
        ENDDO
    ENDIF
ENDDO
    g2table1(:,:) =g2table1(:,:)/gtct_2c
    g2table10a(:,:) =g2table9a(:,:)
    g2table10b(:,:) =g2table9b(:,:)
    g2table9a(:,:) =g2table9a(:,:)/gtct_2c
    g2table9b(:,:) =g2table9b(:,:)/gtct_2c
    g2table10a(:,1) =g2table10a(:,1)/gtct_2c
    g2table10b(:,1) =g2table10b(:,1)/gtct_2c
    g2table11a(:,1) =g2table11a(:,1)/gtct_2c
    g2table11b(:,1) =g2table11b(:,1)/gtct_2c
    g2table12(:,1) =g2table12(:,1)/gtct_2c

    IF (gtct2_eff1>0.0d0) THEN
        g2table10a(:,2) =g2table10a(:,2)/gtct2_eff1
        g2table10b(:,2) =g2table10b(:,2)/gtct2_eff1
        g2table11a(:,2) =g2table11a(:,2)/gtct_2c
        g2table11b(:,2) =g2table11b(:,2)/gtct_2c
        g2table12(:,2) =g2table12(:,2)/gtct_2c
    ELSE
        g2table10a(:,2) =0.0d0
        g2table10b(:,2) =0.0d0
        g2table11a(:,2) =0.0d0
        g2table11b(:,2) =0.0d0
        g2table12(:,2) =0.0d0
    ENDIF
    IF (gtct2_eff2>0.0d0) THEN
        g2table10a(:,3) =g2table10a(:,3)/gtct2_eff2
        g2table10b(:,3) =g2table10b(:,3)/gtct2_eff2
        g2table11a(:,3) =g2table11a(:,3)/gtct_2c
        g2table11b(:,3) =g2table11b(:,3)/gtct_2c
        g2table12(:,3) =g2table12(:,3)/gtct_2c
    ELSE
        g2table10a(:,3) =0.0d0
        g2table10b(:,3) =0.0d0
        g2table11a(:,3) =0.0d0
        g2table11b(:,3) =0.0d0
        g2table12(:,3) =0.0d0
    ENDIF
    IF (gtct2_eff3>0.0d0) THEN
        g2table10a(:,4) =g2table10a(:,4)/gtct2_eff3
        g2table10b(:,4) =g2table10b(:,4)/gtct2_eff3
        g2table11a(:,4) =g2table11a(:,4)/gtct_2c
        g2table11b(:,4) =g2table11b(:,4)/gtct_2c
        g2table12(:,4) =g2table12(:,4)/gtct_2c
    ELSE
        g2table10a(:,4) =0.0d0
        g2table10b(:,4) =0.0d0
        g2table11a(:,4) =0.0d0
        g2table11b(:,4) =0.0d0
        g2table12(:,4) =0.0d0
    ENDIF


g2table13a=0.0d0
g2table13b=0.0d0
g2table13c=0.0d0
g2table14=0.0d0
g2table15=0.0d0
DO m=1,nfam
    IF (nchild(m)==2) THEN
        DO i0=0, 3
        DO i2=0, 11
        DO i4=0, 15
                g2table13a(i4/4,1)=g2table13a(i4/4,1)+gtype2b(m,i4,i2,i0)
                g2table13b(MOD(i4,4),1)=g2table13b(MOD(i4,4),1)+gtype2b(m,i4,i2,i0)
                g2table13c(i4,1)=g2table13c(i4,1)+gtype2b(m,i4,i2,i0)
            IF     (i2==0 .or. i2==1 .or. i2==2 .or. i2==3) THEN
                g2table13a(i4/4,2)=g2table13a(i4/4,2)+gtype2b(m,i4,i2,i0)
                g2table13b(MOD(i4,4),2)=g2table13b(MOD(i4,4),2)+gtype2b(m,i4,i2,i0)
                g2table13c(i4,2)=g2table13c(i4,2)+gtype2b(m,i4,i2,i0)
            ELSEIF (i2==4 .or. i2==5 .or. i2==6 .or. i2==7) THEN
                g2table13a(i4/4,3)=g2table13a(i4/4,3)+gtype2b(m,i4,i2,i0)
                g2table13b(MOD(i4,4),3)=g2table13b(MOD(i4,4),3)+gtype2b(m,i4,i2,i0)
                g2table13c(i4,3)=g2table13c(i4,3)+gtype2b(m,i4,i2,i0)
            ELSEIF (i2==8 .or. i2==9 .or. i2==10 .or. i2==11) THEN
                g2table13a(i4/4,4)=g2table13a(i4/4,4)+gtype2b(m,i4,i2,i0)
                g2table13b(MOD(i4,4),4)=g2table13b(MOD(i4,4),4)+gtype2b(m,i4,i2,i0)
                g2table13c(i4,4)=g2table13c(i4,4)+gtype2b(m,i4,i2,i0)
            ENDIF
        ENDDO
        ENDDO
        ENDDO

        DO i0=0, 3
        DO i2=0, 11
        DO i5=0, 6
        DO i6=0, 5
                g2table14(i6,1)=g2table14(i6,1)+gtype2c(m,i6,i5,i2,i0)
                g2table15(i5,1)=g2table15(i5,1)+gtype2c(m,i6,i5,i2,i0)
            IF     (i2==0 .or. i2==1 .or. i2==2 .or. i2==3) THEN
                g2table14(i6,2)=g2table14(i6,2)+gtype2c(m,i6,i5,i2,i0)
                g2table15(i5,2)=g2table15(i5,2)+gtype2c(m,i6,i5,i2,i0)
            ELSEIF (i2==4 .or. i2==5 .or. i2==6 .or. i2==7) THEN
                g2table14(i6,3)=g2table14(i6,3)+gtype2c(m,i6,i5,i2,i0)
                g2table15(i5,3)=g2table15(i5,3)+gtype2c(m,i6,i5,i2,i0)
            ELSEIF (i2==8 .or. i2==9 .or. i2==10 .or. i2==11) THEN
                g2table14(i6,4)=g2table14(i6,4)+gtype2c(m,i6,i5,i2,i0)
                g2table15(i5,4)=g2table15(i5,4)+gtype2c(m,i6,i5,i2,i0)
            ENDIF
        ENDDO
        ENDDO
        ENDDO
        ENDDO
    ENDIF
ENDDO
    g2table13a(:,1) =g2table13a(:,1)/gtct_2c
    g2table13b(:,1) =g2table13b(:,1)/gtct_2c
    g2table13c(:,1) =g2table13c(:,1)/gtct_2c
    g2table14(:,1) =g2table14(:,1)/gtct_2c
    g2table15(:,1) =g2table15(:,1)/gtct_2c


    IF (gtct2_eff1>0.0d0) THEN
        g2table13a(:,2) =g2table13a(:,2)/gtct2_eff1
        g2table13b(:,2) =g2table13b(:,2)/gtct2_eff1
        g2table13c(:,2) =g2table13c(:,2)/gtct_2c
        g2table14(:,2) =g2table14(:,2)/gtct_2c
        g2table15(:,2) =g2table15(:,2)/gtct_2c
    ELSE
        g2table13a(:,2) =0.0d0
        g2table13b(:,2) =0.0d0
        g2table13c(:,2) =0.0d0
        g2table14(:,2) =0.0d0
        g2table15(:,2) =0.0d0
    ENDIF
    IF (gtct2_eff2>0.0d0) THEN
        g2table13a(:,3) =g2table13a(:,3)/gtct2_eff2
        g2table13b(:,3) =g2table13b(:,3)/gtct2_eff2
        g2table13c(:,3) =g2table13c(:,3)/gtct_2c
        g2table14(:,3) =g2table14(:,3)/gtct_2c
        g2table15(:,3) =g2table15(:,3)/gtct_2c
    ELSE
        g2table13a(:,3) =0.0d0
        g2table13b(:,3) =0.0d0
        g2table13c(:,3) =0.0d0
        g2table14(:,3) =0.0d0
        g2table15(:,3) =0.0d0
    ENDIF
    IF (gtct2_eff3>0.0d0) THEN
        g2table13a(:,4) =g2table13a(:,4)/gtct2_eff3
        g2table13b(:,4) =g2table13b(:,4)/gtct2_eff3
        g2table13c(:,4) =g2table13c(:,4)/gtct_2c
        g2table14(:,4) =g2table14(:,4)/gtct_2c
        g2table15(:,4) =g2table15(:,4)/gtct_2c
    ELSE
        g2table13a(:,4) =0.0d0
        g2table13b(:,4) =0.0d0
        g2table13c(:,4) =0.0d0
        g2table14(:,4) =0.0d0
        g2table15(:,4) =0.0d0
    ENDIF



g2table16a=0.0d0 ; g2table16b=0.0d0
g2table17a=0.0d0 ; g2table17b=0.0d0
DO m=1,nfam
    IF (nchild(m)==2) THEN
        DO i0=0, 3
        DO i1=1, 15
        DO i2=0, 11
        DO i3=0, 16
                IF (i3==0) THEN
                    g2table16a(i3,1)=g2table16a(i3,1)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==1 .or. i3==2 .or. i3==3) THEN
                    g2table16a(1,1)=g2table16a(1,1)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==4 .or. i3==5 .or. i3==6) THEN
                    g2table16a(2,1)=g2table16a(2,1)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==7 .or. i3==8 .or. i3==9) THEN
                    g2table16a(3,1)=g2table16a(3,1)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==10 .or. i3==11 .or. i3==12) THEN
                    g2table16a(4,1)=g2table16a(4,1)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==13 .or. i3==14) THEN
                    g2table16a(5,1)=g2table16a(5,1)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==15 .or. i3==16) THEN
                    g2table16a(6,1)=g2table16a(6,1)+gtype2a(m,i3,i2,i1,i0)
                ENDIF
                g2table16b(i3,1)=g2table16b(i3,1)+gtype2a(m,i3,i2,i1,i0)
                IF (i1==1 .or. i1==4) THEN
                    g2table17a(1,1)=g2table17a(1,1)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==2 .or. i1==5) THEN
                    g2table17a(2,1)=g2table17a(2,1)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==3 .or. i1==6) THEN
                    g2table17a(3,1)=g2table17a(3,1)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==7 .or. i1==10) THEN
                    g2table17a(4,1)=g2table17a(4,1)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==8 .or. i1==11) THEN
                    g2table17a(5,1)=g2table17a(5,1)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==9 .or. i1==12) THEN
                    g2table17a(6,1)=g2table17a(6,1)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1>=13) THEN
                    g2table17a(i1-6,1)=g2table17a(i1-6,1)+gtype2a(m,i3,i2,i1,i0)
                ENDIF
                g2table17b(i1,1)=g2table17b(i1,1)+gtype2a(m,i3,i2,i1,i0)
            IF     (MOD(i2,4)==0) THEN
                IF (i3==0) THEN
                    g2table16a(i3,2)=g2table16a(i3,2)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==1 .or. i3==2 .or. i3==3) THEN
                    g2table16a(1,2)=g2table16a(1,2)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==4 .or. i3==5 .or. i3==6) THEN
                    g2table16a(2,2)=g2table16a(2,2)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==7 .or. i3==8 .or. i3==9) THEN
                    g2table16a(3,2)=g2table16a(3,2)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==10 .or. i3==11 .or. i3==12) THEN
                    g2table16a(4,2)=g2table16a(4,2)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==13 .or. i3==14) THEN
                    g2table16a(5,2)=g2table16a(5,2)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==15 .or. i3==16) THEN
                    g2table16a(6,2)=g2table16a(6,2)+gtype2a(m,i3,i2,i1,i0)
                ENDIF
                g2table16b(i3,2)=g2table16b(i3,2)+gtype2a(m,i3,i2,i1,i0)
                IF (i1==1 .or. i1==4) THEN
                    g2table17a(1,2)=g2table17a(1,2)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==2 .or. i1==5) THEN
                    g2table17a(2,2)=g2table17a(2,2)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==3 .or. i1==6) THEN
                    g2table17a(3,2)=g2table17a(3,2)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==7 .or. i1==10) THEN
                    g2table17a(4,2)=g2table17a(4,2)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==8 .or. i1==11) THEN
                    g2table17a(5,2)=g2table17a(5,2)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==9 .or. i1==12) THEN
                    g2table17a(6,2)=g2table17a(6,2)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1>=13) THEN
                    g2table17a(i1-6,2)=g2table17a(i1-6,2)+gtype2a(m,i3,i2,i1,i0)
                ENDIF
                g2table17b(i1,2)=g2table17b(i1,2)+gtype2a(m,i3,i2,i1,i0)
            ELSEIF (MOD(i2,4)==1) THEN
                IF (i3==0) THEN
                    g2table16a(i3,3)=g2table16a(i3,3)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==1 .or. i3==2 .or. i3==3) THEN
                    g2table16a(1,3)=g2table16a(1,3)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==4 .or. i3==5 .or. i3==6) THEN
                    g2table16a(2,3)=g2table16a(2,3)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==7 .or. i3==8 .or. i3==9) THEN
                    g2table16a(3,3)=g2table16a(3,3)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==10 .or. i3==11 .or. i3==12) THEN
                    g2table16a(4,3)=g2table16a(4,3)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==13 .or. i3==14) THEN
                    g2table16a(5,3)=g2table16a(5,3)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==15 .or. i3==16) THEN
                    g2table16a(6,3)=g2table16a(6,3)+gtype2a(m,i3,i2,i1,i0)
                ENDIF
                g2table16b(i3,3)=g2table16b(i3,3)+gtype2a(m,i3,i2,i1,i0)
                IF (i1==1 .or. i1==4) THEN
                    g2table17a(1,3)=g2table17a(1,3)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==2 .or. i1==5) THEN
                    g2table17a(2,3)=g2table17a(2,3)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==3 .or. i1==6) THEN
                    g2table17a(3,3)=g2table17a(3,3)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==7 .or. i1==10) THEN
                    g2table17a(4,3)=g2table17a(4,3)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==8 .or. i1==11) THEN
                    g2table17a(5,3)=g2table17a(5,3)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==9 .or. i1==12) THEN
                    g2table17a(6,3)=g2table17a(6,3)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1>=13) THEN
                    g2table17a(i1-6,3)=g2table17a(i1-6,3)+gtype2a(m,i3,i2,i1,i0)
                ENDIF
                g2table17b(i1,3)=g2table17b(i1,3)+gtype2a(m,i3,i2,i1,i0)
            ELSEIF (MOD(i2,4)==2) THEN
                IF (i3==0) THEN
                    g2table16a(i3,4)=g2table16a(i3,4)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==1 .or. i3==2 .or. i3==3) THEN
                    g2table16a(1,4)=g2table16a(1,4)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==4 .or. i3==5 .or. i3==6) THEN
                    g2table16a(2,4)=g2table16a(2,4)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==7 .or. i3==8 .or. i3==9) THEN
                    g2table16a(3,4)=g2table16a(3,4)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==10 .or. i3==11 .or. i3==12) THEN
                    g2table16a(4,4)=g2table16a(4,4)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==13 .or. i3==14) THEN
                    g2table16a(5,4)=g2table16a(5,4)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==15 .or. i3==16) THEN
                    g2table16a(6,4)=g2table16a(6,4)+gtype2a(m,i3,i2,i1,i0)
                ENDIF
                g2table16b(i3,4)=g2table16b(i3,4)+gtype2a(m,i3,i2,i1,i0)
                IF (i1==1 .or. i1==4) THEN
                    g2table17a(1,4)=g2table17a(1,4)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==2 .or. i1==5) THEN
                    g2table17a(2,4)=g2table17a(2,4)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==3 .or. i1==6) THEN
                    g2table17a(3,4)=g2table17a(3,4)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==7 .or. i1==10) THEN
                    g2table17a(4,4)=g2table17a(4,4)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==8 .or. i1==11) THEN
                    g2table17a(5,4)=g2table17a(5,4)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==9 .or. i1==12) THEN
                    g2table17a(6,4)=g2table17a(6,4)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1>=13) THEN
                    g2table17a(i1-6,4)=g2table17a(i1-6,4)+gtype2a(m,i3,i2,i1,i0)
                ENDIF
                g2table17b(i1,4)=g2table17b(i1,4)+gtype2a(m,i3,i2,i1,i0)
            ELSEIF (MOD(i2,4)==3) THEN
                IF (i3==0) THEN
                    g2table16a(i3,5)=g2table16a(i3,5)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==1 .or. i3==2 .or. i3==3) THEN
                    g2table16a(1,5)=g2table16a(1,5)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==4 .or. i3==5 .or. i3==6) THEN
                    g2table16a(2,5)=g2table16a(2,5)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==7 .or. i3==8 .or. i3==9) THEN
                    g2table16a(3,5)=g2table16a(3,5)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==10 .or. i3==11 .or. i3==12) THEN
                    g2table16a(4,5)=g2table16a(4,5)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==13 .or. i3==14) THEN
                    g2table16a(5,5)=g2table16a(5,5)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i3==15 .or. i3==16) THEN
                    g2table16a(6,5)=g2table16a(6,5)+gtype2a(m,i3,i2,i1,i0)
                ENDIF
                g2table16b(i3,5)=g2table16b(i3,5)+gtype2a(m,i3,i2,i1,i0)
                IF (i1==1 .or. i1==4) THEN
                    g2table17a(1,5)=g2table17a(1,5)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==2 .or. i1==5) THEN
                    g2table17a(2,5)=g2table17a(2,5)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==3 .or. i1==6) THEN
                    g2table17a(3,5)=g2table17a(3,5)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==7 .or. i1==10) THEN
                    g2table17a(4,5)=g2table17a(4,5)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==8 .or. i1==11) THEN
                    g2table17a(5,5)=g2table17a(5,5)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1==9 .or. i1==12) THEN
                    g2table17a(6,5)=g2table17a(6,5)+gtype2a(m,i3,i2,i1,i0)
                ELSEIF (i1>=13) THEN
                    g2table17a(i1-6,5)=g2table17a(i1-6,5)+gtype2a(m,i3,i2,i1,i0)
                ENDIF
                g2table17b(i1,5)=g2table17b(i1,5)+gtype2a(m,i3,i2,i1,i0)
            ENDIF

        ENDDO
        ENDDO
        ENDDO
        ENDDO
    ENDIF
ENDDO
    g2table16a(:,1) =g2table16a(:,1)/gtct_2c
    g2table16a(:,2) =g2table16a(:,2)/gtct2_idf0
    g2table16a(:,3) =g2table16a(:,3)/gtct2_idf1
    g2table16a(:,4) =g2table16a(:,4)/gtct2_idf2
    g2table16a(:,5) =g2table16a(:,5)/gtct2_idf3

    g2table16b(:,1) =g2table16b(:,1)/gtct_2c
    g2table16b(:,2) =g2table16b(:,2)/gtct2_idf0
    g2table16b(:,3) =g2table16b(:,3)/gtct2_idf1
    g2table16b(:,4) =g2table16b(:,4)/gtct2_idf2
    g2table16b(:,5) =g2table16b(:,5)/gtct2_idf3

    g2table17a(:,1) =g2table17a(:,1)/gtct_2c
    g2table17a(:,2) =g2table17a(:,2)/gtct2_idf0
    g2table17a(:,3) =g2table17a(:,3)/gtct2_idf1
    g2table17a(:,4) =g2table17a(:,4)/gtct2_idf2
    g2table17a(:,5) =g2table17a(:,5)/gtct2_idf3

    g2table17b(:,1) =g2table17b(:,1)/gtct_2c
    g2table17b(:,2) =g2table17b(:,2)/gtct2_idf0
    g2table17b(:,3) =g2table17b(:,3)/gtct2_idf1
    g2table17b(:,4) =g2table17b(:,4)/gtct2_idf2
    g2table17b(:,5) =g2table17b(:,5)/gtct2_idf3

g2table18a=0.0d0
g2table18b=0.0d0
g2table18c=0.0d0
g2table19=0.0d0 ; g2table19a=0.0d0
g2table20=0.0d0 ; g2table20a=0.0d0
DO m=1,nfam
    IF (nchild(m)==2) THEN
        DO i0=0, 3
        DO i2=0, 11
        DO i4=0, 15
                g2table18a(i4/4,1)=g2table18a(i4/4,1)+gtype2b(m,i4,i2,i0)
                g2table18b(MOD(i4,4),1)=g2table18b(MOD(i4,4),1)+gtype2b(m,i4,i2,i0)
                g2table18c(i4,1)=g2table18c(i4,1)+gtype2b(m,i4,i2,i0)
            IF     (MOD(i2,4)==0) THEN
                g2table18a(i4/4,2)=g2table18a(i4/4,2)+gtype2b(m,i4,i2,i0)
                g2table18b(MOD(i4,4),2)=g2table18b(MOD(i4,4),2)+gtype2b(m,i4,i2,i0)
                g2table18c(i4,2)=g2table18c(i4,2)+gtype2b(m,i4,i2,i0)
            ELSEIF (MOD(i2,4)==1) THEN
                g2table18a(i4/4,3)=g2table18a(i4/4,3)+gtype2b(m,i4,i2,i0)
                g2table18b(MOD(i4,4),3)=g2table18b(MOD(i4,4),3)+gtype2b(m,i4,i2,i0)
                g2table18c(i4,3)=g2table18c(i4,3)+gtype2b(m,i4,i2,i0)
            ELSEIF (MOD(i2,4)==2) THEN
                g2table18a(i4/4,4)=g2table18a(i4/4,4)+gtype2b(m,i4,i2,i0)
                g2table18b(MOD(i4,4),4)=g2table18b(MOD(i4,4),4)+gtype2b(m,i4,i2,i0)
                g2table18c(i4,4)=g2table18c(i4,4)+gtype2b(m,i4,i2,i0)
            ELSEIF (MOD(i2,4)==3) THEN
                g2table18a(i4/4,5)=g2table18a(i4/4,5)+gtype2b(m,i4,i2,i0)
                g2table18b(MOD(i4,4),5)=g2table18b(MOD(i4,4),5)+gtype2b(m,i4,i2,i0)
                g2table18c(i4,5)=g2table18c(i4,5)+gtype2b(m,i4,i2,i0)
            ENDIF
        ENDDO
        ENDDO
        ENDDO
        DO i0=0, 3
        DO i2=0, 11
        DO i5=0, 6
        DO i6=0, 5
                g2table19(i6,1)=g2table19(i6,1)+gtype2c(m,i6,i5,i2,i0)
                g2table20(i5,1)=g2table20(i5,1)+gtype2c(m,i6,i5,i2,i0)
            IF     (MOD(i2,4)==0) THEN
                g2table19(i6,2)=g2table19(i6,2)+gtype2c(m,i6,i5,i2,i0)
                g2table20(i5,2)=g2table20(i5,2)+gtype2c(m,i6,i5,i2,i0)
            ELSEIF (MOD(i2,4)==1) THEN
                g2table19(i6,3)=g2table19(i6,3)+gtype2c(m,i6,i5,i2,i0)
                g2table20(i5,3)=g2table20(i5,3)+gtype2c(m,i6,i5,i2,i0)
            ELSEIF (MOD(i2,4)==2) THEN
                g2table19(i6,4)=g2table19(i6,4)+gtype2c(m,i6,i5,i2,i0)
                g2table20(i5,4)=g2table20(i5,4)+gtype2c(m,i6,i5,i2,i0)
            ELSEIF (MOD(i2,4)==3) THEN
                g2table19(i6,5)=g2table19(i6,5)+gtype2c(m,i6,i5,i2,i0)
                g2table20(i5,5)=g2table20(i5,5)+gtype2c(m,i6,i5,i2,i0)
            ENDIF
        ENDDO
        ENDDO
        ENDDO
        ENDDO
    ENDIF
ENDDO
    g2table18a(:,1) =g2table18a(:,1)/gtct_2c ;  g2table18a(:,2) =g2table18a(:,2)/gtct2_idf0 ;  g2table18a(:,3) =g2table18a(:,3)/gtct2_idf1 ; g2table18a(:,4) =g2table18a(:,4)/gtct2_idf2 ;  g2table18a(:,5) =g2table18a(:,5)/gtct2_idf3
    g2table18b(:,1) =g2table18b(:,1)/gtct_2c ;  g2table18b(:,2) =g2table18b(:,2)/gtct2_idf0 ;  g2table18b(:,3) =g2table18b(:,3)/gtct2_idf1 ; g2table18b(:,4) =g2table18b(:,4)/gtct2_idf2 ;  g2table18b(:,5) =g2table18b(:,5)/gtct2_idf3
    g2table18c(:,1) =g2table18c(:,1)/gtct_2c ;  g2table18c(:,2) =g2table18c(:,2)/gtct2_idf0 ;  g2table18c(:,3) =g2table18c(:,3)/gtct2_idf1 ; g2table18c(:,4) =g2table18c(:,4)/gtct2_idf2 ;  g2table18c(:,5) =g2table18c(:,5)/gtct2_idf3

    g2table19a(:,:) =g2table19(:,:)
    g2table20a(:,:) =g2table20(:,:)

    g2table19(:,:) = g2table19(:,:)/gtct_2c
    g2table20(:,:) = g2table20(:,:)/gtct_2c

    g2table19a(:,1) =g2table19a(:,1)/gtct_2c
    g2table19a(:,2) =g2table19a(:,2)/gtct2_idf0
    g2table19a(:,3) =g2table19a(:,3)/gtct2_idf1
    g2table19a(:,4) =g2table19a(:,4)/gtct2_idf2
    g2table19a(:,5) =g2table19a(:,5)/gtct2_idf3

    g2table20a(:,1) =g2table20a(:,1)/gtct_2c
    g2table20a(:,2) =g2table20a(:,2)/gtct2_idf0
    g2table20a(:,3) =g2table20a(:,3)/gtct2_idf1
    g2table20a(:,4) =g2table20a(:,4)/gtct2_idf2
    g2table20a(:,5) =g2table20a(:,5)/gtct2_idf3


g2table21=0.0d0
g2table22=0.0d0
g2table23=0.0d0
g2table24=0.0d0
g2table25=0.0d0
g2table26=0.0d0

DO m=1,nfam
    IF (nchild(m)==2) THEN
        DO i0=0, 3
        DO i2=0, 11
        DO i5=0, 6
        DO i6=0, 5
            IF     (i6==0) THEN
                g2table21(i5,1)=g2table21(i5,1)+gtype2c(m,i6,i5,i2,i0)
            ELSEIF (i6==1) THEN
                g2table21(i5,2)=g2table21(i5,2)+gtype2c(m,i6,i5,i2,i0)
            ELSEIF (i6==2) THEN
                g2table21(i5,3)=g2table21(i5,3)+gtype2c(m,i6,i5,i2,i0)
            ELSEIF (i6==3) THEN
                g2table21(i5,4)=g2table21(i5,4)+gtype2c(m,i6,i5,i2,i0)
            ELSEIF (i6==4) THEN
                g2table21(i5,5)=g2table21(i5,5)+gtype2c(m,i6,i5,i2,i0)
            ELSEIF (i6==5) THEN
                g2table21(i5,6)=g2table21(i5,6)+gtype2c(m,i6,i5,i2,i0)
            ENDIF
        ENDDO
        ENDDO
        ENDDO
        ENDDO
        DO i0=0, 3
        DO i2=0, 11
        DO i4=0, 15
                g2table23(i0,1)=g2table23(i0,1)+gtype2b(m,i4,i2,i0)
                g2table24(i2/4,1)=g2table24(i2/4,1)+gtype2b(m,i4,i2,i0)
                g2table25(MOD(i2,4),1)=g2table25(MOD(i2,4),1)+gtype2b(m,i4,i2,i0)
                g2table26(MOD(i4,4),1)=g2table26(MOD(i4,4),1)+gtype2b(m,i4,i2,i0)
            IF     (i4/4==0) THEN
                g2table23(i0,2)=g2table23(i0,2)+gtype2b(m,i4,i2,i0)
                g2table24(i2/4,2)=g2table24(i2/4,2)+gtype2b(m,i4,i2,i0)
                g2table25(MOD(i2,4),2)=g2table25(MOD(i2,4),2)+gtype2b(m,i4,i2,i0)
                g2table26(MOD(i4,4),2)=g2table26(MOD(i4,4),2)+gtype2b(m,i4,i2,i0)
            ELSEIF (i4/4==1) THEN
                g2table23(i0,3)=g2table23(i0,3)+gtype2b(m,i4,i2,i0)
                g2table24(i2/4,3)=g2table24(i2/4,3)+gtype2b(m,i4,i2,i0)
                g2table25(MOD(i2,4),3)=g2table25(MOD(i2,4),3)+gtype2b(m,i4,i2,i0)
                g2table26(MOD(i4,4),3)=g2table26(MOD(i4,4),3)+gtype2b(m,i4,i2,i0)
            ELSEIF (i4/4==2) THEN
                g2table23(i0,4)=g2table23(i0,4)+gtype2b(m,i4,i2,i0)
                g2table24(i2/4,4)=g2table24(i2/4,4)+gtype2b(m,i4,i2,i0)
                g2table25(MOD(i2,4),4)=g2table25(MOD(i2,4),4)+gtype2b(m,i4,i2,i0)
                g2table26(MOD(i4,4),4)=g2table26(MOD(i4,4),4)+gtype2b(m,i4,i2,i0)
            ELSEIF (i4/4==3) THEN
                g2table23(i0,5)=g2table23(i0,5)+gtype2b(m,i4,i2,i0)
                g2table24(i2/4,5)=g2table24(i2/4,5)+gtype2b(m,i4,i2,i0)
                g2table25(MOD(i2,4),5)=g2table25(MOD(i2,4),5)+gtype2b(m,i4,i2,i0)
                g2table26(MOD(i4,4),5)=g2table26(MOD(i4,4),5)+gtype2b(m,i4,i2,i0)
            ENDIF
        ENDDO
        ENDDO
        ENDDO
    ENDIF
ENDDO
    g2table21(:,1) =g2table21(:,1)/gtct2_spt1
    g2table21(:,2) =g2table21(:,2)/gtct2_spt2
    g2table21(:,3) =g2table21(:,3)/gtct2_spt3
    g2table21(:,4) =g2table21(:,4)/gtct2_spt4
    g2table21(:,5) =g2table21(:,5)/gtct2_spt5
    g2table21(:,6) =g2table21(:,6)/gtct2_spt6

    g2table23(:,1) =g2table23(:,1)/gtct_2c ;  g2table23(:,2) =g2table23(:,2)/gtct2_fbla0 ;  g2table23(:,3) =g2table23(:,3)/gtct2_fbla1 ;  g2table23(:,4) =g2table23(:,4)/gtct2_fbla2 ;  g2table23(:,5) =g2table23(:,5)/gtct2_fbla3
    g2table24(:,1) =g2table24(:,1)/gtct_2c ;  g2table24(:,2) =g2table24(:,2)/gtct2_fbla0 ;  g2table24(:,3) =g2table24(:,3)/gtct2_fbla1 ;  g2table24(:,4) =g2table24(:,4)/gtct2_fbla2 ;  g2table24(:,5) =g2table24(:,5)/gtct2_fbla3
    g2table25(:,1) =g2table25(:,1)/gtct_2c ;  g2table25(:,2) =g2table25(:,2)/gtct2_fbla0 ;  g2table25(:,3) =g2table25(:,3)/gtct2_fbla1 ;  g2table25(:,4) =g2table25(:,4)/gtct2_fbla2 ;  g2table25(:,5) =g2table25(:,5)/gtct2_fbla3
    g2table26(:,1) =g2table26(:,1)/gtct_2c ;  g2table26(:,2) =g2table26(:,2)/gtct2_fbla0 ;  g2table26(:,3) =g2table26(:,3)/gtct2_fbla1 ;  g2table26(:,4) =g2table26(:,4)/gtct2_fbla2 ;  g2table26(:,5) =g2table26(:,5)/gtct2_fbla3





WRITE(4,fmt=*) 'Table 2-1: Living arrangement patterns of 2 child families'
WRITE(4,fmt=*) '  By SPNE Efficiency type.'
WRITE(4,fmt=*) '   0. Far , Far  (no one lives near)'
WRITE(4,fmt=*) '   1. Far , Near (only the youngest lives near)'
WRITE(4,fmt=*) '   2. Near, Far '
WRITE(4,fmt=*) '   3. Near, Near'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(5A12)') '       Type','       Total','     KH best','        P.O.','        P.D.'
DO k=0,3
    WRITE(4,fmt='(I12,4F12.7)') k,(g2table1(k,m),m=1,4)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-2a: Normal form classification of 2 child families'
WRITE(4,fmt=*) '  By SPNE (observed) living arrangement type.'
WRITE(4,fmt=*) '   0. No normal form equilibrium'
WRITE(4,fmt=*) '   1. Unique equilibrium (Far ,Far )'
WRITE(4,fmt=*) '   2. Unique equilibrium (Far ,Near)'
WRITE(4,fmt=*) '   3. Unique equilibrium (Near,Far )'
WRITE(4,fmt=*) '   4. Unique equilibrium (Near,Near)'
WRITE(4,fmt=*) '   5. Two equilibria (coordination)'
WRITE(4,fmt=*) '   6. Two equilibria (anti-coordination)'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A12)') '       Type','       Total',' (Far ,Far )',' (Far ,Near)',' (Near,Far )',' (Near,Near)'
DO k=0,6
    WRITE(4,fmt='(I12,5F12.7)') k,(g2table2a(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-2b: Normal form classification of 2 child families'
WRITE(4,fmt=*) '  By SPNE (observed) living arrangement type.'
WRITE(4,fmt=*) '   0. No normal form equilibrium'
WRITE(4,fmt=*) '   1. Unique equilibrium (Far ,Far ): First best'
WRITE(4,fmt=*) '   2. Unique equilibrium (Far ,Far ): Not first best, but not Pareto dominated'
WRITE(4,fmt=*) '   3. Unique equilibrium (Far ,Far ): Pareto dominated'
WRITE(4,fmt=*) '   4. Unique equilibrium (Far ,Near): First best'
WRITE(4,fmt=*) '   5. Unique equilibrium (Far ,Near): Not first best, but not Pareto dominated'
WRITE(4,fmt=*) '   6. Unique equilibrium (Far ,Near): Pareto dominated'
WRITE(4,fmt=*) '   7. Unique equilibrium (Near,Far ): First best'
WRITE(4,fmt=*) '   8. Unique equilibrium (Near,Far ): Not first best, but not Pareto dominated'
WRITE(4,fmt=*) '   9. Unique equilibrium (Near,Far ): Pareto dominated'
WRITE(4,fmt=*) '  10. Unique equilibrium (Near,Near): First best'
WRITE(4,fmt=*) '  11. Unique equilibrium (Near,Near): Not first best, but not Pareto dominated'
WRITE(4,fmt=*) '  12. Unique equilibrium (Near,Near): Pareto dominated'
WRITE(4,fmt=*) '  13. Two equilibria (coordination): Either first best'
WRITE(4,fmt=*) '  14. Two equilibria (coordination): Neither first best'
WRITE(4,fmt=*) '  15. Two equilibria (anti-coordination): Either first best'
WRITE(4,fmt=*) '  16. Two equilibria (anti-coordination): Neither first best'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A12)') '       Type','       Total',' (Far ,Far )',' (Far ,Near)',' (Near,Far )',' (Near,Near)'
DO k=0,16
    WRITE(4,fmt='(I12,5F12.7)') k,(g2table2b(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-3a: Normal form classification of 2 child families'
WRITE(4,fmt=*) '  By SPNE (observed) living arrangement type. (Column share)'
WRITE(4,fmt=*) '   0. No normal form equilibrium'
WRITE(4,fmt=*) '   1. Unique equilibrium (Far ,Far )'
WRITE(4,fmt=*) '   2. Unique equilibrium (Far ,Near)'
WRITE(4,fmt=*) '   3. Unique equilibrium (Near,Far )'
WRITE(4,fmt=*) '   4. Unique equilibrium (Near,Near)'
WRITE(4,fmt=*) '   5. Two equilibria (coordination)'
WRITE(4,fmt=*) '   6. Two equilibria (anti-coordination)'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A12)') '       Type','       Total',' (Far ,Far )',' (Far ,Near)',' (Near,Far )',' (Near,Near)'
DO k=0,6
    WRITE(4,fmt='(I12,5F12.7)') k,(g2table3a(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-3b: Normal form classification of 2 child families'
WRITE(4,fmt=*) '  By SPNE (observed) living arrangement type. (Column share)'
WRITE(4,fmt=*) '   0. No normal form equilibrium'
WRITE(4,fmt=*) '   1. Unique equilibrium (Far ,Far ): First best'
WRITE(4,fmt=*) '   2. Unique equilibrium (Far ,Far ): Not first best, but not Pareto dominated'
WRITE(4,fmt=*) '   3. Unique equilibrium (Far ,Far ): Pareto dominated'
WRITE(4,fmt=*) '   4. Unique equilibrium (Far ,Near): First best'
WRITE(4,fmt=*) '   5. Unique equilibrium (Far ,Near): Not first best, but not Pareto dominated'
WRITE(4,fmt=*) '   6. Unique equilibrium (Far ,Near): Pareto dominated'
WRITE(4,fmt=*) '   7. Unique equilibrium (Near,Far ): First best'
WRITE(4,fmt=*) '   8. Unique equilibrium (Near,Far ): Not first best, but not Pareto dominated'
WRITE(4,fmt=*) '   9. Unique equilibrium (Near,Far ): Pareto dominated'
WRITE(4,fmt=*) '  10. Unique equilibrium (Near,Near): First best'
WRITE(4,fmt=*) '  11. Unique equilibrium (Near,Near): Not first best, but not Pareto dominated'
WRITE(4,fmt=*) '  12. Unique equilibrium (Near,Near): Pareto dominated'
WRITE(4,fmt=*) '  13. Two equilibria (coordination): Either first best'
WRITE(4,fmt=*) '  14. Two equilibria (coordination): Neither first best'
WRITE(4,fmt=*) '  15. Two equilibria (anti-coordination): Either first best'
WRITE(4,fmt=*) '  16. Two equilibria (anti-coordination): Neither first best'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A12)') '       Type','       Total',' (Far ,Far )',' (Far ,Near)',' (Near,Far )',' (Near,Near)'
DO k=0,16
    WRITE(4,fmt='(I12,5F12.7)') k,(g2table3b(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '


WRITE(4,fmt=*) 'Table 2-4: Dominant strategy patterns'
WRITE(4,fmt=*) '  By SPNE (observed) living arrangement type. (Column share)'
WRITE(4,fmt=*) '   1. 1st child: Dominant strategy yes    ; 2nd child: yes '
WRITE(4,fmt=*) '   2. 1st child: Dominant strategy yes    ; 2nd child: none'
WRITE(4,fmt=*) '   3. 1st child: Dominant strategy none   ; 2nd child: yes '
WRITE(4,fmt=*) '   4. 1st child: Dominant strategy none   ; 2nd child: none'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A12)') '       Type','       Total',' (Far ,Far )',' (Far ,Near)',' (Near,Far )',' (Near,Near)'
DO k=1,4
    WRITE(4,fmt='(I12,5F12.7)') k,(g2table4(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-4a: Dominant strategy patterns'
WRITE(4,fmt=*) '  By SPNE (observed) living arrangement type. (Column share)'
WRITE(4,fmt=*) '   1. 1st child: Dominant strategy (Near) ; 2nd child: (Near)'
WRITE(4,fmt=*) '   2. 1st child: Dominant strategy (Near) ; 2nd child: (Far )'
WRITE(4,fmt=*) '   3. 1st child: Dominant strategy (Near) ; 2nd child: none'
WRITE(4,fmt=*) '   4. 1st child: Dominant strategy (Far ) ; 2nd child: (Near)'
WRITE(4,fmt=*) '   5. 1st child: Dominant strategy (Far ) ; 2nd child: (Far )'
WRITE(4,fmt=*) '   6. 1st child: Dominant strategy (Far ) ; 2nd child: none'
WRITE(4,fmt=*) '   7. 1st child: Dominant strategy none   ; 2nd child: (Near)'
WRITE(4,fmt=*) '   8. 1st child: Dominant strategy none   ; 2nd child: (Far )'
WRITE(4,fmt=*) '   9. 1st child: Dominant strategy none   ; 2nd child: none'
WRITE(4,fmt=*) '  NB: 2nd player always plays a dominant strategy if any.'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A12)') '       Type','       Total',' (Far ,Far )',' (Far ,Near)',' (Near,Far )',' (Near,Near)'
DO k=1,9
    WRITE(4,fmt='(I12,5F12.7)') k,(g2table4a(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-4b: Dominant strategy patterns'
WRITE(4,fmt=*) '  By SPNE (observed) living arrangement type. (Column share)'
WRITE(4,fmt=*) '   1. 1st child: Dominant strategy (Near)==SPNE ; 2nd child: (Near)==SPNE'
WRITE(4,fmt=*) '   2. 1st child: Dominant strategy (Near)==SPNE ; 2nd child: (Far )==SPNE'
WRITE(4,fmt=*) '   3. 1st child: Dominant strategy (Near)==SPNE ; 2nd child: none'
WRITE(4,fmt=*) '   4. 1st child: Dominant strategy (Near)/=SPNE ; 2nd child: (Near)==SPNE'
WRITE(4,fmt=*) '   5. 1st child: Dominant strategy (Near)/=SPNE ; 2nd child: (Far )==SPNE'
WRITE(4,fmt=*) '   6. 1st child: Dominant strategy (Near)/=SPNE ; 2nd child: none'
WRITE(4,fmt=*) '   7. 1st child: Dominant strategy (Far )==SPNE ; 2nd child: (Near)==SPNE'
WRITE(4,fmt=*) '   8. 1st child: Dominant strategy (Far )==SPNE ; 2nd child: (Far )==SPNE'
WRITE(4,fmt=*) '   9. 1st child: Dominant strategy (Far )==SPNE ; 2nd child: none'
WRITE(4,fmt=*) '  10. 1st child: Dominant strategy (Far )/=SPNE ; 2nd child: (Near)==SPNE'
WRITE(4,fmt=*) '  11. 1st child: Dominant strategy (Far )/=SPNE ; 2nd child: (Far )==SPNE'
WRITE(4,fmt=*) '  12. 1st child: Dominant strategy (Far )/=SPNE ; 2nd child: none'
WRITE(4,fmt=*) '  13. 1st child: Dominant strategy none         ; 2nd child: (Near)==SPNE'
WRITE(4,fmt=*) '  14. 1st child: Dominant strategy none         ; 2nd child: (Far )==SPNE'
WRITE(4,fmt=*) '  15. 1st child: Dominant strategy none         ; 2nd child: none'
WRITE(4,fmt=*) '  NB: 2nd player always plays a dominant strategy if any.'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A12)') '       Type','       Total',' (Far ,Far )',' (Far ,Near)',' (Near,Far )',' (Near,Near)'
DO k=1,15
    WRITE(4,fmt='(I12,5F12.7)') k,(g2table4b(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-5: Increasing differences'
WRITE(4,fmt=*) '  By SPNE (observed) living arrangement type. (Column share)'
WRITE(4,fmt=*) '   0. 1st child: Decreasing difference ; 2nd child: Decreasing difference'
WRITE(4,fmt=*) '   1. 1st child: Decreasing difference ; 2nd child: Increasing difference'
WRITE(4,fmt=*) '   2. 1st child: Increasing difference ; 2nd child: Decreasing difference'
WRITE(4,fmt=*) '   3. 1st child: Increasing difference ; 2nd child: Increasing difference'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A12)') '       Type','       Total',' (Far ,Far )',' (Far ,Near)',' (Near,Far )',' (Near,Near)'
DO k=0,3
    WRITE(4,fmt='(I12,5F12.7)') k,(g2table5(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-6a: First best arrangements'
WRITE(4,fmt=*) '  By SPNE (observed) living arrangement type.'
WRITE(4,fmt=*) '   0. FB arrangement: (Far ,Far )'
WRITE(4,fmt=*) '   1. FB arrangement: (Far ,Near)'
WRITE(4,fmt=*) '   2. FB arrangement: (Near,Far )'
WRITE(4,fmt=*) '   3. FB arrangement: (Near,Near)'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A12)') '       Type','       Total',' (Far ,Far )',' (Far ,Near)',' (Near,Far )',' (Near,Near)'
DO k=0,3
    WRITE(4,fmt='(I12,5F12.7)') k,(g2table6a(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-6b: Reverse order SPNE arrangements'
WRITE(4,fmt=*) '  By SPNE (observed) living arrangement type.'
WRITE(4,fmt=*) '   0. Reverse SPNE: (Far ,Far )'
WRITE(4,fmt=*) '   1. Reverse SPNE: (Far ,Near)'
WRITE(4,fmt=*) '   2. Reverse SPNE: (Near,Far )'
WRITE(4,fmt=*) '   3. Reverse SPNE: (Near,Near)'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A12)') '       Type','       Total',' (Far ,Far )',' (Far ,Near)',' (Near,Far )',' (Near,Near)'
DO k=0,3
    WRITE(4,fmt='(I12,5F12.7)') k,(g2table6b(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-6c: First best arrangements and reverse order SPNE arrangements'
WRITE(4,fmt=*) '  By SPNE (observed) living arrangement type.'
WRITE(4,fmt=*) '   0. FB arrangement: (Far ,Far ) ; Reverse SPNE: (Far ,Far )'
WRITE(4,fmt=*) '   1. FB arrangement: (Far ,Far ) ; Reverse SPNE: (Far ,Near)'
WRITE(4,fmt=*) '   2. FB arrangement: (Far ,Far ) ; Reverse SPNE: (Near,Far )'
WRITE(4,fmt=*) '   3. FB arrangement: (Far ,Far ) ; Reverse SPNE: (Near,Near)'
WRITE(4,fmt=*) '   4. FB arrangement: (Far ,Near) ; Reverse SPNE: (Far ,Far )'
WRITE(4,fmt=*) '   5. FB arrangement: (Far ,Near) ; Reverse SPNE: (Far ,Near)'
WRITE(4,fmt=*) '   6. FB arrangement: (Far ,Near) ; Reverse SPNE: (Near,Far )'
WRITE(4,fmt=*) '   7. FB arrangement: (Far ,Near) ; Reverse SPNE: (Near,Near)'
WRITE(4,fmt=*) '   8. FB arrangement: (Near,Far ) ; Reverse SPNE: (Far ,Far )'
WRITE(4,fmt=*) '   9. FB arrangement: (Near,Far ) ; Reverse SPNE: (Far ,Near)'
WRITE(4,fmt=*) '  10. FB arrangement: (Near,Far ) ; Reverse SPNE: (Near,Far )'
WRITE(4,fmt=*) '  11. FB arrangement: (Near,Far ) ; Reverse SPNE: (Near,Near)'
WRITE(4,fmt=*) '  12. FB arrangement: (Near,Near) ; Reverse SPNE: (Far ,Far )'
WRITE(4,fmt=*) '  13. FB arrangement: (Near,Near) ; Reverse SPNE: (Far ,Near)'
WRITE(4,fmt=*) '  14. FB arrangement: (Near,Near) ; Reverse SPNE: (Near,Far )'
WRITE(4,fmt=*) '  15. FB arrangement: (Near,Near) ; Reverse SPNE: (Near,Near)'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A12)') '       Type','       Total',' (Far ,Far )',' (Far ,Near)',' (Near,Far )',' (Near,Near)'
DO k=0,15
    WRITE(4,fmt='(I12,5F12.7)') k,(g2table6c(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-6a2: First best arrangements'
WRITE(4,fmt=*) '  By SPNE (observed) living arrangement type. (Column share)'
WRITE(4,fmt=*) '   0. FB arrangement: (Far ,Far )'
WRITE(4,fmt=*) '   1. FB arrangement: (Far ,Near)'
WRITE(4,fmt=*) '   2. FB arrangement: (Near,Far )'
WRITE(4,fmt=*) '   3. FB arrangement: (Near,Near)'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A12)') '       Type','       Total',' (Far ,Far )',' (Far ,Near)',' (Near,Far )',' (Near,Near)'
DO k=0,3
    WRITE(4,fmt='(I12,5F12.7)') k,(g2table6a2(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-6b2: Reverse order SPNE arrangements'
WRITE(4,fmt=*) '  By SPNE (observed) living arrangement type. (Column share)'
WRITE(4,fmt=*) '   0. Reverse SPNE: (Far ,Far )'
WRITE(4,fmt=*) '   1. Reverse SPNE: (Far ,Near)'
WRITE(4,fmt=*) '   2. Reverse SPNE: (Near,Far )'
WRITE(4,fmt=*) '   3. Reverse SPNE: (Near,Near)'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A12)') '       Type','       Total',' (Far ,Far )',' (Far ,Near)',' (Near,Far )',' (Near,Near)'
DO k=0,3
    WRITE(4,fmt='(I12,5F12.7)') k,(g2table6b2(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-6c2: First best arrangements and reverse order SPNE arrangements'
WRITE(4,fmt=*) '  By SPNE (observed) living arrangement type. (Column share)'
WRITE(4,fmt=*) '   0. FB arrangement: (Far ,Far ) ; Reverse SPNE: (Far ,Far )'
WRITE(4,fmt=*) '   1. FB arrangement: (Far ,Far ) ; Reverse SPNE: (Far ,Near)'
WRITE(4,fmt=*) '   2. FB arrangement: (Far ,Far ) ; Reverse SPNE: (Near,Far )'
WRITE(4,fmt=*) '   3. FB arrangement: (Far ,Far ) ; Reverse SPNE: (Near,Near)'
WRITE(4,fmt=*) '   4. FB arrangement: (Far ,Near) ; Reverse SPNE: (Far ,Far )'
WRITE(4,fmt=*) '   5. FB arrangement: (Far ,Near) ; Reverse SPNE: (Far ,Near)'
WRITE(4,fmt=*) '   6. FB arrangement: (Far ,Near) ; Reverse SPNE: (Near,Far )'
WRITE(4,fmt=*) '   7. FB arrangement: (Far ,Near) ; Reverse SPNE: (Near,Near)'
WRITE(4,fmt=*) '   8. FB arrangement: (Near,Far ) ; Reverse SPNE: (Far ,Far )'
WRITE(4,fmt=*) '   9. FB arrangement: (Near,Far ) ; Reverse SPNE: (Far ,Near)'
WRITE(4,fmt=*) '  10. FB arrangement: (Near,Far ) ; Reverse SPNE: (Near,Far )'
WRITE(4,fmt=*) '  11. FB arrangement: (Near,Far ) ; Reverse SPNE: (Near,Near)'
WRITE(4,fmt=*) '  12. FB arrangement: (Near,Near) ; Reverse SPNE: (Far ,Far )'
WRITE(4,fmt=*) '  13. FB arrangement: (Near,Near) ; Reverse SPNE: (Far ,Near)'
WRITE(4,fmt=*) '  14. FB arrangement: (Near,Near) ; Reverse SPNE: (Near,Far )'
WRITE(4,fmt=*) '  15. FB arrangement: (Near,Near) ; Reverse SPNE: (Near,Near)'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A12)') '       Type','       Total',' (Far ,Far )',' (Far ,Near)',' (Near,Far )',' (Near,Near)'
DO k=0,15
    WRITE(4,fmt='(I12,5F12.7)') k,(g2table6c2(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-7: Utility changes in SPNE compared with normal form N.E.'
WRITE(4,fmt=*) '  By SPNE (observed) living arrangement type. (Column share)'
WRITE(4,fmt=*) '   0. No change (SPNE==Normal form unique N.E.)'
WRITE(4,fmt=*) '   1. 2nd child: (-) ; Total: (-)'
WRITE(4,fmt=*) '   2. 2nd child: (-) ; Total: (+)'
WRITE(4,fmt=*) '   3. 2nd child: (+) ; Total: (+)'
WRITE(4,fmt=*) '   4. 2 equilibria'
WRITE(4,fmt=*) '   5. No equilibrium'
WRITE(4,fmt=*) '  NB: When there is a normal form N.E., the 1st mover can always'
WRITE(4,fmt=*) '      achieve that outcome. --> 1st child never worses off.'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A12)') '       Type','       Total',' (Far ,Far )',' (Far ,Near)',' (Near,Far )',' (Near,Near)'
DO k=0,5
    WRITE(4,fmt='(I12,5F12.7)') k,(g2table7(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-8: Utility changes in reverse SPNE (compared with SPNE)'
WRITE(4,fmt=*) '  By SPNE (observed) living arrangement type. (Column share)'
WRITE(4,fmt=*) '   0. No change (SPNE==ReverseSPNE)'
WRITE(4,fmt=*) '   1. Total: (-) ; 1st child: (-) ; 2nd child: (-)'
WRITE(4,fmt=*) '   2. Total: (-) ; 1st child: (-) ; 2nd child: (+)'
WRITE(4,fmt=*) '   3. Total: (-) ; 1st child: (+) ; 2nd child: (-)'
WRITE(4,fmt=*) '   4. Total: (+) ; 1st child: (-) ; 2nd child: (+)'
WRITE(4,fmt=*) '   5. Total: (+) ; 1st child: (+) ; 2nd child: (-)'
WRITE(4,fmt=*) '   6. Total: (+) ; 1st child: (+) ; 2nd child: (+)'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A12)') '       Type','       Total',' (Far ,Far )',' (Far ,Near)',' (Near,Far )',' (Near,Near)'
DO k=0,6
    WRITE(4,fmt='(I12,5F12.7)') k,(g2table8(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '



WRITE(4,fmt=*) 'Table 2-9a: Normal form classification of 2 child families'
WRITE(4,fmt=*) '  By SPNE Efficiency type.'
WRITE(4,fmt=*) '   0. No normal form equilibrium'
WRITE(4,fmt=*) '   1. Unique equilibrium (Far ,Far )'
WRITE(4,fmt=*) '   2. Unique equilibrium (Far ,Near)'
WRITE(4,fmt=*) '   3. Unique equilibrium (Near,Far )'
WRITE(4,fmt=*) '   4. Unique equilibrium (Near,Near)'
WRITE(4,fmt=*) '   5. Two equilibria (coordination)'
WRITE(4,fmt=*) '   6. Two equilibria (anti-coordination)'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(5A12)') '       Type','       Total','     KH best','        P.O.','        P.D.'
DO k=0,6
    WRITE(4,fmt='(I12,4F12.7)') k,(g2table9a(k,m),m=1,4)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-9b: Normal form classification of 2 child families'
WRITE(4,fmt=*) '  By SPNE Efficiency type.'
WRITE(4,fmt=*) '   0. No normal form equilibrium'
WRITE(4,fmt=*) '   1. Unique equilibrium (Far ,Far ): First best'
WRITE(4,fmt=*) '   2. Unique equilibrium (Far ,Far ): Not first best, but not Pareto dominated'
WRITE(4,fmt=*) '   3. Unique equilibrium (Far ,Far ): Pareto dominated'
WRITE(4,fmt=*) '   4. Unique equilibrium (Far ,Near): First best'
WRITE(4,fmt=*) '   5. Unique equilibrium (Far ,Near): Not first best, but not Pareto dominated'
WRITE(4,fmt=*) '   6. Unique equilibrium (Far ,Near): Pareto dominated'
WRITE(4,fmt=*) '   7. Unique equilibrium (Near,Far ): First best'
WRITE(4,fmt=*) '   8. Unique equilibrium (Near,Far ): Not first best, but not Pareto dominated'
WRITE(4,fmt=*) '   9. Unique equilibrium (Near,Far ): Pareto dominated'
WRITE(4,fmt=*) '  10. Unique equilibrium (Near,Near): First best'
WRITE(4,fmt=*) '  11. Unique equilibrium (Near,Near): Not first best, but not Pareto dominated'
WRITE(4,fmt=*) '  12. Unique equilibrium (Near,Near): Pareto dominated'
WRITE(4,fmt=*) '  13. Two equilibria (coordination): Either first best'
WRITE(4,fmt=*) '  14. Two equilibria (coordination): Neither first best'
WRITE(4,fmt=*) '  15. Two equilibria (anti-coordination): Either first best'
WRITE(4,fmt=*) '  16. Two equilibria (anti-coordination): Neither first best'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(5A12)') '       Type','       Total','     KH best','        P.O.','        P.D.'
DO k=0,16
    WRITE(4,fmt='(I12,4F12.7)') k,(g2table9b(k,m),m=1,4)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-10a: Normal form classification of 2 child families'
WRITE(4,fmt=*) '  By SPNE Efficiency type. (Column share)'
WRITE(4,fmt=*) '   0. No normal form equilibrium'
WRITE(4,fmt=*) '   1. Unique equilibrium (Far ,Far )'
WRITE(4,fmt=*) '   2. Unique equilibrium (Far ,Near)'
WRITE(4,fmt=*) '   3. Unique equilibrium (Near,Far )'
WRITE(4,fmt=*) '   4. Unique equilibrium (Near,Near)'
WRITE(4,fmt=*) '   5. Two equilibria (coordination)'
WRITE(4,fmt=*) '   6. Two equilibria (anti-coordination)'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(5A12)') '       Type','       Total','     KH best','        P.O.','        P.D.'
DO k=0,6
    WRITE(4,fmt='(I12,4F12.7)') k,(g2table10a(k,m),m=1,4)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-10b: Normal form classification of 2 child families'
WRITE(4,fmt=*) '  By SPNE Efficiency type. (Column share)'
WRITE(4,fmt=*) '   0. No normal form equilibrium'
WRITE(4,fmt=*) '   1. Unique equilibrium (Far ,Far ): First best'
WRITE(4,fmt=*) '   2. Unique equilibrium (Far ,Far ): Not first best, but not Pareto dominated'
WRITE(4,fmt=*) '   3. Unique equilibrium (Far ,Far ): Pareto dominated'
WRITE(4,fmt=*) '   4. Unique equilibrium (Far ,Near): First best'
WRITE(4,fmt=*) '   5. Unique equilibrium (Far ,Near): Not first best, but not Pareto dominated'
WRITE(4,fmt=*) '   6. Unique equilibrium (Far ,Near): Pareto dominated'
WRITE(4,fmt=*) '   7. Unique equilibrium (Near,Far ): First best'
WRITE(4,fmt=*) '   8. Unique equilibrium (Near,Far ): Not first best, but not Pareto dominated'
WRITE(4,fmt=*) '   9. Unique equilibrium (Near,Far ): Pareto dominated'
WRITE(4,fmt=*) '  10. Unique equilibrium (Near,Near): First best'
WRITE(4,fmt=*) '  11. Unique equilibrium (Near,Near): Not first best, but not Pareto dominated'
WRITE(4,fmt=*) '  12. Unique equilibrium (Near,Near): Pareto dominated'
WRITE(4,fmt=*) '  13. Two equilibria (coordination): Either first best'
WRITE(4,fmt=*) '  14. Two equilibria (coordination): Neither first best'
WRITE(4,fmt=*) '  15. Two equilibria (anti-coordination): Either first best'
WRITE(4,fmt=*) '  16. Two equilibria (anti-coordination): Neither first best'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(5A12)') '       Type','       Total','     KH best','        P.O.','        P.D.'
DO k=0,16
    WRITE(4,fmt='(I12,4F12.7)') k,(g2table10b(k,m),m=1,4)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-11a: Dominant strategy patterns'
WRITE(4,fmt=*) '  By SPNE Efficiency type. (Column share)'
WRITE(4,fmt=*) '   1. 1st child: Dominant strategy (Near) ; 2nd child: (Near)'
WRITE(4,fmt=*) '   2. 1st child: Dominant strategy (Near) ; 2nd child: (Far )'
WRITE(4,fmt=*) '   3. 1st child: Dominant strategy (Near) ; 2nd child: none'
WRITE(4,fmt=*) '   4. 1st child: Dominant strategy (Far ) ; 2nd child: (Near)'
WRITE(4,fmt=*) '   5. 1st child: Dominant strategy (Far ) ; 2nd child: (Far )'
WRITE(4,fmt=*) '   6. 1st child: Dominant strategy (Far ) ; 2nd child: none'
WRITE(4,fmt=*) '   7. 1st child: Dominant strategy none   ; 2nd child: (Near)'
WRITE(4,fmt=*) '   8. 1st child: Dominant strategy none   ; 2nd child: (Far )'
WRITE(4,fmt=*) '   9. 1st child: Dominant strategy none   ; 2nd child: none'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(5A12)') '       Type','       Total','     KH best','        P.O.','        P.D.'
DO k=1,9
    WRITE(4,fmt='(I12,4F12.7)') k,(g2table11a(k,m),m=1,4)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-11b: Dominant strategy patterns'
WRITE(4,fmt=*) '  By SPNE Efficiency type. (Column share)'
WRITE(4,fmt=*) '   1. 1st child: Dominant strategy (Near)==SPNE ; 2nd child: (Near)==SPNE'
WRITE(4,fmt=*) '   2. 1st child: Dominant strategy (Near)==SPNE ; 2nd child: (Far )==SPNE'
WRITE(4,fmt=*) '   3. 1st child: Dominant strategy (Near)==SPNE ; 2nd child: none'
WRITE(4,fmt=*) '   4. 1st child: Dominant strategy (Near)/=SPNE ; 2nd child: (Near)==SPNE'
WRITE(4,fmt=*) '   5. 1st child: Dominant strategy (Near)/=SPNE ; 2nd child: (Far )==SPNE'
WRITE(4,fmt=*) '   6. 1st child: Dominant strategy (Near)/=SPNE ; 2nd child: none'
WRITE(4,fmt=*) '   7. 1st child: Dominant strategy (Far )==SPNE ; 2nd child: (Near)==SPNE'
WRITE(4,fmt=*) '   8. 1st child: Dominant strategy (Far )==SPNE ; 2nd child: (Far )==SPNE'
WRITE(4,fmt=*) '   9. 1st child: Dominant strategy (Far )==SPNE ; 2nd child: none'
WRITE(4,fmt=*) '  10. 1st child: Dominant strategy (Far )/=SPNE ; 2nd child: (Near)==SPNE'
WRITE(4,fmt=*) '  11. 1st child: Dominant strategy (Far )/=SPNE ; 2nd child: (Far )==SPNE'
WRITE(4,fmt=*) '  12. 1st child: Dominant strategy (Far )/=SPNE ; 2nd child: none'
WRITE(4,fmt=*) '  13. 1st child: Dominant strategy none         ; 2nd child: (Near)==SPNE'
WRITE(4,fmt=*) '  14. 1st child: Dominant strategy none         ; 2nd child: (Far )==SPNE'
WRITE(4,fmt=*) '  15. 1st child: Dominant strategy none         ; 2nd child: none'
WRITE(4,fmt=*) '  NB: 2nd player always plays a dominant strategy if any.'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(5A12)') '       Type','       Total','     KH best','        P.O.','        P.D.'
DO k=1,15
    WRITE(4,fmt='(I12,4F12.7)') k,(g2table11b(k,m),m=1,4)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-12 Increasing differences'
WRITE(4,fmt=*) '  By SPNE Efficiency type. (Column share)'
WRITE(4,fmt=*) '   0. 1st child: Decreasing difference ; 2nd child: Decreasing difference'
WRITE(4,fmt=*) '   1. 1st child: Decreasing difference ; 2nd child: Increasing difference'
WRITE(4,fmt=*) '   2. 1st child: Increasing difference ; 2nd child: Decreasing difference'
WRITE(4,fmt=*) '   3. 1st child: Increasing difference ; 2nd child: Increasing difference'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(5A12)') '       Type','       Total','     KH best','        P.O.','        P.D.'
DO k=0,3
    WRITE(4,fmt='(I12,4F12.7)') k,(g2table12(k,m),m=1,4)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-13a: First best arrangements'
WRITE(4,fmt=*) '  By SPNE Efficiency type. (Column share)'
WRITE(4,fmt=*) '   0. FB arrangement: (Far ,Far )'
WRITE(4,fmt=*) '   1. FB arrangement: (Far ,Near)'
WRITE(4,fmt=*) '   2. FB arrangement: (Near,Far )'
WRITE(4,fmt=*) '   3. FB arrangement: (Near,Near)'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(5A12)') '       Type','       Total','     KH best','        P.O.','        P.D.'
DO k=0,3
    WRITE(4,fmt='(I12,4F12.7)') k,(g2table13a(k,m),m=1,4)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-13b: Reverse order SPNE arrangements'
WRITE(4,fmt=*) '  By SPNE Efficiency type. (Column share)'
WRITE(4,fmt=*) '   0. Reverse SPNE: (Far ,Far )'
WRITE(4,fmt=*) '   1. Reverse SPNE: (Far ,Near)'
WRITE(4,fmt=*) '   2. Reverse SPNE: (Near,Far )'
WRITE(4,fmt=*) '   3. Reverse SPNE: (Near,Near)'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(5A12)') '       Type','       Total','     KH best','        P.O.','        P.D.'
DO k=0,3
    WRITE(4,fmt='(I12,4F12.7)') k,(g2table13b(k,m),m=1,4)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-13c: First best arrangements and reverse order SPNE arrangements'
WRITE(4,fmt=*) '  By SPNE Efficiency type. (Column share)'
WRITE(4,fmt=*) '   0. FB arrangement: (Far ,Far ) ; Reverse SPNE: (Far ,Far )'
WRITE(4,fmt=*) '   1. FB arrangement: (Far ,Far ) ; Reverse SPNE: (Far ,Near)'
WRITE(4,fmt=*) '   2. FB arrangement: (Far ,Far ) ; Reverse SPNE: (Near,Far )'
WRITE(4,fmt=*) '   3. FB arrangement: (Far ,Far ) ; Reverse SPNE: (Near,Near)'
WRITE(4,fmt=*) '   4. FB arrangement: (Far ,Near) ; Reverse SPNE: (Far ,Far )'
WRITE(4,fmt=*) '   5. FB arrangement: (Far ,Near) ; Reverse SPNE: (Far ,Near)'
WRITE(4,fmt=*) '   6. FB arrangement: (Far ,Near) ; Reverse SPNE: (Near,Far )'
WRITE(4,fmt=*) '   7. FB arrangement: (Far ,Near) ; Reverse SPNE: (Near,Near)'
WRITE(4,fmt=*) '   8. FB arrangement: (Near,Far ) ; Reverse SPNE: (Far ,Far )'
WRITE(4,fmt=*) '   9. FB arrangement: (Near,Far ) ; Reverse SPNE: (Far ,Near)'
WRITE(4,fmt=*) '  10. FB arrangement: (Near,Far ) ; Reverse SPNE: (Near,Far )'
WRITE(4,fmt=*) '  11. FB arrangement: (Near,Far ) ; Reverse SPNE: (Near,Near)'
WRITE(4,fmt=*) '  12. FB arrangement: (Near,Near) ; Reverse SPNE: (Far ,Far )'
WRITE(4,fmt=*) '  13. FB arrangement: (Near,Near) ; Reverse SPNE: (Far ,Near)'
WRITE(4,fmt=*) '  14. FB arrangement: (Near,Near) ; Reverse SPNE: (Near,Far )'
WRITE(4,fmt=*) '  15. FB arrangement: (Near,Near) ; Reverse SPNE: (Near,Near)'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(5A12)') '       Type','       Total','     KH best','        P.O.','        P.D.'
DO k=0,15
    WRITE(4,fmt='(I12,4F12.7)') k,(g2table13c(k,m),m=1,4)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-14: Utility changes in SPNE compared with normal form N.E.'
WRITE(4,fmt=*) '  By SPNE Efficiency type. (Column share)'
WRITE(4,fmt=*) '   0. No change (SPNE==Normal form unique N.E.)'
WRITE(4,fmt=*) '   1. 2nd child: (-) ; Total: (-)'
WRITE(4,fmt=*) '   2. 2nd child: (-) ; Total: (+)'
WRITE(4,fmt=*) '   3. 2nd child: (+) ; Total: (+)'
WRITE(4,fmt=*) '   4. 2 equilibria'
WRITE(4,fmt=*) '   5. No equilibrium'
WRITE(4,fmt=*) '  NB: When there is a normal form N.E., the 1st mover can always'
WRITE(4,fmt=*) '      achieve that outcome. --> 1st child never worses off.'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(5A12)') '       Type','       Total','     KH best','        P.O.','        P.D.'
DO k=0,5
    WRITE(4,fmt='(I12,4F12.7)') k,(g2table14(k,m),m=1,4)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-15: Utility changes in reverse SPNE (compared with SPNE)'
WRITE(4,fmt=*) '  By SPNE Efficiency type. (Column share)'
WRITE(4,fmt=*) '   0. No change (SPNE==ReverseSPNE)'
WRITE(4,fmt=*) '   1. Total: (-) ; 1st child: (-) ; 2nd child: (-)'
WRITE(4,fmt=*) '   2. Total: (-) ; 1st child: (-) ; 2nd child: (+)'
WRITE(4,fmt=*) '   3. Total: (-) ; 1st child: (+) ; 2nd child: (-)'
WRITE(4,fmt=*) '   4. Total: (+) ; 1st child: (-) ; 2nd child: (+)'
WRITE(4,fmt=*) '   5. Total: (+) ; 1st child: (+) ; 2nd child: (-)'
WRITE(4,fmt=*) '   6. Total: (+) ; 1st child: (+) ; 2nd child: (+)'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(5A12)') '       Type','       Total','     KH best','        P.O.','        P.D.'
DO k=0,6
    WRITE(4,fmt='(I12,4F12.7)') k,(g2table15(k,m),m=1,4)
ENDDO
WRITE(4,fmt=*) ' '


WRITE(4,fmt=*) 'Table 2-16a: Normal form classification of 2 child families'
WRITE(4,fmt=*) '  By Decreasing difference status (Column share)'
WRITE(4,fmt=*) '   0. No normal form equilibrium'
WRITE(4,fmt=*) '   1. Unique equilibrium (Far ,Far )'
WRITE(4,fmt=*) '   2. Unique equilibrium (Far ,Near)'
WRITE(4,fmt=*) '   3. Unique equilibrium (Near,Far )'
WRITE(4,fmt=*) '   4. Unique equilibrium (Near,Near)'
WRITE(4,fmt=*) '   5. Two equilibria (coordination)'
WRITE(4,fmt=*) '   6. Two equilibria (anti-coordination)'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A12)') '       Type','       Total',' (DDif,DDif)',' (DDif,IDif)',' (IDif,DDif)',' (IDif,IDif)'
DO k=0,6
    WRITE(4,fmt='(I12,5F12.7)') k,(g2table16a(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-16b: Normal form classification of 2 child families'
WRITE(4,fmt=*) '  By Decreasing difference status (Column share)'
WRITE(4,fmt=*) '   0. No normal form equilibrium'
WRITE(4,fmt=*) '   1. Unique equilibrium (Far ,Far ): First best'
WRITE(4,fmt=*) '   2. Unique equilibrium (Far ,Far ): Not first best, but not Pareto dominated'
WRITE(4,fmt=*) '   3. Unique equilibrium (Far ,Far ): Pareto dominated'
WRITE(4,fmt=*) '   4. Unique equilibrium (Far ,Near): First best'
WRITE(4,fmt=*) '   5. Unique equilibrium (Far ,Near): Not first best, but not Pareto dominated'
WRITE(4,fmt=*) '   6. Unique equilibrium (Far ,Near): Pareto dominated'
WRITE(4,fmt=*) '   7. Unique equilibrium (Near,Far ): First best'
WRITE(4,fmt=*) '   8. Unique equilibrium (Near,Far ): Not first best, but not Pareto dominated'
WRITE(4,fmt=*) '   9. Unique equilibrium (Near,Far ): Pareto dominated'
WRITE(4,fmt=*) '  10. Unique equilibrium (Near,Near): First best'
WRITE(4,fmt=*) '  11. Unique equilibrium (Near,Near): Not first best, but not Pareto dominated'
WRITE(4,fmt=*) '  12. Unique equilibrium (Near,Near): Pareto dominated'
WRITE(4,fmt=*) '  13. Two equilibria (coordination): Either first best'
WRITE(4,fmt=*) '  14. Two equilibria (coordination): Neither first best'
WRITE(4,fmt=*) '  15. Two equilibria (anti-coordination): Either first best'
WRITE(4,fmt=*) '  16. Two equilibria (anti-coordination): Neither first best'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A12)') '       Type','       Total',' (DDif,DDif)',' (DDif,IDif)',' (IDif,DDif)',' (IDif,IDif)'
DO k=0,16
    WRITE(4,fmt='(I12,5F12.7)') k,(g2table16b(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-17a: Dominant strategy patterns'
WRITE(4,fmt=*) '  By Decreasing difference status (Column share)'
WRITE(4,fmt=*) '   1. 1st child: Dominant strategy (Near) ; 2nd child: (Near)'
WRITE(4,fmt=*) '   2. 1st child: Dominant strategy (Near) ; 2nd child: (Far )'
WRITE(4,fmt=*) '   3. 1st child: Dominant strategy (Near) ; 2nd child: none'
WRITE(4,fmt=*) '   4. 1st child: Dominant strategy (Far ) ; 2nd child: (Near)'
WRITE(4,fmt=*) '   5. 1st child: Dominant strategy (Far ) ; 2nd child: (Far )'
WRITE(4,fmt=*) '   6. 1st child: Dominant strategy (Far ) ; 2nd child: none'
WRITE(4,fmt=*) '   7. 1st child: Dominant strategy none   ; 2nd child: (Near)'
WRITE(4,fmt=*) '   8. 1st child: Dominant strategy none   ; 2nd child: (Far )'
WRITE(4,fmt=*) '   9. 1st child: Dominant strategy none   ; 2nd child: none'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A12)') '       Type','       Total',' (DDif,DDif)',' (DDif,IDif)',' (IDif,DDif)',' (IDif,IDif)'
DO k=1,9
    WRITE(4,fmt='(I12,5F12.7)') k,(g2table17a(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-17b: Dominant strategy patterns'
WRITE(4,fmt=*) '  By Decreasing difference status (Column share)'
WRITE(4,fmt=*) '   1. 1st child: Dominant strategy (Near)==SPNE ; 2nd child: (Near)==SPNE'
WRITE(4,fmt=*) '   2. 1st child: Dominant strategy (Near)==SPNE ; 2nd child: (Far )==SPNE'
WRITE(4,fmt=*) '   3. 1st child: Dominant strategy (Near)==SPNE ; 2nd child: none'
WRITE(4,fmt=*) '   4. 1st child: Dominant strategy (Near)/=SPNE ; 2nd child: (Near)==SPNE'
WRITE(4,fmt=*) '   5. 1st child: Dominant strategy (Near)/=SPNE ; 2nd child: (Far )==SPNE'
WRITE(4,fmt=*) '   6. 1st child: Dominant strategy (Near)/=SPNE ; 2nd child: none'
WRITE(4,fmt=*) '   7. 1st child: Dominant strategy (Far )==SPNE ; 2nd child: (Near)==SPNE'
WRITE(4,fmt=*) '   8. 1st child: Dominant strategy (Far )==SPNE ; 2nd child: (Far )==SPNE'
WRITE(4,fmt=*) '   9. 1st child: Dominant strategy (Far )==SPNE ; 2nd child: none'
WRITE(4,fmt=*) '  10. 1st child: Dominant strategy (Far )/=SPNE ; 2nd child: (Near)==SPNE'
WRITE(4,fmt=*) '  11. 1st child: Dominant strategy (Far )/=SPNE ; 2nd child: (Far )==SPNE'
WRITE(4,fmt=*) '  12. 1st child: Dominant strategy (Far )/=SPNE ; 2nd child: none'
WRITE(4,fmt=*) '  13. 1st child: Dominant strategy none         ; 2nd child: (Near)==SPNE'
WRITE(4,fmt=*) '  14. 1st child: Dominant strategy none         ; 2nd child: (Far )==SPNE'
WRITE(4,fmt=*) '  15. 1st child: Dominant strategy none         ; 2nd child: none'
WRITE(4,fmt=*) '  NB: 2nd player always plays a dominant strategy if any.'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A12)') '       Type','       Total',' (DDif,DDif)',' (DDif,IDif)',' (IDif,DDif)',' (IDif,IDif)'
DO k=1,15
    WRITE(4,fmt='(I12,5F12.7)') k,(g2table17b(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-18a First best arrangements'
WRITE(4,fmt=*) '  By Decreasing difference status (Column share)'
WRITE(4,fmt=*) '   0. FB arrangement: (Far ,Far )'
WRITE(4,fmt=*) '   1. FB arrangement: (Far ,Near)'
WRITE(4,fmt=*) '   2. FB arrangement: (Near,Far )'
WRITE(4,fmt=*) '   3. FB arrangement: (Near,Near)'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A12)') '       Type','       Total',' (DDif,DDif)',' (DDif,IDif)',' (IDif,DDif)',' (IDif,IDif)'
DO k=0,3
    WRITE(4,fmt='(I12,5F12.7)') k,(g2table18a(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-18b Reverse order SPNE arrangements'
WRITE(4,fmt=*) '  By Decreasing difference status (Column share)'
WRITE(4,fmt=*) '   0. Reverse SPNE: (Far ,Far )'
WRITE(4,fmt=*) '   1. Reverse SPNE: (Far ,Near)'
WRITE(4,fmt=*) '   2. Reverse SPNE: (Near,Far )'
WRITE(4,fmt=*) '   3. Reverse SPNE: (Near,Near)'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A12)') '       Type','       Total',' (DDif,DDif)',' (DDif,IDif)',' (IDif,DDif)',' (IDif,IDif)'
DO k=0,3
    WRITE(4,fmt='(I12,5F12.7)') k,(g2table18b(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-18c First best arrangements and reverse order SPNE arrangements'
WRITE(4,fmt=*) '  By Decreasing difference status (Column share)'
WRITE(4,fmt=*) '   0. FB arrangement: (Far ,Far ) ; Reverse SPNE: (Far ,Far )'
WRITE(4,fmt=*) '   1. FB arrangement: (Far ,Far ) ; Reverse SPNE: (Far ,Near)'
WRITE(4,fmt=*) '   2. FB arrangement: (Far ,Far ) ; Reverse SPNE: (Near,Far )'
WRITE(4,fmt=*) '   3. FB arrangement: (Far ,Far ) ; Reverse SPNE: (Near,Near)'
WRITE(4,fmt=*) '   4. FB arrangement: (Far ,Near) ; Reverse SPNE: (Far ,Far )'
WRITE(4,fmt=*) '   5. FB arrangement: (Far ,Near) ; Reverse SPNE: (Far ,Near)'
WRITE(4,fmt=*) '   6. FB arrangement: (Far ,Near) ; Reverse SPNE: (Near,Far )'
WRITE(4,fmt=*) '   7. FB arrangement: (Far ,Near) ; Reverse SPNE: (Near,Near)'
WRITE(4,fmt=*) '   8. FB arrangement: (Near,Far ) ; Reverse SPNE: (Far ,Far )'
WRITE(4,fmt=*) '   9. FB arrangement: (Near,Far ) ; Reverse SPNE: (Far ,Near)'
WRITE(4,fmt=*) '  10. FB arrangement: (Near,Far ) ; Reverse SPNE: (Near,Far )'
WRITE(4,fmt=*) '  11. FB arrangement: (Near,Far ) ; Reverse SPNE: (Near,Near)'
WRITE(4,fmt=*) '  12. FB arrangement: (Near,Near) ; Reverse SPNE: (Far ,Far )'
WRITE(4,fmt=*) '  13. FB arrangement: (Near,Near) ; Reverse SPNE: (Far ,Near)'
WRITE(4,fmt=*) '  14. FB arrangement: (Near,Near) ; Reverse SPNE: (Near,Far )'
WRITE(4,fmt=*) '  15. FB arrangement: (Near,Near) ; Reverse SPNE: (Near,Near)'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A12)') '       Type','       Total',' (DDif,DDif)',' (DDif,IDif)',' (IDif,DDif)',' (IDif,IDif)'
DO k=0,15
    WRITE(4,fmt='(I12,5F12.7)') k,(g2table18c(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-19: Utility changes in SPNE compared with normal form N.E.'
WRITE(4,fmt=*) '  By Decreasing difference status'
WRITE(4,fmt=*) '   0. No change (SPNE==Normal form unique N.E.)'
WRITE(4,fmt=*) '   1. 2nd child: (-) ; Total: (-)'
WRITE(4,fmt=*) '   2. 2nd child: (-) ; Total: (+)'
WRITE(4,fmt=*) '   3. 2nd child: (+) ; Total: (+)'
WRITE(4,fmt=*) '   4. 2 equilibria'
WRITE(4,fmt=*) '   5. No equilibrium'
WRITE(4,fmt=*) '  NB: When there is a normal form N.E., the 1st mover can always'
WRITE(4,fmt=*) '      achieve that outcome. --> 1st child never worses off.'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A12)') '       Type','       Total',' (DDif,DDif)',' (DDif,IDif)',' (IDif,DDif)',' (IDif,IDif)'
DO k=0,5
    WRITE(4,fmt='(I12,5F12.7)') k,(g2table19(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-19a: Utility changes in SPNE compared with normal form N.E.'
WRITE(4,fmt=*) '  By Decreasing difference status (Column share)'
WRITE(4,fmt=*) '   0. No change (SPNE==Normal form unique N.E.)'
WRITE(4,fmt=*) '   1. 2nd child: (-) ; Total: (-)'
WRITE(4,fmt=*) '   2. 2nd child: (-) ; Total: (+)'
WRITE(4,fmt=*) '   3. 2nd child: (+) ; Total: (+)'
WRITE(4,fmt=*) '   4. 2 equilibria'
WRITE(4,fmt=*) '   5. No equilibrium'
WRITE(4,fmt=*) '  NB: When there is a normal form N.E., the 1st mover can always'
WRITE(4,fmt=*) '      achieve that outcome. --> 1st child never worses off.'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A12)') '       Type','       Total',' (DDif,DDif)',' (DDif,IDif)',' (IDif,DDif)',' (IDif,IDif)'
DO k=0,5
    WRITE(4,fmt='(I12,5F12.7)') k,(g2table19a(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-20: Utility changes in reverse SPNE (compared with SPNE)'
WRITE(4,fmt=*) '  By Decreasing difference status'
WRITE(4,fmt=*) '   0. No change (SPNE==ReverseSPNE)'
WRITE(4,fmt=*) '   1. Total: (-) ; 1st child: (-) ; 2nd child: (-)'
WRITE(4,fmt=*) '   2. Total: (-) ; 1st child: (-) ; 2nd child: (+)'
WRITE(4,fmt=*) '   3. Total: (-) ; 1st child: (+) ; 2nd child: (-)'
WRITE(4,fmt=*) '   4. Total: (+) ; 1st child: (-) ; 2nd child: (+)'
WRITE(4,fmt=*) '   5. Total: (+) ; 1st child: (+) ; 2nd child: (-)'
WRITE(4,fmt=*) '   6. Total: (+) ; 1st child: (+) ; 2nd child: (+)'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A12)') '       Type','       Total',' (DDif,DDif)',' (DDif,IDif)',' (IDif,DDif)',' (IDif,IDif)'
DO k=0,6
    WRITE(4,fmt='(I12,5F12.7)') k,(g2table20(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-20a: Utility changes in reverse SPNE (compared with SPNE)'
WRITE(4,fmt=*) '  By Decreasing difference status (Column share)'
WRITE(4,fmt=*) '   0. No change (SPNE==ReverseSPNE)'
WRITE(4,fmt=*) '   1. Total: (-) ; 1st child: (-) ; 2nd child: (-)'
WRITE(4,fmt=*) '   2. Total: (-) ; 1st child: (-) ; 2nd child: (+)'
WRITE(4,fmt=*) '   3. Total: (-) ; 1st child: (+) ; 2nd child: (-)'
WRITE(4,fmt=*) '   4. Total: (+) ; 1st child: (-) ; 2nd child: (+)'
WRITE(4,fmt=*) '   5. Total: (+) ; 1st child: (+) ; 2nd child: (-)'
WRITE(4,fmt=*) '   6. Total: (+) ; 1st child: (+) ; 2nd child: (+)'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A12)') '       Type','       Total',' (DDif,DDif)',' (DDif,IDif)',' (IDif,DDif)',' (IDif,IDif)'
DO k=0,6
    WRITE(4,fmt='(I12,5F12.7)') k,(g2table20a(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-21: Utility changes in reverse SPNE (compared with SPNE)'
WRITE(4,fmt=*) '  Across [SPNE vs normal form NE comparison], (Column share)'
WRITE(4,fmt=*) '   0. No change (SPNE==ReverseSPNE)'
WRITE(4,fmt=*) '   1. Total: (-) ; 1st child: (-) ; 2nd child: (-)'
WRITE(4,fmt=*) '   2. Total: (-) ; 1st child: (-) ; 2nd child: (+)'
WRITE(4,fmt=*) '   3. Total: (-) ; 1st child: (+) ; 2nd child: (-)'
WRITE(4,fmt=*) '   4. Total: (+) ; 1st child: (-) ; 2nd child: (+)'
WRITE(4,fmt=*) '   5. Total: (+) ; 1st child: (+) ; 2nd child: (-)'
WRITE(4,fmt=*) '   6. Total: (+) ; 1st child: (+) ; 2nd child: (+)'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(7A11)') '       Type','  SPNE=N.E.',' 2nd- sum:-',' 2nd- sum:+',' 2nd+  sum:+','     2 N.E.','    No N.E.'
DO k=0,6
    WRITE(4,fmt='(I11,6F11.6)') k,(g2table21(k,m),m=1,6)
ENDDO
WRITE(4,fmt=*) ' '


WRITE(4,fmt=*) 'Table 2-22: Normal form classification of 2 child families'
WRITE(4,fmt=*) '  By First best living arrangement type.'
WRITE(4,fmt=*) '   0. No normal form equilibrium'
WRITE(4,fmt=*) '   1. Unique equilibrium (Far ,Far ): First best'
WRITE(4,fmt=*) '   2. Unique equilibrium (Far ,Far ): Not first best, but not Pareto dominated'
WRITE(4,fmt=*) '   3. Unique equilibrium (Far ,Far ): Pareto dominated'
WRITE(4,fmt=*) '   4. Unique equilibrium (Far ,Near): First best'
WRITE(4,fmt=*) '   5. Unique equilibrium (Far ,Near): Not first best, but not Pareto dominated'
WRITE(4,fmt=*) '   6. Unique equilibrium (Far ,Near): Pareto dominated'
WRITE(4,fmt=*) '   7. Unique equilibrium (Near,Far ): First best'
WRITE(4,fmt=*) '   8. Unique equilibrium (Near,Far ): Not first best, but not Pareto dominated'
WRITE(4,fmt=*) '   9. Unique equilibrium (Near,Far ): Pareto dominated'
WRITE(4,fmt=*) '  10. Unique equilibrium (Near,Near): First best'
WRITE(4,fmt=*) '  11. Unique equilibrium (Near,Near): Not first best, but not Pareto dominated'
WRITE(4,fmt=*) '  12. Unique equilibrium (Near,Near): Pareto dominated'
WRITE(4,fmt=*) '  13. Two equilibria (coordination): Either first best'
WRITE(4,fmt=*) '  14. Two equilibria (coordination): Neither first best'
WRITE(4,fmt=*) '  15. Two equilibria (anti-coordination): Either first best'
WRITE(4,fmt=*) '  16. Two equilibria (anti-coordination): Neither first best'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A12)') '       Type','       Total',' (Far ,Far )',' (Far ,Near)',' (Near,Far )',' (Near,Near)'
WRITE(4,fmt=*) ' Not available now.'
!DO k=0,16
!    WRITE(4,fmt='(I12,5F12.7)') k,(g2table22b(k,m),m=1,5)
!ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-23: First best arrangements'
WRITE(4,fmt=*) '  By First best living arrangement type. (Column share)'
WRITE(4,fmt=*) '   0. SPNE (observed) arrangement: (Far ,Far )'
WRITE(4,fmt=*) '   1. SPNE (observed) arrangement: (Far ,Near)'
WRITE(4,fmt=*) '   2. SPNE (observed) arrangement: (Near,Far )'
WRITE(4,fmt=*) '   3. SPNE (observed) arrangement: (Near,Near)'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A12)') '       Type','       Total',' (Far ,Far )',' (Far ,Near)',' (Near,Far )',' (Near,Near)'
DO k=0,3
    WRITE(4,fmt='(I12,5F12.7)') k,(g2table23(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-24: SPNE efficiency type'
WRITE(4,fmt=*) '  By First best living arrangement type. (Column share)'
WRITE(4,fmt=*) '   0. Joint utility best (KH efficient)'
WRITE(4,fmt=*) '   1. KH inefficient, not Pareto dominated'
WRITE(4,fmt=*) '   2. Pareto Dominated (P.D.)'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A12)') '       Type','       Total',' (Far ,Far )',' (Far ,Near)',' (Near,Far )',' (Near,Near)'
DO k=0,2
    WRITE(4,fmt='(I12,5F12.7)') k,(g2table24(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-25: Increasing differences'
WRITE(4,fmt=*) '  By First best living arrangement type. (Column share)'
WRITE(4,fmt=*) '   0. 1st child: Decreasing difference ; 2nd child: Decreasing difference'
WRITE(4,fmt=*) '   1. 1st child: Decreasing difference ; 2nd child: Increasing difference'
WRITE(4,fmt=*) '   2. 1st child: Increasing difference ; 2nd child: Decreasing difference'
WRITE(4,fmt=*) '   3. 1st child: Increasing difference ; 2nd child: Increasing difference'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A12)') '       Type','       Total',' (Far ,Far )',' (Far ,Near)',' (Near,Far )',' (Near,Near)'
DO k=0,3
    WRITE(4,fmt='(I12,5F12.7)') k,(g2table25(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '

WRITE(4,fmt=*) 'Table 2-26: Reverse order SPNE arrangements'
WRITE(4,fmt=*) '  By First best living arrangement type. (Column share)'
WRITE(4,fmt=*) '   0. Reverse SPNE: (Far ,Far )'
WRITE(4,fmt=*) '   1. Reverse SPNE: (Far ,Near)'
WRITE(4,fmt=*) '   2. Reverse SPNE: (Near,Far )'
WRITE(4,fmt=*) '   3. Reverse SPNE: (Near,Near)'
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(6A12)') '       Type','       Total',' (Far ,Far )',' (Far ,Near)',' (Near,Far )',' (Near,Near)'
DO k=0,3
    WRITE(4,fmt='(I12,5F12.7)') k,(g2table26(k,m),m=1,5)
ENDDO
WRITE(4,fmt=*) ' '



CLOSE(4)



END PROGRAM MAIN
